/**
 * @desc Declaration of elements of VAM (signals, functional nodes, storages).
 */
#include "elements.h"
#include <algorithm> // find

#define BITWIDTH_ERR(msg) THROW("bitwidth problem: " << msg)

namespace vam
{

const string tbl_oper_to_str[] = {
    "OP_NOP",    // 0
    "OP_CONST",  // 1
    "OP_VAR",    // 2
    "OP_REF",    // 3
    "OP_ASSIGN", // 4

    "OP_NOT",    // 5
    "OP_OR",     // 6
    "OP_AND",    // 7
    "OP_ITE",    // 8

    "OP_INV",    // 9
    "OP_BOR",    // 10
    "OP_BAND",   // 11
    "OP_XOR",    // 12
    "OP_SHL",    // 13
    "OP_SHR",    // 14
    "OP_SAR",    // 15
    "OP_ORJOIN", // 16

    "OP_EQ",     // 17
    "OP_NEQ",    // 18
    "OP_LT",     // 19
    "OP_GT",     // 20
    "OP_LE",     // 21
    "OP_GE",     // 22
    "OP_SLT",    // 23
    "OP_SGT",    // 24
    "OP_SLE",    // 25
    "OP_SGE",    // 26

    "OP_UTS",    // 27
    "OP_STU",    // 28

    "OP_CASE",   // 29
    "OP_COND",   // 30

    "OP_NEG",    // 31
    "OP_PLUS",   // 32
    "OP_MINUS",  // 33
    "OP_MUL",    // 34
    "OP_DIV",    // 35
    "OP_MOD",    // 36

    "OP_IDX",    // 37
    "OP_BITEX",  // 38
    "OP_CONCAT", // 39
    "OP_ZEX",    // 40
    "OP_SEX",    // 41
    "OP_SLICE",  // 42

    "OP_ABS_LOGIC",  // 43
    "OP_ABS_BITWISE",// 44
    "OP_ABS_REL",    // 45
    "OP_ABS_VECTOR", // 46
    "OP_ABS_ARITH",  // 47

    "OP_VAM_STOP", // 48
};

const string tbl_oper_to_group[] = {
    "unique",    // NOP      = 0, // no-operation represents some constant
    "unique",    // CONST    = 1, // constant
    "unique",    // VAR      = 2, // reference to a variable
    "unique",    // REF      = 3, // reference to another expression
    "unique",    // ASSIGN   = 4, // assignment
    // logic
    "logic",    // NOT      = 5,
    "logic",    // OR       = 6,
    "logic",    // AND      = 7,
    "logic",    // ITE      = 8,
    // bitwise
    "bitwise",    // INV      = 9,
    "bitwise",    // BOR      = 10, // bitwise or
    "bitwise",    // BAND     = 11, // bitwise and
    "bitwise",    // XOR      = 12,
    "bitwise",    // SHL      = 13,
    "bitwise",    // SHR      = 14,
    "bitwise",    // SAR      = 15,
    "bitwise",    // ORJOIN   = 16, // or-join over multiple operands
    // relational
    "relational",    // EQ       = 17,
    "relational",    // NEQ      = 18,
    "relational",    // LT       = 19, // unsigned less-than
    "relational",    // GT       = 20, // unsigned greater-than
    "relational",    // LE       = 21, // unsigned less-or-equal
    "relational",    // GE       = 22, // unsigned less-or-equal
    "relational",    // SLT      = 23, // signed less-than
    "relational",    // SGT      = 24, // signed greater-than
    "relational",    // SLE      = 25, // signed less-or-equal
    "relational",    // SGE      = 26, // signed greater-or-equal
    // casting
    "casting",    // UTS      = 27, // unsigned-to-signed
    "casting",    // STU      = 28, // signed-to-unsigned
    // conditional
    "conditional",    // CASE     = 29,
    "conditional",    // SWITCH   = 30,
    // arithmetic
    "arithmetic",    // NEG      = 31,
    "arithmetic",    // PLUS     = 32,
    "arithmetic",    // MINUS    = 33,
    "arithmetic",    // MUL      = 34,
    "arithmetic",    // DIV      = 35,
    "arithmetic",    // MOD      = 36,
    // vector
    "vector",    // IDX      = 37, // indexing
    "vector",    // BITEX    = 38, // bit extraction
    "vector",    // CONCAT   = 39, // concatenation
    "vector",    // ZEX      = 40, // zero extension
    "vector",    // SEX      = 41, // sign extension
    "vector",    // SLICE    = 42, // vector slice
    // abstract
    "abstract",    // ABS_LOGIC    = 43, // abstract logic
    "abstract",    // ABS_BITWISE  = 44, // abstract bitwise
    "abstract",    // ABS_REL      = 45, // abstract relational
    "abstract",    // ABS_VECTOR   = 46, // abstract vector or casting
    "abstract",    // ABS_ARITH    = 47, // abstract arithmetic
    // VAM stop
    "vam_stop",
};

void VAMElement::Accept(VAMVisitor *v) { v->VisitVAMElement(this); }

/****************************************************************************
 * Meta
 ****************************************************************************/
void Meta::Accept(VAMVisitor *v)
{
    v->VisitMeta(this);
}


Meta::~Meta()
{
    for (auto m : metas)
        delete m;
}

/// add a new node
Meta *Meta::add(MetaType t, const string &v)
{
    Meta* m = new Meta(t,v);
    add(m);
    return m;
}

/// add meta node
void Meta::add(Meta *m)
{
    if (type == MT_NODE || type == MT_UNDEFINED)
    {
        type = MT_NODE;
        metas.push_back(m);
    }
    else
        THROW("incompatible meta to add")
}

void Meta::add(const string &v)
{
    if (type == MT_NODE || type == MT_UNDEFINED)
    {
        type = MT_NODE;
        metas.push_back(new Meta(MT_VALUE, v));
    }
    else
        THROW("incompatible meta to add")
}

void Meta::add(long int i)
{
    if (type == MT_NODE || type == MT_UNDEFINED)
    {
        type = MT_NODE;
        ostringstream os;
        os << i;
        metas.push_back(new Meta(MT_INTVALUE, os.str()));
    }
    else
        THROW("incompatible meta to add")
}

/// add key-value nodes
void Meta::addkv(const string &key, const string &val)
{
    add(MT_NODE, key)->add(MT_VALUE, val);
}

void Meta::erase(unsigned pos)
{
    delete metas[pos];
    metas.erase(metas.begin() + pos);
}

void Meta::erase(vector<Meta*>::iterator i)
{
    delete *i;
    metas.erase(i);
}

/// Find a node with name `node'. Return nullptr if not found.
Meta *Meta::find(const string &node, Meta *last)
{
    auto i = metas.begin();
    // skip all nodes till last (including)
    if (last != nullptr)
        for (; i != metas.end(); i++)
            if (((*i)->type == MT_NODE ||
                 (*i)->type == MT_VALUE) && (*i)->value == node)
            {
                i++;
                break;
            }
    // find the next node
    for (; i != metas.end(); i++)
    {
        if (((*i)->type == MT_NODE ||
             (*i)->type == MT_VALUE) && (*i)->value == node)
            return *i;
    }
    return nullptr;
}

Meta &Meta::node(const string &nodename)
{
    Meta *m = find(nodename);
    if (m == nullptr)
        return *add(MT_NODE, nodename);
    else
        return *m;
}

string &Meta::keyval(const string &key)
{
    Meta *node = find(key);
    if (node)
    {
        if (node->metas.empty())
            return node->add(MT_VALUE, "")->value;
        else
            return node->metas[0]->value;
    }
    else
        return add(MT_NODE, key)->add(MT_VALUE, "")->value;
}

string Meta::keyval_nocreat(const string &key)
{
    Meta *node = find(key);
    if (node && !node->metas.empty())
        return node->metas[0]->value;
    else
        return "";
}

string &Meta::operator[](const string &key)
{
    return keyval(key);
}

Meta *Meta::find_erase(const string &node, Meta *last)
{
    auto i = metas.begin();
    // skip all nodes till last (including)
    if (last != nullptr)
        for (; i != metas.end(); i++)
            if (((*i)->type == MT_NODE ||
                 (*i)->type == MT_VALUE) && (*i)->value == node)
            {
                i++;
                break;
            }
    // find the next node
    for (; i != metas.end(); i++)
    {
        if (((*i)->type == MT_NODE ||
             (*i)->type == MT_VALUE) && (*i)->value == node)
        {
            erase(i);
            return last;
        }
    }
    return nullptr;
}

/// move all metadata from this object to target
void Meta::move_to(Meta *target)
{
    for (auto m : metas)
        target->add(m);
    metas.erase(metas.begin(), metas.end());
}

string &Meta::operator[](unsigned pos)
{
    return metas[pos]->value;
}

Meta &Meta::operator <<(Meta *m)
{
    add(m);
    return *this;
}

Meta &Meta::operator <<(const string &v)
{
    add(v);
    return *this;
}

Meta &Meta::operator <<(long int i)
{
    add(i);
    return *this;
}

void Meta::clear()
{
    type = MT_UNDEFINED;
    value = "";
    for (auto m : metas)
        delete m;
    metas.clear();
}

void Meta::copy(const Meta &m)
{
    if (type != MT_UNDEFINED)
        clear();
    type = m.type;
    value = m.value;
    for (auto meta : m.metas)
        metas.push_back(new Meta(*meta));
}

void Meta::AcceptAll(VAMVisitor *visitor)
{
    for (auto m : metas)
        visitor->VisitMeta(m);
}

/****************************************************************************
 * Variable
 ****************************************************************************/
void Variable::Accept(VAMVisitor *v)
{
    v->VisitVariable(this);
}


Variable::~Variable()
{
    if (references.size() > 0)
        WARNING("variable " << id << " (width=" << bitwidth << ") has " <<
                references.size() << " back reference" <<
                (references.size() > 1 ? "s" : ""));
}

/**
 * @brief   Find and remove back reference from references.
 * @param   expr    Expression referencing to the variable to be removed.
 */
void Variable::unlink(Expression *expr)
{
    for (auto i = references.begin(); i != references.end(); i++)
        if (*i == expr)
        {
            references.erase(i);
            break;
        }
}

/****************************************************************************
 * Expression
 ****************************************************************************/
string Expression::default_tag;

void Expression::Accept(VAMVisitor *v)
{
    v->VisitExpression(this);
}


Expression::~Expression()
{
    if (outputs.size() > 0)
        WARNING("Expression (op=" << op << ", bw=" << bitwidth << ") has "
                << outputs.size() << " unlinked output(s).");
}

int Expression::reset_bitwidth()
{
    return bitwidth;
}

/**
 * @brief   Bind a new expression output.
 * @param   expr    Binded expression (reads value of this expression)
 * @return  `this' pointer
 */
Expression *Expression::bind(Expression *expr)
{
    outputs.push_back(expr);
    return this;
}

/**
 * @brief   Find and remove output reference.
 * @param   expr    Expression referencing to the variable to be removed.
 */
void Expression::unlink(const Expression *expr)
{
    for (auto i = outputs.begin(); i != outputs.end(); i++)
        if (*i == expr)
        {
            outputs.erase(i);
            break;
        }
}

/**
 * @brief   Unlink the operand from this expression. If the operand has no
 * other reference, delete it from the memory.
 * @param   operand Operand to unlink from.
 */
void Expression::remove(Expression *operand)
{
    if (operand != nullptr)
    {
        operand->unlink(this);
        if (operand->outputs.size() == 0)
            delete operand;
    }
}

/**
 * @brief   Replaces every occurence of a given expression by a new
 *          expression.
 * @param   orig        original expression to be replaces
 * @param   replacement expression replacement
 */
void Expression::replace_operand(Expression *orig, Expression *replacement)
{
    (void)orig;
    (void)replacement;
}

/**
 * @brief   Returns the first operand of the expression. Used for iterators
 * over expression operands.
 * @return  first operand.
 */
Expression **Expression::first_operand()
{
    return NULL_expression_pointer;
}

/**
 * @brief   Returns operand of the expression next to a given operand.
 * @param   _op     reference operand next to which find the next operand
 * @return  next operand (nullptr if no more operand)
 */
Expression **Expression::next_operand(Expression **_op)
{
    (void)_op;
    return NULL_expression_pointer;
}

bool Expression::op_is_logic()
{
    // depends on order of operator values
    return op >= OP_NOT && op <= OP_ITE;
}

bool Expression::op_is_bitwise()
{
    // depends on order of operator values
    return op >= OP_INV && op <= OP_ORJOIN;
}

bool Expression::op_is_relational()
{
    // depends on order of operator values
    return op >= OP_EQ && op <= OP_SGE;
}

bool Expression::op_is_casting()
{
    // depends on order of operator values
    return op >= OP_UTS && op <= OP_STU;
}

bool Expression::op_is_conditional()
{
    // depends on order of operator values
    return op >= OP_CASE && op <= OP_COND;
}

bool Expression::op_is_arithmetic()
{
    // depends on order of operator values
    return op >= OP_NEG && op <= OP_MOD;
}

bool Expression::op_is_vector()
{
    // depends on order of operator values
    return op >= OP_IDX && op <= OP_SLICE;
}

bool Expression::op_is_abstract()
{
    // depends on order of operator values
    return op >= OP_ABS_LOGIC && op <= OP_ABS_ARITH;
}

int replace_outputs(Expression *e, Expression *replacement,
       Expression *except_for)
{
/*  Simply do this:

    ExprRef holder{e};
    int i = e->outputs.size();
    for (auto op : e->outputs)
        op->replace_operand(e, replacement);
    return i;

    But since replace_operand() invalidates e->outputs iterators, one
    must go around.
*/

    ExprRef holder{e};  // adds 'holder' to the and of e->outputs
    unsigned prev_size = 0;
    unsigned skip = 1;
    unsigned rplc_counter = 0;
    while (e->outputs.size() > skip)
    {
        if (e->outputs[e->outputs.size()-1-skip] == except_for)
        {
            skip++;
            continue;
        }

        ASSERT(prev_size != e->outputs.size(),
                "nothing changed -> possible endless loop");
        prev_size = e->outputs.size();
        e->outputs[e->outputs.size()-1-skip]->replace_operand(e, replacement);
        rplc_counter++;
    }
    return rplc_counter;
}

Expression **NULL_expression_pointer = nullptr;

OperandIterator end_operand_iterator(nullptr, nullptr);

/****************************************************************************
 * Individual types of expression
 ****************************************************************************/

/*************************************************
 * Auxiliary functions for reset_bitwidth methods
 *************************************************/

/**
 * @brief Reset, check, and return one-bit width of a logic operation.
 * Requires one or more one-bit operands.
 * @return  1 if success, or -1 if an error
 */
int get_onebit_multiple_operands(ExpressionList &operands)
{
    if (! operands.empty())
    {
        for (auto o : operands)
            if (o->get_bitwidth() != 1)
                BITWIDTH_ERR("invalid");
//DBG                return -1;
        return 1;
    }
    else
        BITWIDTH_ERR("no operand");
//DBG        return -1;
}

/********************************************
 * ExprConst
 */
void ExprConst::Accept(VAMVisitor *v)
{
    v->VisitExprConst(this);
}


ExprConst::ExprConst()
{
    op = OP_CONST;
}

ExprConst::ExprConst(string v, int bw) : value(v)
{
    op = OP_CONST;
    bitwidth = bw;
}

/********************************************
 * ExprVariable
 */
void ExprVariable::Accept(VAMVisitor *v)
{
    v->VisitExprVariable(this);
}


ExprVariable::ExprVariable()
{
    op = OP_VAR;
    v = nullptr;
}

ExprVariable::ExprVariable(Variable *var)
{
    op = OP_VAR;
    v = var;
    bitwidth = var->bitwidth;
    var->references.push_back(this);
}

ExprVariable::~ExprVariable()
{
    if (v)
        v->unlink(this);
}

int ExprVariable::reset_bitwidth()
{
    bitwidth = (v != nullptr) ? v->bitwidth : -1;
    if (bitwidth < 0)
        BITWIDTH_ERR("undefined in variable");
    return bitwidth;
}

/********************************************
 * UnaryExpression
 */
UnaryExpression::~UnaryExpression()
{
    remove(expr);
}

void UnaryExpression::add_operand(Expression *e)
{
    ASSERT(expr == nullptr, "expression already set");
    expr = e->bind(this);
}

void UnaryExpression::replace_operand(Expression *o, Expression *r)
{
    replace_op(expr, o, r);
}

Expression **UnaryExpression::first_operand()
{
    return &expr;
}

/********************************************
 * ExprRef
 */
void ExprRef::Accept(VAMVisitor *v)
{
    v->VisitExprRef(this);
}


int ExprRef::reset_bitwidth()
{
    if (expr && expr->get_bitwidth() > 0)
        return bitwidth = expr->bitwidth;
    BITWIDTH_ERR("invalid");
}

/********************************************
 * ExprNot
 */
void ExprNot::Accept(VAMVisitor *v)
{
    v->VisitExprNot(this);
}


int ExprNot::reset_bitwidth()
{
    if (expr && expr->get_bitwidth() == 1)
        return bitwidth = 1;
    BITWIDTH_ERR("invalid");
}

/********************************************
 * ExpressionPool
 */
void ExpressionPool::Accept(VAMVisitor *v)
{
    v->VisitExpressionPool(this);
}


ExpressionPool::~ExpressionPool()
{
    for (auto o : operands)
        remove(o);
}

void ExpressionPool::add_operand(Expression *o)
{
    operands.push_back(o->bind(this));
    get_bitwidth();
}

void ExpressionPool::erase_operand(Expression *o)
{
    for (auto i = operands.begin(); i != operands.end(); i++)
        if (*i == o)
        {
            operands.erase(i);
            remove(o);
            break;
        }
}

void ExpressionPool::replace_operand(Expression *o, Expression *r)
{
    for (auto &expr : operands)
        replace_op(expr, o, r);
}

Expression **ExpressionPool::first_operand()
{
    if (operands.size())
        return &operands[0];
    else
        return NULL_expression_pointer;
}

Expression **ExpressionPool::next_operand(Expression **e)

{
    unsigned i;
    for (i = 0; i < operands.size(); i++)
        if (e == &operands[i])
            break;
    if (i+1 < operands.size())
        return &operands[i+1];
    else
        return NULL_expression_pointer;
}

/********************************************
 * ExprOr
 */
void ExprOr::Accept(VAMVisitor *v)
{
    v->VisitExprOr(this);
}


ExprOr::ExprOr()
{
    op = OP_OR;
}

ExprOr::ExprOr(Expression *o1, Expression *o2)
{
    op = OP_OR;
    operands.push_back(o1->bind(this));
    operands.push_back(o2->bind(this));
    get_bitwidth();
}

int ExprOr::reset_bitwidth()
{
    bitwidth = get_onebit_multiple_operands(operands);
    ASSERT(bitwidth == 1, "ExprOr requires one-bit operands");
    return bitwidth;
}

/********************************************
 * ExprAnd
 */
void ExprAnd::Accept(VAMVisitor *v)
{
    v->VisitExprAnd(this);
}

ExprAnd::ExprAnd()
{
    op = OP_AND;
}

ExprAnd::ExprAnd(Expression *o1, Expression *o2)
{
    op = OP_AND;
    operands.push_back(o1->bind(this));
    operands.push_back(o2->bind(this));
    get_bitwidth();
}

int ExprAnd::reset_bitwidth()
{
    bitwidth = get_onebit_multiple_operands(operands);
    ASSERT(bitwidth == 1, "ExprAnd requires one-bit operands");
    return bitwidth;
}

/********************************************
 * ExprOrJoin
 */
void ExprOrJoin::Accept(VAMVisitor *v)
{
    v->VisitExprOrJoin(this);
}


ExprOrJoin::ExprOrJoin()
{
    op = OP_ORJOIN;
}

ExprOrJoin::ExprOrJoin(Expression *o1, Expression *o2)
{
    op = OP_ORJOIN;
    operands.push_back(o1->bind(this));
    operands.push_back(o2->bind(this));
    get_bitwidth();
}

int ExprOrJoin::reset_bitwidth()
{
    if (! operands.empty())
    {
        bitwidth = operands[0]->get_bitwidth();
        for (auto o : operands)
            if (o->get_bitwidth() != bitwidth)
                BITWIDTH_ERR("incompatible bitwidth of OR-join operands");
        return bitwidth;
    }
    else
        BITWIDTH_ERR("ExprOrJoin requires positive-bitwidth operands");
}

/********************************************
 * ExprITE
 */
void ExprITE::Accept(VAMVisitor *v)
{
    v->VisitExprITE(this);
}


ExprITE::ExprITE()
{
    op = OP_ITE;
    cond = e_true = e_false = nullptr;
}

ExprITE::ExprITE(Expression *c, Expression *o1, Expression *o2)
{
    op = OP_ITE;
    cond = c->bind(this);
    e_true = o1->bind(this);
    e_false = o2->bind(this);
    get_bitwidth();
}

ExprITE::~ExprITE()
{
    remove(cond);
    remove(e_true);
    remove(e_false);
}

void ExprITE::add_cond(Expression *e)
{
    ASSERT(cond == nullptr, "condition already set");
    cond = e->bind(this);
}

void ExprITE::add_true(Expression *e)
{
    ASSERT(e_true == nullptr, "true branch already set");
    e_true = e->bind(this);
}

void ExprITE::add_false(Expression *e)
{
    ASSERT(e_false == nullptr, "false branch already set");
    e_false = e->bind(this);
}

int ExprITE::reset_bitwidth()
{
    if (cond && e_true && e_false)
    {
        if (cond->get_bitwidth() == 1 &&
            e_true->get_bitwidth() > 0 &&
            e_false->get_bitwidth() == e_true->bitwidth)
            return bitwidth = e_true->bitwidth;
        else
            BITWIDTH_ERR("incompatible operands in ExprITE")
    }
    BITWIDTH_ERR("invalid operands in ExprITE")
}

void ExprITE::replace_operand(Expression *o, Expression *r)
{
    replace_op(cond, o, r);
    replace_op(e_true, o, r);
    replace_op(e_false, o, r);
}

Expression **ExprITE::first_operand()
{
    return &cond;
}

Expression **ExprITE::next_operand(Expression **e)
{
    if (e == &cond)
        return &e_true;
    if (e == &e_true)
        return &e_false;
    return NULL_expression_pointer;
}

/********************************************
 * ExprInv
 */
void ExprInv::Accept(VAMVisitor *v)
{
    v->VisitExprInv(this);
}


int ExprInv::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0)
            return bitwidth = expr->bitwidth;
    BITWIDTH_ERR("invalid operand in ExprInv")
}

/********************************************
 * ExprNeg
 */
void ExprNeg::Accept(VAMVisitor *v)
{
    v->VisitExprNeg(this);
}


int ExprNeg::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0)
            return bitwidth = expr->bitwidth;
    BITWIDTH_ERR("invalid operand in ExprNeg")
}

/********************************************
 * ExprUnsignedToSigned
 */
void ExprUnsignedToSigned::Accept(VAMVisitor *v)
{
    v->VisitExprUnsignedToSigned(this);
}


int ExprUnsignedToSigned::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0)
            return bitwidth = expr->bitwidth;
    BITWIDTH_ERR("invalid operand in ExprUnsignedToSigned")
}

/********************************************
 * ExprSignedToUnsigned
 */
void ExprSignedToUnsigned::Accept(VAMVisitor *v)
{
    v->VisitExprSignedToUnsigned(this);
}


int ExprSignedToUnsigned::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0)
            return bitwidth = expr->bitwidth;
    BITWIDTH_ERR("invalid operand in ExprSignedToUnsigned")
}

/********************************************
 * ExprControl
 */
ControlExpression::~ControlExpression()
{
    remove(def);
    for (auto c : cases)
    {
        remove(c.first);
        remove(c.second);
    }
}

/// Add case expressions to operands.
void ControlExpression::add_case(Expression *cond, Expression *value)
{
    cond->bind(this);
    value->bind(this);
    cases.push_back(make_pair(cond,value));
    get_bitwidth();
}

/// Add default case to operands.
void ControlExpression::add_default(Expression *d)
{
    if (def)
        THROW("default case already specified in control expression");
    def = d->bind(this);
    get_bitwidth();
}

void ControlExpression::replace_operand(Expression *o, Expression *r)
{
    replace_op(def, o, r);
    for (auto &c : cases)
    {
        replace_op(c.first, o, r);
        replace_op(c.second, o, r);
    }
}

Expression **ControlExpression::first_operand()
{
    if (cases.size())
        return &cases[0].first;
    return NULL_expression_pointer;
}

Expression **ControlExpression::next_operand(Expression **e)
{
    unsigned i;
    for (i = 0; i < cases.size(); i++)
        if (e == &cases[i].first)
            return &cases[i].second;
        else if (e == &cases[i].second)
        {
            if (i+1 < cases.size())
                return &cases[i+1].first;
            else
                return def ? &def : NULL_expression_pointer;
        }
    return NULL_expression_pointer;
}

/********************************************
 * ExprCase
 */
void ExprCase::Accept(VAMVisitor *v)
{
    v->VisitExprCase(this);
}


ExprCase::ExprCase()
{
    op = OP_CASE;
    ctrl = nullptr;
}

ExprCase::~ExprCase()
{
    remove(ctrl);
}

int ExprCase::reset_bitwidth()
{
    if (ctrl && cases.size() > 0)
        if (ctrl->get_bitwidth() > 0)
        {
            int case1bw = cases[0].second->get_bitwidth();
            if (case1bw <= 0)
                BITWIDTH_ERR("invalid ExprCase case #1")
            for (auto c : cases)
            {
                if (!(c.first->get_bitwidth() > 0 &&
                    c.first->bitwidth == ctrl->bitwidth &&
                    c.second->get_bitwidth() == case1bw))
                    BITWIDTH_ERR("invalid ExprCase case")
            }
            if (def)
                if (def->get_bitwidth() > 0)
                    if (def->bitwidth == case1bw)
                        return bitwidth = case1bw;
                    else
                        BITWIDTH_ERR("invalid default in ExprCase")
                else
                    BITWIDTH_ERR("invalid default in ExprCase")
            else
                return bitwidth = case1bw;
        }
    BITWIDTH_ERR("invalid ExprCase operands")
}

/// Add controlling expression to operands.
void ExprCase::add_ctrl_expr(Expression *ctrlexpr)
{
    if (ctrl)
        THROW("controlling expression already defined in ExprCase");
    ctrl = ctrlexpr->bind(this);
}

void ExprCase::replace_operand(Expression *o, Expression *r)
{
    ControlExpression::replace_operand(o, r);
    replace_op(ctrl, o, r);
}

Expression **ExprCase::next_operand(Expression **e)
{
    unsigned i;
    for (i = 0; i < cases.size(); i++)
        if (e == &cases[i].first)
            return &cases[i].second;
        else if (e == &cases[i].second)
        {
            if (i+1 < cases.size())
                return &cases[i+1].first;
            else
                break;
        }
    if (e == &def)
        return &ctrl;
    if (e == &ctrl)
        return NULL_expression_pointer;
    if (def)
        return &def;
    else
        return &ctrl;
}

/********************************************
 * ExprCond
 */
void ExprCond::Accept(VAMVisitor *v)
{
    v->VisitExprCond(this);
}


ExprCond::ExprCond()
{
    op = OP_COND;
}

int ExprCond::reset_bitwidth()
{
    if (cases.size() > 0)
    {
        int case1bw = cases[0].second->get_bitwidth();
        if (case1bw <= 0)
            BITWIDTH_ERR("invalid ExprCond case #1")
        for (auto c : cases)
        {
            if (!(c.first->get_bitwidth() == 1 &&
                c.second->get_bitwidth() == case1bw))
                BITWIDTH_ERR("invalid ExprCond case")
        }
        if (def)
            if (def->get_bitwidth() > 0)
                if (def->bitwidth == case1bw)
                    return bitwidth = case1bw;
                else
                    BITWIDTH_ERR("invalid default in ExprCond")
            else
                BITWIDTH_ERR("invalid default in ExprCond")
        else
            return bitwidth = case1bw;
    }
    BITWIDTH_ERR("invalid ExprCond operands")
}

/********************************************
 * ExprSlice
 */
void ExprSlice::Accept(VAMVisitor *v)
{
    v->VisitExprSlice(this);
}


ExprSlice::ExprSlice()
{
    op = OP_SLICE;
    expr = nullptr;
    from = to = -1;
}

ExprSlice::ExprSlice(Expression *o, int f, int t)
{
    op = OP_SLICE;
    expr = o->bind(this);
    from = f;
    to = t;
    get_bitwidth();
}

int ExprSlice::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0 &&
            0 <= from && from < to && to < expr->bitwidth)
            return bitwidth = to-from+1;
    BITWIDTH_ERR("invalid operand in ExprSlice")
}

/********************************************
 * ExprZeroExt
 */
void ExprZeroExt::Accept(VAMVisitor *v)
{
    v->VisitExprZeroExt(this);
}


ExprZeroExt::ExprZeroExt()
{
    op = OP_ZEX;
    expr = nullptr;
    req_bitwidth = -1;
}

ExprZeroExt::ExprZeroExt(Expression *o, int bw)
{
    op = OP_ZEX;
    expr = o->bind(this);
    req_bitwidth = bw;
    get_bitwidth();
}

int ExprZeroExt::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0 &&
            req_bitwidth >= expr->bitwidth)
            return bitwidth = req_bitwidth;
    BITWIDTH_ERR("invalid operand in ExprZeroExt")
}

/********************************************
 * ExprSignExt
 */
void ExprSignExt::Accept(VAMVisitor *v)
{
    v->VisitExprSignExt(this);
}


ExprSignExt::ExprSignExt()
{
    op = OP_SEX;
    expr = nullptr;
    req_bitwidth = -1;
}

ExprSignExt::ExprSignExt(Expression *o, int bw)
{
    op = OP_SEX;
    expr = o->bind(this);
    req_bitwidth = bw;
    get_bitwidth();
}

int ExprSignExt::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0 &&
            req_bitwidth >= expr->bitwidth)
            return bitwidth = req_bitwidth;
    BITWIDTH_ERR("invalid operand in ExprSignExt")
}

/********************************************
 * ExprBitExtract
 */
void ExprBitExtract::Accept(VAMVisitor *v)
{
    v->VisitExprBitExtract(this);
}


ExprBitExtract::ExprBitExtract()
{
    op = OP_BITEX;
    expr = nullptr;
    index = -1;
}

ExprBitExtract::ExprBitExtract(Expression *o, int idx)
{
    op = OP_BITEX;
    expr = o->bind(this);
    index = idx;
    get_bitwidth();
}

int ExprBitExtract::reset_bitwidth()
{
    if (expr)
        if (expr->get_bitwidth() > 0 &&
            index >= 0 &&
            index < expr->bitwidth)
            return bitwidth = 1;
    BITWIDTH_ERR("invalid operand in ExprBitExtract")
}

/********************************************
 * ExprIndex
 */
void ExprIndex::Accept(VAMVisitor *v)
{
    v->VisitExprIndex(this);
}


int ExprIndex::reset_bitwidth()
{
    if (lhs->get_bitwidth() > 0 &&
        (rhs->get_bitwidth() > 0 || rhs->op == OP_CONST))
        return bitwidth = 1;
    BITWIDTH_ERR("invalid operands in ExprIndex");
}

/********************************************
 * ExprConcatenate
 */
void ExprConcatenate::Accept(VAMVisitor *v)
{
    v->VisitExprConcatenate(this);
}


int ExprConcatenate::reset_bitwidth()
{
    if (lhs->get_bitwidth() > 0 &&
        rhs->get_bitwidth() > 0)
        return bitwidth = lhs->bitwidth + rhs->bitwidth;
    BITWIDTH_ERR("invalid operands in ExprConcatenate");
}

/********************************************
 * ExprAbstract
 */
void ExprAbstract::Accept(VAMVisitor *v)
{
    v->VisitExprAbstract(this);
}


void ExprAbstract::join_operands(ExpressionList &ops)
{
    for (auto o : ops)
        operands.push_back(o->bind(this));
    get_bitwidth();
}

/********************************************
 * individual ExprAbstract derived classes
 */
void ExprAbsLogic::Accept(VAMVisitor *v)
{
    v->VisitExprAbsLogic(this);
}

void ExprAbsBitwise::Accept(VAMVisitor *v)
{
    v->VisitExprAbsBitwise(this);
}

void ExprAbsRelational::Accept(VAMVisitor *v)
{
    v->VisitExprAbsRelational(this);
}

void ExprAbsVector::Accept(VAMVisitor *v)
{
    v->VisitExprAbsVector(this);
}

void ExprAbsArithmetic::Accept(VAMVisitor *v)
{
    v->VisitExprAbsArithmetic(this);
}


ExprAbsLogic::ExprAbsLogic()
{
    op = OP_ABS_LOGIC;
}

ExprAbsBitwise::ExprAbsBitwise()
{
    op = OP_ABS_BITWISE;
}

ExprAbsRelational::ExprAbsRelational()
{
    op = OP_ABS_REL;
}

ExprAbsVector::ExprAbsVector()
{
    op = OP_ABS_VECTOR;
}

ExprAbsArithmetic::ExprAbsArithmetic()
{
    op = OP_ABS_ARITH;
}

/********************************************
 * Binary expression classes
 */

namespace {

/// Auxiliary function for reset_bitwidth() of symetric rhs,lhs operations
inline int get_samewidth_operands(Expression *o1, Expression *o2)
{
    if (o1 && o2)
    {
        if (o1->get_bitwidth() > 0 &&
            o2->get_bitwidth() == o1->bitwidth)
            return o1->bitwidth;
        BITWIDTH_ERR("not same for both operands, left " << o1->bitwidth
                << " != right " << o2->bitwidth);
    }
    else
        BITWIDTH_ERR("undefined operand");
    return -1;
}

/// Auxiliary function for reset_bitwidth() of shift operations
inline int shift_reset_bitwidth(Expression *lhs, Expression *rhs)
{
    if (lhs && rhs)
    {
        if (lhs->get_bitwidth() > 0 &&
            rhs->get_bitwidth() > 0)
            return lhs->bitwidth;
        BITWIDTH_ERR("invalid");
    }
    else
        BITWIDTH_ERR("undefined operand");
    return -1;
}

} // anonymous namespace

/**
 * BinaryExpression
 */
BinaryExpression::BinaryExpression(Expression *_lhs, Expression *_rhs)
{
    lhs = _lhs->bind(this);
    rhs = _rhs->bind(this);
}

BinaryExpression::~BinaryExpression()
{
    remove(lhs);
    remove(rhs);
}

void BinaryExpression::add_lhs(Expression *e)
{
    ASSERT(lhs == nullptr, "lhs value already set");
    lhs = e->bind(this);
}

void BinaryExpression::add_rhs(Expression *e)
{
    ASSERT(rhs == nullptr, "rhs value already set");
    rhs = e->bind(this);
}

void BinaryExpression::replace_operand(Expression *o, Expression *r)
{
    replace_op(lhs, o, r);
    replace_op(rhs, o, r);
}

Expression **BinaryExpression::first_operand()
{
    return &lhs;
}

Expression **BinaryExpression::next_operand(Expression **e)
{
    if (e == &lhs)
        return &rhs;
    return NULL_expression_pointer;
}


#define DEF_BINARY_EXPR_SHIFT(clname, visitmethod) \
    void clname::Accept(VAMVisitor *v)\
    { \
        v->visitmethod(this); \
    } \
    int clname::reset_bitwidth() \
    { \
        bitwidth = shift_reset_bitwidth(lhs, rhs); \
        ASSERT(bitwidth > 0, "invalid operands in " #clname); \
        return bitwidth; \
    }

DEF_BINARY_EXPR_SHIFT(ExprAssign, VisitExprAssign)
DEF_BINARY_EXPR_SHIFT(ExprShiftLeft, VisitExprShiftLeft)
DEF_BINARY_EXPR_SHIFT(ExprShiftRight, VisitExprShiftRight)
DEF_BINARY_EXPR_SHIFT(ExprShiftArithRight, VisitExprShiftArithRight)

#undef DEF_BINARY_EXPR_SHIFT

#define DEF_BINARY_EXPR_ARITH(clname, visitmethod) \
    void clname::Accept(VAMVisitor *v)\
    { \
        v->visitmethod(this); \
    } \
    int clname::reset_bitwidth() \
    { \
        bitwidth = get_samewidth_operands(lhs, rhs); \
        ASSERT(bitwidth > 0, "invalid operands in " #clname); \
        return bitwidth; \
    }

DEF_BINARY_EXPR_ARITH(ExprBitwiseOr, VisitExprBitwiseOr)
DEF_BINARY_EXPR_ARITH(ExprBitwiseAnd, VisitExprBitwiseAnd)
DEF_BINARY_EXPR_ARITH(ExprBitwiseXor, VisitExprBitwiseXor)
DEF_BINARY_EXPR_ARITH(ExprPlus, VisitExprPlus)
DEF_BINARY_EXPR_ARITH(ExprMinus, VisitExprMinus)
DEF_BINARY_EXPR_ARITH(ExprMul, VisitExprMul)
DEF_BINARY_EXPR_ARITH(ExprDiv, VisitExprDiv)
DEF_BINARY_EXPR_ARITH(ExprMod, VisitExprMod)

#undef DEF_BINARY_EXPR_ARITH

#define DEF_BINARY_EXPR_REL(clname, visitmethod) \
    void clname::Accept(VAMVisitor *v)\
    { \
        v->visitmethod(this); \
    } \
    int clname::reset_bitwidth() \
    { \
        bitwidth = get_samewidth_operands(lhs, rhs); \
        ASSERT(bitwidth > 0, "invalid operands in " #clname); \
        return bitwidth = 1; \
    }

DEF_BINARY_EXPR_REL(ExprEqual, VisitExprEqual)
DEF_BINARY_EXPR_REL(ExprNotEqual, VisitExprNotEqual)
DEF_BINARY_EXPR_REL(ExprLessThan, VisitExprLessThan)
DEF_BINARY_EXPR_REL(ExprGreaterThan, VisitExprGreaterThan)
DEF_BINARY_EXPR_REL(ExprLessEqual, VisitExprLessEqual)
DEF_BINARY_EXPR_REL(ExprGreaterEqual, VisitExprGreaterEqual)
DEF_BINARY_EXPR_REL(ExprSignedLessThan, VisitExprSignedLessThan)
DEF_BINARY_EXPR_REL(ExprSignedGreaterThan, VisitExprSignedGreaterThan)
DEF_BINARY_EXPR_REL(ExprSignedLessEqual, VisitExprSignedLessEqual)
DEF_BINARY_EXPR_REL(ExprSignedGreaterEqual, VisitExprSignedGreaterEqual)

#undef DEF_BINARY_EXPR_REL

} // namespace vam
