/**
 * @desc Declaration of elements of VAM (signals, functional nodes, storages).
 */
#ifndef _VAMELEMENTS_H
#define _VAMELEMENTS_H

#include "vam_err.h"
#include <array>
#include <vector>
#include <string>
#include <utility> // pair

using namespace std;

namespace vam {

class VAMVisitor;

/**
 * @brief   Superclass of all vam elements. VAM elements should be visitable,
 * e.g., when exporting model or interpretting it. In such a case, one should
 * implement a visitor (cf. class VAMVisitor) and accept a model by it.
 */
class VAMElement
{
public:
    virtual ~VAMElement() noexcept = default;
    virtual void Accept(VAMVisitor *v);
};

enum MetaType {
    MT_UNDEFINED,
    MT_NODE,
    MT_VALUE,
    MT_INTVALUE,
};

class Meta : public VAMElement
{
public:
    /// meta node/value type
    MetaType type = MT_UNDEFINED;
    /// node name or value
    string value;
private:
    /// list of meta nodes or values
    vector<Meta*> metas;

public:
    Meta() = default;
    Meta(MetaType t, const string &v) : type(t), value(v) {}
    Meta(const Meta &m) : type(MT_UNDEFINED) { copy(m); }
    Meta(Meta &&) = default;


    ~Meta();

    void Accept(VAMVisitor *v);

    /// add a new node
    Meta *add(MetaType t, const string &v);
    /// add meta node
    void add(Meta *m);
    /// add meta value (only if this node is <=MT_NODE)
    void add(const string &v);
    /// add meta value (only if this node is <=MT_NODE)
    void add(long int i);
    /// add key-value nodes
    void addkv(const string &key, const string &val);

    /// erase meta node at a given position
    void erase(unsigned pos);
    /// erase 
    void erase(vector<Meta*>::iterator i);

    /// return true if no meta is set
    bool empty()
    {
        return type == MT_UNDEFINED;
    }

    void tag(const string &t)
    {
        Meta *m = find(t);
        if (!m)
            add(t);
    }

    void untag(const string &t)
    {
        find_erase(t);
    }

    /// Find a node with name `node'. Return nullptr if not found.
    Meta *find(const string &node, Meta *last = nullptr);

    /// Find a node. Create a new one if not found.
    Meta &node(const string &nodename);

    /**
     * @brief Find meta node node-key containing node-value and return a
     * reference to the value. If node-key does not exist, create one with
     * default values 0 or "" (depending on type of node value).
     */
    string &keyval(const string &key);

    /**
     * @brief Similar to keyvar(), but do not create an empty meta if key not
     * found.
     */
    string keyval_nocreat(const string &key);

    /**
     * @brief Find a node with name `node' and erase it.
     * @param   node    meta node name to erase
     * @param   last    the last occurence of meta node with the same name
     *                  which must be preserved.
     * @return last, or nullptr if not found
     */
    Meta *find_erase(const string &node, Meta *last = nullptr);

    /// move all metadata from this object to target
    void move_to(Meta *target);

    string &operator[](unsigned pos);
    string &operator[](const string &key);
    Meta &operator<<(Meta *m);
    Meta &operator<<(const string &v);
    Meta &operator<<(long int i);
    Meta &operator=(const Meta &m) { copy(m); return *this; }
    Meta &operator=(Meta&&) = default;

    /// clear all metadata, set this to MT_UNDEFINED
    void clear();
    /// copy metadata from m
    void copy(const Meta &m);

    void AcceptAll(VAMVisitor *visitor);
};

class Expression;

using ExpressionList = vector<Expression*>;

class OperandIterator;

/// Variable access modes.
enum VarAccess {
    VA_READ,
    VA_WRITE,
    VA_RDWR,
    VA_INVALID,
};

/**
 * @brief   Variable type is used for specific purpose libraries and analysis
 * fields. E.g. RTL layer uses signals which are specialization of general
 * variables; software analysis uses local/global variables. Each analysis
 * field should specify its own variable type. New variable types should be
 * greater than VT_GENERAL.
 */
enum VariableType
{
    VT_GENERAL = 0,
};

/**
 * @brief   Variable represents a symbolic name for a simple bit vector. When
 * connecting it to the expression, it should be used via ExprRef.
 */
class Variable : public VAMElement
{
public:
    /// variable type (useful for specific purpose libraries)
    int var_type;

    /// identificator
    string id;

    /// bitwidth; -1 means not-set
    int bitwidth;

    Meta meta;

    /**
     * @brief Back references (which expressions references here). Should not 
     * contain nullptr.
     */
    ExpressionList references;

public:

    Variable(string i = "", int bw = -1)
    {
        var_type = VT_GENERAL;
        id = i;
        bitwidth = bw;
    }

    Variable(const Variable &v) : meta(v.meta)
    {
        var_type = VT_GENERAL;
        id = v.id;
        bitwidth = v.bitwidth;
    }

    ~Variable();

    void Accept(VAMVisitor *v);

    /**
     * @brief   Find and remove back reference from references.
     * @param   expr    Expression referencing to the variable to be removed.
     */
    void unlink(Expression *expr);
};

/** 
 * @brief   Operation code for an expression. Different types of expressions
 * have different classes. Opcode bypasses dynamic casting to explicitly get
 * the type of expression.
 */
enum Operator {
    OP_NOP      = 0, // no-operation represents some constant
    OP_CONST    = 1, // constant
    OP_VAR      = 2, // reference to a variable
    OP_REF      = 3, // reference to another expression
    OP_ASSIGN   = 4, // assignment
    // logic
    OP_NOT      = 5,
    OP_OR       = 6,
    OP_AND      = 7,
    OP_ITE      = 8,
    // bitwise
    OP_INV      = 9,
    OP_BOR      = 10, // bitwise or
    OP_BAND     = 11, // bitwise and
    OP_XOR      = 12,
    OP_SHL      = 13,
    OP_SHR      = 14,
    OP_SAR      = 15,
    OP_ORJOIN   = 16, // or-join over multiple operands
    // relational
    OP_EQ       = 17,
    OP_NEQ      = 18,
    OP_LT       = 19, // unsigned less-than
    OP_GT       = 20, // unsigned greater-than
    OP_LE       = 21, // unsigned less-or-equal
    OP_GE       = 22, // unsigned less-or-equal
    OP_SLT      = 23, // signed less-than
    OP_SGT      = 24, // signed greater-than
    OP_SLE      = 25, // signed less-or-equal
    OP_SGE      = 26, // signed greater-or-equal
    // casting
    OP_UTS      = 27, // unsigned-to-signed
    OP_STU      = 28, // signed-to-unsigned
    // conditional
    OP_CASE     = 29,
    OP_COND     = 30,
    // arithmetic
    OP_NEG      = 31,
    OP_PLUS     = 32,
    OP_MINUS    = 33,
    OP_MUL      = 34,
    OP_DIV      = 35,
    OP_MOD      = 36,
    // vector
    OP_IDX      = 37, // indexing
    OP_BITEX    = 38, // bit extraction
    OP_CONCAT   = 39, // concatenation
    OP_ZEX      = 40, // zero extension
    OP_SEX      = 41, // sign extension
    OP_SLICE    = 42, // vector slice
    // abstract
    OP_ABS_LOGIC    = 43, // abstract logic
    OP_ABS_BITWISE  = 44, // abstract bitwise
    OP_ABS_REL      = 45, // abstract relational
    OP_ABS_VECTOR   = 46, // abstract vector or casting
    OP_ABS_ARITH    = 47, // abstract arithmetic
    // VAM operator stop
    // Operators with value greater than OP_VAM_STOP represents other than
    // VAM-known operators.
    OP_VAM_STOP     = 48,
};

/// table transforming operation code to string
extern const string tbl_oper_to_str[];

/// table transforming operation code to operation group name
extern const string tbl_oper_to_group[];

class Expression : public VAMElement
{
public:
    /// operator (operation code)
    Operator op;

    /// bit width; -1 means not-set
    int bitwidth;

    /// Non nullptr outputs (which expressions uses value of this expression).
    ExpressionList outputs;
    using outputs_iterator = ExpressionList::iterator;

    /// Metadata
    Meta meta;

    static string default_tag;

public:

    Expression()
    {
        op = OP_NOP;
        bitwidth = -1;
        if (!this->default_tag.empty())
            meta.tag(default_tag);
    }

    ~Expression();

    int get_bitwidth()
    {
        return (bitwidth < 0) ? reset_bitwidth() : bitwidth;
    }

    void Accept(VAMVisitor *v);

    virtual int reset_bitwidth();

    /**
     * @brief   Bind a new expression output.
     * @param   expr    Binded expression (reads value of this expression)
     * @return  `this' pointer
     */
    Expression *bind(Expression *expr);

    /**
     * @brief   Find and remove output reference.
     * @param   expr    Expression to be removed from list of references.
     */
    void unlink(const Expression *expr);

    /**
     * @brief   Unlink the operand from this expression. If the operand has no
     * other reference, delete it from the memory.
     * @param   operand Operand to unlink from.
     */
    void remove(Expression *operand);

    /**
     * @brief   Replaces every occurence of a given expression among operands
     *          by a new expression.
     * @param   orig        original expression to be replaced (non nullptr)
     * @param   replacement expression replacement
     */
    virtual void replace_operand(Expression *orig, Expression *replacement);

    /**
     * @brief   Returns the first operand of the expression. Used for
     * iterators over expression operands.
     * @param   it      operand iterator (FIXME: should set it.aux_int)
     * @return  first operand.
     */
    virtual Expression **first_operand();

    /**
     * @brief   Returns pointer of operand of the expression next to the
     *          operand given by the pointer.
     * @param   _op     pointer to operand next to which find the next operand
     * @param   it      operand iterator (FIXME: should use this instead of
     *                  _op, and read&write it.aux_int, because... what if two
     *                  or more operands point to the same expression?)
     * @return  next operand (nullptr if no more operand)
     */
    virtual Expression **next_operand(Expression **_op);

    using iterator = OperandIterator;

    iterator begin();
    iterator &end();

    /**
     * @brief   Returns string of operator.
     */
    string operator_to_str()
    {
        if (op < OP_VAM_STOP)
            return tbl_oper_to_str[op];
        else
            return "VAM_unknown_operator";
    }

    /**
     * \{
     * @brief   Returns true if expression (wrt. operator) is a given type.
     */
    virtual bool op_is_logic();
    virtual bool op_is_bitwise();
    virtual bool op_is_relational();
    virtual bool op_is_casting();
    virtual bool op_is_conditional();
    virtual bool op_is_arithmetic();
    virtual bool op_is_vector();
    virtual bool op_is_abstract();

protected:
    /// Intended for use by inherited classes within replace_operand().
    void replace_op(Expression *&operand, Expression *orig,
            Expression *replacement)
    {
        if (operand == orig)
        {
            remove(operand);
            operand = replacement->bind(this);
        }
    }

};

/**
 * @desc Replace outputs of the expression 'e' by the replacement, which
 * positively leads to delete the expression 'e' by unlinking its outputs.
 * @param   e           Expression which outputs should be replaced
 * @param   replacement Expression replacement
 * @param   except_for  Do not replace output of e, if it is equal to
 *                      except_for
 * @return  number of performed replacements
 */
int replace_outputs(Expression *e, Expression *replacement,
       Expression *except_for = nullptr);

extern Expression **NULL_expression_pointer;

class OperandIterator
{
public:
    Expression *refexpr;
    Expression **pos;
    int aux_int = -1;
public:
    OperandIterator(Expression *ref, Expression **p) :
        refexpr(ref), pos(p) {}

    bool operator==(const OperandIterator &i) const
    {
        return (i.pos == pos) &&
            ((i.refexpr == refexpr) ||
             (i.refexpr == nullptr) ||
             (refexpr == nullptr));
    }

    bool operator!=(const OperandIterator &i) const
    {
        return !(*this == i);
    }

    OperandIterator &operator=(const OperandIterator &i)
    {
        refexpr = i.refexpr;
        pos = i.pos;
        return *this;
    }

    OperandIterator &operator++()
    {
        if (refexpr)
            pos = refexpr->next_operand(pos);
        return *this;
    }

    Expression *&operator*()
    {
        return *pos;
    }

    Expression *&operator->()
    {
        return *pos;
    }
};

extern OperandIterator end_operand_iterator;

// inline here because OperandIterator is unknown during declaration
// of class Expression.
inline OperandIterator Expression::begin()
{
    return iterator(this, first_operand());
}

inline OperandIterator &Expression::end()
{
    return end_operand_iterator;
}

class ExprConst : public Expression
{
public:
    string value;

    ExprConst();
    ExprConst(string v, int bw);

    void Accept(VAMVisitor *v);
};

class ExprVariable : public Expression
{
public:
    Variable *v;

    ExprVariable();
    ExprVariable(Variable *var);
    ~ExprVariable();
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExpressionPool : public Expression
{
public:
    ExpressionList operands;
public:
    ExpressionPool() { op = OP_NOP; }
    ~ExpressionPool();
    
    void Accept(VAMVisitor *v);
    /// append operand to the pool
    void add_operand(Expression *o);
    /// find and remove operand from the pool
    void erase_operand(Expression *o);
    /// replace (and rebind) operand o in the pool by r
    void replace_operand(Expression *o, Expression *r);

    Expression **first_operand();
    Expression **next_operand(Expression **e);
};

class ExprOr : public ExpressionPool
{
public:
    ExprOr();
    ExprOr(Expression *o1, Expression *o2);
    ~ExprOr() noexcept = default;
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprAnd : public ExpressionPool
{
public:
    ExprAnd();
    ExprAnd(Expression *o1, Expression *o2);
    ~ExprAnd() noexcept = default;
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprOrJoin : public ExpressionPool
{
public:
    ExprOrJoin();
    ExprOrJoin(Expression *o1, Expression *o2);
    ~ExprOrJoin() noexcept = default;
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprITE : public Expression
{
public:
    Expression *cond;
    Expression *e_true;
    Expression *e_false;

    ExprITE();
    ExprITE(Expression *c, Expression *o1, Expression *o2);
    ~ExprITE();
    void Accept(VAMVisitor *v);
    void add_cond(Expression *e);
    void add_true(Expression *e);
    void add_false(Expression *e);
    int reset_bitwidth();
    void replace_operand(Expression *o, Expression *r);
    Expression **first_operand();
    Expression **next_operand(Expression **e);
};

/// Auxiliary super-class for control structure 
class ControlExpression : public Expression
{
public:
    vector<pair<Expression*, Expression*> > cases;
    /// default-case value; nullptr represents undefined default-case
    Expression *def;
protected:
    ControlExpression() { def = nullptr; }
    ~ControlExpression();
public:
    /// Add case expressions to operands.
    virtual void add_case(Expression *cond, Expression *value);
    /// Add default case to operands.
    virtual void add_default(Expression *def);
    
    void replace_operand(Expression *o, Expression *r);
    Expression **first_operand();
    Expression **next_operand(Expression **e);
};

/** 
 * @brief ExprCase represents control structure with multiple choices
 * controlled by a specific value of controlling expression.
 */
class ExprCase : public ControlExpression
{
public:
    Expression *ctrl;

    ExprCase();
    ~ExprCase();
    void Accept(VAMVisitor *v);
    int reset_bitwidth();

    /// Add controlling expression to operands.
    void add_ctrl_expr(Expression *ctrlexpr);
    
    void replace_operand(Expression *o, Expression *r);
    
    Expression **next_operand(Expression **e);
};

/** 
 * @brief ExprCond represents control structure with multiple choices
 * controlled by general expressions.
 */
class ExprCond : public ControlExpression
{
public:
    ExprCond();
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprAbstract : public ExpressionPool
{
public:
    void Accept(VAMVisitor *v);
    void join_operands(ExpressionList &ops);
};

class ExprAbsLogic : public ExprAbstract
{
public:
    void Accept(VAMVisitor *v);
    ExprAbsLogic();
};

class ExprAbsBitwise : public ExprAbstract
{
public:
    void Accept(VAMVisitor *v);
    ExprAbsBitwise();
};

class ExprAbsRelational : public ExprAbstract
{
public:
    void Accept(VAMVisitor *v);
    ExprAbsRelational();
};

class ExprAbsVector : public ExprAbstract
{
public:
    void Accept(VAMVisitor *v);
    ExprAbsVector();
};

class ExprAbsArithmetic : public ExprAbstract
{
public:
    void Accept(VAMVisitor *v);
    ExprAbsArithmetic();
};

class UnaryExpression : public Expression
{
public:
    Expression *expr;
protected:
    UnaryExpression() { expr = nullptr; }
    UnaryExpression(Expression *e) { expr = e->bind(this); }
    ~UnaryExpression();
public:
    void add_operand(Expression *e);
    void replace_operand(Expression *o, Expression *r);
    Expression **first_operand();
};

#define UNARY_EXPR_CLASS(clname, opcode) \
    class clname : public UnaryExpression \
    { \
    public: \
        clname() { op = opcode; } \
        clname(Expression *e) : UnaryExpression(e) { op = opcode; } \
        void Accept(VAMVisitor *v); \
        int reset_bitwidth(); \
    };

class ExprSlice : public UnaryExpression
{
public:
    int from;
    int to;

    ExprSlice();
    ExprSlice(Expression *o, int f, int t);
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprZeroExt : public UnaryExpression
{
public:
    int req_bitwidth;

    ExprZeroExt();
    /// create ZEX of required bit width
    ExprZeroExt(Expression *o, int bw);
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprSignExt : public UnaryExpression
{
public:
    int req_bitwidth;

    ExprSignExt();
    /// create SEX of required bit width
    ExprSignExt(Expression *o, int bw);
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

class ExprBitExtract : public UnaryExpression
{
public:
    int index;

    ExprBitExtract();
    /// create bit extraction of required index
    ExprBitExtract(Expression *o, int idx);
    void Accept(VAMVisitor *v);
    int reset_bitwidth();
};

UNARY_EXPR_CLASS(ExprRef, OP_REF)
UNARY_EXPR_CLASS(ExprNot, OP_NOT)
UNARY_EXPR_CLASS(ExprInv, OP_INV)
UNARY_EXPR_CLASS(ExprNeg, OP_NEG)
UNARY_EXPR_CLASS(ExprUnsignedToSigned, OP_UTS)
UNARY_EXPR_CLASS(ExprSignedToUnsigned, OP_STU)

#undef UNARY_EXPR_CLASS

class BinaryExpression : public Expression
{
public:
    Expression *lhs;
    Expression *rhs;
protected:
    BinaryExpression() { lhs = rhs = nullptr; }
    BinaryExpression(Expression *_lhs, Expression *_rhs);
    ~BinaryExpression();
public:
    void add_lhs(Expression *e);
    void add_rhs(Expression *e);
    void replace_operand(Expression *o, Expression *r);
    Expression **first_operand();
    Expression **next_operand(Expression **e);
};

#define BINARY_EXPR_CLASS(clname, opcode) \
    class clname : public BinaryExpression \
    { \
    public: \
        clname() { op = opcode; } \
        clname(Expression *_lhs, Expression *_rhs) : \
            BinaryExpression(_lhs, _rhs) { op = opcode; reset_bitwidth() ; } \
        void Accept(VAMVisitor *v); \
        int reset_bitwidth(); \
    }; \

BINARY_EXPR_CLASS(ExprAssign, OP_ASSIGN)
BINARY_EXPR_CLASS(ExprShiftLeft, OP_SHL)
BINARY_EXPR_CLASS(ExprShiftRight, OP_SHR)
BINARY_EXPR_CLASS(ExprShiftArithRight, OP_SAR)
BINARY_EXPR_CLASS(ExprBitwiseOr, OP_BOR)
BINARY_EXPR_CLASS(ExprBitwiseAnd, OP_BAND)
BINARY_EXPR_CLASS(ExprBitwiseXor, OP_XOR)
BINARY_EXPR_CLASS(ExprEqual, OP_EQ)
BINARY_EXPR_CLASS(ExprNotEqual, OP_NEQ)
BINARY_EXPR_CLASS(ExprLessThan, OP_LT)
BINARY_EXPR_CLASS(ExprGreaterThan, OP_GT)
BINARY_EXPR_CLASS(ExprLessEqual, OP_LE)
BINARY_EXPR_CLASS(ExprGreaterEqual, OP_GE)
BINARY_EXPR_CLASS(ExprSignedLessThan, OP_SLT)
BINARY_EXPR_CLASS(ExprSignedGreaterThan, OP_SGT)
BINARY_EXPR_CLASS(ExprSignedLessEqual, OP_SLE)
BINARY_EXPR_CLASS(ExprSignedGreaterEqual, OP_SGE)
BINARY_EXPR_CLASS(ExprPlus, OP_PLUS)
BINARY_EXPR_CLASS(ExprMinus, OP_MINUS)
BINARY_EXPR_CLASS(ExprMul, OP_MUL)
BINARY_EXPR_CLASS(ExprDiv, OP_DIV)
BINARY_EXPR_CLASS(ExprMod, OP_MOD)
BINARY_EXPR_CLASS(ExprIndex, OP_IDX)
BINARY_EXPR_CLASS(ExprConcatenate, OP_CONCAT)

#undef BINARY_EXPR_CLASS

//#define NA { throw VAMException("method not implemented"); }
#define NA { }

class VAMVisitor
{
public:
    virtual void VisitVAMElement(VAMElement *) NA
    virtual void VisitExpression(Expression *) NA

    virtual void VisitMeta(Meta *) NA
    virtual void VisitVariable(Variable *) NA
    virtual void VisitExprConst(ExprConst *) NA
    virtual void VisitExprVariable(ExprVariable *) NA
    virtual void VisitExprRef(ExprRef *) NA
    virtual void VisitExprAssign(ExprAssign *) NA
    virtual void VisitExprNot(ExprNot *) NA
    virtual void VisitExprOr(ExprOr *) NA
    virtual void VisitExprAnd(ExprAnd *) NA
    virtual void VisitExprOrJoin(ExprOrJoin *) NA
    virtual void VisitExprITE(ExprITE *) NA
    virtual void VisitExprInv(ExprInv *) NA
    virtual void VisitExprBitwiseOr(ExprBitwiseOr *) NA
    virtual void VisitExprBitwiseAnd(ExprBitwiseAnd *) NA
    virtual void VisitExprBitwiseXor(ExprBitwiseXor *) NA
    virtual void VisitExprShiftLeft(ExprShiftLeft *) NA
    virtual void VisitExprShiftRight(ExprShiftRight *) NA
    virtual void VisitExprShiftArithRight(ExprShiftArithRight *) NA
    virtual void VisitExprEqual(ExprEqual *) NA
    virtual void VisitExprNotEqual(ExprNotEqual *) NA
    virtual void VisitExprLessThan(ExprLessThan *) NA
    virtual void VisitExprGreaterThan(ExprGreaterThan *) NA
    virtual void VisitExprLessEqual(ExprLessEqual *) NA
    virtual void VisitExprGreaterEqual(ExprGreaterEqual *) NA
    virtual void VisitExprSignedLessThan(ExprSignedLessThan *) NA
    virtual void VisitExprSignedGreaterThan(ExprSignedGreaterThan *) NA
    virtual void VisitExprSignedLessEqual(ExprSignedLessEqual *) NA
    virtual void VisitExprSignedGreaterEqual(ExprSignedGreaterEqual *) NA
    virtual void VisitExprUnsignedToSigned(ExprUnsignedToSigned *) NA
    virtual void VisitExprSignedToUnsigned(ExprSignedToUnsigned *) NA
    virtual void VisitExprCase(ExprCase *) NA
    virtual void VisitExprCond(ExprCond *) NA
    virtual void VisitExprNeg(ExprNeg *) NA
    virtual void VisitExprPlus(ExprPlus *) NA
    virtual void VisitExprMinus(ExprMinus *) NA
    virtual void VisitExprMul(ExprMul *) NA
    virtual void VisitExprDiv(ExprDiv *) NA
    virtual void VisitExprMod(ExprMod *) NA
    virtual void VisitExprIndex(ExprIndex *) NA
    virtual void VisitExprConcatenate(ExprConcatenate *) NA
    virtual void VisitExprZeroExt(ExprZeroExt *) NA
    virtual void VisitExprSignExt(ExprSignExt *) NA
    virtual void VisitExprBitExtract(ExprBitExtract *) NA
    virtual void VisitExprSlice(ExprSlice *) NA
    virtual void VisitExpressionPool(ExpressionPool *) NA
    virtual void VisitExprAbstract(ExprAbstract *) NA
    virtual void VisitExprAbsLogic(ExprAbsLogic *) NA
    virtual void VisitExprAbsBitwise(ExprAbsBitwise *) NA
    virtual void VisitExprAbsRelational(ExprAbsRelational *) NA
    virtual void VisitExprAbsVector(ExprAbsVector *) NA
    virtual void VisitExprAbsArithmetic(ExprAbsArithmetic *) NA
};

} // namespace vam

#endif
