/**
 * @desc Routines for import/export of VAM.
 */
#include "input.h"
#include <fstream>
#include <ctype.h> // isspace, isdigit
#include "switch-parser-macro.h"

using namespace std;

namespace vam {

/**********************************
 * Lexer
 */
int Lexer::get_char_with_comment()
{
    if (pos < input.size())
    {
        int c = input[pos++];
        if (c == '\n')
        {
            row++;
            col = 1;
        }
        else
            col++;
        return c;
    }
    else
        return -1;
}

int Lexer::get_char()
{
    int s = 0;
        // 0=no_comment
        // 1=;
        // 2=;; or // comment
        // 3=/
        // 4=/* comment
        // 5=/* comment *
        // 6=# comment
    int c;
    while (1)
    {
        c = get_char_with_comment();
        if (c == -1)
            return c;
        if (s == 0)
            if (c == ';')
                s = 1;
            else if (c == '/')
                s = 3;
            else if (c == '#')
                s = 6;
            else
                return c;
        else if (s == 1)
            if (c == ';')
                s = 2;
            else
            {
                unget();
                return ';';
            }
        else if (s == 2)
        {
            if (c == '\n')
                return ' ';
        }
        else if (s == 3)
        {
            if (c == '/')
                s = 2;
            else if (c == '*')
                s = 4;
            else
            {
                unget();
                return '/';
            }
        }
        else if (s == 4)
        {
            if (c == '*')
                s = 5;
        }
        else if (s == 5)
        {
            if (c == '/')
                return ' ';
            else if (c != '*')
                s = 4;
        }
        else // if (s == 6)
        {
            if (c == '\n')
                return ' ';
        }
    }
}

void Lexer::unget()
{
    if (pos)
    {
        pos--;
        if (col == 1)
        {
            row--;
            // don't know previous column, so 0 it is
            col = 0;
        }
        else
            col--;
    }
}

//enum TokenType
//{
//    T_UNDEF,        // undefined token
//    T_LPAR,         // (
//    T_RPAR,         // )
//    T_IDENT,        // <identifier>
//    T_CONST,        // <integer constant>
//    T_STRING,       // "string"
//};
Token Lexer::next_token()
{
    Token t;
    t.filename = filename;
    int s = 0;
    int c;
    while (1)
    {
        c = get_char();
        // init
        if (s == 0)
        {
            if (isspace(c))
                continue;
            else if (c == '"')
            {
                t.row = row;
                t.col = col-1;
                s = 1;
            }
            else if (isdigit(c))
            {
                t.value += c;
                t.row = row;
                t.col = col-1;
                s = 4;
            }
            else if (c == '+' || c == '-')
            {
                t.value += c;
                t.row = row;
                t.col = col-1;
                s = 3;
            }
            else if (c == -1)
                return Token(T_EOF, "", row, col);
            else if (c == '(')
                return Token(T_LPAR, "(", row, col-1);
            else if (c == ')')
                return Token(T_RPAR, ")", row, col-1);
            else
            {
                t.value += c;
                t.row = row;
                t.col = col-1;
                s = 5;
            }
        }
        // T_STRING
        else if (s == 1)
        {
            ASSERT(c != -1, row << ":" << col << ": unexpected eof")
            if (c == '\\')
                s = 2;
            else if (c == '"')
            {
                t.type = T_STRING;
                break;
            }
            else
                t.value += c;
        }
        // T_STRING/escape
        else if (s == 2)
        {
            ASSERT(c != -1, row << ":" << col << ": unexpected eof")
            t.value += c;
            s = 1;
        }
        // T_CONST/+-
        else if (s == 3)
        {
            if (isspace(c))
            {
                t.type = T_IDENT;
                break;
            }
            else if (c == -1)
            {
                t.type = T_IDENT;
                break;
            }
            else if (c == '(' || c == ')')
            {
                unget();
                t.type = T_IDENT;
                break;
            }
            else if (isdigit(c))
            {
                t.value += c;
                s = 4;
            }
            else
            {
                t.value += c;
                s = 5;
            }
        }
        // T_CONST/0-9
        else if (s == 4)
        {
            if (isspace(c))
            {
                t.type = T_CONST;
                break;
            }
            else if (c == -1)
            {
                t.type = T_CONST;
                break;
            }
            else if (c == '(' || c == ')')
            {
                unget();
                t.type = T_CONST;
                break;
            }
            else if (isdigit(c))
                t.value += c;
            else
            {
                t.value += c;
                s = 5;
            }
        }
        // T_IDENT
        else // if (s == 5)
        {
            if (isspace(c))
            {
                t.type = T_IDENT;
                break;
            }
            else if (c == -1)
            {
                t.type = T_IDENT;
                break;
            }
            else if (c == '(' || c == ')')
            {
                unget();
                t.type = T_IDENT;
                break;
            }
            else
                t.value += c;
        }
    }
    return t;
}

bool Lexer::read_from_file(const string &name)
{
    ifstream i(name);
    if (i.is_open())
    {
        input.assign(istreambuf_iterator<char>(i),
                     istreambuf_iterator<char>());
        filename = name;
        return true;
    }
    else
        return false;
}

void Lexer::read_from(istream &i)
{
    input.assign(istreambuf_iterator<char>(i),
                 istreambuf_iterator<char>());
    filename = "";
}

const Token tokens1[] =
{
    {T_VCONST,  "_"},
    {T_NOT,     "!"},
    {T_ITE,     "?"},
    {T_INV,     "~"},
    {T_BOR,     "|"},
    {T_BAND,    "&"},
    {T_BXOR,    "^"},
    {T_LT,      "<"},
    {T_GT,      ">"},
    {T_CONCAT,  "."},
    {T_PLUS,    "+"},
    {T_MINUS,   "-"},
    {T_MUL,     "*"},
    {T_DIV,     "/"},
    {T_MOD,     "%"},
    {T_D,       "d"},
    {T_Q,       "q"},
    {T_EOF, ""},
};

const Token tokens2[] =
{
    {T_EASSIGN, ":="},
    {T_OR,      "||"},
    {T_AND,     "&&"},
    {T_ORJOIN,  "|."},
    {T_SHL,     "<<"},
    {T_SHR,     ">>"},
    {T_EQ,      "=="},
    {T_NEQ,     "!="},
    {T_LE,      "<="},
    {T_GE,      ">="},
    {T_SLT,     "<."},
    {T_SGT,     ">."},
    {T_UTS,     "-."},
    {T_STU,     ".-"},
    {T_IDX,     "[]"},
    {T_ZEX,     "^^"},
    {T_NEG,     "--"},
    {T_ABSLOGIC,    "!!"},
    {T_ABSBITWISE,  "~~"},
    {T_ABSREL,      "<>"},
    {T_ABSVECTOR,   "]["},
    {T_ABSARITH,    "-+"},
    {T_VT,      "vt"},
    {T_WE,      "we"},
    {T_EN,      "en"},
    {T_RO,      "ro"},
    {T_WO,      "wo"},
    {T_RW,      "rw"},
    {T_EOF, ""},
};

const Token tokens3[] =
{
    {T_SAR,     ">>."},
    {T_SLE,     "<=."},
    {T_SGE,     ">=."},
    {T_SEX,     "^^."},
    {T_BITEX,   "[.]"},
    {T_CLR,     "clr"},
    {T_DIR,     "dir"},
    {T_MEM,     "mem"},
    {T_REG,     "reg"},
    {T_RST,     "rst"},
    {T_SIG,     "sig"},
    {T_EOF, ""},
};

const Token tokens4[] =
{
    {T_SLICE,   "[..]"},
    {T_COND,    "cond"},
    {T_META,    "meta"},
    {T_PORT,    "port"},
    {T_ADDR,    "addr"},
    {T_DATA,    "data"},
    {T_EOF, ""},
};

const Token tokens5[] =
{
    {T_MODEL, "model"},
    {T_STAGE, "stage"},
    {T_STALL, "stall"},
    {T_PORTS, "ports"},
    {T_FNODE, "fnode"},
    {T_INPUT, "input"},
    {T_EOF, ""},
};

const Token tokens6[] =
{
    {T_FNODES,  "fnodes"},
    {T_CASE,    "switch"},
    {T_INPUTS,  "inputs"},
    {T_OUTPUT,  "output"},
    {T_ASSIGN,  "assign"},
    {T_TIMING,  "timing"},
    {T_EOF, ""},
};

const Token tokensX[] =
{
    {T_ADDRESS,     "address"},
    {T_CONTROL,     "control"},
    {T_DEFAULT,     "default"},
    {T_UNKNOWN,     "unknown"},
    {T_CONNECTIONS, "connections"},
    {T_PIPESTAGES,  "pipeline-stages"},
    {T_EOF, ""},
};

const Token *tokstbl[] =
    {nullptr, tokens1, tokens2, tokens3, tokens4, tokens5, tokens6};

Token RTLParser::tok2keyword(Token t)
{
    if (t.type == T_IDENT)
    {
        const Token *tokens = tokensX;
        if (0 < t.value.size() && t.value.size() <= 6)
            tokens = tokstbl[t.value.size()];
        for (int i = 0; tokens[i].type != T_EOF; i++)
            if (tokens[i].value == t.value)
            {
                t.type = tokens[i].type;
                return t;
            }
        return t;
    }
    else
        return t;
}


/**********************************
 * RTLParser
 */

void RTLParser::parse(System *s)
{
    system = s;
    p_model();
}

void RTLParser::p_meta(Meta *meta, int was_lpar)
{
    Token t;
    t = lexer.next_token();
    PASSERT(meta != nullptr, t, "unexpected meta information");
    PASSERT(t.type != T_EOF, t, "unexpected end-of-file");
    PASSERT(t.type != T_UNDEF, t, "undefined token");
    while (t.type != T_RPAR)
    {
        switch (t.type)
        {
            case T_IDENT:
                if (was_lpar)
                {
                    meta = meta->add(MT_NODE, t.value);
                    was_lpar = 0;
                }
                else
                    meta->add(t.value);
                break;
            case T_CONST:
                meta->add(stoi(t.value));
                break;
            case T_LPAR:
                p_meta(meta, 1);
                break;
            case T_STRING:
                meta->add(t.value);
                break;
            default:
                PEXPECT(0, "meta information");
        }
        t = lexer.next_token();
        PASSERT(t.type != T_EOF, t, "unexpected end-of-file");
        PASSERT(t.type != T_UNDEF, t, "undefined token");
    }
}

void RTLParser::p_model()
{
    TOKINIT(&system->meta);
    TOKNEW; PEXPECT(T_LPAR, "'('");
    TOKNEW; REXPECT(T_MODEL, "'model'");
    TOKNEW; PEXPECT(T_IDENT, "model identifier");
    system->name = t.value;
    TOKNEW; /*PEXPECT(T_LPAR, "'('");
    TOKNEW; REXPECT(T_PIPESTAGES, "pipeline stages definition");
    p_pipeline();
    TOKNEW;*/
    while (t.type != T_RPAR) // end of model
    {
        PEXPECT(T_LPAR, "(");
        TOKRNEW;
        switch (t.type)
        {
            case T_PIPESTAGES: p_pipeline(); break;
            case T_SIG: system->insert(p_sig(false)); break;
            case T_REG: system->insert(p_reg(false)); break;
            case T_MEM: system->insert(p_mem(false)); break;
            case T_FNODE: system->insert(p_fnode(false)); break;
            default: PEXPECT(0, "'sig', 'reg', 'mem', 'fnode', or 'pipeline-stages'");
        };
        TOKNEW;
    }
}

void RTLParser::p_pipeline()
{
    TOKINIT(&system->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "pipeline stage identifier");
        system->stages.push_back(t.value);
        TOKNEW;
    }
}

Signal *RTLParser::p_sig(bool include_head)
{
    Signal *s = new Signal();
    TOKINIT(&s->meta);
    if (include_head)
    {
        TOKNEW; PEXPECT(T_LPAR, "(");
        TOKRNEW; PEXPECT(T_SIG, "'sig' keyword");
    }
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    s->id = t.value;
    TOKNEW; PEXPECT(T_CONST, "signal bitwidth");
    s->bitwidth = stoi(t.value);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "'('");
        TOKRNEW;
        switch (t.type)
        {
            case T_VT: p_sig_vt(s); break;
            case T_DIR: p_sig_dir(s); break;
            case T_INPUTS: p_sig_inputs(s); break;
            case T_STAGE: p_sig_stage(s); break;
            case T_CONNECTIONS: p_sig_connections(s); break;
            default: PEXPECT(0, "direction, inputs, or stage definition");
        }
        TOKNEW;
    }
    return s;
}

void RTLParser::p_sig_vt(Signal *s)
{
    TOKINIT(&s->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        t = tok2keyword(t);
        switch (t.type)
        {
            case T_DATA: s->vt |= VT_DATA; break;
            case T_ADDRESS: s->vt |= VT_ADDRESS; break;
            case T_CONTROL: s->vt |= VT_CONTROL; break;
            case T_UNKNOWN: break;
            default:
                PEXPECT(0, "'data', 'address', 'control', or 'unknown'");
        }
        TOKNEW;
    }
}

void RTLParser::p_sig_dir(Signal *s)
{
    TOKINIT(&s->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        t = tok2keyword(t);
        switch (t.type)
        {
            case T_RW: s->set_write();
            case T_RO: s->set_read(); break;
            case T_WO: s->set_write(); break;
            default: PEXPECT(0, "'ro', 'wo', 'rw'");
        }
        TOKNEW;
    }
}

void RTLParser::p_sig_inputs(Signal *s)
{
    TOKINIT(&s->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "storage identifier");
        s->inputs_name.insert(t.value);
        TOKNEW;
    }
}

void RTLParser::p_sig_stage(Signal *s)
{
    TOKINIT(&s->meta);
    TOKNEW; PEXPECT(T_IDENT, "stage identifier");
    s->stage = system->find_stage(t.value);
    PASSERT(s->stage >= 0, t, "stage identifier");
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_sig_connections(Signal *s)
{
    // ignores the input file because of the connections are made during
    // storage definitions
    TOKINIT(&s->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "new connection");
        TOKNEW; PEXPECT(T_IDENT, "identifier");
        TOKNEW; PEXPECT(T_IDENT, "connection type");
        TOKNEW; PEXPECT(T_RPAR, "')'");
        TOKNEW;
    }
}

Register *RTLParser::p_reg(bool include_head)
{
    Register *r = new Register();
    TOKINIT(&r->meta);
    if (include_head)
    {
        TOKNEW; PEXPECT(T_LPAR, "(");
        TOKRNEW; PEXPECT(T_SIG, "'reg' keyword");
    }
    TOKNEW; PEXPECT(T_IDENT, "register identifier");
    r->id = t.value;
    TOKNEW; PEXPECT(T_CONST, "register bitwidth");
    r->bitwidth = stoi(t.value);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "'('");
        TOKRNEW;
        switch (t.type)
        {
            case T_WE: p_reg_we(r); break;
            case T_D: p_reg_d(r); break;
            case T_Q: p_reg_q(r); break;
            case T_STALL: p_reg_stall(r); break;
            case T_CLR: p_reg_clr(r); break;
            case T_RST: p_reg_rst(r); break;
            case T_STAGE: p_reg_stage(r); break;
            default:
                PEXPECT(0, "register port or stage definition");
        }
        TOKNEW;
    }
    r->check_bitwidth();
    return r;
}

void RTLParser::p_reg_we(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { r->add_we(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_reg_d(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { r->add_d(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_reg_q(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "signal identifier");
        Signal *s = system->signals.find(t.value);
        PASSERT(s, t, "unknown signal identifier");
        try { r->add_q(s); }
        catch (VAMException e) { PASSERT(0, t, e.msg); }
        TOKNEW;
    }
}

void RTLParser::p_reg_stall(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { r->add_stall(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_reg_clr(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { r->add_clr(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_reg_rst(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { r->add_rst(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_reg_stage(Register *r)
{
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_IDENT, "stage identifier");
    r->stage = system->find_stage(t.value);
    PASSERT(r->stage >= 0, t, "stage identifier");
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

Memory *RTLParser::p_mem(bool include_head)
{
    Memory *mem = new Memory();
    TOKINIT(&mem->meta);
    if (include_head)
    {
        TOKNEW; PEXPECT(T_LPAR, "(");
        TOKRNEW; PEXPECT(T_SIG, "'mem' keyword");
    }
    TOKNEW; PEXPECT(T_IDENT, "memory identifier");
    mem->id = t.value;
    TOKNEW; PEXPECT(T_CONST, "memory cell bitwidth");
    mem->bitwidth = stoi(t.value);
    TOKNEW; PEXPECT(T_CONST, "memory address bitwidth");
    mem->addr_bitwidth = stoi(t.value);
    TOKNEW; PEXPECT(T_CONST, "number of memory cells");
    mem->size = stoi(t.value);
    p_mem_ports(mem);
    TOKRNEW;
    if (t.type == T_RST)
    {
        p_mem_rst(mem);
        TOKRNEW;
    }
    PEXPECT(T_RPAR, "')'");
    return mem;
}

void RTLParser::p_mem_ports(Memory *mem)
{
    TOKINIT(&mem->meta);
    TOKNEW; PEXPECT(T_LPAR, "ports definition section");
    TOKNEW; REXPECT(T_PORTS, "ports definition section");
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "port definition section");
        mem->add_port(p_mem_port(false));
        TOKNEW;
    }
}

MemPort *RTLParser::p_mem_port(bool include_head)
{
    MemPort *port = new MemPort();
    TOKINIT(&port->meta);
    if (include_head)
    {
        TOKNEW; PEXPECT(T_LPAR, "(");
        TOKRNEW; PEXPECT(T_SIG, "'port' keyword");
    }
    TOKNEW; REXPECT(T_PORT, "port definition section");
    TOKNEW; PEXPECT(T_IDENT, "port identifier");
    port->id = t.value;
    p_mem_port_dir(port);
    p_mem_port_timing(port);
    p_mem_port_en(port);
    p_mem_port_addr(port);
    p_mem_port_data(port);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "'('");
        TOKRNEW;
        switch (t.type)
        {
            case T_CONTROL: p_mem_port_control(port); break;
            case T_STAGE: p_mem_port_stage(port); break;
            default:
                PEXPECT(0, "memory port signals definition or ')'");
        }
        TOKNEW;
    }
    return port;
}

void RTLParser::p_mem_port_dir(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_LPAR, "data access direction specification");
    TOKNEW; REXPECT(T_DIR, "data access direction specification");
    TOKNEW;
    while (t.type != T_RPAR)
    {
        t = tok2keyword(t);
        switch (t.type)
        {
            case T_RW: p->set_write();
            case T_RO: p->set_read(); break;
            case T_WO: p->set_write(); break;
            default: PEXPECT(0, "'ro', 'wo', 'rw'");
        }
        TOKNEW;
    }
}

void RTLParser::p_mem_port_timing(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_LPAR, "timing specification");
    TOKNEW; REXPECT(T_TIMING, "timing specification");
    TOKNEW; PEXPECT(T_CONST, "timing delay");
    if (p->is_write())
        p->write_timing = stoi(t.value);
    else
        p->read_timing = stoi(t.value);
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_mem_port_en(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_LPAR, "EN signal definition");
    TOKNEW; REXPECT(T_EN, "EN signal definition");
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { p->add_enable(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_mem_port_addr(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_LPAR, "ADDR signal definition");
    TOKNEW; REXPECT(T_ADDR, "ADDR signal definition");
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { p->add_address(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_mem_port_data(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_LPAR, "DATA signal definition");
    TOKNEW; REXPECT(T_DATA, "DATA signal definition");
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "signal identifier");
        Signal *s = system->signals.find(t.value);
        PASSERT(s, t, "unknown signal identifier");
        try { p->add_data(s); }
        catch (VAMException e) { PASSERT(0, t, e.msg); }
        TOKNEW;
    }
}

void RTLParser::p_mem_port_control(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "signal identifier");
        Signal *s = system->signals.find(t.value);
        PASSERT(s, t, "unknown signal identifier");
        try { p->add_control(s); }
        catch (VAMException e) { PASSERT(0, t, e.msg); }
        TOKNEW;
    }
}

void RTLParser::p_mem_port_stage(MemPort *p)
{
    TOKINIT(&p->meta);
    TOKNEW; PEXPECT(T_IDENT, "stage identifier");
    p->stage = system->find_stage(t.value);
    PASSERT(p->stage >= 0, t, "stage identifier");
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

void RTLParser::p_mem_rst(Memory *mem)
{
    TOKINIT(&mem->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    try { mem->add_rst(s); }
    catch (VAMException e) { PASSERT(0, t, e.msg); }
    TOKNEW; PEXPECT(T_RPAR, "')'");
}

FunNode *RTLParser::p_fnode(bool include_head)
{
    FunNode *f = new FunNode();
    TOKINIT(&f->meta);
    if (include_head)
    {
        TOKNEW; PEXPECT(T_LPAR, "(");
        TOKRNEW; PEXPECT(T_SIG, "'fnode' keyword");
    }
    TOKNEW; PEXPECT(T_IDENT, "functional node identifier");
    f->id = t.value;
    p_fnode_input(f);
    p_fnode_output(f);
    TOKNEW; PEXPECT(T_LPAR, "'(' of stage or assign definition");
    TOKRNEW;
    if (t.type == T_STAGE)
    {
        p_fnode_stage(f);
        TOKNEW; PEXPECT(T_LPAR, "'(' of assignment");
        TOKNEW;
    }
    REXPECT(T_ASSIGN, "'assign'");
    p_fnode_assign(f);
    TOKNEW; PEXPECT(T_RPAR, "')' of functional node");
    return f;
}

void RTLParser::p_fnode_input(FunNode *f)
{
    TOKINIT(&f->meta);
    TOKNEW; PEXPECT(T_LPAR, "fun. node inputs definition");
    TOKNEW; REXPECT(T_INPUT, "fun. node inputs definition");
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "input signal identifier");
        Signal *s = system->signals.find(t.value);
        PASSERT(s, t, "unknown signal identifier");
        f->inputs.insert(s);
        TOKNEW;
    }
}

void RTLParser::p_fnode_output(FunNode *f)
{
    TOKINIT(&f->meta);
    TOKNEW; PEXPECT(T_LPAR, "fun. node output definition");
    TOKNEW; REXPECT(T_OUTPUT, "fun. node output definition");
    TOKNEW;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_IDENT, "output signal identifier");
        Signal *s = system->signals.find(t.value);
        PASSERT(s, t, "unknown signal identifier");
        f->outputs.insert(s);
        TOKNEW;
    }
}

void RTLParser::p_fnode_stage(FunNode *f)
{
    TOKINIT(&f->meta);
    TOKNEW; PEXPECT(T_IDENT, "stage identifier");
    f->stage = system->find_stage(t.value);
    PASSERT(f->stage >= 0, t, "stage identifier");
    TOKNEW; PEXPECT(T_RPAR, "enclosed stage definition");
}

void RTLParser::p_fnode_assign(FunNode *f)
{
    TOKINIT(&f->meta);
    TOKNEW;
    ExprAssign *e;
    while (t.type != T_RPAR)
    {
        PEXPECT(T_LPAR, "'(' of an assignment");
        TOKNEW; REXPECT(T_EASSIGN, "':='");
        e = p_e_assign();
        CHECKWIDTH(e);
        Signal *s;
        if (e->lhs->op == OP_VAR &&
            (s = dynamic_cast<Signal*>(dynamic_cast<ExprVariable*>(e->lhs)->v)))
            s->owner = f;
        f->assignments.add_operand(e);
        TOKNEW; // next (:= or end of (assign ?
    }
}

ExprAssign* RTLParser::p_e_assign()
{
    ExprAssign *e = new ExprAssign();
    TOKINIT(&e->meta);
    TOKNEW; PEXPECT(T_IDENT, "signal identifier");
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier");
    ExprVariable *lhs = new ExprVariable(s);
    Expression *rhs = p_expr(e);
    PASSERT(lhs, t, "error when parsing lhs expression");
    PASSERT(rhs, t, "error when parsing rhs expression");
    e->add_lhs(lhs);
    e->add_rhs(rhs);
    CHECKWIDTH(e);
    TOKNEW; PEXPECT(T_RPAR, "enclosed assignment"); // end of (:=
    return e;
}

Expression* RTLParser::p_expr(Expression *parent, bool accept_rpar)
{
    TOKINIT(parent ? &parent->meta : nullptr);
    TOKRNEW;
    if (accept_rpar && t.type == T_RPAR)
        // absorb RPAR of the parent expression and return nullptr indicating
        // the end of the expression
        return nullptr;
    else if (t.type == T_IDENT)
        return p_e_var(t);
    else if (t.type == T_CONST)
        return new ExprConst(t.value, -1);
    else
    {
        try {
        PEXPECT(T_LPAR, "'('");
        TOKRNEW;
        switch (t.type)
        {
            case T_VCONST: return p_e_vconst();
            // unary
            case T_NOT: return p_e_unary(new ExprNot);
            case T_INV: return p_e_unary(new ExprInv);
            case T_NEG: return p_e_unary(new ExprNeg);
            case T_UTS: return p_e_unary(new ExprUnsignedToSigned);
            case T_STU: return p_e_unary(new ExprSignedToUnsigned);
            // binary
            case T_BOR: return p_e_binary(new ExprBitwiseOr);
            case T_BAND: return p_e_binary(new ExprBitwiseAnd);
            case T_BXOR: return p_e_binary(new ExprBitwiseXor);
            case T_SHL: return p_e_binary(new ExprShiftLeft);
            case T_SHR: return p_e_binary(new ExprShiftRight);
            case T_SAR: return p_e_binary(new ExprShiftArithRight);
            case T_EQ: return p_e_binary(new ExprEqual);
            case T_NEQ: return p_e_binary(new ExprNotEqual);
            case T_LT: return p_e_binary(new ExprLessThan);
            case T_GT: return p_e_binary(new ExprGreaterThan);
            case T_LE: return p_e_binary(new ExprLessEqual);
            case T_GE: return p_e_binary(new ExprGreaterEqual);
            case T_SLT: return p_e_binary(new ExprSignedLessThan);
            case T_SGT: return p_e_binary(new ExprSignedGreaterThan);
            case T_SLE: return p_e_binary(new ExprSignedLessEqual);
            case T_SGE: return p_e_binary(new ExprSignedGreaterEqual);
            case T_PLUS: return p_e_binary(new ExprPlus);
            case T_MINUS: return p_e_binary(new ExprMinus);
            case T_MUL: return p_e_binary(new ExprMul);
            case T_DIV: return p_e_binary(new ExprDiv);
            case T_MOD: return p_e_binary(new ExprMod);
            case T_IDX: return p_e_binary(new ExprIndex);
            case T_CONCAT: return p_e_binary(new ExprConcatenate);
            // logic
            case T_OR: return p_e_or(t);
            case T_AND: return p_e_and(t);
            case T_ORJOIN: return p_e_orjoin(t);
            // special
            case T_ZEX: return p_e_zex();
            case T_SEX: return p_e_sex();
            case T_BITEX: return p_e_bitex();
            case T_SLICE: return p_e_slice();
            case T_ITE: return p_e_ite();
            case T_CASE: return p_e_switch();
            case T_COND: return p_e_cond();
            // abstract
            case T_ABSLOGIC: return p_e_abstract(new ExprAbsLogic);
            case T_ABSBITWISE: return p_e_abstract(new ExprAbsBitwise);
            case T_ABSREL: return p_e_abstract(new ExprAbsRelational);
            case T_ABSVECTOR: return p_e_abstract(new ExprAbsVector);
            case T_ABSARITH: return p_e_abstract(new ExprAbsArithmetic);
            // variable reference with meta
            case T_IDENT: return p_e_var_with_meta(t);
            default:
                cerr << "neco je spatne\n";
                PEXPECT(0, "expression operator");
                return nullptr;
        }
        }
        catch (RTLException &e)
        {
            e.msg += ", encapsulated\n" + t.loc() + "here";
            throw;
        }
    }
}

Expression* RTLParser::p_e_var(Token t)
{
    Signal *s = system->signals.find(t.value);
    PASSERT(s, t, "unknown signal identifier " << t.value);
    auto r = new ExprVariable(s);
    CHECKWIDTH(r);
    return r;
}

Expression* RTLParser::p_e_var_with_meta(Token tok)
{
    Signal *s = system->signals.find(tok.value);
    PASSERT(s, tok, "unknown signal identifier " << tok.value);
    auto r = new ExprVariable(s);
    TOKINIT(&r->meta);
    TOKNEW; PEXPECT(T_RPAR, "enclosed variable reference");
    CHECKWIDTH(r);
    return r;
}

Expression* RTLParser::p_e_vconst()
{
    ExprConst *e = new ExprConst;
    TOKINIT(&e->meta);
    TOKNEW; PEXPECT(T_CONST, "integer value");
    e->value = t.value;
    TOKNEW; PEXPECT(T_CONST, "integer value bitwidth");
    e->bitwidth = stoi(t.value);
    TOKNEW; PEXPECT(T_RPAR, "enclosed vector constant");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_unary(UnaryExpression *e)
{
    TOKINIT(&e->meta);
    e->add_operand(p_expr(e));
    TOKNEW; PEXPECT(T_RPAR, "enclosed unary expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_binary(BinaryExpression *e)
{
    TOKINIT(&e->meta);
    e->add_lhs(p_expr(e));
    e->add_rhs(p_expr(e));
    TOKNEW; PEXPECT(T_RPAR, "enclosed binary expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_or(Token t)
{
    ExprOr *e = new ExprOr;
    Expression *o;
    while ((o = p_expr(e, true))) // true means: absorb T_RPAR
        e->add_operand(o);
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_and(Token t)
{
    ExprAnd *e = new ExprAnd;
    Expression *o;
    while ((o = p_expr(e, true))) // true means: absorb T_RPAR
        e->add_operand(o);
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_orjoin(Token t)
{
    ExprOrJoin *e = new ExprOrJoin;
    Expression *o;
    while ((o = p_expr(e, true))) // true means: absorb T_RPAR
        e->add_operand(o);
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_zex()
{
    ExprZeroExt *e = new ExprZeroExt;
    TOKINIT(&e->meta);
    e->add_operand(p_expr(e));
    TOKNEW; PEXPECT(T_CONST, "bit width specification");
    e->req_bitwidth = stoi(t.value);
    TOKNEW; PEXPECT(T_RPAR, "enclosed binary expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_sex()
{
    ExprSignExt *e = new ExprSignExt;
    TOKINIT(&e->meta);
    e->add_operand(p_expr(e));
    TOKNEW; PEXPECT(T_CONST, "bit width specification");
    e->req_bitwidth = stoi(t.value);
    TOKNEW; PEXPECT(T_RPAR, "enclosed binary expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_bitex()
{
    ExprBitExtract *e = new ExprBitExtract;
    TOKINIT(&e->meta);
    e->add_operand(p_expr(e));
    TOKNEW; PEXPECT(T_CONST, "bit width specification");
    e->index = stoi(t.value);
    TOKNEW; PEXPECT(T_RPAR, "enclosed binary expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_slice()
{
    ExprSlice *e = new ExprSlice;
    TOKINIT(&e->meta);
    e->add_operand(p_expr(e));
    TOKNEW; PEXPECT(T_CONST, "vector lower bound");
    e->from = stoi(t.value);
    TOKNEW; PEXPECT(T_CONST, "vector upper bound");
    e->to = stoi(t.value);
    PASSERT(0 <= e->from && e->from < e->to, t,
            "invalid combination of lower and upper bounds of the vector");
    TOKNEW; PEXPECT(T_RPAR, "enclosed slice expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_ite()
{
    ExprITE *e = new ExprITE;
    TOKINIT(&e->meta);
    e->add_cond(p_expr(e));
    e->add_true(p_expr(e));
    e->add_false(p_expr(e));
    TOKNEW; PEXPECT(T_RPAR, "enclosed ITE expression");
    CHECKWIDTH(e);
    return e;
}

Expression* RTLParser::p_e_switch()
{
    ExprCase *e = new ExprCase;
    TOKINIT(&e->meta);
    e->add_ctrl_expr(p_expr(e));
    while (1)
    {
        TOKNEW;
        if (t.type == T_LPAR)
            p_e_case(e);
        else
        {
            PEXPECT(T_RPAR, "case definition or enclosed switch-expression");
            CHECKWIDTH(e);
            return e;
        }
    }
}

void RTLParser::p_e_case(ControlExpression *e)
{
    TOKINIT(&e->meta);
    Expression *e1 = p_expr(e);
    Expression *e2 = p_expr(e, true); // return nullptr if ')'-token
    if (e2)
    {
        e->add_case(e1, e2);
        TOKNEW; PEXPECT(T_RPAR, "enclosed case");
    }
    else
    {
        PASSERT(e->def == nullptr, t, "already defined default case");
        e->add_default(e1);
    }

}

Expression* RTLParser::p_e_cond()
{
    ExprCond *e = new ExprCond;
    TOKINIT(&e->meta);
    while (1)
    {
        TOKNEW;
        if (t.type == T_LPAR)
            p_e_case(e);
        else
        {
            PEXPECT(T_RPAR, "case definition or enclosed cond-expression");
            CHECKWIDTH(e);
            return e;
        }
    }
}

ExprAbstract* RTLParser::p_e_abstract(ExprAbstract *e)
{
    TOKINIT(&e->meta);
    TOKNEW; PEXPECT(T_CONST, "bitwidth of abstract function");
    e->bitwidth = stoi(t.value);
    Expression *o;
    while ((o = p_expr(e, true))) // true means: absorb T_RPAR
        e->add_operand(o);
    CHECKWIDTH(e);
    return e;
}

/**********************************
 * RTLStringParser
 */

RTLObject RTLStringParser::declare(const string &s)
{
    lexer.init(s);
    TOKINIT(&system->meta);
    TOKNEW; PEXPECT(T_LPAR, "'('");
    TOKRNEW;
    RTLObject obj;
    switch (t.type)
    {
        case T_SIG: system->insert(obj = p_sig(false)); break;
        case T_REG: system->insert(obj = p_reg(false)); break;
        case T_MEM: system->insert(obj = p_mem(false)); break;
        case T_PORT: obj = p_mem_port(false); break;
        case T_FNODE: system->insert(obj = p_fnode(false)); break;
        default: PEXPECT(0, "'sig', 'reg', 'mem', 'port', or 'fnode'");
    };
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return obj;
}

Expression *RTLStringParser::compile_expression(const string &s)
{
    lexer.init(s);
    return p_expr();
}

}

#include "switch-parser-macro.h"
