/**
 * @desc Routines for import of VAM.
 */
#ifndef _VAMINPUT_H
#define _VAMINPUT_H

#include "vamrtl.h"

namespace vam {

enum TokenType
{
    T_UNDEF,        // undefined token
    T_LPAR,         // (
    T_RPAR,         // )
    T_IDENT,        // <identifier>
    T_CONST,        // <integer constant>
    T_STRING,       // "string"
    T_EOF,          // end-of-file
};

enum RTLKeywords
{
    T_META = T_EOF+1,         // meta
    T_MODEL,        // model
    T_PIPESTAGES,   // pipeline-stages
    T_SIG,          // sig
    T_REG,          // reg
    T_MEM,          // mem
    T_FNODES,       // fnodes
    T_VT,           // vt (value type)
    T_DIR,          // dir
    T_INPUTS,       // inputs
    T_STAGE,        // stage
    T_CONNECTIONS,  // connections
    T_WE,           // we
    T_D,            // d
    T_Q,            // q
    T_STALL,        // stall
    T_CLR,          // clr
    T_RST,          // rst
    T_PORTS,        // ports
    T_PORT,         // port
    T_EN,           // en
    T_ADDR,         // addr
    T_ADDRESS,      // address
    T_TIMING,       // timing
    T_CONTROL,      // control
    T_UNKNOWN,      // unknown
    T_DATA,         // data
    T_FNODE,        // fnode
    T_INPUT,        // input
    T_OUTPUT,       // output
    T_ASSIGN,       // assign
    T_RO,           // ro
    T_WO,           // wo
    T_RW,           // rw
    T_EXPR_OPERATOR_BEGIN,  // abstract begin() of expression operators
    T_VCONST = T_EXPR_OPERATOR_BEGIN,       // _
    T_EASSIGN,      // :=
    T_NOT,          // !
    T_OR,           // ||
    T_AND,          // &&
    T_ORJOIN,       // |.
    T_ITE,          // ?
    T_INV,          // ~
    T_BOR,          // |
    T_BAND,         // &
    T_BXOR,         // ^
    T_SHL,          // <<
    T_SHR,          // >>
    T_SAR,          // >>.
    T_EQ,           // ==
    T_NEQ,          // !=
    T_LT,           // <
    T_GT,           // >
    T_LE,           // <=
    T_GE,           // >=
    T_SLT,          // <.
    T_SGT,          // >.
    T_SLE,          // <=.
    T_SGE,          // >=.
    T_UTS,          // -.
    T_STU,          // .-
    T_IDX,          // []
    T_BITEX,        // [.]
    T_CONCAT,       // .
    T_ZEX,          // ^^
    T_SEX,          // ^^.
    T_SLICE,        // [..]
    T_CASE,         // switch
    T_COND,         // cond
    T_DEFAULT,      // default
    T_NEG,          // --
    T_PLUS,         // +
    T_MINUS,        // -
    T_MUL,          // *
    T_DIV,          // /
    T_MOD,          // %
    T_ABSLOGIC,     // !!
    T_ABSBITWISE,   // ~~
    T_ABSREL,       // <>
    T_ABSVECTOR,    // ][
    T_ABSARITH,     // -+
    T_EXPR_OPERATOR_END, // abstract end() of expression operators
};

class Token
{
public:
    int type;
    string value;
    unsigned row;
    unsigned col;
    string filename;
public:
    Token() : type(T_UNDEF), row(0), col(0) {}
    Token(int t, string v = "", unsigned r = 0, unsigned c = 0) :
        type(t), value(v), row(r), col(c) {}

    string loc()
    {
        ostringstream out;
        if (!filename.empty())
            out << filename << ":";
        out << row << ":" << col << ": ";
        return out.str();
    }
};

/**
 * @brief A lexer of LISP-like language. Returns tokens of type TokenType.
 * Numerical constant are of type [-+]?[0-9]+. String literals can be across 
 * multiple lines, escape character is '\'. Anything else (even sequence of
 * symbols) is an identifier. Ignores comments:
 *   a) beginning with ;; and ending with \n
 *   b) beginning with // and ending with \n
 *   c) beginning with '/','*' and ending with '*','/'
 */
class Lexer
{
public:
    string input;
    unsigned pos = 0;
    unsigned row = 1;
    unsigned col = 1;
    string filename;

protected:
    int get_char_with_comment();
    int get_char();
    void unget();

public:
    Lexer(const string &i) : input(i)
    { }
    Lexer(const char *s = "") : input(s)
    { }

    void init(string s = "")
    {
        input = s;
        pos = 0;
        row = col = 1;
        filename = "";
    }

    /// reads content from the file
    bool read_from_file(const string &name);

    /// reads content from input stream
    void read_from(istream &i);

    /// get next token
    Token next_token();

    /// returns true if there is no character to parse
    bool empty()
    {
        return pos >= input.size();
    }
};

/**
 * @brief Generate a system from an input (lexer, file, or a string).
 * Data already set in the system is preserved.
 */
class RTLParser
{
public:
    Lexer lexer;

public:
    /**
     * @brief Load the system via given lexer.
     * @param   s   system to which to load
     */
    void parse(System *s);

    /**
     * @brief Load the system via given lexer.
     * @param   l   lexer
     * @param   s   system to which to load
     */
    void parse(Lexer &l, System *s)
    {
        lexer = l;
        parse(s);
    }

    /**
     * @brief   Load the system from the given file.
     * @param   ifile   input file name
     * @param   s       system to which to load
     */
    void parse_from_file(const string &ifile, System *s)
    {
        ASSERT(lexer.read_from_file(ifile), "unable to open " << ifile);
        parse(s);
    }

    /**
     * @brief   Load the system from the given string.
     * @param   i   input string
     * @param   s   system to which to load.
     */
    void parse_str(const string &i, System *s)
    {
        lexer.init(i);
        parse(s);
    }

protected:
    System *system;

    Token tmptok;   // for possible "(meta..."

    static Token tok2keyword(Token t);

    void p_meta(Meta *meta, int was_lpar = 0);

    /**
     * @brief Return next token, deal with three situations:
     * 1. next tok is not '(' => return it
     * 2. next toks are "(meta" => accept meta rule and return the next tok
     * 3. next toks are "(<no-meta>" => return '(' and then return <no-meta>
     */
    Token next_token(Meta *meta)
    {
        Token t;
        while (1)
        {
            if (tmptok.type == T_UNDEF)
                t = lexer.next_token();
            else
            {
                t = tmptok;
                tmptok.type = T_UNDEF;
            }
            if (t.type == T_LPAR)
            {
                tmptok = lexer.next_token();
                Token keyword = tok2keyword(tmptok);
                if (keyword.type == T_META)
                {
                    tmptok.type = T_UNDEF;
                    p_meta(meta);
                }
                else
                    return t;
            }
            else
                return t;
        }
    }

    void p_model();
    void p_pipeline();
    Signal *p_sig(bool include_head);
    void p_sig_vt(Signal *s);
    void p_sig_dir(Signal *s);
    void p_sig_inputs(Signal *s);
    void p_sig_stage(Signal *s);
    void p_sig_connections(Signal *s);
    Register *p_reg(bool include_head);
    void p_reg_we(Register *r);
    void p_reg_d(Register *r);
    void p_reg_q(Register *r);
    void p_reg_stall(Register *r);
    void p_reg_clr(Register *r);
    void p_reg_rst(Register *r);
    void p_reg_stage(Register *r);
    Memory *p_mem(bool include_head);
    void p_mem_ports(Memory *mem);
    MemPort *p_mem_port(bool include_head);
    void p_mem_port_dir(MemPort *p);
    void p_mem_port_timing(MemPort *p);
    void p_mem_port_en(MemPort *p);
    void p_mem_port_addr(MemPort *p);
    void p_mem_port_data(MemPort *p);
    void p_mem_port_control(MemPort *p);
    void p_mem_port_stage(MemPort *p);
    void p_mem_rst(Memory *mem);
    FunNode *p_fnode(bool include_head);
    void p_fnode_input(FunNode *f);
    void p_fnode_output(FunNode *f);
    void p_fnode_stage(FunNode *f);
    void p_fnode_assign(FunNode *f);
    ExprAssign* p_e_assign();
    Expression* p_expr(Expression *parent = nullptr, bool accept_rpar = false);
    Expression* p_e_var(Token t);
    Expression* p_e_var_with_meta(Token tok);
    Expression* p_e_vconst();
    Expression* p_e_unary(UnaryExpression *e);
    Expression* p_e_binary(BinaryExpression *e);
    Expression* p_e_or(Token t);
    Expression* p_e_and(Token t);
    Expression* p_e_orjoin(Token t);
    Expression* p_e_zex();
    Expression* p_e_sex();
    Expression* p_e_bitex();
    Expression* p_e_slice(); // [..]
    Expression* p_e_ite();
    Expression* p_e_switch();
    void p_e_case(ControlExpression *e);
    Expression* p_e_cond();
    ExprAbstract* p_e_abstract(ExprAbstract *e);
};

class RTLStringParser : public RTLParser
{
public:
    RTLStringParser(System &s)
    {
        system = &s;
    }

    /**
     * \{
     * @brief   Methods for creating parts of the system from strings.
     */
    void parse_system(const string &s)
    {
        parse_str(s, system);
    }

    /**
     * @brief Declare an RTLNode and add it to the system.
     * @param   s   string in VAM language
     * @return  RTL object (signal, register, memory, memory port, fun. node)
     */
    RTLObject declare(const string &s);

    /**
     * @brief   Parse the VAML string and compile a new expression.
     * @param   s   string in VAM language representing an expression
     * @return  compiled RTL expression
     */
    Expression *compile_expression(const string &s);
};

}

#endif
