/**
 * @desc Routines for import/export of VAM.
 */
#include "output.h"
#include <fstream>
#include <boost/algorithm/string.hpp> // replace_all

using namespace std;

namespace vam {

inline string sc2str(SignalConnectionType sc)
{
    switch (sc)
    {
        case CT_UNDEFINED: return "undefined";
        case CT_Q: return "q";
        case CT_D: return "d";
        case CT_DATA: return "data";
        case CT_WE: return "we";
        case CT_RE: return "re";
        case CT_EN: return "en";
        case CT_WA: return "wa";
        case CT_RA: return "ra";
        case CT_A: return "address";
        case CT_STALL: return "stall";
        case CT_CLR: return "clr";
        case CT_RST: return "rst";
        default: return "unknown";
    };
}

void RTLVisitor::VisitSignal(Signal *s)
{
    out << indentstr();
    // signal identifier
    out << "(sig " << s->id << " " << s->bitwidth << " ";

    // value type
    out << "(vt ";
    if (s->vt)
    {
        if (s->is_data()) out << "data ";
        if (s->is_address()) out << "address ";
        if (s->is_control()) out << "control";
        out << ") ";
    }
    else
        out << "unknown) ";

    // direction
    if (s->dir)
    {
        if (s->is_read())
            if (s->is_write())
                out << "(dir rw) ";
            else
                out << "(dir ro) ";
        else
            out << "(dir wo) ";
    }

    // influential inputs
    if (s->inputs.size() != 0)
    {
        out << "(inputs ";
        for (auto i : s->inputs)
            out << i->id << " ";
        out << ") ";
    }

    // stage identification
    if (s->stage >= 0)
        out << "(stage " << system->stages[s->stage] << ") ";

    // connections
    if (!s->connections.empty())
    {
        out << "(connections ";
        for (auto c : s->connections)
        {
            if (c.storage)
                out << "(" << c.storage->id << " ";
            else
                out << "(NULL ";
            out << sc2str(SignalConnectionType(c.type)) << ") ";
        }
        out << ") ";
    }
    // meta
    indent++;
    s->meta.Accept(this);
    indent--;
    out << ")";
}

void RTLVisitor::VisitStorage(Storage *st)
{
    (void)st;
}

void RTLVisitor::VisitRegister(Register *r)
{
    out << indentstr();
    out << "(reg " << r->id << " " << r->bitwidth << " ";
    if (r->we)
        out << "(we " << r->we->id << ") ";
    if (r->d)
        out << "(d " << r->d->id << ") ";
    else
        out << "(d N/A) ";
    out << "(q ";
    for (auto s : r->q)
        out << s->id << " ";
    out << ") ";
    if (r->stall)
        out << "(stall " << r->stall->id << ") ";
    if (r->clr)
        out << "(clr " << r->clr->id << ") ";
    if (r->rst)
        out << "(rst " << r->rst->id << ") ";
    if (r->stage >= 0)
        out << "(stage " << r->stage << ") ";
    indent++;
    r->meta.Accept(this);
    indent--;
    out << ")";
}

void RTLVisitor::VisitMemPort(MemPort *port)
{
    out << "(port " << port->id << " ";
    // direction
    if (port->is_read())
        if (port->is_write())
        {
            out << "(dir rw) ";
            out << "(timing " << port->write_timing.min << ") ";
        }
        else
        {
            out << "(dir ro) ";
            out << "(timing " << port->read_timing.min << ") ";
        }
    else
    {
        out << "(dir wo) ";
        out << "(timing " << port->write_timing.min << ") ";
    }
    out << "(en " << port->enable->id << ") ";
    out << "(addr " << port->address->id << ") ";
    // data
    out << "(data ";
    for (auto d : port->data)
        out << d->id << " ";
    out << ") ";
    // control
    if (port->control.size())
    {
        out << "(control ";
        for (auto s : port->control)
            out << s->id << " ";
        out << ") ";
    }
    // stage identification
    if (port->stage >= 0)
        out << "(stage " << system->stages[port->stage] << ") ";
    port->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitMemory(Memory *m)
{
    out << indentstr();
    out << "(mem " << m->id << " " << m->bitwidth << " " << m->addr_bitwidth;
    out << " " << m->size << " ";
    out << "(ports ";
    for (auto port : m->ports)
        VisitMemPort(port);
    out << ") ";
    if (m->rst)
        out << "(rst " << m->rst->id << ") ";
    indent++;
    m->meta.Accept(this);
    indent--;
    out << ") ";
}

void RTLVisitor::VisitFunNode(FunNode *f)
{
    out << indentstr();
    out << "(fnode " << f->id << " ";
    out << "(input ";
    for (auto i : f->inputs)
        out << i->id << " ";
    out << ") ";
    out << "(output ";
    for (auto o : f->outputs)
        out << o->id << " ";
    out << ") ";
    if (f->stage >= 0)
        out << "(stage " << system->stages[f->stage] << ") ";
    indent++;
    out << endl << indentstr() << "(assign ";
    for (ExprAssign *a : f->assignments)
        a->Accept(this);
    out << indentstr() << ")";
    f->meta.Accept(this);
    indent--;
    out << ") ";
}

void RTLVisitor::VisitSystem(System *s)
{
    system = s;
    out << "(model " << s->name << endl;
    out << "  (pipeline-stages";
    for (auto stage : s->stages)
        out << " " << stage;
    out << ")" << endl;
    indent++;
    indent++;
    s->meta.Accept(this);
    indent--;
    for (auto signal : s->signals)
    {
        signal->Accept(this);
        out << endl;
    }
    for (auto reg : s->registers)
    {
        reg->Accept(this);
        out << endl;
    }
    for (auto mem : s->memories)
    {
        mem->Accept(this);
        out << endl;
    }
    for (auto fun : s->fnodes)
    {
        fun->Accept(this);
        out << endl;
    }
    indent--;
    out << ")" << endl;
}

void RTLVisitor::VisitMeta(Meta *meta)
{
    if (meta->type == MT_NODE)
    {
        if (meta->value == "")
            out << endl << indentstr() << "(meta ";
        else
            out << "(" << meta->value << " ";
        indent++;
        meta->AcceptAll(this);
        indent--;
        out << ") ";
    }
    else if (meta->type == MT_INTVALUE)
        out << meta->value << " ";
    else if (meta->type == MT_VALUE)
    {
        string v = meta->value;
        boost::replace_all(v, "\"", "\\\"");
        out << "\"" << v << "\" ";
    }
}

void RTLVisitor::VisitVariable(Variable *var)
{
    out << var->id << " ";
    var->meta.Accept(this);
}

void RTLVisitor::VisitUnaryOp(UnaryExpression *e, string oper)
{
    out << "(" << oper << " ";
    e->expr->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitBinaryOp(BinaryExpression *e, string oper)
{
    out << "(" << oper << " ";
    e->lhs->Accept(this);
    e->rhs->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitAbstract(ExprAbstract *e, string oper)
{
    out << "(" << oper << " " << e->bitwidth << " ";
    for (auto o : e->operands)
        o->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprConst(ExprConst *e)
{
    if (e->bitwidth > 0 || e->meta.type != MT_UNDEFINED)
    {
        out << "(_ " << e->value << " " << e->bitwidth;
        if (e->meta.type != MT_UNDEFINED)
        {
            out << " ";
            e->meta.Accept(this);
        }
        out << ") ";
    }
    else
        out << e->value << " ";
}

void RTLVisitor::VisitExprVariable(ExprVariable *e)
{
    ASSERT(e->v, "Undefined referenced variable");
    if (e->meta.type == MT_UNDEFINED)
        out << e->v->id << " ";
    else
    {
        out << "(" << e->v->id << " ";
        e->meta.Accept(this);
        out << ") ";
    }

}

void RTLVisitor::VisitExprRef(ExprRef *e)
{
    ASSERT(e->bitwidth < 0, "Undefined referenced expression");
    e->expr->Accept(this);
    e->meta.Accept(this);
}

void RTLVisitor::VisitExprAssign(ExprAssign *e)
{
    VisitBinaryOp(e, ":=");
}

void RTLVisitor::VisitExprNot(ExprNot *e)
{
    VisitUnaryOp(e, "!");
}

void RTLVisitor::VisitExprOr(ExprOr *e)
{
    out << "(|| ";
    for (auto o : e->operands)
        o->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprAnd(ExprAnd *e)
{
    out << "(&& ";
    for (auto o : e->operands)
        o->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprOrJoin(ExprOrJoin *e)
{
    out << "(|. ";
    for (auto o : e->operands)
        o->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprITE(ExprITE *e)
{
    out << "(? ";
    e->cond->Accept(this);
    e->e_true->Accept(this);
    e->e_false->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprInv(ExprInv *e)
{
    VisitUnaryOp(e, "~");
}

void RTLVisitor::VisitExprBitwiseOr(ExprBitwiseOr *e)
{
    VisitBinaryOp(e, "|");
}

void RTLVisitor::VisitExprBitwiseAnd(ExprBitwiseAnd *e)
{
    VisitBinaryOp(e, "&");
}

void RTLVisitor::VisitExprBitwiseXor(ExprBitwiseXor *e)
{
    VisitBinaryOp(e, "^");
}

void RTLVisitor::VisitExprShiftLeft(ExprShiftLeft *e)
{
    VisitBinaryOp(e, "<<");
}

void RTLVisitor::VisitExprShiftRight(ExprShiftRight *e)
{
    VisitBinaryOp(e, ">>");
}

void RTLVisitor::VisitExprShiftArithRight(ExprShiftArithRight *e)
{
    VisitBinaryOp(e, ">>.");
}

void RTLVisitor::VisitExprEqual(ExprEqual *e)
{
    VisitBinaryOp(e, "==");
}

void RTLVisitor::VisitExprNotEqual(ExprNotEqual *e)
{
    VisitBinaryOp(e, "!=");
}

void RTLVisitor::VisitExprLessThan(ExprLessThan *e)
{
    VisitBinaryOp(e, "<");
}

void RTLVisitor::VisitExprGreaterThan(ExprGreaterThan *e)
{
    VisitBinaryOp(e, ">");
}

void RTLVisitor::VisitExprLessEqual(ExprLessEqual *e)
{
    VisitBinaryOp(e, "<=");
}

void RTLVisitor::VisitExprGreaterEqual(ExprGreaterEqual *e)
{
    VisitBinaryOp(e, ">=");
}

void RTLVisitor::VisitExprSignedLessThan(ExprSignedLessThan *e)
{
    VisitBinaryOp(e, "<.");
}

void RTLVisitor::VisitExprSignedGreaterThan(ExprSignedGreaterThan *e)
{
    VisitBinaryOp(e, ">.");
}

void RTLVisitor::VisitExprSignedLessEqual(ExprSignedLessEqual *e)
{
    VisitBinaryOp(e, "<=.");
}

void RTLVisitor::VisitExprSignedGreaterEqual(ExprSignedGreaterEqual *e)
{
    VisitBinaryOp(e, ">=.");
}

void RTLVisitor::VisitExprUnsignedToSigned(ExprUnsignedToSigned *e)
{
    VisitUnaryOp(e, "-.");
}

void RTLVisitor::VisitExprSignedToUnsigned(ExprSignedToUnsigned *e)
{
    VisitUnaryOp(e, ".-");
}

void RTLVisitor::VisitExprCase(ExprCase *e)
{
    out << "(switch ";
    e->ctrl->Accept(this);
    for (auto c : e->cases)
    {
        out << "(";
        c.first->Accept(this);
        c.second->Accept(this);
        out << ") ";
    }
    if (e->def)
    {
        out << "(";
        e->def->Accept(this);
        out << ") ";
    }
    out << ") ";
}

void RTLVisitor::VisitExprCond(ExprCond *e)
{
    out << "(cond ";
    for (auto c : e->cases)
    {
        out << "(";
        c.first->Accept(this);
        c.second->Accept(this);
        out << ") ";
    }
    if (e->def)
    {
        out << "(";
        e->def->Accept(this);
        out << ") ";
    }
    out << ") ";
}

void RTLVisitor::VisitExprNeg(ExprNeg *e)
{
    VisitUnaryOp(e, "--");
}

void RTLVisitor::VisitExprPlus(ExprPlus *e)
{
    VisitBinaryOp(e, "+");
}

void RTLVisitor::VisitExprMinus(ExprMinus *e)
{
    VisitBinaryOp(e, "-");
}

void RTLVisitor::VisitExprMul(ExprMul *e)
{
    VisitBinaryOp(e, "*");
}

void RTLVisitor::VisitExprDiv(ExprDiv *e)
{
    VisitBinaryOp(e, "/");
}

void RTLVisitor::VisitExprMod(ExprMod *e)
{
    VisitBinaryOp(e, "%");
}

void RTLVisitor::VisitExprIndex(ExprIndex *e)
{
    VisitBinaryOp(e, "[]");
}

void RTLVisitor::VisitExprConcatenate(ExprConcatenate *e)
{
    VisitBinaryOp(e, ".");
}

void RTLVisitor::VisitExprZeroExt(ExprZeroExt *e)
{
    out << "(^^ ";
    e->expr->Accept(this);
    out << e->req_bitwidth << " ";
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprSignExt(ExprSignExt *e)
{
    out << "(^^. ";
    e->expr->Accept(this);
    out << e->req_bitwidth << " ";
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprBitExtract(ExprBitExtract *e)
{
    out << "([.] ";
    e->expr->Accept(this);
    out << e->index << " ";
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprSlice(ExprSlice *e)
{
    out << "([..] ";
    e->expr->Accept(this);
    out << e->from << " ";
    out << e->to << " ";
    e->meta.Accept(this);
    out << ") ";
}

void RTLVisitor::VisitExprAbstract(ExprAbstract *)
{
    WARNING("should not be here: visiting superabstract operation");
}

void RTLVisitor::VisitExprAbsLogic(ExprAbsLogic *e)
{
    VisitAbstract(e, "!!");
}

void RTLVisitor::VisitExprAbsBitwise(ExprAbsBitwise *e)
{
    VisitAbstract(e, "~~");
}

void RTLVisitor::VisitExprAbsRelational(ExprAbsRelational *e)
{
    VisitAbstract(e, "<>");
}

void RTLVisitor::VisitExprAbsVector(ExprAbsVector *e)
{
    VisitAbstract(e, "][");
}

void RTLVisitor::VisitExprAbsArithmetic(ExprAbsArithmetic *e)
{
    VisitAbstract(e, "-+");
}

//////////////////////////////////////////////////////////////////////////////
void SMTLibVisitor::VisitSignal(Signal *s)
{
    out << indentstr();
    // signal identifier
    out << "(declare-fun " << s->id;
    out << suffix();
    out << " () (_ BitVec " << s->bitwidth << "))";
}

void SMTLibVisitor::VisitStorage(Storage *st)
{
    (void)st;
}

void SMTLibVisitor::VisitRegister(Register *r)
{
    out << indentstr();

    ASSERT(r->read_timing.max == r->read_timing.min, "Undeterministic reading time");
    ASSERT(r->write_timing.max == r->write_timing.min, "Undeterministic writing time");
    int delay = r->read_timing.max + r->write_timing.max;

    /*for (auto s : r->q)
    {
        out << "(declare-fun " << s->id;
        int dprimes = primes + delay;
        if (dprimes != 0)
        {
            if (!tag.empty())
                out << "_" << tag;
            out << "_" << dprimes << (dprimes > 0 ? 'p' : 'm');
        }
        out << " () (_ BitVec " << s->bitwidth << ")) ";
    }*/

    if (r->d)
    {
        /*out << "(declare-fun " << r->d->id;
        out << suffix();
        out << " () (_ BitVec " << r->bitwidth << ")) ";*/

        for (auto s : r->q)
        {
            out << "(assert (= " << s->id;
            int dprimes = primes + delay;
            if (dprimes != 0)
            {
                if (!tag.empty())
                    out << "_" << tag;
                out << "_" << abs(dprimes) <<
                    (dprimes > 0 ? 'p' : 'm');
            }

            out << " " << r->d->id;
            out << suffix();
            out << ")) ";
        }
    }
}

void SMTLibVisitor::VisitMemPort(MemPort *port)
{
    out << "(port " << port->id << " ";
    // direction
    if (port->is_read())
        if (port->is_write())
        {
            out << "(dir rw) ";
            out << "(timing " << port->write_timing.min << ") ";
        }
        else
        {
            out << "(dir ro) ";
            out << "(timing " << port->read_timing.min << ") ";
        }
    else
    {
        out << "(dir wo) ";
        out << "(timing " << port->write_timing.min << ") ";
    }
    out << "(en " << port->enable->id << ") ";
    out << "(addr " << port->address->id << ") ";
    // data
    out << "(data ";
    for (auto d : port->data)
        out << d->id << " ";
    out << ") ";
    // control
    if (port->control.size())
    {
        out << "(control ";
        for (auto s : port->control)
            out << s->id << " ";
        out << ") ";
    }
    // stage identification
    if (port->stage >= 0)
        out << "(stage " << system->stages[port->stage] << ") ";
    port->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitMemory(Memory *m)
{
    out << indentstr();
    out << "(mem " << m->id << " " << m->bitwidth << " " << m->addr_bitwidth;
    out << " " << m->size << " ";
    out << "(ports ";
    for (auto port : m->ports)
        VisitMemPort(port);
    out << ") ";
    if (m->rst)
        out << "(rst " << m->rst->id << ") ";
    indent++;
    m->meta.Accept(this);
    indent--;
    out << ") ";
}

void SMTLibVisitor::VisitFunNode(FunNode *f)
{
    out << indentstr();
    out << "(fnode " << f->id << " ";
    out << "(input ";
    for (auto i : f->inputs)
        out << i->id << " ";
    out << ") ";
    out << "(output ";
    for (auto o : f->outputs)
        out << o->id << " ";
    out << ") ";
    if (f->stage >= 0)
        out << "(stage " << system->stages[f->stage] << ") ";
    indent++;
    out << endl << indentstr() << "(assign ";
    for (ExprAssign *a : f->assignments)
        a->Accept(this);
    out << indentstr() << ")";
    f->meta.Accept(this);
    indent--;
    out << ") ";
}

void SMTLibVisitor::VisitSystem(System *s)
{
    system = s;
    out << "(model " << s->name << endl;
    out << "  (pipeline-stages";
    for (auto stage : s->stages)
        out << " " << stage;
    out << ")" << endl;
    indent++;
    indent++;
    s->meta.Accept(this);
    indent--;
    for (auto signal : s->signals)
    {
        signal->Accept(this);
        out << endl;
    }
    for (auto reg : s->registers)
    {
        reg->Accept(this);
        out << endl;
    }
    for (auto mem : s->memories)
    {
        mem->Accept(this);
        out << endl;
    }
    for (auto fun : s->fnodes)
    {
        fun->Accept(this);
        out << endl;
    }
    indent--;
    out << ")" << endl;
}

void SMTLibVisitor::VisitMeta(Meta *meta)
{
    if (meta->type == MT_NODE)
    {
        if (meta->value == "")
            out << endl << indentstr() << "(meta ";
        else
            out << "(" << meta->value << " ";
        indent++;
        meta->AcceptAll(this);
        indent--;
        out << ") ";
    }
    else if (meta->type == MT_INTVALUE)
        out << meta->value << " ";
    else if (meta->type == MT_VALUE)
    {
        string v = meta->value;
        boost::replace_all(v, "\"", "\\\"");
        out << "\"" << v << "\" ";
    }
}

void SMTLibVisitor::VisitVariable(Variable *var)
{
    out << var->id;
    out << suffix();
    out << " ";
    var->meta.Accept(this);
}

void SMTLibVisitor::VisitUnaryOp(UnaryExpression *e, string oper)
{
    out << "(" << oper << " ";
    e->expr->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitBinaryOp(BinaryExpression *e, string oper)
{
    out << "(" << oper << " ";
    e->lhs->Accept(this);
    e->rhs->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitAbstract(ExprAbstract *e, string oper)
{
    (void)oper;
    // TODO: abstract operator replace by constant '0'
    out << "(_ bv0 " << e->bitwidth;
    //for (auto o : e->operands)
    //    o->Accept(this);
    //e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprConst(ExprConst *e)
{
    if (e->bitwidth > 0 || e->meta.type != MT_UNDEFINED)
    {
        out << "(_ bv" << e->value << " " << e->bitwidth;
        if (e->meta.type != MT_UNDEFINED)
        {
            out << " ";
            e->meta.Accept(this);
        }
        out << ") ";
    }
    else
        out << e->value << " ";
}

void SMTLibVisitor::VisitExprVariable(ExprVariable *e)
{
    ASSERT(e->v, "Undefined referenced variable");
    if (e->meta.type == MT_UNDEFINED)
    {
        out << e->v->id;
        out << suffix();
        out << " ";
    }
    else
    {
        out << "(" << e->v->id;
        out << suffix();
        out << " ";
        e->meta.Accept(this);
        out << ") ";
    }

}

void SMTLibVisitor::VisitExprRef(ExprRef *e)
{
    ASSERT(e->bitwidth < 0, "Undefined referenced expression");
    e->expr->Accept(this);
    e->meta.Accept(this);
}

void SMTLibVisitor::VisitExprAssign(ExprAssign *e)
{
    out << "(assert (= ";
    e->lhs->Accept(this);
    e->rhs->Accept(this);
    //e->meta.Accept(this);
    out << ")) ";
}

void SMTLibVisitor::VisitExprNot(ExprNot *e)
{
    VisitUnaryOp(e, "bvnot");
}

void SMTLibVisitor::VisitExprOr(ExprOr *e)
{
    unsigned n = e->operands.size();
    unsigned i = 0;
    unsigned p = 0;
    for (auto o : e->operands)
    {
        ++i;
        if (i < n)
        {
            out << "(bvor ";
            ++p;
        }
        o->Accept(this);
    }
    for (unsigned j = 0; j < p; j++)
    {
        if (j == p - 1)
            e->meta.Accept(this);
        out << ") ";
    }
}

void SMTLibVisitor::VisitExprAnd(ExprAnd *e)
{
    unsigned n = e->operands.size();
    unsigned i = 0;
    unsigned p = 0;
    for (auto o : e->operands)
    {
        ++i;
        if (i < n)
        {
            out << "(bvand ";
            ++p;
        }
        o->Accept(this);
    }
    for (unsigned j = 0; j < p; j++)
    {
        if (j == p - 1)
            e->meta.Accept(this);
        out << ") ";
    }
}

void SMTLibVisitor::VisitExprOrJoin(ExprOrJoin *e)
{
    unsigned n = e->operands.size();
    unsigned i = 0;
    unsigned p = 0;
    for (auto o : e->operands)
    {
        ++i;
        if (i < n)
        {
            out << "(bvor ";
            ++p;
        }
        o->Accept(this);
    }
    for (unsigned j = 0; j < p; j++)
    {
        if (j == p - 1)
            e->meta.Accept(this);
        out << ") ";
    }
}

void SMTLibVisitor::VisitExprITE(ExprITE *e)
{
    out << "(ite (= ";
    e->cond->Accept(this);
    out << "(_ bv1 1)) ";
    e->e_true->Accept(this);
    e->e_false->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprInv(ExprInv *e)
{
    VisitUnaryOp(e, "bvnot");
}

void SMTLibVisitor::VisitExprBitwiseOr(ExprBitwiseOr *e)
{
    VisitBinaryOp(e, "bvor");
}

void SMTLibVisitor::VisitExprBitwiseAnd(ExprBitwiseAnd *e)
{
    VisitBinaryOp(e, "bvand");
}

void SMTLibVisitor::VisitExprBitwiseXor(ExprBitwiseXor *e)
{
    VisitBinaryOp(e, "bvxor");
}

void SMTLibVisitor::VisitExprShiftLeft(ExprShiftLeft *e)
{
    // SMTLib requires both operands of shift to have the same
    // bitwidth.
    out << "(bvshl ";
    e->lhs->Accept(this);
    out << "((_ extract ";
    out << e->lhs->bitwidth - 1 << " 0) ((_ zero_extend ";
    out << e->lhs->bitwidth << ") ";
    e->rhs->Accept(this);
    out << ")) ";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprShiftRight(ExprShiftRight *e)
{
    // SMTLib requires both operands of shift to have the same
    // bitwidth.
    out << "(bvlshr ";
    e->lhs->Accept(this);
    out << "((_ extract ";
    out << e->lhs->bitwidth - 1 << " 0) ((_ zero_extend ";
    out << e->lhs->bitwidth << ") ";
    e->rhs->Accept(this);
    out << ")) ";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprShiftArithRight(ExprShiftArithRight *e)
{
    VisitBinaryOp(e, "bvashr");
}

void SMTLibVisitor::VisitExprEqual(ExprEqual *e)
{
    out << "(ite (= ";
    e->lhs->Accept(this);
    e->rhs->Accept(this);
    out << ") (_ bv1 1) (_ bv0 1)";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprNotEqual(ExprNotEqual *e)
{
    out << "(ite (not (= ";
    e->lhs->Accept(this);
    e->rhs->Accept(this);
    out << ")) (_ bv1 1) (_ bv0 1)";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprLessThan(ExprLessThan *e)
{
    VisitBinaryOp(e, "<");
}

void SMTLibVisitor::VisitExprGreaterThan(ExprGreaterThan *e)
{
    VisitBinaryOp(e, ">");
}

void SMTLibVisitor::VisitExprLessEqual(ExprLessEqual *e)
{
    VisitBinaryOp(e, "<=");
}

void SMTLibVisitor::VisitExprGreaterEqual(ExprGreaterEqual *e)
{
    VisitBinaryOp(e, ">=");
}

void SMTLibVisitor::VisitExprSignedLessThan(ExprSignedLessThan *e)
{
    VisitBinaryOp(e, "<.");
}

void SMTLibVisitor::VisitExprSignedGreaterThan(ExprSignedGreaterThan *e)
{
    VisitBinaryOp(e, ">.");
}

void SMTLibVisitor::VisitExprSignedLessEqual(ExprSignedLessEqual *e)
{
    VisitBinaryOp(e, "<=.");
}

void SMTLibVisitor::VisitExprSignedGreaterEqual(ExprSignedGreaterEqual *e)
{
    VisitBinaryOp(e, ">=.");
}

void SMTLibVisitor::VisitExprUnsignedToSigned(ExprUnsignedToSigned *e)
{
    VisitUnaryOp(e, "-.");
}

void SMTLibVisitor::VisitExprSignedToUnsigned(ExprSignedToUnsigned *e)
{
    VisitUnaryOp(e, ".-");
}

void SMTLibVisitor::VisitExprCase(ExprCase *e)
{
    unsigned p = 0;
    auto last = *(e->cases.begin());
    for (auto c : e->cases)
    {
        out << "(ite ";
        out << "(= ";
        e->ctrl->Accept(this);
        out << " ";
        c.first->Accept(this);
        out << ") ";
        c.second->Accept(this);
        last = c;
        ++p;
    }
    if (e->def)
        e->def->Accept(this);
    else
        last.second->Accept(this);
    for (unsigned j = 0; j < p; j++)
        out << ") ";
}

void SMTLibVisitor::VisitExprCond(ExprCond *e)
{
    unsigned p = 0;
    auto last = *(e->cases.begin());
    for (auto c : e->cases)
    {
        out << "(ite (= ";
        c.first->Accept(this);
        out << " (_ bv1 1)) ";
        c.second->Accept(this);
        last = c;
        ++p;
    }
    if (e->def)
        e->def->Accept(this);
    else
        last.second->Accept(this);
    for (unsigned j = 0; j < p; j++)
        out << ") ";
}

void SMTLibVisitor::VisitExprNeg(ExprNeg *e)
{
    VisitUnaryOp(e, "--");
}

void SMTLibVisitor::VisitExprPlus(ExprPlus *e)
{
    VisitBinaryOp(e, "bvadd");
}

void SMTLibVisitor::VisitExprMinus(ExprMinus *e)
{
    VisitBinaryOp(e, "bvsub");
}

void SMTLibVisitor::VisitExprMul(ExprMul *e)
{
    VisitBinaryOp(e, "bvmul");
}

void SMTLibVisitor::VisitExprDiv(ExprDiv *e)
{
    VisitBinaryOp(e, "bvudiv");
}

void SMTLibVisitor::VisitExprMod(ExprMod *e)
{
    VisitBinaryOp(e, "%");
}

void SMTLibVisitor::VisitExprIndex(ExprIndex *e)
{
    VisitBinaryOp(e, "[]");
}

void SMTLibVisitor::VisitExprConcatenate(ExprConcatenate *e)
{
    VisitBinaryOp(e, "concat");
}

void SMTLibVisitor::VisitExprZeroExt(ExprZeroExt *e)
{
    out << "((_ extract ";
    out << e->req_bitwidth - 1 << " 0) ((_ zero_extend ";
    out << e->req_bitwidth << ") ";
    e->expr->Accept(this);
    out << ") ";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprSignExt(ExprSignExt *e)
{
    out << "((_ extract ";
    out << e->req_bitwidth - 1 << " 0) ((_ sign_extend ";
    out << e->req_bitwidth << ") ";
    e->expr->Accept(this);
    out << ") ";
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprBitExtract(ExprBitExtract *e)
{
    out << "((_ extract ";
    out << e->index << " " << e->index << ") ";
    e->expr->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprSlice(ExprSlice *e)
{
    out << "((_ extract ";
    out << e->to << " ";
    out << e->from << ") ";
    e->expr->Accept(this);
    e->meta.Accept(this);
    out << ") ";
}

void SMTLibVisitor::VisitExprAbstract(ExprAbstract *)
{
    WARNING("should not be here: visiting superabstract operation");
}

void SMTLibVisitor::VisitExprAbsLogic(ExprAbsLogic *e)
{
    VisitAbstract(e, "!!");
}

void SMTLibVisitor::VisitExprAbsBitwise(ExprAbsBitwise *e)
{
    VisitAbstract(e, "~~");
}

void SMTLibVisitor::VisitExprAbsRelational(ExprAbsRelational *e)
{
    VisitAbstract(e, "<>");
}

void SMTLibVisitor::VisitExprAbsVector(ExprAbsVector *e)
{
    VisitAbstract(e, "][");
}

void SMTLibVisitor::VisitExprAbsArithmetic(ExprAbsArithmetic *e)
{
    VisitAbstract(e, "-+");
}



}
