/**
 * @desc Routines for import/export of VAM.
 */
#ifndef _VAMOUTPUT_H
#define _VAMOUTPUT_H

#include "vamrtl.h"

namespace vam {

/****************************************************************************
 * Output
 ***************************************************************************/
class Signal;
class Storage;
class Register;
class Memory;
class MemPort;
class FunNode;
class System;

class RTLVisitor : public VAMVisitor
{
public:
    ostringstream out;
    int indent;

public:
    RTLVisitor(System *s = nullptr) : system(s) { indent = 0; }

    /// RTL specific objects
    virtual void VisitSignal(Signal *var);
    virtual void VisitStorage(Storage *st);
    virtual void VisitRegister(Register *var);
    virtual void VisitMemory(Memory *m);
    virtual void VisitMemPort(MemPort *p);
    virtual void VisitFunNode(FunNode *f);
    virtual void VisitSystem(System *s);

    /// general meta object
    virtual void VisitMeta(Meta *meta) override;

    /// variable objects
    virtual void VisitVariable(Variable *var) override;

    virtual void VisitUnaryOp(UnaryExpression *e, string oper);
    virtual void VisitBinaryOp(BinaryExpression *e, string oper);
    virtual void VisitAbstract(ExprAbstract *e, string oper);

    /// individual expressions
    virtual void VisitExprConst(ExprConst *e) override;
    virtual void VisitExprVariable(ExprVariable *e) override;
    virtual void VisitExprRef(ExprRef *e) override;
    virtual void VisitExprAssign(ExprAssign *e) override;
    virtual void VisitExprNot(ExprNot *e) override;
    virtual void VisitExprOr(ExprOr *e) override;
    virtual void VisitExprAnd(ExprAnd *e) override;
    virtual void VisitExprOrJoin(ExprOrJoin *e) override;
    virtual void VisitExprITE(ExprITE *e) override;
    virtual void VisitExprInv(ExprInv *e) override;
    virtual void VisitExprBitwiseOr(ExprBitwiseOr *e) override;
    virtual void VisitExprBitwiseAnd(ExprBitwiseAnd *e) override;
    virtual void VisitExprBitwiseXor(ExprBitwiseXor *e) override;
    virtual void VisitExprShiftLeft(ExprShiftLeft *e) override;
    virtual void VisitExprShiftRight(ExprShiftRight *e) override;
    virtual void VisitExprShiftArithRight(ExprShiftArithRight *e) override;
    virtual void VisitExprEqual(ExprEqual *e) override;
    virtual void VisitExprNotEqual(ExprNotEqual *e) override;
    virtual void VisitExprLessThan(ExprLessThan *e) override;
    virtual void VisitExprGreaterThan(ExprGreaterThan *e) override;
    virtual void VisitExprLessEqual(ExprLessEqual *e) override;
    virtual void VisitExprGreaterEqual(ExprGreaterEqual *e) override;
    virtual void VisitExprSignedLessThan(ExprSignedLessThan *e) override;
    virtual void VisitExprSignedGreaterThan(ExprSignedGreaterThan *e) override;
    virtual void VisitExprSignedLessEqual(ExprSignedLessEqual *e) override;
    virtual void VisitExprSignedGreaterEqual(ExprSignedGreaterEqual *e)
        override;
    virtual void VisitExprUnsignedToSigned(ExprUnsignedToSigned *e) override;
    virtual void VisitExprSignedToUnsigned(ExprSignedToUnsigned *e) override;
    virtual void VisitExprCase(ExprCase *e) override;
    virtual void VisitExprCond(ExprCond *e) override;
    virtual void VisitExprNeg(ExprNeg *e) override;
    virtual void VisitExprPlus(ExprPlus *e) override;
    virtual void VisitExprMinus(ExprMinus *e) override;
    virtual void VisitExprMul(ExprMul *e) override;
    virtual void VisitExprDiv(ExprDiv *e) override;
    virtual void VisitExprMod(ExprMod *e) override;
    virtual void VisitExprIndex(ExprIndex *e) override;
    virtual void VisitExprConcatenate(ExprConcatenate *e) override;
    virtual void VisitExprZeroExt(ExprZeroExt *e) override;
    virtual void VisitExprSignExt(ExprSignExt *e) override;
    virtual void VisitExprBitExtract(ExprBitExtract *e) override;
    virtual void VisitExprSlice(ExprSlice *e) override;
    virtual void VisitExprAbstract(ExprAbstract *e) override;
    virtual void VisitExprAbsLogic(ExprAbsLogic *e) override;
    virtual void VisitExprAbsBitwise(ExprAbsBitwise *e) override;
    virtual void VisitExprAbsRelational(ExprAbsRelational *e) override;
    virtual void VisitExprAbsVector(ExprAbsVector *e) override;
    virtual void VisitExprAbsArithmetic(ExprAbsArithmetic *e) override;

private:
    System *system;

    string indentstr()
    {
        static string _indentstr[8] =
        {   "",
            "  ",
            "    ",
            "      ",
            "        ",
            "          ",
            "            ",
            "              ",};
        if (indent < 8)
            return _indentstr[indent];
        else
        {
            string s = _indentstr[7];
            for (int i = 7; i < indent; i++)
                s += "  ";
            return s;
        }
    }
};

class SMTLibVisitor : public RTLVisitor
{
public:
    ostringstream out;

public:
    SMTLibVisitor(System *s = nullptr, int p = 0,
        string t = "") :
        system(s), indent(0), primes(p), tag(t),
        suff("")
        {
            ostringstream o;
            if (primes != 0)
            {
                if (!tag.empty())
                    o << "_" << tag;
                o << "_" << abs(primes) <<
                    (primes > 0 ? 'p' : 'm');
            }

            suff = o.str();
        }

    /// RTL specific objects
    virtual void VisitSignal(Signal *var);
    virtual void VisitStorage(Storage *st);
    virtual void VisitRegister(Register *var);
    virtual void VisitMemory(Memory *m);
    virtual void VisitMemPort(MemPort *p);
    virtual void VisitFunNode(FunNode *f);
    virtual void VisitSystem(System *s);

    /// general meta object
    virtual void VisitMeta(Meta *meta) override;

    /// variable objects
    virtual void VisitVariable(Variable *var) override;

    virtual void VisitUnaryOp(UnaryExpression *e, string oper);
    virtual void VisitBinaryOp(BinaryExpression *e, string oper);
    virtual void VisitAbstract(ExprAbstract *e, string oper);

    /// individual expressions
    virtual void VisitExprConst(ExprConst *e) override;
    virtual void VisitExprVariable(ExprVariable *e) override;
    virtual void VisitExprRef(ExprRef *e) override;
    virtual void VisitExprAssign(ExprAssign *e) override;
    virtual void VisitExprNot(ExprNot *e) override;
    virtual void VisitExprOr(ExprOr *e) override;
    virtual void VisitExprAnd(ExprAnd *e) override;
    virtual void VisitExprOrJoin(ExprOrJoin *e) override;
    virtual void VisitExprITE(ExprITE *e) override;
    virtual void VisitExprInv(ExprInv *e) override;
    virtual void VisitExprBitwiseOr(ExprBitwiseOr *e) override;
    virtual void VisitExprBitwiseAnd(ExprBitwiseAnd *e) override;
    virtual void VisitExprBitwiseXor(ExprBitwiseXor *e) override;
    virtual void VisitExprShiftLeft(ExprShiftLeft *e) override;
    virtual void VisitExprShiftRight(ExprShiftRight *e) override;
    virtual void VisitExprShiftArithRight(ExprShiftArithRight *e) override;
    virtual void VisitExprEqual(ExprEqual *e) override;
    virtual void VisitExprNotEqual(ExprNotEqual *e) override;
    virtual void VisitExprLessThan(ExprLessThan *e) override;
    virtual void VisitExprGreaterThan(ExprGreaterThan *e) override;
    virtual void VisitExprLessEqual(ExprLessEqual *e) override;
    virtual void VisitExprGreaterEqual(ExprGreaterEqual *e) override;
    virtual void VisitExprSignedLessThan(ExprSignedLessThan *e) override;
    virtual void VisitExprSignedGreaterThan(ExprSignedGreaterThan *e) override;
    virtual void VisitExprSignedLessEqual(ExprSignedLessEqual *e) override;
    virtual void VisitExprSignedGreaterEqual(ExprSignedGreaterEqual *e)
        override;
    virtual void VisitExprUnsignedToSigned(ExprUnsignedToSigned *e) override;
    virtual void VisitExprSignedToUnsigned(ExprSignedToUnsigned *e) override;
    virtual void VisitExprCase(ExprCase *e) override;
    virtual void VisitExprCond(ExprCond *e) override;
    virtual void VisitExprNeg(ExprNeg *e) override;
    virtual void VisitExprPlus(ExprPlus *e) override;
    virtual void VisitExprMinus(ExprMinus *e) override;
    virtual void VisitExprMul(ExprMul *e) override;
    virtual void VisitExprDiv(ExprDiv *e) override;
    virtual void VisitExprMod(ExprMod *e) override;
    virtual void VisitExprIndex(ExprIndex *e) override;
    virtual void VisitExprConcatenate(ExprConcatenate *e) override;
    virtual void VisitExprZeroExt(ExprZeroExt *e) override;
    virtual void VisitExprSignExt(ExprSignExt *e) override;
    virtual void VisitExprBitExtract(ExprBitExtract *e) override;
    virtual void VisitExprSlice(ExprSlice *e) override;
    virtual void VisitExprAbstract(ExprAbstract *e) override;
    virtual void VisitExprAbsLogic(ExprAbsLogic *e) override;
    virtual void VisitExprAbsBitwise(ExprAbsBitwise *e) override;
    virtual void VisitExprAbsRelational(ExprAbsRelational *e) override;
    virtual void VisitExprAbsVector(ExprAbsVector *e) override;
    virtual void VisitExprAbsArithmetic(ExprAbsArithmetic *e) override;

private:
    System *system;
    int indent;
    int primes;
    string tag;
    string suff;

    string indentstr()
    {
        static string _indentstr[8] =
        {   "",
            "  ",
            "    ",
            "      ",
            "        ",
            "          ",
            "            ",
            "              ",};
        if (indent < 8)
            return _indentstr[indent];
        else
        {
            string s = _indentstr[7];
            for (int i = 7; i < indent; i++)
                s += "  ";
            return s;
        }
    }

    string suffix()
    {
        return suff;
    }
};



}

#endif
