/**
 * @desc RTL query language
 */
#ifndef _RTLQUERY_H
#define _RTLQUERY_H

#include "vamrtl.h"
#include "input.h" // T_EOF
#include <vector>
#include <utility>
#include <memory>

using namespace std;

namespace vam
{
namespace rtl
{

class Query;

using PQuery = shared_ptr<Query>;

enum QueryType
{
    Q_NONE = T_EOF+1,
    Q_END,
    Q_CONST,
    Q_NOP,
    Q_VAR,
    Q_RTLOBJECT,
    Q_HELP,
    Q_PRINT,
    Q_SMTLIB,
    Q_SMTLIB_PRIME,
    Q_SMTLIB_TAG_PRIME,
    Q_STR,
    Q_SELECT,
    Q_SELECT_SIG,
    Q_SELECT_REG,
    Q_SELECT_MEM,
    Q_SELECT_PORT,
    Q_SELECT_FNODE,
    Q_LET,
    Q_UNDEF,
    Q_EXISTS,
    Q_FORALL,
    Q_FROM,
    Q_SIGNAL,
    Q_REGISTER,
    Q_MEMORY,
    Q_MEMPORT,
    Q_FNODE,
    Q_EXPR,
    Q_DECL,
    Q_SET,
    Q_SET_WHERE,
    Q_TUPLE,
    Q_TUPLE_WHERE,
    Q_WHERE,
    Q_ADD,
    Q_APPEND,
    Q_NOT,
    Q_AND,
    Q_OR,
    Q_IMPLIES,
    Q_AT,
    Q_ATINC,
    Q_IN,
    Q_OCCURENCES,
    Q_SUBSET,
    Q_INTERSECTION,
    Q_SETMINUS,
    Q_BIGCAP,
    Q_UNION,
    Q_BIGCUP,
    Q_IF,
    Q_EQUALS,
    Q_SIZE,
    Q_EMPTY,
    Q_ISINSTANCE,
    Q_ISINSTANCE_SIGNAL,
    Q_ISINSTANCE_REGISTER,
    Q_ISINSTANCE_MEMORY,
    Q_ISINSTANCE_MEMPORT,
    Q_ISINSTANCE_FNODE,
    Q_ISINSTANCE_EXPR,
    Q_GT,
    Q_GE,
    Q_LT,
    Q_LE,
    Q_PLUS,
    Q_MINUS,
    Q_MULTIPLY,
    Q_DIVIDE,
    Q_MODULO,
    Q_RTLID,
    Q_EXPRTYPE,
    Q_INPUTS,
    Q_OUTPUTS,
    Q_RTLOUTPUTS,
    Q_REFERENCES,
    Q_PATH,
    Q_PATH_DATA,
    Q_SIGNALPATH,
    Q_DIRECTPATH,
    Q_COI,
    Q_COIB,
    Q_TARGETS,
    Q_SOURCES,
    Q_DATAIN,
    Q_DATAOUT,
    Q_ADDRESS,
    Q_ENABLE,
    Q_BITWIDTH,
    Q_INITVALUE,
    Q_PORTS,
    Q_ISPORT,
    Q_READPORT,
    Q_WRITEPORT,
    Q_METACLEAR,
    Q_METAGET,
    Q_METASET,
    Q_METAUNSET,
    Q_SCRIPT_BEGIN, // begin mark
    Q_LOAD,
    Q_SAVE,
    Q_DATAFLOW,
    Q_SIMPLIFY,
    Q_UNIFYSELECTORS,
    Q_LOAD_VARS,
    Q_DEBUG_BEGIN, // begin mark
    Q_DEBUG_VARIABLES,
    Q_DEBUG_VARSDEP,
    Q_DEBUG_VARINFO,
    Q_DEBUG_NEXTVALUE,
    Q_DEBUG_SETUP,
    Q_DEBUG_VARGROUPS,
    Q_DEBUG_RMBOUND,
    Q_DEBUG_STOP,
    Q_DEBUG_REPAIR,
    Q_SYSTEM_SHELL,

    Q_EXPR_COI,
    Q_EXPR_COIB,
    Q_STALL,
    Q_CLEAR,
};

class Query
{
public:

    using SubQueries = vector<PQuery>;

    QueryType type;
    SubQueries subqueries;
    RTLObject rtlobject;
    string value;

    /**
     * @brief Purpose specific auxiliary integer variable. For instance, index
     * of 'at' function, or RTL expression keyword of 'expr-type' function.
     */
    int aux_int = 0;

public:
    Query(QueryType t = Q_NONE) noexcept
    {
       type = t;
    }

    Query(QueryType t, const Query &q)
    {
        type = t;
        (*this) << q;
    }

    Query(QueryType t, Query &&q)
    {
        type = t;
        (*this) << move(q);
    }

    Query(QueryType t, PQuery subq)
    {
        type = t;
        subqueries.push_back(subq);
    }

    Query(QueryType t, RTLObject rtl)
    {
        type = t;
        rtlobject = rtl;
    }

    Query(QueryType t, string val)
    {
        type = t;
        value = val;
    }

    Query(QueryType t, int val)
    {
        type = t;
        aux_int = val;
    }

    Query(const Query &) = default;
    Query(Query &&) = default;
    // deleted, undesirable
    //Query &operator=(const Query &q) = default;
    Query &operator=(Query &&) = default;

    bool operator==(const Query &q)
    {
        if ((type == q.type) &&
            (rtlobject == q.rtlobject) &&
            (value == q.value) &&
            (aux_int == q.aux_int))
        {
            if (type == Q_SET)
                return compare_set(q.subqueries);
            else
                return compare_tuple(q.subqueries);
        }
        else
            return false;
    }
    bool operator!=(const Query &q)
    {
        return !operator==(q);
    }

    Query &operator<<(const Query &q)
    {
        if (type == Q_SET || type == Q_TUPLE)
            if (q.type == Q_NONE)
                return *this;
        subqueries.push_back(
                make_shared<Query>(q));
        return *this;
    }
    Query &operator<<(Query &&q)
    {
        if (type == Q_SET || type == Q_TUPLE)
            if (q.type == Q_NONE)
                return *this;
        subqueries.push_back(
                make_shared<Query>(std::move(q)));
        return *this;
    }

    // FIXME: provide hash() function and < operator for sorting
    // non-const, because possible rearrangement of items (in case of Q_SET)
    unsigned hash() { return 0; }

private:
    bool compare_tuple(const SubQueries &subqs)
    {
        if (subqueries.size() != subqs.size())
            return false;
        for (unsigned i = 0; i < subqueries.size(); i++)
            if (*subqueries[i] != *subqs[i])
                return false;
        return true;
    }

    bool compare_set(const SubQueries &subqs)
    {
        if (subqueries.size() != subqs.size())
            return false;
        for (auto &subq : subqueries)
        {
            bool found = false;
            for (auto &subq2 : subqs)
                if (*subq == *subq2)
                {
                    found = true;
                    break;
                }
            if (!found)
                return false;
        }
        return true;
    }

};

string query_to_str(const Query &q);

} // namespace rtl
} // namespace vam

#endif
