/**
 * Interpret of RTL queries.
 */
#include "queryinterpret.h"
#include "rtlmanip.h"
#include "output.h"
#include <boost/algorithm/string/predicate.hpp>
#include <map>
#include <set>
#include <utility>
#include <fstream>
#include <stdlib.h> // system, popen
#include <stdio.h> // fgetc
#include <sstream>
#include <iomanip>

namespace vam
{
namespace rtl
{

const string Interpret::tmpvar_prefix = "_tmp";

string debug_query_to_str(const PQuery &q, int n)
{
    if (q == nullptr)
        return "<not-set>";
    else
    {
        string strval = query_to_str(*q);
        if (n > 0 && strval.size() > (unsigned)n)
            return strval.substr(0, n) + " ...";
        else
            return strval;
    }
}

inline string bool2str(bool b)
{
    return b ? "1" : "0";
}

void Interpret::initialize_from_system()
{
    initialize_rtl_set("signals", system.signals);
    initialize_rtl_set("registers", system.registers);
    initialize_rtl_set("memories", system.memories);
    initialize_rtl_set("fnodes", system.fnodes);
    initialize_memports();
}

void Interpret::initialize_memports()
{
    PQuery q = make_shared<Query>(Q_SET);
    for (auto mem : system.memories)
        for (auto port : mem->ports)
            *q << Query{Q_RTLOBJECT, RTLObject(port)};
    auto v = find_var("memports");
    if (v)
    {
        v->reset();
        v->value = q;
        v->generator = nullptr;
    }
    else
    {
        auto v = new VariableInfo{*this, "memports", nullptr};
        v->value = q;
        variables.insert({"memports", v});
    }
}

int Interpret::exec(PQuery &q)
{
    setup_query(q);
    Query out = evaluate_query(q);
    teardown_query();
    return 1;
}

void Interpret::invalidate_variable(string id)
{
    auto v = find_var(id);
    if (v)
        v->reset();
}

void Interpret::setup_query(PQuery &q)
{
    if (!_debug)
        if (q && q->type != Q_NONE)
        {
            // setup 'query_transducers' + 'variables'
            initialize_symboltable(q);
            // apply query_transducers on all generators of
            // bound evaluators
            replace_generator(q);
            replace_generators();
            // setup 'declarations' of bound variables
            VarSet dummy;
            declarations.clear();
            initialize_declarations(q, dummy);
            // setup all variable's
            // 'dependencies','prerequisites','bound_prerequisites'
            initialize_variable_dependencies();
            // setup 'vgroups'
            initialize_vgroups();
        }
}

void Interpret::teardown_query()
{
    if (!_debug)
    {
        // fom each free variable, remove its generator and dependencies,
        // because - consider this:
        // (let use_memories (exists m (in m memories))), where m will be
        // erased
        // note: dependencies will be erased and updated for all variables
        for (auto vit : variables)
        {
            auto v = vit.second;
            if (!v->bound && !v->temp)
                v->generator = nullptr;
        }

        // vgroups are just for bound variables => clear it
        for (auto &vg : vgroups)
            delete vg.second;
        vgroups.clear();

        // erase all bound variables
        for (auto it = variables.begin(); it != variables.end();)
        {
            auto v = it->second;
            if (v->bound)
            {
                variables.erase(it++);
                delete v;
            }
            else
                ++it;
        }

        // reset prerequisites and dependencies
        for (auto vit : variables)
        {
            auto v = vit.second;
            v->dependencies.clear();
            v->prerequisites.clear();
            v->bound_prerequisites.clear();
            v->update_prerequisites();
        }

        // erase all temporary orphans
        erase_tmp_orphans();

        // remove all declarations
        declarations.clear();

        // remove all query transducers
        query_transducers.clear();
    }
}

Query Interpret::evaluate_query(const PQuery &q)
{
    switch (q->type)
    {
        case Q_NONE:
        case Q_END:
        case Q_HELP:
            return Query{Q_NONE};
        /// \{
        // scripting and debug support
        case Q_NOP:
            return execute_nop(q);
        case Q_DEBUG_VARIABLES:
            return execute_debug_listvariables(q);
        case Q_DEBUG_VARSDEP:
            return execute_debug_printvarsdep(q);
        case Q_DEBUG_VARINFO:
            return execute_debug_varinfo(q);
        case Q_DEBUG_NEXTVALUE:
            return execute_debug_nextvalue(q);
        case Q_DEBUG_SETUP:
            return execute_debug_setup(q);
        case Q_DEBUG_STOP:
            return execute_debug_stop(q);
        case Q_DEBUG_REPAIR:
            return execute_debug_repair(q);
        case Q_DEBUG_VARGROUPS:
            return execute_debug_vgroups(q);
        case Q_SYSTEM_SHELL:
            return execute_system_shell(q);
        /// \}
        case Q_CONST:
        case Q_RTLOBJECT:
            return *q;
        case Q_DATAFLOW: return execute_dataflow(q);
        case Q_SIMPLIFY: return execute_simplify(q);
        case Q_UNIFYSELECTORS: return execute_unifyselectors(q);
        case Q_LOAD_VARS: return execute_load_vars(q);
        case Q_VAR: return execute_var(q);
        case Q_PRINT: return execute_print(q);
        case Q_SMTLIB: return execute_smtlib(q);
        case Q_SMTLIB_PRIME: return execute_smtlib_prime(q);
        case Q_SMTLIB_TAG_PRIME: return execute_smtlib_tag_prime(q);
        case Q_STR: return execute_str(q);
        case Q_LET: return execute_let(q);
        case Q_UNDEF: return execute_undef(q);
        case Q_EXISTS: return execute_exists(q);
        case Q_FORALL: return execute_forall(q);
        case Q_SET: return execute_set(q);
        case Q_SET_WHERE: return execute_set_where(q);
        case Q_TUPLE: return execute_tuple(q);
        case Q_TUPLE_WHERE: return execute_tuple_where(q);
        case Q_DECL: return execute_decl(q);
        case Q_ADD: return execute_add(q);
        case Q_APPEND: return execute_append(q);
        case Q_NOT: return execute_not(q);
        case Q_AND: return execute_and(q);
        case Q_OR: return execute_or(q);
        case Q_IMPLIES: return execute_implies(q);
        case Q_AT: return execute_at(q);
        case Q_ATINC: return execute_atinc(q);
        case Q_IN: return execute_in(q);
        case Q_SUBSET: return execute_subset(q);
        case Q_INTERSECTION: return execute_intersection(q);
        case Q_UNION: return execute_union(q);
        case Q_BIGCAP: return execute_bigcap(q);
        case Q_BIGCUP: return execute_bigcup(q);
        case Q_SETMINUS: return execute_setminus(q);
        case Q_IF: return execute_if(q);
        case Q_EQUALS: return execute_equals(q);
        case Q_SIZE: return execute_size(q);
        case Q_EMPTY: return execute_empty(q);
        case Q_ISINSTANCE: return execute_is_instance(q);
        case Q_GT: return execute_gt(q);
        case Q_GE: return execute_ge(q);
        case Q_LT: return execute_lt(q);
        case Q_LE: return execute_le(q);
        case Q_PLUS: return execute_plus(q);
        case Q_MINUS: return execute_minus(q);
        case Q_MULTIPLY: return execute_multiply(q);
        case Q_DIVIDE: return execute_divide(q);
        case Q_MODULO: return execute_modulo(q);
        case Q_EXPRTYPE: return execute_expr_type(q);
        case Q_RTLID: return execute_rtlid(q);
        case Q_INPUTS: return execute_inputs(q);
        case Q_OUTPUTS: return execute_outputs(q);
        case Q_RTLOUTPUTS: return execute_rtloutputs(q);
        case Q_REFERENCES: return execute_references(q);
        case Q_PATH: return execute_path(q, false);
        case Q_PATH_DATA: return execute_path(q, true);
        case Q_SIGNALPATH: return execute_signal_path(q);
        case Q_DIRECTPATH: return execute_direct_path(q);
        case Q_COI: return execute_coi(q);
        case Q_COIB: return execute_coib(q);
        case Q_TARGETS: return execute_targets(q);
        case Q_SOURCES: return execute_sources(q);
        case Q_DATAIN: return execute_data_in(q);
        case Q_DATAOUT: return execute_data_out(q);
        case Q_ADDRESS: return execute_address(q);
        case Q_ENABLE: return execute_enable(q);
        case Q_BITWIDTH: return execute_bitwidth(q);
        case Q_INITVALUE: return execute_initvalue(q);
        case Q_PORTS: return execute_ports(q);
        case Q_ISPORT: return execute_is_port(q);
        case Q_READPORT: return execute_read_port(q);
        case Q_WRITEPORT: return execute_write_port(q);
        case Q_SELECT_SIG: return execute_select_sig(q);
        case Q_SELECT_REG: return execute_select_reg(q);
        case Q_SELECT_MEM: return execute_select_mem(q);
        case Q_SELECT_PORT: return execute_select_port(q);
        case Q_SELECT_FNODE: return execute_select_fnode(q);
        case Q_ISINSTANCE_SIGNAL:
        case Q_ISINSTANCE_REGISTER:
        case Q_ISINSTANCE_MEMPORT:
        case Q_ISINSTANCE_MEMORY:
        case Q_ISINSTANCE_FNODE:
        case Q_ISINSTANCE_EXPR: return execute_is_instance(q);

        case Q_METACLEAR: return execute_meta_clear(q);
        case Q_METAGET: return execute_meta_get(q);
        case Q_METASET: return execute_meta_set(q);
        case Q_METAUNSET: return execute_meta_unset(q);

        // scripting queries
        case Q_LOAD: return execute_load(q);
        case Q_SAVE: return execute_save(q);

        case Q_EXPR_COI: return execute_expr_coi(q);
        case Q_EXPR_COIB: return execute_expr_coib(q);
        case Q_STALL: return execute_stall(q);
        case Q_CLEAR: return execute_clear(q);

        default:
            throw InterpretException("unknown type of query #" +
                    std::to_string(q->type));
    }
}

string Interpret::visit_rtlobject(RTLObject &o)
{
    if (o.type == RTLObject::NONE)
        throw InterpretException("unknown type of RTL object");
    RTLVisitor v;
    o.Accept(&v);
    return v.out.str();
}

Query Interpret::evaluated_query_to_str(Query q)
{
    Query out{Q_CONST};
    char delim;
    switch (q.type)
    {
        case Q_NONE: break;
        case Q_CONST:
            out.value += q.value;
            break;
        case Q_RTLOBJECT:
            out.value += visit_rtlobject(q.rtlobject);
            break;
        case Q_SET:
            out.value += "(set ";
            delim = q.subqueries.size() >= 5 ? '\n' : ' ';
            for (auto const &item : q.subqueries)
            {
                out.value += evaluated_query_to_str(*item).value;
                out.value += delim;
            }
            out.value += ")";
            break;
        case Q_TUPLE:
            out.value += "(tuple ";
            delim = q.subqueries.size() >= 5 ? '\n' : ' ';
            for (auto const &item : q.subqueries)
            {
                out.value += evaluated_query_to_str(*item).value;
                out.value += delim;
            }
            out.value += ")";
            break;
        default:
            throw InterpretException("unknown type of query #" +
                    std::to_string(q.type));
    }
    return out;
}

Query Interpret::evaluated_query_to_smtlib_str(Query q, int nprimes, const string& tag)
{
    Query out{Q_CONST};
    char delim;
    switch (q.type)
    {
        case Q_NONE: break;
        case Q_CONST:
            out.value += q.value;
            break;
        case Q_RTLOBJECT:
            {
                if (q.rtlobject.type == RTLObject::NONE)
                    throw InterpretException("unknown type of RTL object");
                SMTLibVisitor v(nullptr, nprimes, tag);
                q.rtlobject.Accept(&v);
                out.value += v.out.str();
            }
            break;
        case Q_SET:
            delim = '\n';
            for (auto const &item : q.subqueries)
            {
                out.value += evaluated_query_to_smtlib_str(*item, nprimes, tag).value;
                out.value += delim;
            }
            break;
        case Q_TUPLE:
            delim = '\n';
            for (auto const &item : q.subqueries)
            {
                out.value += evaluated_query_to_smtlib_str(*item, nprimes, tag).value;
                out.value += delim;
            }
            break;
        default:
            throw InterpretException("unknown type of query #" +
                    std::to_string(q.type));
    }
    return out;
}

/****************************************************************************
 * \{
 * Query commands
 */

Query Interpret::execute_print(const PQuery &q)
{
    unsigned i;
    for (i = 0; i < q->subqueries.size(); i++)
    {
        if (i) cout << ' ';
        PQuery &subq = q->subqueries[i];
        cout << evaluated_query_to_str(evaluate_query(subq)).value;
    }
    if (i) cout << '\n';
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_smtlib(const PQuery &q)
{
    unsigned i;
    cout << ";;; Begin of SMTLib export" << endl;
    for (i = 0; i < q->subqueries.size(); i++)
    {
        if (i) cout << '\n';
        PQuery &subq = q->subqueries[i];
        cout << evaluated_query_to_smtlib_str(evaluate_query(subq), 0, "").value;
    }
    cout << endl << ";;; End of SMTLib export" << endl;
    if (i) cout << '\n';
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_smtlib_prime(const PQuery &q)
{
    unsigned i;
    cout << ";;; Begin of SMTLib export" << endl;
    for (i = 0; i < q->subqueries.size(); i++)
    {
        if (i) cout << '\n';
        PQuery &subq = q->subqueries[i];
        cout << evaluated_query_to_smtlib_str(evaluate_query(subq), q->aux_int, "").value;
    }
    cout << endl << ";;; End of SMTLib export" << endl;
    if (i) cout << '\n';
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_smtlib_tag_prime(const PQuery &q)
{
    unsigned i;
    cout << ";;; Begin of SMTLib export" << endl;
    for (i = 0; i < q->subqueries.size(); i++)
    {
        if (i) cout << '\n';
        PQuery &subq = q->subqueries[i];
        cout << evaluated_query_to_smtlib_str(evaluate_query(subq), q->aux_int, q->value).value;
    }
    cout << endl << ";;; End of SMTLib export" << endl;
    if (i) cout << '\n';
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_str(const PQuery &q)
{
    Query out{Q_CONST};
    for (unsigned i = 0; i < q->subqueries.size(); i++)
    {
        if (i) out.value += ' ';
        PQuery &subq = q->subqueries[i];
        cout << evaluated_query_to_str(evaluate_query(subq)).value;
    }
    return out;
}

Query Interpret::execute_decl(const PQuery &q)
{
    (void)q;
    throw InterpretException("not implemented decl");
}

Query Interpret::execute_let(const PQuery &q)
{
    auto v = find_var(q->value);
    if (v)
    {
        v->evaluate();
        return *v->value;
    }
    return Query{Q_NONE};
}

Query Interpret::execute_undef(const PQuery &q)
{
    auto v = find_var(q->value);
    if (v)
    {
        erase_variable(v);
        return Query{Q_CONST, "1"};
    }
    return Query{Q_CONST, "0"};
}

Query Interpret::execute_var(const PQuery &q)
{
    auto v = find_var(q->value);
    if (v)
    {
        v->evaluate();
        return *v->value;
    }
    throw InterpretException("unknown identifier '" + q->value + "'");
    return Query{Q_NONE};
}

Query Interpret::execute_exists(const PQuery &q)
{
    bool out = false;
    auto vgroup_it = vgroups.find(q);
    if (vgroup_it != vgroups.end())
    {
        PVariableGroup vg = vgroup_it->second;
        vg->drive_undriven();
        vg->reset();
        unsigned combinations = 0;

        while (vg->next_value_combination())
        {
            combinations++;
            Query operand_pred = evaluate_query(q->subqueries[0]);
            if (query_to_bool(operand_pred))
            {
                out = true;
                break;
            }
        }
        // someone else have already set the bound-variable value, so check only
        // its value wrt. predicate
        if (combinations == 0)
        {
            Query operand_pred = evaluate_query(q->subqueries[0]);
            if (query_to_bool(operand_pred))
                out = true;
        }
    }
    else
    {
        auto v = find_var(q->value);
        out = !v->notset() && !v->novalue();
    }
    return Query{Q_CONST, bool2str(out)};
}

Query Interpret::execute_forall(const PQuery &q)
{
    bool out = true;
    auto vgroup_it = vgroups.find(q);
    if (vgroup_it != vgroups.end())
    {
        PVariableGroup vg = vgroup_it->second;
        vg->drive_undriven();
        unsigned combinations = 0;

        while (vg->next_value_combination())
        {
            combinations++;
            Query operand_pred = evaluate_query(q->subqueries[0]);
            if (!query_to_bool(operand_pred))
            {
                out = false;
                break;
            }
        }
        // someone else have already set the bound-variable value, so check only
        // its value wrt. predicate
        if (combinations == 0)
        {
            Query operand_pred = evaluate_query(q->subqueries[0]);
            if (!query_to_bool(operand_pred))
                out = false;
        }
    }
    else
        throw InterpretException(
                "unable to evaluate 'forall' over bound variable");
    return Query{Q_CONST, bool2str(out)};
}

/// \}

/****************************************************************************
 * \{
 * Query object operations
 */

Query Interpret::execute_set(const PQuery &q)
{
    Query out{Q_SET};
    for (auto &subq : q->subqueries)
    {
        bool found = false;
        for (auto &item : out.subqueries)
            if (*item == *subq)
            {
                found = true;
                break;
            }
        if (!found)
            out << evaluate_query(subq);
    }
    return out;
}

Query Interpret::execute_set_where(const PQuery &q)
{
    Query out{Q_SET};
    auto vgroup_it = vgroups.find(q);
    if (vgroup_it != vgroups.end())
    {
        PVariableGroup vg = vgroups[q];
        vg->drive_undriven();

        while (vg->next_value_combination())
        {
            // evaluate the condition
            Query operand_pred = evaluate_query(q->subqueries[1]);
            if (query_to_bool(operand_pred))
            {
                // evaluate a new item
                Query item = evaluate_query(q->subqueries[0]);
                // insert it to the output
                if (item.type != Q_NONE)
                {
                    bool found = false;
                    for (auto &subq : out.subqueries)
                        if (*subq == item)
                        {
                            found = true;
                            break;
                        }
                    if (!found)
                        out << item;
                }
            }
        }
    }
    else
        throw InterpretException(
            "missing variable(s) to generate result of '(set ... where)'");
    return out;
}

Query Interpret::execute_tuple(const PQuery &q)
{
    Query out{Q_TUPLE};
    for (auto &subq : q->subqueries)
        out << evaluate_query(subq);
    return out;
}

Query Interpret::execute_tuple_where(const PQuery &q)
{
    Query out{Q_TUPLE};
    auto vgroup_it = vgroups.find(q);
    if (vgroup_it != vgroups.end())
    {
        PVariableGroup vg = vgroups[q];
        vg->drive_undriven();

        while (vg->next_value_combination())
        {
            // evaluate the condition
            Query operand_pred = evaluate_query(q->subqueries[1]);
            if (query_to_bool(operand_pred))
                // evaluate a new item and append it
                out << evaluate_query(q->subqueries[0]);
        }
    }
    else
        throw InterpretException(
            "missing variable(s) to generate result of '(tuple ... where)'");
    return out;
}

Query Interpret::execute_add(const PQuery &q)
{
    Query operand_item = evaluate_query(q->subqueries[0]);

    Query out = query_to_set(evaluate_query(q->subqueries[1]));
    bool found = false;
    for (auto &subq : out.subqueries)
        if (operand_item == *subq)
        {
            found = true;
            break;
        }
    if (!found)
        out << operand_item;
    return out;
}

Query Interpret::execute_append(const PQuery &q)
{
    Query operand_item = evaluate_query(q->subqueries[0]);

    Query out = query_to_set(evaluate_query(q->subqueries[1]));
    out << operand_item;
    return out;
}

Query Interpret::execute_not(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);

    bool subq = query_to_bool(operand);
    return Query{Q_CONST, bool2str(!subq)};
}

Query Interpret::execute_and(const PQuery &q)
{
    bool result = true;
    for (auto &subq : q->subqueries)
    {
        Query operand = evaluate_query(subq);
        result = query_to_bool(operand);
        if (!result)
            break;
    }
    return Query{Q_CONST, bool2str(result)};
}

Query Interpret::execute_or(const PQuery &q)
{
    bool result = true;
    for (auto &subq : q->subqueries)
    {
        Query operand = evaluate_query(subq);
        result = query_to_bool(operand);
        if (result)
            break;
    }
    return Query{Q_CONST, bool2str(result)};
}

Query Interpret::execute_implies(const PQuery &q)
{
    bool result = true;
    Query operand_if = evaluate_query(q->subqueries[0]);

    if (query_to_bool(operand_if))
    {
        Query operand_then = evaluate_query(q->subqueries[1]);
        result = query_to_bool(operand_then);
    }
    return Query{Q_CONST, bool2str(result)};
}

Query Interpret::execute_at(const PQuery &q)
{
    Query tuple = query_to_set(evaluate_query(q->subqueries[0]));

    if (q->aux_int >= 0 && q->aux_int < (int)tuple.subqueries.size())
        return *tuple.subqueries[q->aux_int];
    else
        invalid_operand("at", "index within tuple bound");
    return Query{};
}

Query Interpret::execute_atinc(const PQuery &q)
{
    Query tuple = query_to_set(evaluate_query(q->subqueries[0]));

    if (q->aux_int >= 0 && q->aux_int < (int)tuple.subqueries.size())
        return *tuple.subqueries[q->aux_int++];
    else
        return Query{};
}

Query Interpret::execute_in(const PQuery &q)
{
    // check if '(in...)' is generator, if so, returns 1
    if (q->subqueries[0]->type == Q_VAR)
    {
        auto v = find_bound_var(q->subqueries[0]->value);
        if (v)
            return Query{Q_CONST, bool2str(true)};
    }

    Query operand1 = evaluate_query(q->subqueries[0]);
    Query operand2 = query_to_set(evaluate_query(q->subqueries[1]));

    bool out = false;
    for (auto &subq : operand2.subqueries)
        if (operand1 == *subq)
        {
            out = true;
            break;
        }
    return Query{Q_CONST, bool2str(out)};
}

Query Interpret::execute_occurences(const PQuery &q)
{
    Query operand1 = evaluate_query(q->subqueries[0]);
    Query tuple2 = query_to_set(evaluate_query(q->subqueries[1]));

    int occurences = 0;
    for (auto &subq : tuple2.subqueries)
        if (operand1 == *subq)
            occurences++;
    return Query{Q_CONST, std::to_string(occurences)};
}

Query Interpret::execute_subset(const PQuery &q)
{
    bool result = true;
    Query set1 = query_to_set(evaluate_query(q->subqueries[0]));

    if (set1.subqueries.size())
    {
        Query set2 = query_to_set(evaluate_query(q->subqueries[1]));
        for (auto &subq : set1.subqueries)
            if (std::find(set2.subqueries.begin(),
                        set2.subqueries.end(),
                        subq) == set2.subqueries.end())
            {
                result = false;
                break;
            }
    }
    return Query{Q_CONST, bool2str(result)};
}

Query Interpret::intersection_of_subqueries(
        const Query::SubQueries &set_of_sets, bool evaluated)
{
    Query out{Q_SET};

    // there is no set
    if (set_of_sets.size() == 0)
        return out;

    intersection_init(out, set_of_sets[0], evaluated);

    // there is just one set
    if (set_of_sets.size() == 1)
        return out;

    // more than one sets
    for (unsigned i = 1; i < set_of_sets.size(); i++)
    {
        intersection_with(out, set_of_sets[i], evaluated);
        if (out.subqueries.empty())
            break;
    }
    return out;
}

Query Interpret::execute_intersection(const PQuery &q)
{
    return intersection_of_subqueries(q->subqueries, false);
}

Query Interpret::execute_bigcap(const PQuery &q)
{
    Query operand_set = query_to_set(evaluate_query(q->subqueries[0]));
    return intersection_of_subqueries(operand_set.subqueries, true);
}

Query Interpret::union_of_subqueries(const Query::SubQueries &subqueries,
        bool evaluated)
{
    Query out{Q_SET};
    for (auto &subq : subqueries)
    {
        const Query::SubQueries &items = evaluated ?
            subq->subqueries :
            query_to_set(evaluate_query(subq)).subqueries;
        for (auto &item : items)
        {
            bool found = false;
            for (auto &item_set_out : out.subqueries)
            {
                if (*item == *item_set_out)
                {
                    found = true;
                    break;
                }
            }
            if (!found)
                out << *item;
        }
    }
    return out;
}

Query Interpret::execute_union(const PQuery &q)
{
    return union_of_subqueries(q->subqueries, false);
}

Query Interpret::execute_bigcup(const PQuery &q)
{
    Query operand_set = query_to_set(evaluate_query(q->subqueries[0]));
    return union_of_subqueries(operand_set.subqueries, true);
}

Query Interpret::execute_setminus(const PQuery &q)
{
    Query operand1 = query_to_set(evaluate_query(q->subqueries[0]));
    Query out{Q_SET};
    if (operand1.subqueries.empty())
        return out;
    Query operand2 = query_to_set(evaluate_query(q->subqueries[1]));

    for (auto &op1 : operand1.subqueries)
    {
        bool to_insert = true;
        for (auto &op2 : operand2.subqueries)
            if (*op1 == *op2)
            {
                to_insert = false;
                break;
            }
        if (to_insert)
            out.subqueries.push_back(std::move(op1));
    }
    return out;
}

Query Interpret::execute_if(const PQuery &q)
{
    Query operand_if = evaluate_query(q->subqueries[0]);

    if (query_to_bool(operand_if))
        return evaluate_query(q->subqueries[1]);
    else if (q->subqueries.size() > 2)
        return evaluate_query(q->subqueries[2]);
    else
        return Query{Q_NONE};
}

Query Interpret::execute_equals(const PQuery &q)
{
    Query operand1 = evaluate_query(q->subqueries[0]);
    Query operand2 = evaluate_query(q->subqueries[1]);

    return Query{Q_CONST, bool2str(operand1 == operand2)};
}

Query Interpret::execute_size(const PQuery &q)
{
    Query operand_set = query_to_set(evaluate_query(q->subqueries[0]));

    return Query{Q_CONST, std::to_string(operand_set.subqueries.size())};
}

Query Interpret::execute_empty(const PQuery &q)
{
    Query operand_set = query_to_set(evaluate_query(q->subqueries[0]));

    return Query{Q_CONST, bool2str(operand_set.subqueries.empty())};
}

pair<Query,Query> Interpret::prepare_operands_for_binint_operator(
		const PQuery &q, const string err_msg)
{
    Query operand1 = evaluate_query(q->subqueries[0]);
    if (operand1.type != Q_CONST)
        invalid_operand(err_msg, "1st operand as integer value");
    Query operand2 = evaluate_query(q->subqueries[1]);
    if (operand2.type != Q_CONST)
        invalid_operand(err_msg, "2nd operand as integer value");
    return make_pair(operand1, operand2);
}

Query Interpret::execute_gt(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "greater-than (>)");

    return Query{Q_CONST, bool2str(
            query_to_int(operands.first) > query_to_int(operands.second))};
}

Query Interpret::execute_ge(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q,
            "greater-or-equal-than (>=)");

    return Query{Q_CONST, bool2str(
            query_to_int(operands.first) >= query_to_int(operands.second))};
}

Query Interpret::execute_lt(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "less-than (<)");

    return Query{Q_CONST, bool2str(
            query_to_int(operands.first) < query_to_int(operands.second))};
}

Query Interpret::execute_le(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q,
            "less-or-equal-than (>=)");

    return Query{Q_CONST, bool2str(
            query_to_int(operands.first) <= query_to_int(operands.second))};
}

Query Interpret::execute_plus(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "plus (+)");

    return Query{Q_CONST, std::to_string(
            query_to_int(operands.first) + query_to_int(operands.second))};
}

Query Interpret::execute_minus(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "minus (-)");

    return Query{Q_CONST, std::to_string(
            query_to_int(operands.first) - query_to_int(operands.second))};
}

Query Interpret::execute_multiply(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "multiply (*)");

    return Query{Q_CONST, std::to_string(
            query_to_int(operands.first) * query_to_int(operands.second))};
}

Query Interpret::execute_divide(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "divide (/)");

    return Query{Q_CONST, std::to_string(
            query_to_int(operands.first) / query_to_int(operands.second))};
}

Query Interpret::execute_modulo(const PQuery &q)
{
    auto operands = prepare_operands_for_binint_operator(q, "modulo (%)");

    return Query{Q_CONST, std::to_string(
            query_to_int(operands.first) % query_to_int(operands.second))};
}

/// \}

/****************************************************************************
 * \{
 * RTL operations
 */

Query Interpret::execute_select_sig(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);
    if (operand.type != Q_CONST)
        invalid_operand("select-sig", "signal identifier (string)");
    Signal *o = system.signals[operand.value];
    if (o)
        return Query{Q_RTLOBJECT, RTLObject(o)};
    else
        return Query{Q_NONE};
}

Query Interpret::execute_select_reg(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);
    if (operand.type != Q_CONST)
        invalid_operand("select-reg", "register identifier (string)");
    Register *o = system.registers[operand.value];
    if (o)
        return Query{Q_RTLOBJECT, RTLObject(o)};
    else
        return Query{Q_NONE};
}

Query Interpret::execute_select_mem(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);
    if (operand.type != Q_CONST)
        invalid_operand("select-mem", "memory identifier (string)");
    Memory *o = system.memories[operand.value];
    if (o)
        return Query{Q_RTLOBJECT, RTLObject(o)};
    else
        return Query{Q_NONE};
}

Query Interpret::execute_select_port(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);
    if (operand.type != Q_CONST)
        invalid_operand("select-port", "memory port identifier (string)");
    for (auto m : system.memories)
        for (auto p : m->ports)
            if (p->id == operand.value)
                return Query{Q_RTLOBJECT, RTLObject(p)};
    return Query{Q_NONE};
}

Query Interpret::execute_select_fnode(const PQuery &q)
{
    Query operand = evaluate_query(q->subqueries[0]);
    if (operand.type != Q_CONST)
        invalid_operand("select-port", "functional node identifier (string)");
    FunNode *o = system.fnodes[operand.value];
    if (o)
        return Query{Q_RTLOBJECT, RTLObject(o)};
    else
        return Query{Q_NONE};
}

Query Interpret::execute_is_instance(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
        "is-instance", "RTL object");
    bool out;
    switch (q->aux_int)
    {
        case Q_ISINSTANCE_SIGNAL:
            out = ro.type == RTLObject::SIGNAL;
            break;
        case Q_ISINSTANCE_REGISTER:
            out = ro.type == RTLObject::REGISTER;
            break;
        case Q_ISINSTANCE_MEMORY:
            out = ro.type == RTLObject::MEMORY;
            break;
        case Q_ISINSTANCE_MEMPORT:
            out = ro.type == RTLObject::MEMPORT;
            break;
        case Q_ISINSTANCE_FNODE:
            out = ro.type == RTLObject::FNODE;
            break;
        case Q_ISINSTANCE_EXPR:
            out = ro.type == RTLObject::EXPR;
            break;
        default:
            out = false;
    };
    return Query{Q_CONST, bool2str(out)};
}

Query Interpret::execute_expr_type(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::EXPR,
            "expr-type", "RTL expression");
    bool out = ro.expr->op == q->aux_int;
    return Query{Q_CONST, bool2str(out)};
}

template<typename T>
string int_to_hex(T i)
{
  stringstream stream;
  stream << "0x"
         << std::setfill ('0') << std::setw(sizeof(T)*2)
         << std::hex << i;
  return stream.str();
}

Query Interpret::execute_rtlid(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "rtlid", "RTL object");
    Query out{Q_CONST};
    switch (ro.type)
    {
        case RTLObject::SIGNAL:
            out.value = ro.signal->id;
            break;
        case RTLObject::REGISTER:
        case RTLObject::MEMORY:
        case RTLObject::MEMPORT:
        case RTLObject::STORAGE:
            out.value = ro.reg->id;
            break;
        case RTLObject::FNODE:
            out.value = ro.fnode->id;
            break;
        case RTLObject::EXPR:
            out.value = "expr@" + int_to_hex(ro.expr);
            break;
        default:
            out.value = "rtl@" + int_to_hex(ro.pointer);
    }
    return out;
}

Query Interpret::execute_inputs(const PQuery &q)
{
    Query out{Q_SET};
    if (q->subqueries.size())
    {
        RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
                "inputs", "RTL object");
        SignalSet sset;
        switch (ro.type)
        {
            case RTLObject::SIGNAL:
                // returns set of assignments and storages
                if (ro.signal->owner)
                    out << Query{Q_RTLOBJECT, RTLObject{ro.signal->owner}};
                for (auto connection : ro.signal->connections)
                {
                    if (connection.storage->contains_output(ro.signal))
                        out << Query{Q_RTLOBJECT, RTLObject{connection.storage}};
                }
                break;
            case RTLObject::EXPR:
                inputs_fill_usages(out, ro.expr);
                break;
            case RTLObject::FNODE:
                for (auto sig : ro.fnode->inputs)
                    out << Query{Q_RTLOBJECT, RTLObject{sig}};
                break;
            case RTLObject::MEMORY:
            case RTLObject::MEMPORT:
            case RTLObject::REGISTER:
                ro.mem->get_influenced_inputs(
                        nullptr,
                        tfint{-TF_INF, TF_INF},
                        sset);
                for (auto sig : sset)
                    out << Query{Q_RTLOBJECT, RTLObject{sig}};
                break;
            default:
                break;
        }
    }
    else
        invalid_operand("inputs", "RTL object");
    return out;
}

Query Interpret::execute_outputs(const PQuery &q)
{
    Query out{Q_SET};
    if (q->subqueries.size())
    {
        RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
                "outputs", "RTL object");
        SignalSet sset;
        switch (ro.type)
        {
            case RTLObject::EXPR:
                outputs_fill_usages(out, ro.expr->outputs);
                break;
            case RTLObject::SIGNAL:
                outputs_fill_usages(out, ro.signal->references);
                break;
            case RTLObject::FNODE:
                for (auto sig : ro.fnode->outputs)
                    out << Query{Q_RTLOBJECT, RTLObject{sig}};
                break;
            case RTLObject::MEMORY:
            case RTLObject::MEMPORT:
            case RTLObject::REGISTER:
                // TODO FIXME bug? really ro.mem?
                ro.mem->get_influenced_outputs(
                        nullptr,
                        tfint{-TF_INF, TF_INF},
                        sset);
                for (auto sig : sset)
                    out << Query{Q_RTLOBJECT, RTLObject{sig}};
                break;
            default:
                break;
        }
    }
    else
        invalid_operand("outputs", "RTL object");
    return out;
}

Query Interpret::execute_rtloutputs(const PQuery &q)
{
    Query out{Q_SET};
    if (q->subqueries.size())
    {
        RTLObject ro = operand_to_rtl(q, 0, RTLObject::SIGNAL,
                "rtloutputs", "RTL signal");
        if (ro.type != RTLObject::NONE)
        {
            outputs_fill_usages(out, ro.signal->references);
            for (auto connection : ro.signal->connections)
            {
                if (connection.storage->contains_input(ro.signal))
                    out << Query{Q_RTLOBJECT, RTLObject{connection.storage}};
            }
        }
    }
    else
        invalid_operand("rtloutputs", "RTL signal");
    return out;
}

Query Interpret::execute_references(const PQuery &q)
{
    Query expr_set = query_to_set(evaluate_query(q->subqueries[0]));
    Query out{Q_SET};

    SystemManipulation manipulator(system);
    SignalSet sset;

    for (auto &exprq : expr_set.subqueries)
    {
        RTLObject expr = query_to_rtl(exprq, RTLObject::EXPR,
                "references", "set of expressions");

        manipulator.get_signals_referenced_by(expr.expr, sset);
    }

    for (auto signal : sset)
        out << Query{Q_RTLOBJECT, RTLObject{signal}};

    return out;
}

Query Interpret::execute_path(const PQuery &q, bool is_data)
{
    // TODO join execute_path and execute_direct_path
    Query input_set = query_to_set(evaluate_query(q->subqueries[0]));
    Query output_set = query_to_set(evaluate_query(q->subqueries[1]));
    Query out{Q_SET};

    for (auto &isubq : input_set.subqueries)
    {
        RTLObject input_signal = query_to_rtl(isubq, RTLObject::SIGNAL,
                "path", "set of signals");
        if (input_signal.type == RTLObject::NONE)
            continue;
        for (auto &osubq : output_set.subqueries)
        {
            RTLObject output_signal = query_to_rtl(osubq, RTLObject::SIGNAL,
                    "path", "set of signals");
            if (output_signal.type == RTLObject::NONE)
                continue;

            if (is_data)
                find_expression_path_data(input_signal.signal,
                    output_signal.signal, out);
            else
                find_expression_path(input_signal.signal,
                    output_signal.signal, out);
        }
    }

    return out;
}

Query Interpret::execute_direct_path(const PQuery &q)
{
    Query input_set = query_to_set(evaluate_query(q->subqueries[0]));
    Query output_set = query_to_set(evaluate_query(q->subqueries[1]));
    Query out{Q_SET};

    for (auto &isubq : input_set.subqueries)
    {
        RTLObject input_signal = query_to_rtl(isubq, RTLObject::SIGNAL,
                "direct-path", "set of signals");
        if (input_signal.type == RTLObject::NONE)
            continue;
        for (auto &osubq : output_set.subqueries)
        {
            RTLObject output_signal = query_to_rtl(osubq, RTLObject::SIGNAL,
                    "direct-path", "set of signals");
            if (output_signal.type == RTLObject::NONE)
                continue;

            find_expression_direct_path(input_signal.signal,
                    output_signal.signal, out);
        }
    }

    return out;
}

Query Interpret::execute_signal_path(const PQuery &q)
{
    (void)q;
    throw InterpretException("not implemented signal-path");
}

IdSet Interpret::filter_signals_from_query(const Query &operand_set)
{
    IdSet sigids;
    for (auto &subq : operand_set.subqueries)
    {
        RTLObject ro = query_to_rtl(subq, RTLObject::SIGNAL,
                "coi", "set of signals");
        if (ro.type == RTLObject::NONE)
            continue;
        sigids.insert(ro.signal->id);
    }
    return sigids;
}

Query Interpret::execute_coi_dir(const PQuery &q, DataflowDir dir, bool expr)
{
    SystemManipulation manip(system);

    ConeOfInfluence coi;

    Query operand_set = query_to_set(evaluate_query(q->subqueries[0]));

    IdSet sigids = filter_signals_from_query(operand_set);

    manip.select_coi(sigids, q->aux_int, dir, coi);

    Query out{Q_SET};
    if (expr)
        for (auto &exprs : coi.cone_exprs)
            for (auto &expr : exprs)
                out << Query{Q_RTLOBJECT, RTLObject{expr}};
    else
        for (auto &signals : coi.cone_signals)
            for (auto &signal : signals)
                out << Query{Q_RTLOBJECT, RTLObject{signal}};

    return union_of_subqueries(out.subqueries, false);
}

Query Interpret::execute_coi(const PQuery &q)
{
    return execute_coi_dir(q, DF_FORWARD);
}

Query Interpret::execute_coib(const PQuery &q)
{
    return execute_coi_dir(q, DF_BACKWARD);
}

Query Interpret::execute_expr_coi(const PQuery &q)
{
    return execute_coi_dir(q, DF_FORWARD, true);
}

Query Interpret::execute_expr_coib(const PQuery &q)
{
    return execute_coi_dir(q, DF_BACKWARD, true);
}

Query Interpret::execute_stall(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "stall", "RTL storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::NONE:
        case RTLObject::MEMPORT:
        case RTLObject::MEMORY:
            return out;
        case RTLObject::REGISTER:
            if (ro.reg->stall)
                out << Query{Q_RTLOBJECT, RTLObject(ro.reg->stall)};
            return out;
        default:
            invalid_operand("stall", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_clear(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "clear", "RTL storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::NONE:
        case RTLObject::MEMPORT:
        case RTLObject::MEMORY:
            return out;
        case RTLObject::REGISTER:
            if (ro.reg->clr)
                out << Query{Q_RTLOBJECT, RTLObject(ro.reg->clr)};
            return out;
        default:
            invalid_operand("clear", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_targets(const PQuery &q)
{
    (void)q;
    throw InterpretException("not implemented targets");
}

Query Interpret::execute_sources(const PQuery &q)
{
    (void)q;
    throw InterpretException("not implemented sources");
}

Query Interpret::execute_data_in(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "data-in", "RTL storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::REGISTER:
            out << Query{Q_RTLOBJECT, RTLObject(ro.reg->d)};
        case RTLObject::MEMPORT:
            if (ro.port->is_write())
                // should be just one data-in signal
                for (auto sig : ro.port->data)
                    out << Query{Q_RTLOBJECT, RTLObject(sig)};
            return out;
        case RTLObject::MEMORY:
            for (auto port : ro.mem->ports)
                if (port->is_write())
                    // should be just one data-in signal
                    for (auto sig : port->data)
                        out << Query{Q_RTLOBJECT, RTLObject(sig)};
            return out;
        default:
            invalid_operand("data-in", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_data_out(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "data-out", "RTL storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::REGISTER:
            for (auto sig : ro.reg->q)
                out << Query{Q_RTLOBJECT, RTLObject(sig)};
        case RTLObject::MEMPORT:
            if (ro.port->is_read())
                for (auto sig : ro.port->data)
                    out << Query{Q_RTLOBJECT, RTLObject(sig)};
            return out;
        case RTLObject::MEMORY:
            for (auto port : ro.mem->ports)
                if (port->is_read())
                    for (auto sig : port->data)
                        out << Query{Q_RTLOBJECT, RTLObject(sig)};
            return out;
        default:
            invalid_operand("data-out", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_address(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "address", "RTL addressable storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::MEMPORT:
            if (ro.port->address)
                out << Query{Q_RTLOBJECT, RTLObject(ro.port->address)};
            return out;
        case RTLObject::MEMORY:
            for (auto port : ro.mem->ports)
                out << Query{Q_RTLOBJECT, RTLObject(port->address)};
            return out;
        default:
            //invalid_operand("address", "RTL addressable storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_enable(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "enable", "RTL storage");
    Query out{Q_SET};
    switch (ro.type)
    {
        case RTLObject::NONE:
            return out;
        case RTLObject::REGISTER:
            if (ro.reg->we)
                out << Query{Q_RTLOBJECT, RTLObject(ro.reg->we)};
            return out;
        case RTLObject::MEMPORT:
            if (ro.port->enable)
                out << Query{Q_RTLOBJECT, RTLObject(ro.port->enable)};
            return out;
        case RTLObject::MEMORY:
            for (auto port : ro.mem->ports)
                out << Query{Q_RTLOBJECT, RTLObject(port->enable)};
            return out;
        default:
            invalid_operand("enable", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_bitwidth(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::NONE,
            "bitwidth", "RTL storage");
    switch (ro.type)
    {
        case RTLObject::NONE:
            return Query{Q_NONE};
        case RTLObject::SIGNAL:
            return Query{Q_CONST, std::to_string(ro.signal->bitwidth)};
        case RTLObject::REGISTER:
            return Query{Q_CONST, std::to_string(ro.reg->bitwidth)};
        case RTLObject::MEMPORT:
            return Query{Q_CONST, std::to_string(ro.port->bitwidth)};
        case RTLObject::MEMORY:
            return Query{Q_CONST, std::to_string(ro.mem->bitwidth)};
        default:
            invalid_operand("bitwidth", "RTL storage");
            return Query{Q_NONE};
    }
}

Query Interpret::execute_initvalue(const PQuery &q)
{
    // get register from 1st operand
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::REGISTER,
            "initvalue", "RTL register");

    // find its meta specifying init value
    string init_str = ro.reg->meta.keyval_nocreat("init");

    // convert string meta value to long long
    long long init_val;
    try { init_val = stoll(init_str); }
    catch (...) { init_val = 0; }

    return Query{Q_CONST, std::to_string(init_val)};
}

Query Interpret::execute_ports(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::MEMORY,
            "ports", "RTL memory");
    Query out{Q_SET};
    for (auto port : ro.mem->ports)
        out << Query{Q_RTLOBJECT, RTLObject(port)};
    return out;
}

Query Interpret::execute_is_port(const PQuery &q)
{
    (void)q;
    throw InterpretException("not implemented is-port");
}

Query Interpret::execute_read_port(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::MEMPORT,
            "read-port", "RTL memory port");
    return Query{Q_CONST, bool2str(ro.port->is_read())};
}

Query Interpret::execute_write_port(const PQuery &q)
{
    RTLObject ro = operand_to_rtl(q, 0, RTLObject::MEMPORT,
            "write-port", "RTL memory port");
    return Query{Q_CONST, bool2str(ro.port->is_write())};
}

Meta *Interpret::get_meta_of_rtlnode(const PQuery &q,
        int op_idx, const char *mname)
{
    RTLObject ro = operand_to_rtl(q, op_idx, RTLObject::NONE,
            mname, "RTL object");
    Meta *meta = nullptr;
    switch (ro.type)
    {
        case RTLObject::SIGNAL:
            meta = &ro.signal->meta;
            break;
        case RTLObject::REGISTER:
        case RTLObject::MEMORY:
        case RTLObject::MEMPORT:
        case RTLObject::STORAGE:
            meta = &ro.storage->meta;
            break;
        case RTLObject::FNODE:
            meta = &ro.fnode->meta;
            break;
        case RTLObject::EXPR:
            meta = &ro.expr->meta;
            break;
        default:
            // this should not happen
            invalid_operand(mname, "RTL object");
    }
    return meta;
}

Query Interpret::execute_meta_clear(const PQuery &q)
{
    Meta *meta = get_meta_of_rtlnode(q, 0, "meta-clear");
    assert(meta);
    meta->clear();
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_meta_get(const PQuery &q)
{
    const char *mname = "getattr";
    Meta *meta = get_meta_of_rtlnode(q, 0, mname);
    Query attrname = evaluate_query(q->subqueries[1]);
    if (attrname.type != Q_CONST)
        invalid_operand(mname, "attribute name (string)");

    assert(meta);
    string value = meta->keyval_nocreat(attrname.value);
    return Query{Q_CONST, value};
}

Query Interpret::execute_meta_set(const PQuery &q)
{
    const char *mname = "setattr";
    Meta *meta = get_meta_of_rtlnode(q, 0, mname);
    Query attrname = evaluate_query(q->subqueries[1]);
    if (attrname.type != Q_CONST)
        invalid_operand(mname, "attribute name (string)");
    Query attrval = evaluate_query(q->subqueries[2]);
    if (attrval.type != Q_CONST)
        invalid_operand(mname, "attribute value (string)");

    assert(meta);
    meta->keyval(attrname.value) = attrval.value;
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_meta_unset(const PQuery &q)
{
    const char *mname = "unsetattr";
    Meta *meta = get_meta_of_rtlnode(q, 0, mname);
    Query attrname = evaluate_query(q->subqueries[1]);
    if (attrname.type != Q_CONST)
        invalid_operand(mname, "attribute name (string)");

    assert(meta);
    meta->find_erase(attrname.value);
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_nop(const PQuery &q)
{
    for (auto &subq : q->subqueries)
        evaluate_query(subq);
    return Query{Q_CONST, "1"};
}

void Interpret::_substitute_rtlobjects_with_none(PQuery &q)
{
    if (q->type == Q_RTLOBJECT)
        q->type = Q_NONE;
    else
        for (auto &subq : q->subqueries)
            _substitute_rtlobjects_with_none(subq);
}

void Interpret::substitute_rtlobjects_with_none()
{
    for (auto vit : variables)
    {
        auto v = vit.second;
        if (v->value)
            _substitute_rtlobjects_with_none(v->value);
        if (v->generator)
            _substitute_rtlobjects_with_none(v->generator);
    }
}

Query Interpret::execute_load(const PQuery &q)
{
    substitute_rtlobjects_with_none();
    RTLParser p;
    system.clear();
    p.parse_from_file(q->value, &system);

    initialize_from_system();

    return Query{Q_CONST, "1"};
}

Query Interpret::execute_save(const PQuery &q)
{
    RTLVisitor v;
    system.Accept(&v);

    ofstream o(q->value);
    o << v.out.str() << endl;
    o.close();

    return Query{Q_CONST, "1"};
}

RTLObject Interpret::dequeue_unique(RTLObjectList& rol)
{
    RTLObject ro = rol.front();
    rol.pop_front();
    return ro;
}

void Interpret::enqueue_unique(RTLObjectList& rol, RTLObject ro)
{
    // Check for presence of the currently enlisted item
    for (auto o : rol)
        if (o == ro)
            return;

    // Object is not in list, add it.
    rol.push_back(ro);
}

void Interpret::set_fvs(RTLObjectFVSetMap& fvss, RTLObject ro, const FVSet& fvs)
{
    fvss[ro.pointer] = fvs;
}

Interpret::FVSet Interpret::get_fvs(RTLObjectFVSetMap& fvss, RTLObject ro)
{
    if (fvss.find(ro.pointer) != fvss.end())
        return fvss[ro.pointer];
    else
        return FVSet();
}

string Interpret::join_fvs(const FVSet& fvs, char delim)
{
    stringstream ss;
    bool first = true;
    for (auto fv : fvs)
    {
        if (first)
        {
            ss << to_string(fv);
            first = false;
        }
        else
        {
            ss << delim;
            ss << to_string(fv);
        }
    }

    return ss.str();
}

Interpret::FVSet Interpret::split_fvs(const std::string& meta, char delim)
{
    std::stringstream ss(meta);
    std::string item;
    FVSet fvs;
    while (std::getline(ss, item, delim))
        fvs.insert(atoi(item.c_str()));

    return fvs;
}

string print_set(const set<int>& set)
{
    stringstream ss;
    ss << "{";
    for (int i : set)
        ss << " " << i;
    ss << " }";
    return ss.str();
}

void Interpret::set_fvs_meta(RTLObject ro, const FVSet& fvs,
    const df_type& type)
{
    // meta tag string
    string meta_name = DF_STR_MAP.find(type)->second;
    set_meta(ro, meta_name, join_fvs(fvs, ' '));
}

Interpret::FVSet Interpret::get_fvs_meta(RTLObject ro, const df_type& type)
{
    // meta tag string
    string meta_name = DF_STR_MAP.find(type)->second;
    return split_fvs(get_meta(ro, meta_name), ' ');
}

bool Interpret::update_fvs(RTLObjectFVSetMap& fvss, RTLObject ro,
    const FVSet& fvs, const df_type& type)
{
    FVSet old_fvs = get_fvs(fvss, ro);
    FVSet new_fvs;
    // compute new value according to type of data-flow analysis
    switch(type)
    {
        case DF_PHS:
            // select minimum from old and new value
            if (fvs.empty())
            {
                cerr << "warning: Unexpected flow value of RTLObject occurred!"
                    << endl;
                return false;
            }
            if (old_fvs.empty())
                new_fvs.insert(*(fvs.begin()));
            else
                new_fvs.insert(min(*(old_fvs.begin()), *(fvs.begin())));
            break;
        case DF_BCK:
        case DF_FWD:
            // join elements of the old and new set
            for (const fv_type& fv : old_fvs)
                new_fvs.insert(fv);
            for (const fv_type& fv : fvs)
                new_fvs.insert(fv);
            break;
        default:
            // handle uknown type of data-flow operation
            cerr << "warning: Unimplemented update operation!"
                << endl;
            return false;

            break;
    }
    set_fvs(fvss, ro, new_fvs);
    bool updated = old_fvs != new_fvs;

    // Meta update
    if (updated)
        set_fvs_meta(ro, new_fvs, type);

    return updated;
}

bool Interpret::phs_dataflow(RTLObject pc)
{
    // Initialize a work list with the program counter
    RTLObjectList wl;
    enqueue_unique(wl, pc);

    // Flow value map
    RTLObjectFVSetMap fvss;
    // Set initial flow value of the program counter
    update_fvs(fvss, pc, {1}, DF_PHS);

    // Mark program counter with meta tag 'pc'
    set_meta(pc, S_PC, "1");

    // Main loop until the work list is empty
    while(!wl.empty())
    {
        // Pop the first object
        RTLObject ro = dequeue_unique(wl);
        // Get data-flow value
        FVSet fvs = get_fvs(fvss, ro);
        fv_type fv = *(fvs.begin());
        // Influenced signals
        SignalSet sset;
        switch (ro.type)
        {
            case RTLObject::SIGNAL:
                // Check inputs of registers
                // Update is of the form:
                //
                //      signal       register
                //     -- x -->| min(old, x + wt) |
                //
                for (Register* r : system.registers)
                    if (r->contains_input(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                    {
                        bool updated = update_fvs(fvss, r,
                            {fv + r->write_timing.min}, DF_PHS);
                        if (updated && !is_arch(r))
                            enqueue_unique(wl, r);
                    }

                // Check inputs of memory ports
                // Update is of the form:
                //
                //      signal       wport
                //     -- x -->| min(old, x + wt) |
                //
                //      signal       rport
                //     -- x -->| min(old, x)      |
                //
                for (Memory* m : system.memories)
                    for (MemPort* p : m->ports)
                        if (p->contains_input(ro.signal,
                            tfint{-TF_INF, TF_INF}))
                        {
                            fv_type new_fv = p->is_read() ? fv :
                                (fv + p->write_timing.min);
                            // TODO: Consider stopping propagation for write ports ...
                            if (update_fvs(fvss, p, {new_fv}, DF_PHS))
                                enqueue_unique(wl, p);
                        }

                // Check inputs of function nodes
                // Update is of the form:
                //
                //      signal       fnode
                //     -- x -->(  min(old, x)   )
                //
                for (FunNode* f : system.fnodes)
                    if (f->contains_input(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                        if (update_fvs(fvss, f, {fv}, DF_PHS))
                            enqueue_unique(wl, f);
                break;

            case RTLObject::REGISTER:
                // Update is of the form:
                //
                //     register    signal
                //     |  x  |-- min(old, x + rt) -->
                //
                ro.reg->get_influenced_outputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                    if (update_fvs(fvss, s,
                        {fv + ro.reg->read_timing.min}, DF_PHS))
                        enqueue_unique(wl, s);
                break;

            case RTLObject::MEMPORT:
                // Update is of the form:
                //
                //      wport      rport
                //     |  x  || min(old, x) |
                //
                //      rport      signal
                //     |  x  |-- min(old, x + rt) -->
                //
                if (ro.port->is_read())
                {
                    ro.port->get_influenced_outputs(nullptr,
                        tfint{-TF_INF, TF_INF}, sset);
                    for (Signal *s : sset)
                        if (update_fvs(fvss, s,
                            {fv + ro.port->read_timing.min}, DF_PHS))
                            enqueue_unique(wl, s);
                }
                else
                    for (MemPort *p : ro.port->owner->ports)
                        if (p->is_read())
                            if (update_fvs(fvss, p, {fv}, DF_PHS))
                                enqueue_unique(wl, p);
                break;

            case RTLObject::FNODE:
                // Update is of the form:
                //
                //      fnode      signal
                //     (  x  )-- min(old, x) -->
                //
                ro.fnode->get_influenced_outputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                    if (update_fvs(fvss, s, {fv}, DF_PHS))
                        enqueue_unique(wl, s);
                break;

            // Remaining types are unexpected
            case RTLObject::STORAGE:
            case RTLObject::EXPR:
            case RTLObject::MEMORY:
            case RTLObject::SYSTEM:
                cerr << "error: Unexpected type of RTLObject occurred!"
                    << endl;
                return false;
            default:
                cerr << "error: Unknown type of RTLObject occurred!"
                    << endl;
                return false;
        }
    }

    return true;
}

bool Interpret::bck_dataflow()
{
    // Compute a set of stages reading from storage, e.g.,
    //
    // | x | ---+--> | k |
    //          |
    //          +--> ( l ) --> | m |
    //
    // rphs(x) := { ph(k)-wt(k), ph(l), ph(m)-wt(m) }
    // rphs(l) := { ph(m)-wt(m) }

    // Work list
    RTLObjectList wl;
    // Flow value set map
    RTLObjectFVSetMap fvss;

    // Initialize a work list with all registers and memports
    for (Register* r : system.registers)
        enqueue_unique(wl, r);
    for (Memory* m : system.memories)
        for (MemPort* p : m->ports)
            enqueue_unique(wl, p);

    // Main loop until the work list is empty
    while(!wl.empty())
    {
        // Pop the first object
        RTLObject ro = dequeue_unique(wl);
        // Get data-flow value
        FVSet fvs = get_fvs(fvss, ro);
        // Influenced signals
        SignalSet sset;
        switch (ro.type)
        {
            case RTLObject::SIGNAL:
                // Check inputs of registers
                // Update is of the form:
                //
                //         register      signal
                //     | old U X-rt(r) |-- X -->
                //
                for (Register* r : system.registers)
                    if (r->contains_output(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                    {
                        FVSet new_fvs;
                        for (fv_type fv : fvs)
                            new_fvs.insert(max(0, fv - r->read_timing.min));
                        if (update_fvs(fvss, r, new_fvs, DF_BCK))
                            enqueue_unique(wl, r);
                    }

                // Check inputs of memory ports
                // Update is of the form:
                //
                //           rport       signal
                //     | old U X-rt(p) |-- X -->
                //
                // Note: case of wport is impossible
                //
                for (Memory* m : system.memories)
                    for (MemPort* p : m->ports)
                        if (p->contains_output(ro.signal,
                            tfint{-TF_INF, TF_INF}))
                        {
                            if (p->is_read())
                            {
                                FVSet new_fvs;
                                for (fv_type fv : fvs)
                                    new_fvs.insert(max(0, fv - p->read_timing.min));
                                if (update_fvs(fvss, p, new_fvs, DF_BCK))
                                    enqueue_unique(wl, p);
                            }
                            /*else
                            {
                                cerr << "error: Unexpected type of RTLObject occurred!"
                                    << endl;
                                return false;
                            }*/
                        }

                // Check inputs of function nodes
                // Update is of the form:
                //
                //         fnode     signal
                //     (  old U X  )-- X -->
                //
                for (FunNode* f : system.fnodes)
                    if (f->contains_output(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                        if (update_fvs(fvss, f, fvs, DF_BCK))
                            enqueue_unique(wl, f);
                break;

            case RTLObject::REGISTER:
                // Update is of the form:
                //
                //               signal            register
                //         old U ph(r) == 0 ?      |     |
                //     --       X-wt(r) :       -->|  X  |
                //           { ph(r)-wt(r) }       |     |
                //
                // ad: X-wr(r) ... decrement is pointwise
                //
                ro.reg->get_influenced_inputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                {
                    // determine phase of the processed register
                    FVSet phs = get_fvs_meta(ro, DF_PHS);

                    // compute data-flow value set
                    FVSet new_fvs;
                    if (phs.empty())
                        for (fv_type fv : fvs)
                            new_fvs.insert(max(0, fv - ro.reg->write_timing.min));
                    else
                    {
                        fv_type ph = *(phs.begin());
                        new_fvs = { ph - ro.reg->write_timing.min };
                    }

                    // update the flow value
                    if (update_fvs(fvss, s, new_fvs, DF_BCK))
                        enqueue_unique(wl, s);
                }
                break;

            case RTLObject::MEMPORT:
                // Update is of the form:
                //
                //                   wport                     rport
                //     |            old U X                  ||  X  |
                //
                //                   signal                    rport
                //     -- old U ph(p) == 0 ? X : { ph(p) } -->|  X  |
                //
                //                   signal                    wport
                //                old U ph(p) == 0 ?          |     |
                //     --             X-wt(p) :            -->|  X  |
                //                 { ph(p)-wt(p) }            |     |
                //
                if (ro.port->is_write())
                {
                    ro.port->get_influenced_inputs(nullptr,
                        tfint{-TF_INF, TF_INF}, sset);
                    for (Signal *s : sset)
                    {
                        // determine phase of the processed port
                        FVSet phs = get_fvs_meta(ro, DF_PHS);

                        // compute data-flow value set
                        FVSet new_fvs;
                        if (phs.empty())
                            for (fv_type fv : fvs)
                                new_fvs.insert(max(0, fv - ro.port->write_timing.min));
                        else
                        {
                            fv_type ph = *(phs.begin());
                            new_fvs = { ph - ro.port->write_timing.min };
                        }

                        // update the flow value
                        if (update_fvs(fvss, s, new_fvs, DF_BCK))
                            enqueue_unique(wl, s);
                    }
                }
                else /* read port */
                {
                    // add write port first
                    for (MemPort *p : ro.port->owner->ports)
                        if (p->is_write())
                            if (update_fvs(fvss, p, fvs, DF_BCK))
                                enqueue_unique(wl, p);

                    // FIXME: maybe change get_influenced_inputs() of memport
                    // to conform to the following code
                    if (ro.port->enable)
                        sset.insert(ro.port->enable);
                    if (ro.port->address)
                        sset.insert(ro.port->address);
                    sset.merge(ro.port->control);
                    sset.merge(ro.port->data);

                    for (Signal *s : sset)
                    {
                        // determine phase of the processed register
                        FVSet phs = get_fvs_meta(ro, DF_PHS);

                        // compute data-flow value set
                        FVSet new_fvs;
                        if (phs.empty())
                            for (fv_type fv : fvs)
                                new_fvs.insert(fv);
                        else
                        {
                            fv_type ph = *(phs.begin());
                            new_fvs = { ph };
                        }

                        // update the flow value
                        if (update_fvs(fvss, s, new_fvs, DF_BCK))
                            enqueue_unique(wl, s);
                    }
                }
                break;

            case RTLObject::FNODE:
                // Update is of the form:
                //
                //         signal     fnode
                //     -- old U X -->(  X  )
                //
                ro.fnode->get_influenced_inputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                    if (update_fvs(fvss, s, fvs, DF_BCK))
                        enqueue_unique(wl, s);
                break;

            // Remaining types are unexpected
            case RTLObject::STORAGE:
                cerr << "error: Unexpected type 'STORAGE' of RTLObject occurred!"
                    << endl;
                return false;
            case RTLObject::EXPR:
                cerr << "error: Unexpected type 'EXPR' of RTLObject occurred!"
                    << endl;
                return false;
            case RTLObject::MEMORY:
                cerr << "error: Unexpected type 'MEMORY' of RTLObject occurred!"
                    << endl;
                return false;
            case RTLObject::SYSTEM:
                cerr << "error: Unexpected type 'SYSTEM' of RTLObject occurred!"
                    << endl;
                return false;
            default:
                cerr << "error: Unknown type of RTLObject occurred!"
                    << endl;
                return false;
        }
    }

    return true;
}

bool Interpret::fwd_dataflow()
{
    // Compute a set of stages writing to a storage, e.g.,
    //
    // | x | ---+--> | k |
    //          |
    //          +--> ( l ) --> | m |
    //
    // wphs(k) := { ph(x)+rt(x) }
    // wphs(m) := { ph(l), ph(x)+rt(x) }

    // Work list
    RTLObjectList wl;
    // Flow value set map
    RTLObjectFVSetMap fvss;

    // Initialize a work list with all pipeline registers
    for (Register* r : system.registers)
        if (is_pipe(r))
            enqueue_unique(wl, r);
    /*for (Memory* m : system.memories)
        for (MemPort* p : m->ports)
            enqueue_unique(wl, p);*/

    // Main loop until the work list is empty
    while(!wl.empty())
    {
        // Pop the first object
        RTLObject ro = dequeue_unique(wl);
        // Get data-flow value
        FVSet fvs = get_fvs(fvss, ro);
        // Influenced signals
        SignalSet sset;
        switch (ro.type)
        {
            case RTLObject::SIGNAL:
                // Check inputs of registers
                // Update is of the form:
                //
                //      signal     register
                //     -- X -->| old U X+wt(r) |
                //
                for (Register* r : system.registers)
                    if (r->contains_input(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                    {
                        FVSet new_fvs;
                        for (fv_type fv : fvs)
                            new_fvs.insert(fv + r->write_timing.min);
                        update_fvs(fvss, r, new_fvs, DF_FWD);
                    }

                // Check inputs of memory ports
                // Update is of the form:
                //
                //      signal       wport
                //     -- X -->| old U X+wt(p) |
                //
                //      signal       rport
                //     -- X -->|    old U X    |
                //
                for (Memory* m : system.memories)
                    for (MemPort* p : m->ports)
                        if (p->contains_input(ro.signal,
                            tfint{-TF_INF, TF_INF}))
                        {
                            FVSet new_fvs;
                            for (fv_type fv : fvs)
                                new_fvs.insert(p->is_read() ? fv :
                                    (fv + p->write_timing.min));
                            update_fvs(fvss, p, new_fvs, DF_FWD);
                        }

                // Check inputs of function nodes
                // Update is of the form:
                //
                //      signal       fnode
                //     -- X -->(    old U X    )
                //
                for (FunNode* f : system.fnodes)
                    if (f->contains_input(ro.signal,
                        tfint{-TF_INF, TF_INF}))
                        if (update_fvs(fvss, f, fvs, DF_FWD))
                            enqueue_unique(wl, f);
                break;

            case RTLObject::REGISTER:
                // Update is of the form:
                //
                //     register         signal
                //     |  X  |-- old U { ph(r)+rt(r) } -->
                //
                ro.reg->get_influenced_outputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                {
                    // determine phase of the processed register
                    FVSet phs = get_fvs_meta(ro, DF_PHS);

                    // compute data-flow value set
                    FVSet new_fvs;
                    if (!phs.empty())
                    {
                        fv_type ph = *(phs.begin());
                        new_fvs = { ph + ro.reg->read_timing.min };
                    }
                    if (update_fvs(fvss, s, new_fvs, DF_FWD))
                        enqueue_unique(wl, s);
                }
                break;

            case RTLObject::MEMPORT:
                // Update is of the form:
                //
                //      rport      signal
                //     |  X  |-- old U { ph(p)+rt(p) } -->
                //
                // NOTE: Now omitted:
                //      wport      rport
                //     |  X  ||  old U X  |
                //
                if (ro.port->is_read())
                {
                    ro.port->get_influenced_outputs(nullptr,
                        tfint{-TF_INF, TF_INF}, sset);
                    for (Signal *s : sset)
                    {
                        // determine phase of the processed port
                        FVSet phs = get_fvs_meta(ro, DF_PHS);

                        // compute data-flow value set
                        FVSet new_fvs;
                        if (!phs.empty())
                        {
                            fv_type ph = *(phs.begin());
                            new_fvs = { ph + ro.port->read_timing.min };
                        }
                        if (update_fvs(fvss, s, new_fvs, DF_FWD))
                            enqueue_unique(wl, s);
                    }
                }
                //else /* write port */
                //    for (MemPort *p : ro.port->owner->ports)
                //        if (p->is_read())
                //            if (update_fvs(fvss, p, fvs, DF_FWD))
                //                enqueue_unique(wl, p);
                break;

            case RTLObject::FNODE:
                // Update is of the form:
                //
                //      fnode     signal
                //     (  X  )-- old U X -->
                //
                ro.fnode->get_influenced_outputs(nullptr,
                    tfint{-TF_INF, TF_INF}, sset);
                for (Signal *s : sset)
                    if (update_fvs(fvss, s, fvs, DF_FWD))
                        enqueue_unique(wl, s);
                break;

            // Remaining types are unexpected
            case RTLObject::STORAGE:
            case RTLObject::EXPR:
            case RTLObject::MEMORY:
            case RTLObject::SYSTEM:
                cerr << "error: Unexpected type of RTLObject occurred!"
                    << endl;
                return false;
            default:
                cerr << "error: Unknown type of RTLObject occurred!"
                    << endl;
                return false;
        }
    }

    return true;
}

bool Interpret::finalize_dataflow()
{
    list<RTLObject> ros;
    for (Register* r : system.registers)
        ros.push_back(r);
    for (Memory* m : system.memories)
        for (MemPort* p : m->ports)
            ros.push_back(p);
    for (Signal* s : system.signals)
        ros.push_back(s);
    for (FunNode* f : system.fnodes)
        ros.push_back(f);

    for (RTLObject ro : ros)
    {
        FVSet phs  = get_fvs_meta(ro, DF_PHS);
        FVSet rphs = get_fvs_meta(ro, DF_BCK);

        // Compute phase from reading phase
        if (phs.empty() && !rphs.empty())
            set_fvs_meta(ro, { *min_element(rphs.begin(), rphs.end()) },
                DF_PHS);
    }

    // Determine potential hazards for architectural registers
    for (Register* r : system.registers)
    {
        RTLObject ro(r);

        FVSet rphs = get_fvs_meta(ro, DF_BCK);
        FVSet wphs = get_fvs_meta(ro, DF_FWD);

        if (is_arch(ro))
        {
            // RAW hazards
            for (fv_type rph : rphs)
                for (fv_type wph : wphs)
                    if (rph < wph)
                    {
                        set_meta(ro, "rawh", "1");
                        break;
                    }

            // WAW hazards
            if (wphs.size() > 1)
                set_meta(ro, "wawh", "1");

            // WAR hazards
            for (fv_type rph : rphs)
                for (fv_type wph : wphs)
                    if (rph >= wph)
                    {
                        set_meta(ro, "warh", "1");
                        break;
                    }
        }
    }

    // Determine potential hazards for memories
    for (Memory* m : system.memories)
    {
        RTLObject rom(m);
        // expect no hazards at the beginning
        bool has_hazard[HT_END] = {false, false, false};
        for (MemPort* p1 : m->ports)
            for (MemPort* p2 : m->ports)
            {
                RTLObject rop1(p1);
                RTLObject rop2(p2);

                FVSet rphs1 = get_fvs_meta(rop1, DF_BCK);
                FVSet wphs1 = get_fvs_meta(rop1, DF_FWD);
                FVSet wphs2 = get_fvs_meta(rop2, DF_FWD);

                for (fv_type rph : rphs1)
                    for (fv_type wph : wphs2)
                        if (p1->is_read() && p2->is_write())
                        {
                            // RAW hazards
                            if (!has_hazard[HT_RAW] && rph < wph)
                            {
                                set_meta(rom, "rawh", "1");
                                has_hazard[HT_RAW] = true;
                            }

                            // WAR hazards
                            if (!has_hazard[HT_WAR] && rph >= wph)
                            {
                                set_meta(rom, "warh", "1");
                                has_hazard[HT_WAR] = true;
                            }
                        }

                for (fv_type wph1 : wphs1)
                    for (fv_type wph2 : wphs2)
                        if (p1->is_write() && p2->is_write())
                        {
                            // WAW hazards
                            if (!has_hazard[HT_WAW] && wph1 != wph2)
                            {
                                set_meta(rom, "wawh", "1");
                                has_hazard[HT_WAW] = true;
                            }
                        }
            }
    }

    return true;
}

Query Interpret::execute_dataflow(const PQuery &q)
{
    // Extract PC from the query
    RTLObject pc = operand_to_rtl(q, 0, RTLObject::NONE,
        "dataflow", "RTL register");
    if (pc.type != RTLObject::REGISTER)
    {
        invalid_operand("dataflow", "RTL register");
        return Query{Q_CONST, "0"};
    }

    // Call methods implementing data-flow analysis
    bool df_ok = phs_dataflow(pc) && bck_dataflow() && fwd_dataflow() &&
        finalize_dataflow();
    return Query{Q_CONST, df_ok ? "1" : "0"};
}

Query Interpret::execute_simplify(const PQuery &q)
{
    (void)q;
    SystemManipulation manip{system};
    manip.simplify();
    initialize_from_system();
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_unifyselectors(const PQuery &q)
{
    (void)q;
    SystemManipulation manip{system};
    manip.unify_selectors();
    initialize_from_system();
    return Query{Q_CONST, "1"};
}

Query Interpret::execute_load_vars(const PQuery &q)
{
    (void)q;
    int no_variables = 0;

    for (auto signal : system.signals)
        add_new_rtl_variable(signal, "s_");
    no_variables += system.signals.size();

    for (auto reg : system.registers)
        add_new_rtl_variable(reg, "r_");
    no_variables += system.registers.size();

    for (auto mem : system.memories)
        add_new_rtl_variable(mem, "m_");
    no_variables += system.memories.size();

    for (auto mem : system.memories)
    {
        for (auto port : mem->ports)
            add_new_rtl_variable(port, "p_");
        no_variables += mem->ports.size();
    }

    return Query{Q_CONST, std::to_string(no_variables)};
}

Query Interpret::execute_debug_listvariables(const PQuery &q)
{
    (void)q;
    string out;
    out += "===== Variables =====\n";
    for (auto &i : variables)
    {
        out += varinfo_to_str(i.second);
        out += '\n';
    }
    out += "===== Declarations =====\n";
    for (auto &decl : declarations)
    {
        out += decl_to_str(decl.first, decl.second);
        out += '\n';
    }

    return Query{Q_CONST, out};
}

Query Interpret::execute_debug_printvarsdep(const PQuery &q)
{
    string dot_fname = q->value;
    std::ofstream out{dot_fname};
    out << "digraph {\n";

    // styles definition
    int some = 0;
    for (auto &i : variables)
        if (i.second->bound)
        {
            some++;
            break;
        }
    if (some)
    {
        out << "  node [style = bold]; ";
        for (auto &i : variables)
        {
            if (i.second->bound)
                out << " " << i.second->id;
        }
        out << ";\n";
    }
    some = 0;
    for (auto &i : variables)
        if (!i.second->bound && !i.second->temp)
        {
            some++;
            break;
        }
    if (some)
    {
        out << "  node [style = solid]; ";
        for (auto &i : variables)
        {
            if (!i.second->bound && !i.second->temp)
                out << " " << i.second->id;
        }
        out << ";\n";
    }
    some = 0;
    for (auto &i : variables)
        if (i.second->temp)
        {
            some++;
            break;
        }
    if (some)
    {
        out << "  node [style = dashed]; ";
        for (auto &i : variables)
        {
            if (i.second->temp)
                out << " " << i.second->id;
        }
        out << ";\n";
    }

    for (auto &i : variables)
    {
        for (auto &dep : i.second->prerequisites)
            out << "  " << i.second->id << " -> " << dep->id << ";\n";
        for (auto &dep : i.second->bound_prerequisites)
            out << "  " << i.second->id << " -> " << dep->id <<
                "[style=dotted];\n";
    }

    if (some)
    {
        out << "label=\"";
        for (auto &i : variables)
            if (i.second->temp)
            {
                out << i.second->id << ".generator = ";
                out << debug_query_to_str(i.second->generator);
                out << "\\n";
            }
        out << "\";\n";
    }

    out << "}\n";
    out.close();

    return Query{Q_CONST, "written to " + dot_fname};
}

Query Interpret::execute_system_shell(const PQuery &q)
{
    FILE *result = popen(q->value.c_str(), "r");
    string out;
    for (int c = fgetc(result); c != EOF; c = fgetc(result))
        out += (char)c;
    fclose(result);
    return Query{Q_CONST, out};
}

string Interpret::decl_to_str(const PQuery &q, const VarSet &vars)
{
    string out;
    out += "Declaration of: ";
    for (auto v : vars)
        out += v->id + ' ';
    out += '\n';
    out += "declarator    : ";
    out += debug_query_to_str(q);
    out += '\n';
    return out;
}

string Interpret::varinfo_to_str(PVariableInfo v)
{
    string out;
    out = "VariableInfo for '" + v->id + "'\n";
    out += "type         : ";
    out += (v->bound) ? "bound" : "free";
    if (v->temp) out += " - temporary";
    out += '\n';
    out += "value        : ";
    out += debug_query_to_str(v->value);
    out += '\n';
    out += "generator    : ";
    out += debug_query_to_str(v->generator);
    out += '\n';
    if (v->dependencies.size())
    {
        out += "dependencies :";
        for (auto subv : v->dependencies)
            out += ' ' + subv->id;
        out += '\n';
    }
    if (v->prerequisites.size())
    {
        out += "prerequisites:";
        for (auto subv : v->prerequisites)
            out += ' ' + subv->id;
        out += '\n';
    }
    if (v->bound && v->bound_prerequisites.size())
    {
        out += "bound-prereq :";
        for (auto subv : v->bound_prerequisites)
            out += ' ' + subv->id;
        out += '\n';
    }
    return out;
}

Query Interpret::execute_debug_varinfo(const PQuery &q)
{
    Query out{Q_CONST};
    auto v = find_var(q->value);
    if (!v)
        throw InterpretException("unknown identifier '" + q->value + "'");
    return Query{Q_CONST, varinfo_to_str(v)};
}

Query Interpret::execute_debug_nextvalue(const PQuery &q)
{
    _debug = true;
    auto v = find_var(q->value);
    if (!v)
        throw InterpretException("unknown identifier '" +
                q->value + "'");
    v->evaluate_next(v->driver);
    return *v->value;
}

Query Interpret::execute_debug_vgroups(const PQuery &q)
{
    (void)q;
    string out;
    int i = 0;
    for (auto &vgp : vgroups)
    {
        out += "Variable group #" + std::to_string(i++) + '\n';
        out += "variables:";
        for (auto v : vgp.second->variables)
        {
            out += ' ' + v->id;
        }
        out += '\n';
        out += "driver    : ";
        out += debug_query_to_str(vgp.first);
        out += '\n';
        for (auto v : vgp.second->variables)
        {
            out += "  driver for " + v->id + " : ";
            out += debug_query_to_str(v->driver);
            out += '\n';
        }
        out += '\n';
    }
    return Query{Q_CONST, out};
}

Query Interpret::execute_debug_setup(const PQuery &q)
{
    (void)q;
    _debug = true;

    return Query{Q_NONE};
}

Query Interpret::execute_debug_stop(const PQuery &q)
{
    (void)q;
    _debug = false;
    return Query{Q_NONE};
}

Query Interpret::execute_debug_repair(const PQuery &q)
{
    (void)q;
    // 1. Repair I/O connections
    SystemManipulation manip{system};
    manip.update_fnode_io();
    initialize_from_system();

    // 2. Recompute the number of memory cells
    for (Memory* m : system.memories)
    {
        int sz = 1;
        // sz = 2^addr_bw
        for (int i = 0; i < m->addr_bitwidth; i++)
            sz <<= 1;
        m->size = sz;

        m->check_bitwidth();
    }

    // 3. Add default cases to (case ) and (switch )
    // control expressions
    for (FunNode* fn : system.fnodes)
        for (ExprAssign* assign : fn->assignments)
        {
            TraverseInputs traverse{
                assign,
                // take only control expressions
                [](Expression*& expr) {
                    return (expr->op == OP_COND) || (expr->op == OP_CASE);
                }
            };

            // traverse inputs of each assignment that are control exprs
            for (auto it = traverse.begin(); it != traverse.end(); ++it)
            {
                ControlExpression* ctrl = static_cast<ControlExpression*>(*it);
                int bw = ctrl->bitwidth;
                // create a new default case for ctrl exprs without one
                if (ctrl->def == nullptr)
                {
                    if (bw > 0)
                        ctrl->add_default(new ExprConst("0", bw));
                    else
                        cerr << "warning: Unable to determine bitwidth "
                            "of constrol expression!" << endl;
                }
            }
        }
    return Query{Q_CONST, "1"};
}

Query Interpret::make_query_expr_tuple(ExpressionList &exprs, Expression *last)
{
    Query out{Q_TUPLE};

    bool skip = true;
    for (auto expr : exprs)
    {
        if (skip)
        {
            skip = false;
            continue;
        }
        else
            out << Query{Q_RTLOBJECT, RTLObject{expr}};
    }

    out << Query{Q_RTLOBJECT, RTLObject{last}};

    return out;
}

void Interpret::_find_expression_direct_path(set<Expression*> inputs,
        Expression* output, Query &paths)
{
    for (auto input : inputs)
    {
        TraverseOutputs traverse{input};

        for (auto it = traverse.begin(); it != traverse.end(); ++it)
            // if *it is one of 'output's ExprAssing(s)
            if (vector_contains(output->outputs, *it))
            {
                paths << make_query_expr_tuple(it.bt, *it);
            }
    }
}

set<Expression*> Interpret::get_signal_read_references(Signal *signal)
{
    set<Expression*> references;
    for (auto exprvar : signal->references)
        if (exprvar != signal->writing_expr)
            references.insert(exprvar);
    return references;
}

void Interpret::find_expression_direct_path(Signal *input, Signal *output,
        Query &paths)
{
    if (input == output)
        paths << Query{Q_RTLOBJECT, RTLObject(input)};
    else
    {
        if (output->writing_expr == nullptr)
            return;
        _find_expression_direct_path(get_signal_read_references(input),
                output->writing_expr, paths);
    }
}

Query Interpret::make_query_exprs_tuple(SignalPath &spath)
{
    Query out{Q_TUPLE};

    for (unsigned i = 0; i < spath.traces.size(); i++)
    {
        bool skip = true;
        for (auto expr : *spath.traces[i])
        {
            if (skip)
            {
                skip = false;
                continue;
            }
            else
                out << Query{Q_RTLOBJECT, RTLObject{expr}};
        }

        out << Query{Q_RTLOBJECT, RTLObject{spath.lasts[i]}};
    }

    return out;
}

unsigned Interpret::get_data_passing_outputs(Signal* input,
    SignalSet &outputs)
{
    for (auto connection: input->connections)
        // TODO: include register delay
        if (connection.storage->contains_input(input, TF_ALL))
        {
            // We do not want pass through architectural storages.
            if (connection.storage->meta.find("architectural") != nullptr)
                continue;

            // Exclude memories. (1) Write-ports are considered
            // architectural. (2) Read-port has only address and enable input
            // signals which both do not propagate data.
            MemPort* port = dynamic_cast<MemPort*>(connection.storage);
            if (port != nullptr && port->is_write())
                continue;

            // We want to propagate only through data signal of a register.
            Register* reg = dynamic_cast<Register*>(connection.storage);
            if (reg != nullptr && reg->d != input)
                continue;

            connection.storage->get_influenced_outputs(
                input, TF_ALL, outputs);
        }

    return outputs.size();
}

Query Interpret::make_query_expression_storage_tuple(
    AssignTrace &assign_trace, SignalTrace &trans_signal_trace)
{
    // Constructs and returns tuple of the form that be, e.g.
    // (tuple (:= ...) (:= ...) (reg ...) (:= ...) ... )
    //
    // Presence of (reg ...) denotes trans-storage propagation.
    Query out{Q_TUPLE};
    unsigned i = 0;
    for (auto assign : assign_trace)
    {
        // Emit current assignment.
        out << Query{Q_RTLOBJECT, RTLObject{assign}};

        // Check if it is needed to add register part.
        if (i < trans_signal_trace.size())
        {
            // Get all trans-storage outputs for signal
            // on the LHS of the assignment.
            Signal* signal = static_cast<Signal*>(
                static_cast<ExprVariable*>(assign->lhs)->v);
            SignalSet outputs;
            get_data_passing_outputs(signal, outputs);

            // Try to find trans-storage signal output in the trans-storage
            // signal trace.
            for (auto output : outputs)
            {
                if (output == trans_signal_trace[i])
                {
                    // We have found the signal, now we must determine
                    // corresponding storage.
                    for (auto connection: output->connections)
                    {
                        if (connection.storage->contains_output(output))
                        {
                            // Emit storage.
                            out << Query{Q_RTLOBJECT,
                                RTLObject{connection.storage}};

                            break;
                        }
                    }

                    // Move to the next signal in the trans-storage trace.
                    ++i;
                }
            }
        }
    }

    return out;
}

bool Interpret::is_trace_through_condition(ExpressionList &trace)
{
    // Check each expression in a trace for ExprITE.
    // If found, then further investigate a connection to
    // conditional part of the processed ExprITE.
    Expression* last = nullptr;
    for (auto expr : trace)
    {
        if (last == nullptr)
        {
            last = expr;
            continue;
        }

        if (expr->op == OP_ITE &&
            static_cast<ExprITE*>(expr)->cond == last)
            return true;

        if (expr->op == OP_CASE &&
            static_cast<ExprCase*>(expr)->ctrl == last)
            return true;

        if (expr->op == OP_COND)
        {
            ExprCond* e = static_cast<ExprCond*>(expr);
            for (auto case_pair : e->cases)
                if (case_pair.first == last)
                    return true;
        }

        last = expr;
    }

    return false;
}

Query Interpret::make_query_condition_set(ExprTrace &expr_trace,
    SignalTrace &trans_signal_trace)
{
    // Extract conditions by examination of backtrace.
    //
    // TODO: FIXME: Cannot handle "(cond ... )" expressions. Now requires
    // normalization to "only-ite" or "only-switch" scheme.
    Query conds{Q_SET};
    Expression* last = nullptr;
    unsigned stage = 0;
    unsigned i = 0;
    for (auto expr_list : expr_trace)
    {
        if (expr_list == nullptr)
        {
            // nullptr serves as a storage place holder
            // and stage separator in case that storage is not
            // read port
            if (i < trans_signal_trace.size())
            {
                // Output of delimiting storage
                Signal* output = trans_signal_trace[i];

                // We have output transtorage signal and we must determine
                // corresponding storage.
                for (auto connection: output->connections)
                {
                    if (connection.storage->contains_output(output))
                    {
                        // Check wheter the found storage is read port
                        MemPort* port = dynamic_cast<MemPort*>(connection.storage);
                        if (port != nullptr && port->is_read())
                        {
                            /*cerr << port->enable->id << "==" << "1" << endl;

                            // Equality tuple
                            Query equ{Q_TUPLE};
                            equ << Query{Q_RTLOBJECT, RTLObject{port->enable}};
                            // WARNING: hack, string that looks like numeric constant
                            // TODO: FIXME: Find a better solution.
                            equ << Query{Q_CONST, "(_ 1 1)" };

                            // Stage no. constant
                            Query stage_num{Q_CONST, std::to_string(stage)};

                            // Construction of a new condition from the previous queries
                            Query cond{Q_TUPLE};
                            cond << equ;
                            cond << stage_num;


                            conds << cond;*/
                        }
                        else
                        {
                            // Increment stage (delay) if not
                            stage++;
                        }
                        break;
                    }
                }
            }
            i++;
            continue;
        }

        for (auto expr : *expr_list)
        {
            if (last == nullptr)
            {
                last = expr;
                continue;
            }

            if (expr->op == OP_ITE)
            {
                ExprITE* ite = static_cast<ExprITE*>(expr);
                if (ite->e_true == last || ite->e_false == last)
                {
                    // Equality tuple
                    Query equ{Q_TUPLE};
                    equ << Query{Q_RTLOBJECT, RTLObject{ite->cond}};
                    // WARNING: hack, string that looks like numeric constant
                    // TODO: FIXME: Find a better solution.
                    equ << Query{Q_CONST, ite->e_true == last ? "(_ 1 1)" : "(_ 0 1)" };

                    // Stage no. constant
                    Query stage_num{Q_CONST, std::to_string(stage)};

                    // Construction of a new condition from the previous queries
                    Query cond{Q_TUPLE};
                    cond << equ;
                    cond << stage_num;

                    conds << cond;
                }
            }

            if (expr->op == OP_CASE)
            {
                ExprCase* e = static_cast<ExprCase*>(expr);
                for (auto case_pair : e->cases)
                    if (case_pair.second == last)
                    {
                        // Equality tuple
                        Query equ{Q_TUPLE};
                        equ << Query{Q_RTLOBJECT, RTLObject{e->ctrl}};
                        equ << Query{Q_RTLOBJECT, RTLObject{case_pair.first}};

                        // Stage no. constant
                        Query stage_num{Q_CONST, std::to_string(stage)};

                        // Construction of a new condition from the previous queries
                        Query cond{Q_TUPLE};
                        cond << equ;
                        cond << stage_num;

                        conds << cond;
                        break;
                    }
            }

            last = expr;
        }
    }

    return conds;
}

void Interpret::_find_expression_path_data_signal(Signal *input,
        Signal* output, ExprTrace &expr_trace,
        AssignTrace &assign_trace,
        SignalTrace &trans_signal_trace, Query &paths)
{
    if (output->writing_expr == nullptr)
        return;

    bool found = false;
    set<Expression*> inputs = get_signal_read_references(input);
    for (auto in : inputs)
    {
        TraverseOutputs traverse{in};

        for (auto it = traverse.begin(); it != traverse.end(); ++it)
        {
            if (it->op == OP_ASSIGN)
            {
                expr_trace.push_back(&it.bt);
                assign_trace.push_back(static_cast<ExprAssign*>(*it));
                if (vector_contains(output->writing_expr->outputs, *it))
                {
                    found = true;

                    // Eliminate paths going through condition parts
                    // of conditional expressions.
                    //if (!is_trace_through_condition(it.bt))
                    //{
                        // Result of the form (tuple (tuple ... "path" ...)
                        // (set ... "conditions" ... ))
                        Query out{Q_TUPLE};

                        // Adjoin the found path as (tuple ...) to result.
                        out << make_query_expression_storage_tuple(
                            assign_trace, trans_signal_trace);

                        // Adjoin the found conditions along the path as
                        // (set ... ) to the result.
                        out << make_query_condition_set(expr_trace,
                            trans_signal_trace);

                        // Add path and conditions to final results.
                        paths << out;
                    //}
                }
                else
                {
                    ExprAssign *assignment = static_cast<ExprAssign*>(*it);
                    Signal *newinput;
                    newinput = static_cast<Signal*>(
                            static_cast<ExprVariable*>(assignment->lhs)->v);

                    _find_expression_path_data_signal(newinput, output,
                        expr_trace, assign_trace, trans_signal_trace, paths);

                }
                assign_trace.pop_back();
                expr_trace.pop_back();
            }
        }
    }

    // If no path was found, try to continue through connected storages.
    if (!found)
        _find_expression_path_data_storage(input, output, expr_trace,
            assign_trace, trans_signal_trace, paths);
}

void Interpret::_find_expression_path_data_storage(Signal *input,
        Signal* output, ExprTrace &expr_trace,
        AssignTrace &assign_trace,
        SignalTrace &trans_signal_trace, Query &paths)
{
    SignalSet trans_storage_signals;
    get_data_passing_outputs(input, trans_storage_signals);

    for (auto signal : trans_storage_signals)
    {
        if (vector_contains(trans_signal_trace, signal))
            continue;

        trans_signal_trace.push_back(signal);
        expr_trace.push_back(nullptr); // nullptr used as separator
        _find_expression_path_data_signal(signal, output, expr_trace,
            assign_trace, trans_signal_trace, paths);
        expr_trace.pop_back();
        trans_signal_trace.pop_back();
    }
}

void Interpret::_find_expression_path(Signal *input,
        Signal* output, SignalPath &spath, Query &paths)
{
    if (output->writing_expr == nullptr)
        return;

    set<Expression*> inputs = get_signal_read_references(input);
    for (auto input : inputs)
    {
        TraverseOutputs traverse{input};

        for (auto it = traverse.begin(); it != traverse.end(); ++it)
            if (it->op == OP_ASSIGN)
            {
                spath.traces.push_back(&it.bt);
                spath.lasts.push_back(*it);
                if (vector_contains(output->writing_expr->outputs, *it))
                {
                    paths << make_query_exprs_tuple(spath);
                }
                else
                {
                    ExprAssign *assignment = static_cast<ExprAssign*>(*it);
                    Signal *newinput;
                    newinput = static_cast<Signal*>(
                            static_cast<ExprVariable*>(assignment->lhs)->v);

                    _find_expression_path(newinput, output, spath, paths);
                }
                spath.traces.pop_back();
                spath.lasts.pop_back();
            }
    }
}

void Interpret::find_expression_path(Signal *input, Signal *output,
        Query &paths)
{
    if (input == output)
        paths << Query{Q_RTLOBJECT, RTLObject(input)};
    else
    {
        SignalPath spath;
        _find_expression_path(input, output, spath, paths);
    }
}

void Interpret::find_expression_path_data(Signal *input, Signal *output,
        Query &paths)
{
    if (input == output)
        paths << Query{Q_RTLOBJECT, RTLObject(input)};
    else
    {
        ExprTrace expr_trace;
        AssignTrace assign_trace;
        SignalTrace trans_signal_trace;
        _find_expression_path_data_signal(input, output, expr_trace,
            assign_trace, trans_signal_trace, paths);
    }
}

void Interpret::initialize_symboltable(PQuery &q)
{
    if (q->type == Q_IN && q->subqueries[0]->type == Q_VAR)
    {
        initialize_symboltable_in(q);
        for (unsigned i = 1; i < q->subqueries.size(); i++)
            initialize_symboltable(q->subqueries[i]);
        return;
    }
    else if (q->type == Q_LET)
        initialize_symboltable_let(q);

    for (auto &subq : q->subqueries)
        initialize_symboltable(subq);
}

void Interpret::initialize_symboltable_let(PQuery &q)
{
    auto v = find_var(q->value);
    if (!v)
    {
        // not yet defined => create it
        auto nvinfo = new VariableInfo{*this, q->value, q->subqueries[0]};
        variables.insert({q->value, nvinfo});
    }
    else if (v->bound)
    {
        // previously declared by '(in v ...)' query) =>
        // take control of originally intended bound variable
        v->clear();
        v->value = q->subqueries[0];
        // find and erase the record of transducer:
        // (in v ...) to (<at-inc> ... #0)
        for (auto it = query_transducers.begin();
                it != query_transducers.end(); ++it)
            if (it->second.var == v)
            {
                query_transducers.erase(it);
                break;
            }
    }
    else
    {
        // redefinition of free variable
        v->clear();
        v->generator = q->subqueries[0];
    }
}

void Interpret::initialize_symboltable_in(PQuery &q)
{
    string vid = q->subqueries[0]->value;
    auto v = find_var(vid);
    if (!v)
    {
        // not yet defined => create bound variable

        // 1. encapsulate in.subqueries[1] to a tmp variable
        string tmpvar = generate_tmp_varid();
        auto nvinfo = new VariableInfo{*this, tmpvar, q->subqueries[1]};
        nvinfo->temp = true;
        variables.insert({tmpvar, nvinfo});

        // 2. make incremental-at generator over tmp variable
        PQuery generator = make_shared<Query>(Q_ATINC, 0);
        generator->value = vid;
        *generator << Query{Q_VAR, tmpvar};

        // 3. create a bound variable with Q_ATINC generator
        nvinfo = new VariableInfo{*this, vid, generator};
        nvinfo->bound = true;
        variables.insert({vid, nvinfo});

        // record info about generator
        query_transducers[q] = GeneratorTransducer{generator, nvinfo};
    }
    else if (v->bound)
        throw InterpretException(
                "redefinition of variable '" + vid + "'");
}

void Interpret::replace_generator(PQuery &q)
{
    // if the query has its replacement in the given map, replace it
    auto it = query_transducers.find(q);
    if (it != query_transducers.end())
        q = make_shared<Query>(Q_CONST, "1");
//        q = it->second.generator;
    else
        for (auto &subq : q->subqueries)
            replace_generator(subq);
}

void Interpret::replace_generators()
{
    for (auto v : variables)
        if (v.second->generator)
            replace_generator(v.second->generator);
}

void Interpret::get_all_variable_references(const PQuery &q, VarSet &vars)
{
    if (q->type == Q_VAR)
    {
        auto v = find_var(q->value);
        if (v)
            vars.insert(v);
    }
    else
        for (auto &subq : q->subqueries)
            get_all_variable_references(subq, vars);
}

// setminus (returns s1 \ s2)
inline VarSet operator-(const VarSet &s1, const VarSet &s2)
{
    VarSet difference;
    std::set_difference(s1.begin(), s1.end(), s2.begin(), s2.end(),
            std::inserter(difference, difference.end()));
    return difference;
}

// setunion (returns s1 U s2)
inline VarSet operator+(const VarSet &s1, const VarSet &s2)
{
    VarSet result;
    std::set_union(s1.begin(), s1.end(), s2.begin(), s2.end(),
            std::inserter(result, result.begin()));
    return result;
}

void Interpret::fill_query_prerequisites_generator(const PQuery &q,
        VarSet &prerequisites)
{
    VarSet *declared_here = nullptr;
    for (auto &declaration : declarations)
        if (declaration.first == q)
        {
            declared_here = &declaration.second;
            break;
        }
    assert(declared_here);

    for (unsigned i = 1; i < q->subqueries.size(); i++)
        fill_query_prerequisites(q->subqueries[i], prerequisites);

    prerequisites = prerequisites - *declared_here;
}

void Interpret::fill_query_prerequisites_quantifier(const PQuery &q,
        VarSet &prerequisites)
{
    auto bound_var = find_var(q->value);
    assert(bound_var);
    for (auto &subq : q->subqueries)
        fill_query_prerequisites(subq, prerequisites);
    prerequisites.erase(bound_var);
}

void Interpret::fill_query_prerequisites(const PQuery &q,
        VarSet &prerequisites)
{
    VarSet declared_here;
    if (is_set_generator(q))
        fill_query_prerequisites_generator(q, prerequisites);
    else if (is_quantifier(q))
        fill_query_prerequisites_quantifier(q, prerequisites);
    else if (q->type == Q_LET)
        fill_query_prerequisites(q->subqueries[0], prerequisites);
    else if (q->type == Q_VAR)
    {
        auto v = find_var(q->value);
        assert(v);
        prerequisites.insert(v);
    }
    else
        for (auto &subq : q->subqueries)
            fill_query_prerequisites(subq, prerequisites);
}

void Interpret::initialize_declarations(const PQuery &q,
        VarSet &already_declared)
{
    if (is_set_generator(q))
    {
        VarSet in_definition;
        get_all_variable_references(q->subqueries[0], in_definition);
        VarSet declared_here = in_definition - already_declared;

        declarations.push_back(make_pair(q, declared_here));

        VarSet declared_further = already_declared + declared_here;
        for (unsigned i = 1; i < q->subqueries.size(); i++)
            initialize_declarations(q->subqueries[i], declared_further);
    }
    else if (is_quantifier(q))
    {
        auto v = find_var(q->value);
        assert(v);

        declarations.push_back(make_pair(q, VarSet{v}));

        already_declared.insert(v);
        for (auto &subq : q->subqueries)
            initialize_declarations(subq, already_declared);
        already_declared.erase(v);
    }
    else
        for (auto &subq : q->subqueries)
            initialize_declarations(subq, already_declared);
}

inline void update_prerequisites_and_dependencies(PVariableInfo v,
        VarSet &prerequisites)
{
    for (auto pre : prerequisites)
    {
        v->prerequisites.insert(pre);
        pre->dependencies.insert(v);
    }
}

void Interpret::initialize_variable_dependencies()
{
    // clear old dependencies
    for (auto vit : variables)
    {
        auto v = vit.second;
        v->prerequisites.clear();
        v->dependencies.clear();
        v->bound_prerequisites.clear();
    }

    // for each variable, create a new prerequisites depending on its
    // generator
    for (auto vit : variables)
    {
        auto v = vit.second;

        // find-out prerequisites of a generator
        if (!v->generator)
            continue;
        VarSet generator_pre;
        fill_query_prerequisites(v->generator, generator_pre);

        // add the prerequisites and dependencies to the variables
        update_prerequisites_and_dependencies(v, generator_pre);
    }

    // update bound prerequisites for generating vgroups
    for (auto vit : variables)
    {
        auto v = vit.second;
        if (v->bound)
        {
            v->check_prerequisites_cycle();
            v->update_bound_prerequisites();
        }
    }
}

VarSet Interpret::_get_vgroup_variables(const VarSet &vars_init)
{
    VarSet vars = vars_init;
    for (auto v : vars_init)
        vars = vars + v->bound_prerequisites;

    for (auto &vgroup : vgroups)
        for (auto v : vgroup.second->variables)
            vars.erase(v);

    return vars;
}

void Interpret::initialize_vgroups()
{
    vgroups.clear();
    for (auto &declaration : declarations)
    {
        const PQuery &q = declaration.first;

        VarSet vars_group = _get_vgroup_variables(declaration.second);

        if (!vars_group.empty())
        {
            PVariableGroup vg = new VariableGroup{q, vars_group};
            vgroups.insert({q, vg});
        }
    }
}

// called by erase_variable() and teardown_query()
void Interpret::erase_tmp_orphans()
{
    vector<PVariableInfo> to_delete;
    for (auto &i : variables)
    {
        PVariableInfo &v = i.second;
        if (v->temp)
        {
            if (v->dependencies.empty())
            {
                to_delete.push_back(v);
                for (auto pre : v->prerequisites)
                {
                    auto it = pre->dependencies.find(v);
                    if (it != pre->dependencies.end())
                        pre->dependencies.erase(it);
                }
                v->prerequisites.clear();
            }
        }
    }
    for (auto v : to_delete)
    {
        variables.erase(v->id);
        delete v;
    }
}

// called just by (undef) query command
void Interpret::erase_variable(PVariableInfo v)
{
    // check if no variable depends on this
    if (v->dependencies.size())
    {
        string deps;
        for (auto var : v->dependencies)
        {
            if (deps.size()) deps += ',';
            deps += var->id;
        }
        throw InterpretException("variable(s) " + deps +
                " depend on this.");
    }

    // erase the variable
    variables.erase(v->id);

    // erase all temporary orphans (possibly there was one)
    if (v->bound)
        erase_tmp_orphans();

    // erase all vgroups containing the variable
    for (auto it = vgroups.begin(); it != vgroups.end();)
    {
        auto vg = it->second;
        if (std::find(vg->variables.begin(), vg->variables.end(), v)
                != vg->variables.end())
        {
            // erase all references from bound-variables to vgroup
            for (auto vfromvg : vg->variables)
                vfromvg->vgroups.erase(vg);

            // erase the vgroup
            vgroups.erase(it++);
            // map.erase: Iterators, pointers and references referring to
            // elements removed by the function are invalidated.
            // All other iterators, pointers and references keep their
            // validity.

            // free it from the memory
            delete vg;
        }
        else
            ++it;
    }

    // finally, delete the variable from the memory
    delete v;
}

} // namespace rtl
} // namespace vam
