/**
 * Interpret of RTL queries.
 */
#ifndef _QUERYINTERPRET_H
#define _QUERYINTERPRET_H

#include "query.h"
#include "queryintbasictypes.h"
#include "queryvariableinfo.h"
#include "queryvargroup.h"
#include "queryevalinfo.h"
#include "rtlmanip.h" // DataflowDir
#include <memory>

using namespace std;

namespace vam
{
namespace rtl
{

string debug_query_to_str(const PQuery &q, int n = 55);

/**
 * @brief   Interpreter of queries.
 */
class Interpret
{
    map<string, PVariableInfo> variables;

    map<PQuery, PVariableGroup> vgroups;

    map<PQuery, vector<PVariableInfo>> visible_variables;

public:
    System &system;

    Interpret(System &s) : system(s)
    {
        initialize_from_system();
    }

    ~Interpret()
    {
        delete_variables();
    }

    /**
     * @brief   Execute a given query. Containers of variables may be
     *          invalidated.
     * @param   q   query to execute
     * @return  0 if query evaluates to false, zero, empty container, or some
     *          error occurs.
     * @post    query will be changed and invalid for further execution
     */
    int exec(PQuery &q);

    /// debug-mode tag disabling some cleanup after query evaluation
    bool _debug = false;

private:

    void initialize_from_system();

    /**
     * @brief   Initialize query variables like signals, registers, memories
     * @param   id  variable name
     * @param   set set of rtl objects
     * @post    variables contains the given variable info with value query
     */
    template<typename T>
    void initialize_rtl_set(string id, T &set)
    {
        auto q = make_shared<Query>(Q_SET);
        for (auto o : set)
            *q << Query{Q_RTLOBJECT, RTLObject(o)};
        auto v = find_var(id);
        if (v)
        {
            v->reset();
            v->value = q;
            v->generator = nullptr;
        }
        else
        {
            auto v = new VariableInfo{*this, id, nullptr};
            v->value = q;
            variables.insert({id, v});
        }
    }

    void delete_variables()
    {
        for (auto vg : vgroups)
            delete vg.second;
        vgroups.clear();
        for (auto v : variables)
            delete v.second;
        variables.clear();
    }

    /**
     * @brief   Finds either free or bound variable information.
     * @param   id  variable identifier
     * @return  pointer to VariableInfo, or nullptr if not found
     */
    PVariableInfo find_var(string id)
    {
        auto it = variables.find(id);
        if (it == variables.end())
            return nullptr;
        else
            return it->second;
    }

    /**
     * @brief   Finds variable wrt. id
     * @param   id  variable identifier
     * @return  pointer to free VariableInfo, or nullptr if not found
     */
    PVariableInfo find_free_var(string id)
    {
        auto vi = find_var(id);
        if (vi && !vi->bound)
            return vi;
        else
            return nullptr;
    }

    /**
     * @brief   Finds bound variable wrt. id
     * @param   id  bound variable identifier
     * @return  pointer to bound VariableInfo, or nullptr if not found
     */
    PVariableInfo find_bound_var(string id)
    {
        auto vi = find_var(id);
        if (vi && vi->bound)
            return vi;
        else
            return nullptr;
    }

    /**
     * @brief   initialize query variable 'memports'
     * @post    variables contains 'memports' with the given value query
     */
    void initialize_memports();

    /**
     * @brief   Invalidates value of (bound) variable.
     * @param   id  variable identifier
     * @post    find_var(id)->notset() == true
     */
    void invalidate_variable(string id);

    /**
     * @brief   Setup query to execute.
     * @param   q   query to setup
     * @post    initialized variables,vgroups,declarations,query_transducers
     */
    void setup_query(PQuery &q);

    /**
     * @brief   Teardown auxiliary structures after query evaluating
     */
    void teardown_query();

    Query evaluate_query(const PQuery &q);
    Query evaluated_query_to_str(Query q);
    Query evaluated_query_to_smtlib_str(Query q, int nprimes, const string& tag);
    Query execute_print(const PQuery &q);
    Query execute_smtlib(const PQuery &q);
    Query execute_smtlib_prime(const PQuery &q);
    Query execute_smtlib_tag_prime(const PQuery &q);
    Query execute_str(const PQuery &q);
    Query execute_decl(const PQuery &q);
    Query execute_select_sig(const PQuery &q);
    Query execute_select_reg(const PQuery &q);
    Query execute_select_mem(const PQuery &q);
    Query execute_select_port(const PQuery &q);
    Query execute_select_fnode(const PQuery &q);
    Query execute_let(const PQuery &q);
    Query execute_undef(const PQuery &q);
    Query execute_var(const PQuery &q);
    Query execute_exists(const PQuery &q);
    Query execute_forall(const PQuery &q);
    Query execute_set(const PQuery &q);
    Query execute_set_where(const PQuery &q);
    Query execute_tuple(const PQuery &q);
    Query execute_tuple_where(const PQuery &q);
    Query execute_add(const PQuery &q);
    Query execute_append(const PQuery &q);
    Query execute_not(const PQuery &q);
    Query execute_and(const PQuery &q);
    Query execute_or(const PQuery &q);
    Query execute_implies(const PQuery &q);
    Query execute_at(const PQuery &q);
    Query execute_atinc(const PQuery &q);
    Query execute_in(const PQuery &q);
    Query execute_occurences(const PQuery &q);
    Query execute_subset(const PQuery &q);
    Query execute_intersection(const PQuery &q);
    Query execute_union(const PQuery &q);
    Query execute_bigcap(const PQuery &q);
    Query execute_bigcup(const PQuery &q);
    Query execute_setminus(const PQuery &q);
    Query execute_if(const PQuery &q);
    Query execute_equals(const PQuery &q);
    Query execute_size(const PQuery &q);
    Query execute_empty(const PQuery &q);
    Query execute_is_instance(const PQuery &q);
    Query execute_gt(const PQuery &q);
    Query execute_ge(const PQuery &q);
    Query execute_lt(const PQuery &q);
    Query execute_le(const PQuery &q);
    Query execute_plus(const PQuery &q);
    Query execute_minus(const PQuery &q);
    Query execute_multiply(const PQuery &q);
    Query execute_divide(const PQuery &q);
    Query execute_modulo(const PQuery &q);
    Query execute_expr_type(const PQuery &q);
    Query execute_rtlid(const PQuery &q);
    Query execute_inputs(const PQuery &q);
    Query execute_outputs(const PQuery &q);
    Query execute_rtloutputs(const PQuery &q);
    Query execute_references(const PQuery &q);
    Query execute_path(const PQuery &q, bool is_data = false);
    Query execute_signal_path(const PQuery &q);
    Query execute_direct_path(const PQuery &q);
    Query execute_coi_dir(const PQuery &q, DataflowDir dir, bool expr = false);
    Query execute_coi(const PQuery &q);
    Query execute_coib(const PQuery &q);
    Query execute_expr_coi(const PQuery &q);
    Query execute_expr_coib(const PQuery &q);
    Query execute_stall(const PQuery &q);
    Query execute_clear(const PQuery &q);
    Query execute_targets(const PQuery &q);
    Query execute_sources(const PQuery &q);
    Query execute_data_in(const PQuery &q);
    Query execute_data_out(const PQuery &q);
    Query execute_address(const PQuery &q);
    Query execute_enable(const PQuery &q);
    Query execute_bitwidth(const PQuery &q);
    Query execute_initvalue(const PQuery &q);
    Query execute_ports(const PQuery &q);
    Query execute_is_port(const PQuery &q);
    Query execute_read_port(const PQuery &q);
    Query execute_write_port(const PQuery &q);
    Query execute_meta_clear(const PQuery &q);
    Query execute_meta_get(const PQuery &q);
    Query execute_meta_set(const PQuery &q);
    Query execute_meta_unset(const PQuery &q);

    /// \{
    /// executing queries useful for scripting
    Query execute_nop(const PQuery &q);
    Query execute_debug_listvariables(const PQuery &q);
    Query execute_debug_printvarsdep(const PQuery &q);
    string decl_to_str(const PQuery &q, const VarSet &vars);
    string varinfo_to_str(PVariableInfo v);
    Query execute_debug_varinfo(const PQuery &q);
    Query execute_debug_nextvalue(const PQuery &q);
    Query execute_debug_vgroups(const PQuery &q);
    Query execute_debug_setup(const PQuery &q);
    Query execute_debug_stop(const PQuery &q);
    Query execute_debug_repair(const PQuery &q);
    Query execute_system_shell(const PQuery &q);
    Query execute_load(const PQuery &q);
    Query execute_save(const PQuery &q);
    Query execute_dataflow(const PQuery &q);
    Query execute_simplify(const PQuery &q);
    Query execute_unifyselectors(const PQuery &q);
    Query execute_load_vars(const PQuery &q);
    /// \}

    // typedefs for data-flow analysis
    using RTLObjectList = deque<RTLObject>;
    using fv_type = int;            // data-flow value type
    using FVSet = set<fv_type>;     // data-flow value set
    using RTLObjectFVSetMap = unordered_map<void*, FVSet>;

    // string constant for meta tags
    const string S_PH = "ph";       // PHase
    const string S_WPHS = "wphs";   // Write PHaseS
    const string S_RPHS = "rphs";   // Read  PHaseS
    const string S_PC = "pc";       // Program Counter

    // type of data-flow analysis
    enum df_type {
        DF_PHS, // phase determination
        DF_FWD, // write phases
        DF_BCK  // read phases
    };

    enum hazard_type {
        HT_RAW = 0,
        HT_WAW,
        HT_WAR,
        // TODO:
        // HT_CTL
        HT_END
    };

    // constant mapping between string constansts and
    // the type of data-flow analysis
    const std::map<df_type, string> DF_STR_MAP = {
        { DF_PHS, S_PH },
        { DF_FWD, S_WPHS },
        { DF_BCK, S_RPHS }
    };

    /**
     * @brief   Pops the first item from the queue, and returns it.
     *
     * @param   rol     RTL Object List (or Queue)
     * @return          first item of the input list
     */
    RTLObject dequeue_unique(RTLObjectList& rol);

    /**
     * @brief   Add new item to the end of the queue if not present.
     *
     * @param   rol     RTL Object List (or Queue)
     * @param   ro      RTL Object to add
     */
    void enqueue_unique(RTLObjectList& rol, RTLObject ro);

    /**
     * @brief   Sets data-flow value of the given RTL Object.
     *
     * @param   fvss    map containing data-flow values of RTL Objects
     * @param   ro      RTL Object to be updated
     * @param   fvs     new set of flow values
     */
    void set_fvs(RTLObjectFVSetMap& fvss, RTLObject ro, const FVSet& fvs);

    /**
     * @brief   Gets data-flow value of the given RTL Object.
     *
     * @param   fvss   map containing data-flow values of RTL Objects
     * @param   ro      RTL Object
     * @return          set of flow values
     */
    FVSet get_fvs(RTLObjectFVSetMap& fvss, RTLObject ro);

    /**
     * @brief   Joins elements of a set consisting of data-flow values into
     *  into single string separated by the specified delimiter.
     *
     * @param   fvs     flow value set
     * @param   delim   delimiter
     * @return          string consisting of joined elements
     *                  separated by delimiter
     */
    string join_fvs(const FVSet& fvs, char delim);

    /**
     * @brief   Splits elements of a meta string separated by the specified
     *  delimiter into a set of data-flow values.
     *
     * @param   meta    meta string with flow values
     * @param   delim   delimiter
     * @return          set of flow values
     */
    FVSet split_fvs(const string& meta, char delim);

    /**
     * @brief   Sets data-flow value of the given RTL Object to its meta tag.
     *
     * @param   ro      RTL Object
     * @param   type    data-flow analysis type
     * @param   fvs     set of flow values
     */
    void set_fvs_meta(RTLObject ro, const FVSet& fvs, const df_type& type);

    /**
     * @brief   Gets data-flow value of the given RTL Object from its meta tag.
     *
     * @param   ro      RTL Object
     * @param   type    data-flow analysis type
     * @return          set of flow values
     */
    FVSet get_fvs_meta(RTLObject ro, const df_type& type);

    /**
     * @brief   Updated data-flow value of given RTL Object by taking
     * minimum from the new value and current value. Minimum is store in the
     * form of meta tag. Returns true when the value is changed during
     * update process.
     *
     * @param   fvss    map containing data-flow values of RTL Objects
     * @param   ro      RTL Object to be updated
     * @param   fvs     set of new flow values
     * @param   type    data-flow analysis type
     * @return          true iff the flow value is changed
     */
    bool update_fvs(RTLObjectFVSetMap& fvss, RTLObject ro, const FVSet& fvs,
        const df_type& type);

    /**
     * @brief   Performs simple dataflow calculating delay on each
     *  node of VAM graph from program counter. Results are stored in
     *  the form of nodes' meta tags.
     *
     * @param   pc      register representing program counter
     * @return          true iff the analysis is ok
     */
    bool phs_dataflow(RTLObject pc);

    /**
     * @brief   Performs simple dataflow calculating reading phases
     *  of each element of VAM graph. Results are stored in
     *  the form of nodes' meta tags. Expects phs_dataflow() to
     *  be done first.
     *
     * @return          true iff the analysis is ok
     */
    bool bck_dataflow();

    /**
     * @brief   Performs simple dataflow calculating writing phases
     *  of each element of VAM graph. Results are stored in
     *  the form of nodes' meta tags. Expects phs_dataflow() to
     *  be done first.
     *
     * @return          true iff the analysis is ok
     */
    bool fwd_dataflow();

    /**
     * @brief   Finalizes the data-flow. Namely, it marks elements with
     *  potential hazards, and determines phases for elements that are
     *  only read. Expects previous analysis implemented by
     *  {phs, bck, fwd}_dataflow() functions to be done.
     *
     * @return          true iff the finalization was ok
     */
    bool finalize_dataflow();

    /**
     * @brief   Resets the generator query. Supported just Q_ATINC as
     *          generator.
     */
    void reset_query(const PQuery &q)
    {
        if (q->type == Q_ATINC)
            q->aux_int = 0;
    }

    string visit_rtlobject(RTLObject &o);

    void throw_invalid_type(string req = "", string got = "")
    {
        if (got.size() && req.size())
            throw InterpretException("incompatible valus within query. "
                   "Excepted " + req + ", got " + got);
        else
            throw InterpretException("incompatible value(s) within query");
    }

    int query_to_int(const Query &q)
    {
        if (q.type == Q_CONST)
            try { return std::stoi(q.value); }
            catch(...) { throw_invalid_type("integer", q.value); }
        else
            throw_invalid_type("integer",
                   "query type #" + std::to_string(q.type));
        return 0;
    }

    RTLObject query_to_rtlobject(const Query &q)
    {
        switch (q.type)
        {
            case Q_RTLOBJECT:
                return q.rtlobject;
            case Q_SET:
            case Q_TUPLE:
                if (q.subqueries.size() == 1)
                    return query_to_rtlobject(*q.subqueries[0]);
                else
                    throw_invalid_type("container with one rtl object",
                            "empty container");
            default:
                throw_invalid_type("query of rtl object",
                        "query type #" + std::to_string(q.type));
        }
        return RTLObject{};
    }

    Query query_to_set(Query &&q)
    {
        switch (q.type)
        {
            case Q_SET:
            case Q_TUPLE:
                return std::move(q);
            default:
                Query qset{Q_SET};
                qset << q;
                return qset;
        }
    }

    bool query_to_bool(const Query &q)
    {
        switch (q.type)
        {
            case Q_CONST:
                try { return std::stoi(q.value); }
                catch(...) { return !q.value.empty(); }
            case Q_SET:
            case Q_TUPLE:
                return q.subqueries.size();
            case Q_RTLOBJECT:
                return q.rtlobject.type != RTLObject::NONE;
            default:
                return false;
        }
    }

    /**
     * @brief   Evaluate the first two operands for their integer value.
     * @param   q   query with operands to evaluate
     * @param   err_msg error message identifying the problem.
     * @return  pair of queries of integer value.
     */
    pair<Query,Query> prepare_operands_for_binint_operator(
		const PQuery &q, const string err_msg);

    /**
     * @brief   Returns set of signal identifiers from query of set type.
     *          Fails if the query set contains object which is not a signal.
     * @param   operand_set query (of set type)
     * @return  set of identifiers
     */
    IdSet filter_signals_from_query(const Query &operand_set);

    /**
     * @brief   Checks the given query for the given RTL
     *          object and returns it. Throws InterpretException if the operand is
     *          invalid.
     * @param   q               query supposedly being the RTL object
     * @param   expected_type   expected type of RTLObject
     * @param   operation       operation identifier for error message
     * @param   msg_expected    error message for what operand is expected
     * @return  RTLObject
     */
    RTLObject query_to_rtl(const PQuery &q, int expected_type,
            string operation, string msg_expected)
    {
        if (q->type == Q_NONE)
            return RTLObject{};
        if (q->type != Q_RTLOBJECT)
            invalid_operand(operation, msg_expected + ", got '" +
                    debug_query_to_str(q) + "'");
        if (expected_type != RTLObject::NONE)
            if (q->rtlobject.type != expected_type)
                invalid_operand(operation, msg_expected + ", got " +
                        q->rtlobject.get_typestr());
        return q->rtlobject;
    }

    /**
     * @brief   Evaluates the given operand, checks the result for the given RTL
     *          object and returns it. Throws InterpretException if the operand is
     *          invalid.
     * @param   q               query to evaluate
     * @param   op_index        index of the operand
     * @param   expected_type   expected type of RTLObject
     * @param   operation       operation identifier for error message
     * @param   msg_expected    error message for what operand is expected
     * @return  RTLObject
     */
    RTLObject operand_to_rtl(const PQuery &q,
            int op_index, int expected_type,
            string operation, string msg_expected)
    {
        Query operand = evaluate_query(q->subqueries[op_index]);
        if (operand.type == Q_NONE)
            return RTLObject{};
        if (operand.type != Q_RTLOBJECT)
            invalid_operand(operation, msg_expected + ", got '" +
                    debug_query_to_str(make_shared<Query>(operand)) + "'");
        if (expected_type != RTLObject::NONE)
            if (operand.rtlobject.type != expected_type)
                invalid_operand(operation, msg_expected + ", got " +
                        operand.rtlobject.get_typestr());
        return operand.rtlobject;
    }

    /**
     * \{
     * @brief   Helpers for execute_load_vars
     */
    template<typename T>
    void add_new_rtl_variable(T o, string prefix)
    {
        string vid = prefix + o->id;
        auto v = new VariableInfo{*this, vid, nullptr};
        v->value = make_shared<Query>(Q_RTLOBJECT, RTLObject(o));
        variables.insert({vid, v});
    }

    /// substitutes all retl objects with none objects in the query
    void _substitute_rtlobjects_with_none(PQuery &q);
    /// substitute all rtl objects with none objects in each variable
    void substitute_rtlobjects_with_none();
    /// \}

    /**
     * @brief   Initializes intersection query for evaluation.
     * @param   out         target query
     * @param   item        query representing the set of items
     * @param   evaluated   if the item parameter is already evaluated
     */
    void intersection_init(Query &out,
            const PQuery &item, bool evaluated)
    {
        const Query::SubQueries &items = evaluated ?
            item->subqueries :
            query_to_set(evaluate_query(item)).subqueries;
        for (auto &item : items)
            out << *item;
    }

    /**
     * @brief   Intersects just evaluated intersection query with another set.
     * @param   out         target query
     * @param   item        query representing the set of items
     * @param   evaluated   if the item parameter is already evaluated
     */
    void intersection_with(Query &out,
            const PQuery &item, bool evaluated)
    {
        const Query::SubQueries &items = evaluated ?
            item->subqueries :
            query_to_set(evaluate_query(item)).subqueries;
        for (unsigned i = 0; i < out.subqueries.size();)
        {
            bool found = false;
            for (auto &item : items)
                if (*item == *out.subqueries[i])
                {
                    found = true;
                    break;
                }
            if (!found)
                out.subqueries.erase(out.subqueries.begin() + i);
            else
                i++;
        }
    }

    /**
     * @brief   Evaluates intersection upon subqueries. Generalizes
     *          execute_intersection() and execute_bigcap(). Naive, therefore
     *          O(n^2 * m), where n is the number of items of the largest set,
     *          and m is the number of sets.
     *          UGLY, UGLY, I know! Missing Query::hash() to sort and merge.
     *          FIXME: O(m * n log n) = (m*sort() + m*set_intersection)
     * @param   set_of_sets     subqueries representing set over which to
     *                          perform bigcap operation.
     * @param   evaluated       if the set_of_sets is already evaluated.
     */
    Query intersection_of_subqueries(const Query::SubQueries &set_of_sets,
            bool evaluated);

    /**
     * @brief   Evaluates union upon subqueries. Generalizes
     *          execute_union() and execute_bigcup()
     * @param   set_of_sets     subqueries representing set over which to
     *                          perform bigcup operation.
     * @param   evaluated       if the set_of_sets is already evaluated.
     */
    Query union_of_subqueries(const Query::SubQueries &subq, bool evaluated);

    /**
     * @brief   Fills set of RTL expressions by output references of a VAM
     *          expresions.
     * @param   out         output query to fill
     * @param   references  expression list like Expression::outputs or
     *                      Variable::references
     */
    void outputs_fill_usages(Query &out, ExpressionList &references,
            Expression *var = nullptr)
    {
        for (auto expr : references)
            if (expr->op == OP_VAR)
                outputs_fill_usages(out, expr->outputs, expr);
            else if (expr->op == OP_REF)
                outputs_fill_usages(out, expr->outputs, expr);
            else if (expr->op == OP_ASSIGN)
            {
                if (static_cast<BinaryExpression*>(expr)->lhs != var)
                    out << Query{Q_RTLOBJECT, RTLObject{expr}};
            }
            else
                out << Query{Q_RTLOBJECT, RTLObject{expr}};
    }

    /**
     * @brief   Fills set of RTL expressions and storages by inputs references of a VAM
     *          expresions.
     * @param   out         output query to fill
     * @param   references  expression list like Expression::outputs or
     *                      Variable::references
     */
    void inputs_fill_usages(Query &out, Expression *expr)
    {
        for (auto operand : *expr)
            if (operand->op == OP_REF)
                inputs_fill_usages(out, operand);
            else
                out << Query{Q_RTLOBJECT, RTLObject{operand}};
    }

    /**
     * @brief   Finds all references of the given signal.
     *
     * @param   s       input signal
     * @return          set of the signal's references (expressions)
     */
    set<Expression*> get_signal_read_references(Signal *signal);

    /**
     * \{
     * @brief   Find expression direct path between two signals.
     * @param   input   input signal
     * @param   output  output signal
     * @param   paths   query (set) to which add new tuples of expression path
     */
    void find_expression_direct_path(Signal *input, Signal *output,
            Query &paths);
    void _find_expression_direct_path(set<Expression*> inputs,
            Expression *output, Query &paths);
    /// \}

    struct SignalPath {
        vector<ExpressionList*> traces;
        ExpressionList lasts;
    };

    /**
     * \{
     * @brief   Find expression path between two signals.
     * @param   input   input signal
     * @param   output  output signal
     * @param   paths   query (set) to which add new tuples of expression path
     */
    void find_expression_path(Signal *input, Signal *output, Query &paths);
    void _find_expression_path(Signal *input, Signal *output,
            SignalPath &spath, Query &paths);
    /// \}

    using ExprTrace = vector<ExpressionList*>;
    using AssignTrace = vector<ExprAssign*>;
    using SignalTrace = vector<Signal*>;

    /**
     * \{
     * @brief   Find expression path between two signals. Jumps over storages.
     * @param   input   input signal
     * @param   output  output signal
     * @param   paths   query (set) to which add new tuples of expression path
     */
    void find_expression_path_data(Signal *input, Signal *output,
        Query &paths);
    void _find_expression_path_data_signal(Signal *input, Signal *output,
        ExprTrace &expr_trace,
        AssignTrace &assign_trace,
        SignalTrace &trans_signal_trace,
        Query &paths);
    void _find_expression_path_data_storage(Signal *input, Signal *output,
        ExprTrace &expr_trace,
        AssignTrace &assign_trace,
        SignalTrace &trans_signal_trace,
        Query &paths);
    /// \}

    /**
     * @brief   Based on input signal, determines all influnced outputs.
     * Only data propagation through non-architectural registers
     * is considered.
     */
    unsigned get_data_passing_outputs(Signal* input,
        SignalSet &outputs);

    /**
     * @brief   Makes a query tuple of RTLObject expressions and storages.
     * Example of constructed tuple can be:
     *
     *      (tuple (:= ...) (:= ...) (reg ...) (:= ...) ... ).
     *
     * Presence of (reg ...) denotes trans-storage propagation.
     */
    Query make_query_expression_storage_tuple(AssignTrace &assign_trace,
        SignalTrace &trans_signal_trace);

    /**
     * @brief   Extract required values of all conditions for
     * propagation of data over a given trace. Resulti is a set of
     * tuples of the form:
     *
     *      (set (tuple "cond" "value") ... ).
     */
    Query make_query_condition_set(ExprTrace &expr_trace,
        SignalTrace &trans_signal_trace);

    /**
     * @brief   Predicate is true if a given trace does not pass
     * any ite, switch, or cond expression through as its
     * condition.
     */
    bool is_trace_through_condition(ExpressionList &trace);

    /**
     * @brief   Makes a query tuple of RTLObject expressions. Append the last
     * expression at the end.
     */
    Query make_query_expr_tuple(ExpressionList &exprs, Expression *last);
    Query make_query_exprs_tuple(SignalPath &spath);

    /**
     * @brief   Transducer of original '(in ...)' query to incremental-at
     * query.
     */
    struct GeneratorTransducer
    {
        // incremental-at generator
        PQuery generator;
        // which variable is set by the result of incremental-at
        PVariableInfo var;
    };

    // map of original subquery in the query to private generator
    // e.g. (in x y) ==> [(Q_ATINC x y #0), x]
    map<PQuery, GeneratorTransducer> query_transducers;

    /**
     * \{
     * @brief   Recursively search query and its subqueries for Q_IN query
     *          and fill the symbol table accordingly.
     * @param   q   query in which search for bounded variables
     * @post    query_transducers initialized
     */
    void initialize_symboltable(PQuery &q);
    void initialize_symboltable_let(PQuery &q);
    void initialize_symboltable_in(PQuery &q);
    /// \}

    void replace_generator(PQuery &q);
    void replace_generators();

    /**
     * @brief   Query declarations. Which variables are declared by the given
     * query.
     */
    vector<pair<PQuery, VarSet>> declarations;

    /**
     * @brief   Recursively walkthrough query and its subquery and setup
     *          declarations map.
     * @param   already_declared    set of variables already declared for the
     *                              given query.
     * @post    declarations updated
     */
    void initialize_declarations(const PQuery &q, VarSet &already_declared);

    /**
     * @brief   Setup dependencies of all variables.
     * @post    each variable of 'variables' map has initialized:
     *          prerequisites, depedencies, and bound_prerequisites.
     */
    void initialize_variable_dependencies();

    void fill_query_prerequisites(const PQuery &q,
            VarSet &prerequisites);
    void get_all_variable_references(const PQuery &q, VarSet &vars);

    void fill_query_prerequisites_generator(const PQuery &q,
            VarSet &prerequisites);
    void fill_query_prerequisites_quantifier(const PQuery &q,
            VarSet &prerequisites);

    /**
     * @brief   Search each query of declarations map and fill the vgroups
     *          accordingly.
     * @pre     declarations updated
     * @post    vgroups updated
     */
    void initialize_vgroups();

    /**
     * @brief   Helper for initialize_vgroups(). Return set of bound variables
     *          which makes the variable group (i.e. bound variables declared
     *          by a given query plus all their bound variables prerequisites
     *          which do not occur in already defined variable group.
     */
    VarSet _get_vgroup_variables(const VarSet &vars_init);

    /**
     * @brief   Helper for execute_meta_*(). Finds a meta object of the RTL
     * node given by the op_idx-th operand of the given query.
     * @param   q       query
     * @param   op_idx  index of operand to be evaluated to rtl object
     * @param   mname   string of the parent query (for error messages)
     * @return  pointer to the meta or null if bad rtl object or query operand
     */
    Meta *get_meta_of_rtlnode(const PQuery &q, int op_idx, const char *mname);
    /**
     * \{
     * @brief   Generate temporary names for generator of bound variables.
     */
    static const string tmpvar_prefix;
    int tmp_idx = 0;

    string generate_tmp_varid()
    {
        return tmpvar_prefix + to_string(tmp_idx++);
    }

    /**
     * @brief   Erase all temporary variables which are orphans
     */
    void erase_tmp_orphans();
    /// \}

    /**
     * @brief   Undefines variable. Erases it from variables and vgroups
     */
    void erase_variable(PVariableInfo v);

    friend VariableInfo;
};

inline bool is_set_generator(const PQuery &q)
{
    return q && (q->type == Q_SET_WHERE || q->type == Q_TUPLE_WHERE);
}

inline bool is_quantifier(const PQuery &q)
{
    return q && (q->type == Q_EXISTS || q->type == Q_FORALL);
}

inline bool is_bound_evaluator(const PQuery &q)
{
    return is_set_generator(q) || is_quantifier(q);
}

} // namespace rtl
} // namespace vam

#endif
