/**
 * @desc RTL manipulation query language
 */
#include "querylanguage.h"
#include "input.h"
#include "switch-parser-macro.h"

#define IDENTNOTFOUND(id) \
    PASSERT(0, t, "unknown identifier " << id)

namespace vam
{
namespace rtl
{

const Token qtokens[] = {
    // TODO sort for binary search
    {Q_NOP, "nop"},
    {Q_HELP, "help"},
    {Q_PRINT, "print"},
    {Q_SMTLIB, "smtlib"},
    {Q_SMTLIB_PRIME, "smtlib-prime"},
    {Q_SMTLIB_TAG_PRIME, "smtlib-tag-prime"},
    {Q_STR, "str"},
    {Q_DECL, "decl"},
    {Q_SELECT, "select"},
    {Q_SELECT_SIG, "select-signal"},
    {Q_SELECT_REG, "select-register"},
    {Q_SELECT_MEM, "select-memory"},
    {Q_SELECT_PORT, "select-memport"},
    {Q_SELECT_FNODE, "select-fnode"},
    {Q_LET, "let"},
    {Q_UNDEF, "undef"},
    {Q_SIGNAL, "signal"},
    {Q_REGISTER, "register"},
    {Q_MEMORY, "memory"},
    {Q_MEMPORT, "memport"},
    {Q_FNODE, "fnode"},
    {Q_EXPR, "expr"},
    {Q_SET, "set"},
    {Q_SET_WHERE, "set"},
    {Q_TUPLE, "tuple"},
    {Q_TUPLE_WHERE, "tuple"},
    {Q_WHERE, "where"},
    {Q_WHERE, "|"},
    {Q_ADD, "add"},
    {Q_APPEND, "append"},
    {Q_NOT, "not"},
    {Q_AND, "and"},
    {Q_OR, "or"},
    {Q_IMPLIES, "implies"},
    {Q_AT, "at"},
    {Q_ATINC, "<incremental-at>"},
    {Q_IN, "in"},
    {Q_OCCURENCES, "occurences"},
    {Q_SUBSET, "subset"},
    {Q_INTERSECTION, "intersection"},
    {Q_INTERSECTION, "cap"},
    {Q_SETMINUS, "setminus"},
    {Q_BIGCAP, "bigcap"},
    {Q_UNION, "union"},
    {Q_UNION, "cup"},
    {Q_BIGCUP, "bigcup"},
    {Q_IF, "if"},
    {Q_EQUALS, "equals"},
    {Q_EQUALS, "="},
    {Q_EXISTS, "exists"},
    {Q_FORALL, "forall"},
    {Q_SIZE, "size"},
    {Q_EMPTY, "empty"},
    {Q_ISINSTANCE, "is-instance"},
    {Q_FROM, ":"},
    {Q_GT, ">"},
    {Q_GE, ">="},
    {Q_LT, "<"},
    {Q_LE, "<="},
    {Q_PLUS, "+"},
    {Q_MINUS, "-"},
    {Q_MULTIPLY, "*"},
    {Q_DIVIDE, "/"},
    {Q_MODULO, "%"},
    {Q_EXPRTYPE, "expr-type"},
    {Q_RTLID, "rtlid"},
    {Q_INPUTS, "inputs"},
    {Q_OUTPUTS, "outputs"},
    {Q_RTLOUTPUTS, "rtloutputs"},
    {Q_REFERENCES, "references"},
    {Q_PATH, "path"},
    {Q_PATH_DATA, "data-path"},
    {Q_SIGNALPATH, "signal-path"},
    {Q_DIRECTPATH, "direct-path"},
    {Q_COI, "coi"},
    {Q_COIB, "coib"},
    {Q_TARGETS, "targets"},
    {Q_SOURCES, "sources"},
    {Q_DATAIN, "data-in"},
    {Q_DATAOUT, "data-out"},
    {Q_ADDRESS, "address"},
    {Q_ENABLE, "enable"},
    {Q_BITWIDTH, "bitwidth"},
    {Q_INITVALUE, "initvalue"},
    {Q_PORTS, "ports"},
    {Q_ISPORT, "is-port"},
    {Q_READPORT, "read-port"},
    {Q_WRITEPORT, "write-port"},
    {Q_METACLEAR, "meta-clear"},
    {Q_METAGET, "getattr"},
    {Q_METASET, "setattr"},
    {Q_METAUNSET, "unsetattr"},
    {Q_LOAD, "load"},
    {Q_SAVE, "save"},
    {Q_DATAFLOW, "dataflow"},
    {Q_SIMPLIFY, "simplify"},
    {Q_UNIFYSELECTORS, "unifyselectors"},
    {Q_LOAD_VARS, "load-vars"},
    {Q_DEBUG_VARIABLES, "debug-vars"},
    {Q_DEBUG_VARSDEP, "debug-varsdep"},
    {Q_DEBUG_VARINFO, "debug-var"},
    {Q_DEBUG_NEXTVALUE, "debug-next"},
    {Q_DEBUG_SETUP, "debug-setup"},
    {Q_DEBUG_VARGROUPS, "debug-vgroups"},
    {Q_DEBUG_STOP, "debug-stop"},
    {Q_DEBUG_REPAIR, "debug-repair"},
    {Q_SYSTEM_SHELL, "system"},

    {Q_EXPR_COIB, "expr-coib"},
    {Q_EXPR_COI, "expr-coi"},
    {Q_STALL, "stall"},
    {Q_CLEAR, "clear"},

    {T_EOF, ""},
};

Token tok2querykeyword(Token t)
{
    if (t.type == T_IDENT)
    {
        for (int i = 0; qtokens[i].type != T_EOF; i++)
            if (qtokens[i].value == t.value)
            {
                t.type = qtokens[i].type;
                return t;
            }
        return t;
    }
    else
        return t;
}

void QueryParser::load_command(const string &s)
{
    lexer.init(s);
}

void QueryParser::load_script(const string &filename)
{
    lexer.init("");
    if (!filename.empty())
        lexer.read_from_file(filename);
}

Query QueryParser::parse_next_command()
{
    // discard all white-spaces and comments
    Token t = lexer.next_token();
    if (t.type == T_EOF)
        return Query{Q_NONE};
    else
        unget_token(t);
    return p_query();
}

Query QueryParser::p_query(bool accept_rpar)
{
    TOKDECL;
    TOKNEW;
    if (accept_rpar && t.type == T_RPAR)
        return Query{Q_NONE};
    else if (t.type == T_IDENT)
        return Query{Q_VAR, t.value};
    else if (t.type == T_CONST)
        return Query{Q_CONST, t.value};
    else if (t.type == T_STRING)
        return Query{Q_CONST, t.value};

    PEXPECT(T_LPAR, "'(', variable identifier, or constant");
    TOKQNEW;
    switch (t.type)
    {
        case Q_HELP:
            TOKNEW; PEXPECT(T_RPAR, "')'");
            return Query{Q_HELP};
        case Q_PRINT:
        case Q_STR:
            return p_str(QueryType(t.type));
        case Q_SELECT:
            return p_select();
        case Q_SELECT_SIG:
        case Q_SELECT_REG:
        case Q_SELECT_MEM:
        case Q_SELECT_PORT:
        case Q_SELECT_FNODE:
            return p_unary(QueryType(t.type));
        case Q_LET:
            return p_let();
        case Q_UNDEF:
            return p_undef();
        case Q_DATAFLOW:
            return p_unary(QueryType(t.type));
        case Q_SIMPLIFY:
        case Q_UNIFYSELECTORS:
        case Q_LOAD_VARS:
            return p_0ary(QueryType(t.type));
        /// \{
        // debug queries
        case Q_DEBUG_VARIABLES:
        case Q_DEBUG_VARGROUPS:
        case Q_DEBUG_STOP:
        case Q_DEBUG_REPAIR:
            return p_0ary(QueryType(t.type));
        case Q_DEBUG_VARINFO:
        case Q_DEBUG_NEXTVALUE:
            return p_debug_variable(QueryType(t.type));
        case Q_DEBUG_VARSDEP:
            return p_debug_varsdep();
        case Q_SYSTEM_SHELL:
            return p_system_shell();
        case Q_DEBUG_SETUP:
            return p_unary(QueryType(t.type));
        /// \}
        case Q_SET:
        case Q_TUPLE:
            return p_set(QueryType(t.type));
        case Q_EXISTS:
        case Q_FORALL:
            return p_fol(QueryType(t.type));
        case Q_NOT:
        case Q_SIZE:
        case Q_EMPTY:
        case Q_INPUTS:
        case Q_OUTPUTS:
        case Q_RTLOUTPUTS:
        case Q_REFERENCES:
        case Q_BIGCAP:
        case Q_BIGCUP:
        case Q_TARGETS:
        case Q_SOURCES:
        case Q_SMTLIB:
            return p_unary(QueryType(t.type));
        case Q_ADD:
        case Q_APPEND:
        case Q_IMPLIES:
        case Q_IN:
        case Q_OCCURENCES:
        case Q_SUBSET:
        case Q_GT:
        case Q_GE:
	case Q_LT:
	case Q_LE:
        case Q_PLUS:
        case Q_MINUS:
	case Q_MULTIPLY:
	case Q_DIVIDE:
	case Q_MODULO:
        case Q_SETMINUS:
            return p_binary(QueryType(t.type));
        case Q_NOP:
        case Q_INTERSECTION:
        case Q_UNION:
        case Q_AND:
        case Q_OR:
        case Q_EQUALS:
            return p_nary_operation(QueryType(t.type));
        case Q_IF:
            return p_if();
        case Q_AT:
            return p_at();
        // RTL manipulations
        case Q_ISINSTANCE:
            return p_isinstance();
        case Q_EXPRTYPE:
            return p_exprtype();
        case Q_RTLID:
            return p_unary(QueryType(t.type));
        case Q_PATH:
        case Q_PATH_DATA:
        case Q_DIRECTPATH:
            return p_binary(QueryType(t.type));
        case Q_SIGNALPATH:
            return p_path(QueryType(t.type));
        case Q_COI:
        case Q_COIB:
            return p_coi(QueryType(t.type));
        case Q_EXPR_COI:
        case Q_EXPR_COIB:
            return p_expr_coi(QueryType(t.type));
        case Q_SMTLIB_PRIME:
            return p_smtlib_binary(QueryType(t.type));
        case Q_SMTLIB_TAG_PRIME:
            return p_smtlib_ternary(QueryType(t.type));
        case Q_STALL:
        case Q_CLEAR:
        case Q_DATAIN:
        case Q_DATAOUT:
        case Q_ADDRESS:
        case Q_ENABLE:
        case Q_BITWIDTH:
	case Q_INITVALUE:
        case Q_PORTS:
        case Q_READPORT:
        case Q_WRITEPORT:
            return p_unary(QueryType(t.type));
        case Q_ISPORT:
            return p_binary(QueryType(t.type));
        case Q_METACLEAR:
            return p_unary(QueryType(t.type));
        case Q_METAGET:
        case Q_METAUNSET:
            return p_binary(QueryType(t.type));
        case Q_METASET:
            return p_ternary(QueryType(t.type));
        // scripting queries
        case Q_LOAD:
        case Q_SAVE:
            return p_loadsave_file(QueryType(t.type));
        default:
            PEXPECT(0, "query keyword");
    }
    return Query{Q_NONE};
}

Query QueryParser::p_str(QueryType type)
{
    Query q{type};
    while (1)
    {
        Query subq = p_query(true);
        if (subq.type == Q_NONE)
            break;
        q << move(subq);
    }
    return q;
}

Query QueryParser::p_select()
{
    TOKDECL;
    Query q{Q_SELECT};
    TOKQNEW;
    switch (t.type)
    {
        case Q_SIGNAL:
            q.type = Q_SELECT_SIG;
            break;
        case Q_REGISTER:
            q.type = Q_SELECT_REG;
            break;
        case Q_MEMORY:
            q.type = Q_SELECT_MEM;
            break;
        case Q_MEMPORT:
            q.type = Q_SELECT_PORT;
            break;
        case Q_FNODE:
            q.type = Q_SELECT_FNODE;
            break;
        default:
            PASSERT(0, t, "expected keyword specifying rtl object");
    }
    TOKNEW; PEXPECT(T_IDENT, "RTL object identifier");
    q << Query{Q_CONST, t.value};
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_let()
{
    TOKDECL;
    Query q{Q_LET};
    TOKNEW; PEXPECT(T_IDENT, "identifier");
    q.value = t.value;
    q << p_query();
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_undef()
{
    TOKDECL;
    Query q{Q_UNDEF};
    TOKNEW; PEXPECT(T_IDENT, "identifier");
    q.value = t.value;
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_debug_variable(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_IDENT, "variable identifier");
    q.value = t.value;
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_debug_varsdep()
{
    TOKDECL;
    Query q{Q_DEBUG_VARSDEP};
    TOKNEW; PEXPECT(T_STRING, "output filename (string)");
    q.value = t.value;
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_system_shell()
{
    TOKDECL;
    Query q{Q_SYSTEM_SHELL};
    TOKNEW; PEXPECT(T_STRING, "system shell command");
    q.value = t.value;
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_set(QueryType type)
{
    TOKDECL;
    (void)_tok_meta;
    Query q{type};
    Query firstq = p_query(true);
    if (firstq.type == Q_NONE)
        return q;
    q << move(firstq);
    TOKQNEW;
    if (t.type == Q_WHERE)
    {
        q.type = (type == Q_SET ? Q_SET_WHERE :
                 (type == Q_TUPLE ? Q_TUPLE_WHERE : type));
        Query q_and{Q_AND};
        // syntactic sugar:
        // (set <q> where <q1> <q2> ...) ==>
        // (set <q> where (and <q1> <q2> ...))
        p_nary_subqueries(q_and); // where-constraints
        if (q_and.subqueries.size() == 1)
            q << *q_and.subqueries[0];
        else
            q << q_and;
    }
    else
    {
        // set/tuple of queries
        unget_token(t);
        while (1)
        {
            Query subq = p_query(true);
            if (subq.type == Q_NONE)
                break;
            q << move(subq);
        }
    }
    return q;
}

Query QueryParser::p_fol(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_IDENT, "variable identifier");
    q.value = t.value;

    // syntactic sugar:
    // (quant. var q1 q2 ... qN) ==> (quant. var (and q1 q2 ... qN))
    // (quant. var q1) ==> (quant. var q1)
    Query q_and{Q_AND};
    p_nary_subqueries(q_and);
    if (q_and.subqueries.size() == 1)
        q << *q_and.subqueries[0];
    else
        q << q_and;
    return q;
}


Query QueryParser::p_0ary(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_unary(QueryType type)
{
    TOKDECL;
    Query q{type};
    q << p_query();
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_binary(QueryType type)
{
    TOKDECL;
    Query q{type};
    q << p_query(); // lhs
    q << p_query(); // rhs
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_ternary(QueryType type)
{
    TOKDECL;
    Query q{type};
    q << p_query(); // 1
    q << p_query(); // 2
    q << p_query(); // 3
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

void QueryParser::p_nary_subqueries(Query &q)
{
    TOKDECL;
    (void)_tok_meta;
    while (1)
    {
        Query subq = p_query(true);
        if (subq.type == Q_NONE)
            break;
        q << move(subq);
    }
}

Query QueryParser::p_nary_operation(QueryType type)
{
    Query q{type};
    p_nary_subqueries(q);
    return q;
}

Query QueryParser::p_at()
{
    TOKDECL;
    Query q{Q_AT};
    q << p_query();
    TOKNEW; PEXPECT(T_CONST, "integer constant");
    q.value = t.value;
    q.aux_int = std::stoi(t.value);
    PASSERT(q.aux_int >= 0, t, "expected nonnegative integer constant");
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_if()
{
    TOKDECL;
    Query q{Q_IF};
    q << p_query(); // cond
    q << p_query(); // true-expr
    //q << p_query(); // false-expr
        // false is optional
        Query subq = p_query(true);
        if (subq.type == Q_NONE)
            return q;
        q << move(subq);

    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_isinstance()
{
    TOKDECL;
    Query q{Q_ISINSTANCE};
    q << p_query(); // cond
    TOKQNEW;
    switch (t.type)
    {
        case Q_SIGNAL:
            q.type = Q_ISINSTANCE_SIGNAL;
            break;
        case Q_REGISTER:
            q.type = Q_ISINSTANCE_REGISTER;
            break;
        case Q_MEMORY:
            q.type = Q_ISINSTANCE_MEMORY;
            break;
        case Q_MEMPORT:
            q.type = Q_ISINSTANCE_MEMPORT;
            break;
        case Q_FNODE:
            q.type = Q_ISINSTANCE_FNODE;
            break;
        case Q_EXPR:
            q.type = Q_ISINSTANCE_EXPR;
            break;
        default:
            PASSERT(0, t, "expected keyword specifying rtl object");
    }
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Operator rtltok2operator(int type)
{
    switch (type)
    {
        case T_VCONST: return OP_CONST;
        case T_EASSIGN: return OP_ASSIGN;
        case T_NOT: return OP_NOT;
        case T_OR: return OP_OR;
        case T_AND: return OP_AND;
        case T_ORJOIN: return OP_ORJOIN;
        case T_ITE: return OP_ITE;
        case T_INV: return OP_INV;
        case T_BOR: return OP_BOR;
        case T_BAND: return OP_BAND;
        case T_BXOR: return OP_XOR;
        case T_SHL: return OP_SHL;
        case T_SHR: return OP_SHR;
        case T_SAR: return OP_SAR;
        case T_EQ: return OP_EQ;
        case T_NEQ: return OP_NEQ;
        case T_LT: return OP_LT;
        case T_GT: return OP_GT;
        case T_LE: return OP_LE;
        case T_GE: return OP_GE;
        case T_SLT: return OP_SLT;
        case T_SGT: return OP_SGT;
        case T_SLE: return OP_SLE;
        case T_SGE: return OP_SGE;
        case T_UTS: return OP_UTS;
        case T_STU: return OP_STU;
        case T_IDX: return OP_IDX;
        case T_BITEX: return OP_BITEX;
        case T_CONCAT: return OP_CONCAT;
        case T_ZEX: return OP_ZEX;
        case T_SEX: return OP_SEX;
        case T_SLICE: return OP_SLICE;
        case T_CASE: return OP_CASE;
        case T_COND: return OP_COND;
        case T_NEG: return OP_NEG;
        case T_PLUS: return OP_PLUS;
        case T_MINUS: return OP_MINUS;
        case T_MUL: return OP_MUL;
        case T_DIV: return OP_DIV;
        case T_MOD: return OP_MOD;
        case T_ABSLOGIC: return OP_ABS_LOGIC;
        case T_ABSBITWISE: return OP_ABS_BITWISE;
        case T_ABSREL: return OP_ABS_REL;
        case T_ABSVECTOR: return OP_ABS_VECTOR;
        case T_ABSARITH: return OP_ABS_ARITH;
        default: return OP_NOP;
    };
}

Query QueryParser::p_exprtype()
{
    TOKDECL;
    Query q{Q_EXPRTYPE};
    q << p_query(); // rtl object of expr. type
    TOKRNEW;
    PASSERT(t.type >= T_EXPR_OPERATOR_BEGIN && t.type < T_EXPR_OPERATOR_END,
            t, "expected RTL operator keyword before '" + t.value + "'");
    q.aux_int = rtltok2operator(t.type);
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_path(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_CONST, "specification of number of cycles");
    q.aux_int = std::stoi(t.value);
    q << p_query(); // inputs
    q << p_query(); // outputs
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_coi(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_CONST, "specification of number of cycles");
    q.aux_int = std::stoi(t.value);
    q << p_query(); // set of signals
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_expr_coi(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_CONST, "specification of number of cycles");
    q.aux_int = std::stoi(t.value);
    q << p_query(); // set of signals
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_smtlib_binary(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_CONST, "specification of number of primes");
    q.aux_int = std::stoi(t.value);
    q << p_query(); // set of signals or registers
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_smtlib_ternary(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_STRING, "tag (string)");
    q.value = t.value;
    TOKNEW; PEXPECT(T_CONST, "specification of number of primes");
    q.aux_int = std::stoi(t.value);
    q << p_query(); // set of signals or registers
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

Query QueryParser::p_loadsave_file(QueryType type)
{
    TOKDECL;
    Query q{type};
    TOKNEW; PEXPECT(T_STRING, "filename (string)");
    q.value = t.value;
    TOKNEW; PEXPECT(T_RPAR, "')'");
    return q;
}

} // namespace rtl
} // namespace vam

#include "switch-parser-macro.h"

