/**
 * VariableInfo class for interpret of RTL queries.
 */
#include "queryvariableinfo.h"
#include "queryinterpret.h"
#include "queryvargroup.h"
#include <assert.h>

namespace vam
{
namespace rtl
{

extern VariableInfo varinfo_end;

bool VariableInfo::evaluate_next(const PQuery &wanna_be_driver)
{
    // for bound variable, only driver may ask for another value
    if (bound)
    {
        if (driver == nullptr)
            // if no driver, here it is
            driver = wanna_be_driver;
        else if (driver != wanna_be_driver)
            return false;
    }

    // detector of recursive evaluation
    if (evaluating_tag)
        throw VariableException(
                string{"unable to resolve cyclic dependencies: "} + id);

    try
    {
        if (!generator)
            throw InterpretException("evaluating '" + id +
                    "' with no generator");
        evaluating_tag = true;
        value = make_shared<Query>(interpret->evaluate_query(generator));
        evaluating_tag = false;
    }
    catch (VariableException &e)
    {
        evaluating_tag = false;
        e.msg += "<-" + id;
        throw;
    }
    catch (...)
    {
        evaluating_tag = false;
        throw;
    }

    invalide_dependencies();

    return true;
}

void VariableInfo::reset()
{
    // if already reseted, don't continue;
    // prevents cyclic reseting for runtime cyclic dependencies
    if (value == nullptr)
        return;
    value = nullptr;
    if (generator)
        interpret->reset_query(generator);
    invalide_dependencies();
}

void VariableInfo::check_prerequisites_cycle()
{
    vector<PVariableInfo> trace{this};
    try
    {
        for (auto v : prerequisites)
            _check_prerequisites_cycle(trace, v);
    }
    catch(CyclicDependencyException &e)
    {
        e.msg += id;
        for (unsigned i = 1; i < trace.size(); i++)
            e.msg += "<-" + trace[i]->id;
        throw;
    }
}

void VariableInfo::_check_prerequisites_cycle(vector<PVariableInfo> &trace,
        PVariableInfo v)
{
    if (vector_contains(trace, v))
    {
        trace.push_back(v);
        throw CyclicDependencyException("cycle in variable dependencies: ");
    }

    trace.push_back(v);
    for (auto subv : v->prerequisites)
        _check_prerequisites_cycle(trace, subv);
    trace.pop_back();
}

void VariableInfo::_update_bound_variable_on_tmp_prerequisite()
{
    if (bound)
    {
        assert(generator->type == Q_ATINC);
        assert(generator->subqueries[0]->type == Q_VAR);
        auto v = interpret->find_var(generator->subqueries[0]->value);
        if (v)
        {
            prerequisites.insert(v);
            v->dependencies.insert(this);
        }
    }
}

void VariableInfo::_update_prerequisites(const PQuery &dep)
{
    // skip finding prerequisites in incremental-at generator;
    // prerequisite for the given variable is already set by 
    // _update_bound_variable_on_tmp_prerequisite()
    if (dep->type == Q_ATINC)
        return;
    // found reference to variable...
    else if (dep->type == Q_VAR)
    {
        auto v = interpret->find_var(dep->value);
        // if variable exists,
        if (v)
        {
/*            if (v->temp ||
                    vector_contains(interpret->visible_variables[dep], v))*/
            {
                prerequisites.insert(v);
                v->dependencies.insert(this);
            }
        }
    }
    else
        for (auto &subq : dep->subqueries)
            _update_prerequisites(subq);
}

void VariableInfo::update_prerequisites()
{
    try
    {
        if (generator)
        {
            // TODO wtf? 
            // if bound variable, add prerequisite on temporary variable used
            // for incremental-at generator
            _update_bound_variable_on_tmp_prerequisite();
            _update_prerequisites(generator);
            for (auto v : prerequisites)
                _check_dependency_cycle(v);
        }
        if (bound)
        {
            for (auto v : prerequisites)
                _update_bound_prerequisites(v);
        }
    }
    catch (VariableException &e)
    {
        e.msg += "<-" + id;
        throw;
    }
    initialized = true;
}

void VariableInfo::_check_dependency_cycle(const PVariableInfo &dep)
{
    if (dep == this)
        throw VariableException(string{
            "unable to resolve cyclic dependencies of bound variables: "}
            + id);
    for (auto v : dep->prerequisites)
    {
        if (v->evaluating_tag)
            throw VariableException(string{
                "unable to resolve cyclic dependencies of bound variables: "}
                + id);
        try
        {
            v->evaluating_tag = true;
            _check_dependency_cycle(v);
            v->evaluating_tag = false;
        }
        catch (VariableException &e)
        {
            e.msg += "<-" + dep->id;
            v->evaluating_tag = false;
            throw;
        }
        catch (...)
        {
            v->evaluating_tag = false;
            throw;
        }
    }
}

void VariableInfo::_update_bound_prerequisites(PVariableInfo pre)
{
    if (pre->bound)
        bound_prerequisites.insert(pre);

    for (auto v : pre->prerequisites)
        _update_bound_prerequisites(v); 
}

void VariableInfo::invalide_dependencies()
{
    for (auto v : dependencies)
        v->reset();
}

} // namespace rtl
} // namespace vam
