/**
 * VariableInfo class for interpret of RTL queries.
 */
#ifndef _QUERYVARIABLEINFO_H
#define _QUERYVARIABLEINFO_H

#include "query.h"
#include "queryintbasictypes.h"
#include <set>
#include <vector>

using namespace std;

namespace vam
{
namespace rtl
{

using VarSet = set<PVariableInfo>;

class CyclicDependencyException : public RTLException
{
public:
    CyclicDependencyException(string m) noexcept : RTLException(m) {};
    const char* what() const noexcept { return "CyclicDependencyException"; }
};

/********************************************************/
/**
 * @brief Data structure for information about query variable.
 */
class VariableInfo
{
private:
    Interpret *interpret;
    VariableInfo();

public:
    /// variable identifier
    string id;

    /**
     * @brief   Variable value as query holding the constant. If the value is
     * nullptr, it means not-set. If the value is novalueptr, it means
     * no-value. In other words, "not-set yet" and "no-next-value" if used as
     * bounded variable. Use notset() and novalue().
     */
    PQuery value = nullptr;

    /// variable value generator as query evaluating to variable value
    PQuery generator;

    /**
     * @brief   List of variables which value depends on the value of this
     * variable. Thus, if this variable is invalidated, the dependent
     * variables should be invalidated too.
     */
    VarSet dependencies;

    /**
     * @brief   List of variables which directly influence the value of this
     *          variable.
     */
    VarSet prerequisites;

    /**
     * @brief   Tag if the variable is bound.
     */
    bool bound = false;

    /**
     * @brief   Tag if the variable is temporary.
     */
    bool temp = false;

    /**
     * @brief   List of bound-variables which influence the value of this
     *          variable. Relevant only if bound == true.
     */
    VarSet bound_prerequisites;

    /**
     * @brief   Query which drives the generation of bound-variable values.
     *          Relevant only if bound == true.
     */
    PQuery driver = nullptr;

    set<VariableGroup*> vgroups;

    /**
     * @brief   Makes variable information object
     * @param   _intrprt    owner-interpret reference
     * @param   i   variable identifier
     * @param   g   pointer to query which evaluates to a new value of
     *              the variable
     */
    VariableInfo(Interpret &_intrprt, string i, PQuery g)
        : interpret(&_intrprt), id(i), generator(g) {}

    /**
     * @brief   Clear value-related attributes as just after ctor.
     */
    void clear()
    {
        value = generator = nullptr;
        dependencies.clear();
        prerequisites.clear();
        bound_prerequisites.clear();
        driver = nullptr;
        evaluating_tag = false;
        initialized = false;
    }

    /**
     * @brief   Evaluates the variable. If value is not set, set the
     *          appropriate value from the generator query.
     * @post    value != nullptr
     */
    void evaluate()
    {
        if (!value)
            evaluate_next(driver);
    }

    /**
     * @brief   Evaluates the variable from the generator
     * @return  true if wanna_be_driver is a valid driver
     * @post    value != nullprt
     */
    bool evaluate_next(const PQuery &wanna_be_driver);

    /**
     * @brief   Resets the value, the generator query, and values of all
     *          depending variables.
     */
    void reset();

    /**
     * @brief   Asks the value for not-set.
     * @returns true if value == nullptr
     */
    bool notset()
    {
        return value == nullptr;
    }

    /**
     * @brief   Asks the value for no-value.
     * @returns true if value type is Q_NONE
     */
    bool novalue()
    {
        return (value && value->type == Q_NONE);
    }

    /**
     * @brief   Checks if the variable depends on the given variable.
     * @param   v   variable info to be check for dependency.
     * @return  true if varid is in dependencies.
     */
    bool depends_on(PVariableInfo v)
    {
        return prerequisites.count(v);
    }

    /**
     * @brief   Checks whether prerequisites are cycle free.
     * @throw   throws CyclicDependencyException if found cycle
     */
    void check_prerequisites_cycle();

    /**
     * @brief   Updates bound_prerequisites.
     * @pre     prerequisites initialized; prerequisites cycle-free (cf.
     *          check_prerequisites_cycle)
     * @post    bound_prerequisites initialized
     */
    void update_bound_prerequisites()
    {
        for (auto v : prerequisites)
            _update_bound_prerequisites(v);
    }

    /**
     * @brief   Updates the list of prerequisites.
     * @post    initialized = true
     */
    void update_prerequisites();

    /**
     * @brief   Invalides all the dependencies.
     */
    void invalide_dependencies();

private:
    /// recursive helper for check_prerequisites_cycle();
    void _check_prerequisites_cycle(vector<PVariableInfo> &trace,
            PVariableInfo v);

    void _update_bound_prerequisites(PVariableInfo pre);
    void _update_bound_variable_on_tmp_prerequisite();
    void _update_prerequisites(const PQuery &dep);

    void _check_dependency_cycle(const PVariableInfo &ref);
    
    // lock mechanism preventing cyclic dependencies
    bool evaluating_tag = false;

    // tag for Q_LET that the variable (and its prerequisites) is initialized
    bool initialized = false;

};

} // namespace rtl
} // namespace vam

#endif
