/**
 * @brief VAM-RTL manipulations. CLI user interface
 */
#include "language.h" // RTLVisitor
#include "querylanguage.h"  // RTL + manipulations + QueryParser
#include "queryinterpret.h" // query Interpret
#include <iomanip>
#include <fstream>
#include <readline/readline.h>  // interactive interpret
#include <readline/history.h>

using namespace vam;
using namespace std;

void usage()
{
    cout << "RLT manipulator (build " << __DATE__<<", "<< __TIME__ << ")\n";
    cout <<
    "syntax:\n"
    "   rtlmanip [command [options]]\n"
    "   rtlmanip input.vam\n"
    "commands:\n"
    "   simplify input.vam output.vam\n"
    "       Simplifies the model and makes it uniform\n"
    "   histogram input.vam output.vam\n"
    "       Tags all signals with their usages\n"
    "   traverse input.vam signal-id [depth]\n"
    "       Traverses all signal usages in the system for a given depth\n"
    "       (default depth: unlimited)\n"
    "   coi input.vam signal-ids time-frame [out-file]\n"
    "       Select cone-of-influence of the signals for a given time frame\n"
    "       signals-ids is comma separated list of signal identifiers.\n"
    "   coib input.vam signal-ids time-frame [out-file]\n"
    "       Select backward cone-of-influence of the signals for a given time\n"
    "       frame.\n"
    "if run with no command, readline ui will ask the user\n";
}

void simplify(string input_fname, string output_fname)
{
    RTLParser p;
    System s;
    p.parse_from_file(input_fname, &s);

    SystemManipulation m(s);
    m.simplify();

    RTLVisitor v;
    s.Accept(&v);

    ofstream o(output_fname);
    o << v.out.str() << endl;
    o.close();
}

void histogram(string input_fname, string output_fname)
{
    RTLParser p;
    System s;
    p.parse_from_file(input_fname, &s);

    SystemManipulation m(s);
    int usages = m.count_signal_usages();
    s.meta.addkv("variable_usages", std::to_string(usages));
    s.meta.addkv("signals_count", std::to_string(s.signals.size()));
    s.meta.addkv("usage_ratio", std::to_string(1.0*usages/s.signals.size()));

    RTLVisitor v;
    s.Accept(&v);
    ofstream o(output_fname);
    o << v.out.str() << endl;
    o.close();
}

void traverse_signal(string input_fname, string signame, int depth = -1)
{
    RTLParser p;
    System system;
    p.parse_from_file(input_fname, &system);

    Signal *s = system.signals[signame];
    if (s == nullptr)
    {
        cerr << "unknown signal " << signame << endl;
        return;
    }

    RTLVisitor v;

    int d;

    for (Expression *ref : s->references)
    {
        d = 1;
        TraverseOutputs outputs_of_ref(ref, 
                [](Expression *&e){ return e->op != OP_NOP; });
        for (Expression *e : outputs_of_ref)
        {
            v.out << "depth " << setw(2) << d << ": ";
            e->Accept(&v);
            v.out << endl;
            d++;
            if (depth > 0 && d > depth)
                break;
        }
    }

    cout << v.out.str();

}

void print_coi(ConeOfInfluence &coi)
{
    for (unsigned i = 0; i < coi.cone_signals.size(); i++)
    {
        cout << "CYCLE " << i << ":" << endl << "~~~~~~~~" << endl;
        if (coi.cone_signals[i].size())
            cout << "Signals:" << endl;
        for (auto signal : coi.cone_signals[i])
        {
            cout << "  " << setw(7) << left << signal->id;
            cout << " | " << signal->meta.keyval("comment");
            cout << endl;
        }
        if (coi.cone_storages[i].size())
            cout << "Storages:" << endl;
        for (auto st : coi.cone_storages[i])
            cout << "  " << st->id << endl;
        cout << endl;
    }
}

void print_coi_to_file(ConeOfInfluence &coi, string &ofile)
{
    ofstream o(ofile);
    o << "(cone-of-influence ";
    o << (coi.dir < 0 ? "backward\n" : "forward\n");

    for (unsigned i = 0; i < coi.cone_signals.size(); i++)
    {
        RTLVisitor v;

        v.out << "(cycle " << i;
        v.indent++;

        if (coi.dir < 0)
        {
            for (auto signal : coi.cone_signals[i])
            {
                v.VisitSignal(signal);
                ExprAssign *assign = find_setting_assignment(signal);
                if (assign)
                {
                    v.out << "\n    ";
                    v.VisitExprAssign(assign);
                }
                v.out << endl;
            }
        }
        else
        {
            for (auto signal : coi.cone_signals[i])
            {
                v.VisitSignal(signal);
                set<ExprAssign*> asss;
                for (Expression *ref : signal->references)
                {
                    TraverseOutputs outputs_of_ref(ref,
                           [](Expression *&e){return e->op == OP_ASSIGN;});
                    for (Expression *assign : outputs_of_ref)
                        if (lhs(assign) != ref)
                            asss.insert(static_cast<ExprAssign*>(assign));
                }
                for (auto assign : asss)
                {
                    v.out << "\n    ";
                    v.VisitExprAssign(assign);
                }
                v.out << endl; 
            }
        }

        v.out << endl;
        for (auto st : coi.cone_storages[i])
        {
            Register *reg = dynamic_cast<Register*>(st);
            MemPort *port = dynamic_cast<MemPort*>(st);
            Memory *mem = dynamic_cast<Memory*>(st);
            if (reg)
                v.VisitRegister(reg);
            if (port)
            {
                v.out << "\n  ";
                v.VisitMemPort(port);
            }
            if (mem)
                v.VisitMemory(mem);
            v.out << endl;
        }

        v.out << ")\n";

        o << v.out.str();
    }
    o << ")\n";
    o.close();
}

void coi(string input_fname, string signames, int cycles, DataflowDir dir,
        string ofile)
{
    RTLParser p;
    System system;
    p.parse_from_file(input_fname, &system);
    SystemManipulation manip(system);

    ConeOfInfluence coi;

    IdSet sigids;
    istringstream i_s_signames(signames);
    for (string signame; std::getline(i_s_signames, signame, ',');)
        sigids.insert(signame);

    manip.select_coi(sigids, cycles, dir, coi);

    print_coi(coi);

    if (!ofile.empty())
        print_coi_to_file(coi, ofile);
}

void execute_command(rtl::Interpret &interpret, const string &command,
        const string &script_file, bool quit_on_error)
{
    rtl::QueryParser qparser{interpret.system};
    if (!script_file.empty())
        qparser.load_script(script_file);
    else
        qparser.load_command(command);
    while (1)
    {
        rtl::Query query;
        try
        {
            query = qparser.parse_next_command();
        }
        catch(VAMException &e)
        {
            cerr << "error: " << e.msg << endl;
            if (quit_on_error)
                break;
            else
                continue;
        }
        if (query.type == rtl::Q_NONE)
            break;
        try
        {
            if (command.size() && query.type != rtl::Q_PRINT)
            {
                rtl::PQuery q = make_shared<rtl::Query>(rtl::Q_PRINT);
                *q << std::move(query);
                interpret.exec(q);
            }
            else
            {
                rtl::PQuery query_to_execute = make_shared<rtl::Query>(query);
                interpret.exec(query_to_execute);
            }
        }
        catch(VAMException &e)
        {
            cerr << "error: " << e.msg << endl;
            if (quit_on_error)
                break;
        }
    }
}

void query_command(const string &command, const string &ifile)
{
    RTLParser p;
    System system;
    if (ifile.size())
        p.parse_from_file(ifile, &system);
    rtl::Interpret interpret{system};
    execute_command(interpret, command, "", true);
}

void query_script(const string &script_file, const string &ifile)
{
    RTLParser p;
    System system;
    if (ifile.size())
        p.parse_from_file(ifile, &system);
    rtl::Interpret interpret{system};
    execute_command(interpret, "", script_file, true);
}

void readline_ui(string input_fname)
{
    RTLParser p;
    System system;
    if (input_fname.size())
        p.parse_from_file(input_fname, &system);
    rtl::Interpret interpret{system};

    char *input;
    const char *prompt = "rtl> ";
    string command;
    while ((input = readline(prompt)))
    {
        string command{input};
        add_history(input);
        free(input);
        if (command == "quit" || command == "(quit)" || command == "exit")
            break;
        execute_command(interpret, command, "", false);
        prompt = interpret._debug ? "debug-rtl> " : "rtl> ";
    }
    putchar('\n');
}

int main(int argc, char **argv)
{
    if (argc <= 1)
    {
        readline_ui("");
        return 0;
    }
    string command = argv[1];
    try
    {
        if (command == "simplify")
        {
            if (argc < 4)
            {
                usage();
                return 1;
            }
            simplify(argv[2], argv[3]);
            return 0;
        }
        else if (command == "histogram")
        {
            if (argc < 4)
            {
                usage();
                return 1;
            }
            histogram(argv[2], argv[3]);
            return 0;
        }
        else if (command == "traverse")
        {
            if (argc < 4)
            {
                usage();
                return 1;
            }
            int depth = -1;
            if (argc > 4)
                depth = stoi(argv[4]);
            traverse_signal(argv[2], argv[3], depth);
            return 0;
        }
        else if (command == "coi")
        {
            if (argc < 5)
            {
                usage();
                return 1;
            }
            int cycles = stoi(argv[4]);
            string ofile = (argc > 5) ? argv[5] : "";
            coi(argv[2], argv[3], cycles, DF_FORWARD, ofile);
            return 0;
        }
        else if (command == "coib")
        {
            if (argc < 5)
            {
                usage();
                return 1;
            }
            int cycles = stoi(argv[4]);
            string ofile = (argc > 5) ? argv[5] : "";
            coi(argv[2], argv[3], cycles, DF_BACKWARD, ofile);
            return 0;
        }
        else if (command == "--help")
            usage();
        else if (command == "-q" || command == "--query")
        {
            string command = argv[2];
            string ifile = argv[3] ? argv[3] : "";
            query_command(command, ifile);
        }
        else if (command == "-s")
        {
            string script_file = argv[2];
            string ifile = argv[3] ? argv[3] : "";
            query_script(script_file, ifile);
        }
        else
            readline_ui(command);
    }
    catch(VAMException &e)
    {
        e.print();
    }
    return 0;
}
