/**
 * @desc RTL manipulation
 */
#include "rtlmanip.h"
#include <stdlib.h> // abs
#include <assert.h>

namespace vam
{

// anonymous namespace to hide the following functions
namespace {

// The following inline-static-cast functions exists just for more readable
// syntax. Use with extreme care! Be sure, that 'e' is of a given type.

int tr_debug;

inline void mtag(Expression *e, const string &t)
{
    if (tr_debug)
        e->meta.tag(t);
}

// 1a. (== (_ a w) (_ b w)) for a==b ==> (_ 1 1)
// 1b. (== (_ a w) (_ b w)) for a!=b ==> (_ 0 1)
int tr1ab_eq_const(Expression *&e)
{
    if (e->op == OP_EQ)
    {
        if (lhs(e)->op == OP_CONST && rhs(e)->op == OP_CONST)
        {
            ExprConst *one = new ExprConst("1", 1);
            mtag(one, "simplified r1");
            if (value(lhs(e)) != value(rhs(e)))
                one->value = "0";
            return replace_outputs(e, one);
        }
    }
    return 0;
}

// 2. (== (_ a w) e) ==> (== e (_ a w)), the same for '!=','&','|'
int tr2_const_order(Expression *&e)
{
    if (e->op == OP_EQ || e->op == OP_NEQ ||
        e->op == OP_BAND || e->op == OP_BOR)
    {
        if (lhs(e)->op == OP_CONST && rhs(e)->op != OP_CONST)
        {
            Expression *l = lhs(e);
            lhs(e) = rhs(e);
            rhs(e) = l;
            mtag(e, "unified r2");
            return 1;
        }
    }
    return 0;
}

// -3. (== v (_ 1 1)) and (!= v (_ 0 1)) for one bit 'v' ==> v
int tr3_1bit_eq(Expression *&e)
{
    if (e->op == OP_EQ || e->op == OP_NEQ)
    {
        if (lhs(e)->bitwidth != 1) return 0;
        if (rhs(e)->op == OP_CONST)
        {
            if ((e->op == OP_EQ && value(rhs(e)) == "1") ||
                (e->op == OP_NEQ && value(rhs(e)) == "0"))
            {
                mtag(lhs(e), "simplified r3");
                return -replace_outputs(e, lhs(e));
            }
        }
    }
    return 0;
}

// -4. (!= v (_ 1 1)) and (== v (_ 0 1)) for one bit 'v' ==> (! v)
int tr4_1bit_eq(Expression *&e)
{
    if (e->op == OP_EQ || e->op == OP_NEQ)
    {
        if (lhs(e)->bitwidth != 1) return 0;
        if (rhs(e)->op == OP_CONST)
        {
            if ((e->op == OP_NEQ && value(rhs(e)) == "1") ||
                (e->op == OP_EQ && value(rhs(e)) == "0"))
            {
                ExprNot *neq = new ExprNot(lhs(e));
                mtag(neq, "simplified r4");
                return -replace_outputs(e, neq);
            }
        }
    }
    return 0;
}

// -5a. (! (_ 0 1)) ==> (_ 1 1), the same for '~'
// -5b. (! (_ 1 1)) ==> (_ 0 1), the same for '~'
int tr5ab_not_const(Expression *&e)
{
    if (e->op == OP_NOT || e->op == OP_INV)
    {
        if (operand(e)->bitwidth != 1) return 0;
        if (operand(e)->op == OP_CONST)
        {
            if (value(operand(e)) == "0")
                value(operand(e)) = "1";
            else if (value(operand(e)) == "1")
                value(operand(e)) = "0";
            else
                return 0;
            mtag(operand(e), "simplified r5");
            return -replace_outputs(e, operand(e));
        }
    }
    return 0;
}

// -6. (! (! e)) ==> e, the same for '~'
int tr6_not_not(Expression *&e)
{
    if ((e->op == OP_NOT && operand(e)->op == OP_NOT) ||
        (e->op == OP_INV && operand(e)->op == OP_INV))
    {
        mtag(operand(operand(e)), "simplified r6");
        return -replace_outputs(e, operand(operand(e)));
    }
    return 0;
}

// -7. (^^ (_ x y) z) ==> (_ x z)
int tr7_zex_const(Expression *&e)
{
    if (e->op == OP_ZEX && operand(e)->op == OP_CONST)
    {
        operand(e)->bitwidth = e->bitwidth;
        mtag(operand(e), "simplified r7");
        return -replace_outputs(e, operand(e));
    }
    return 0;
}

// -8. (^^ e w) for width of e=w ==> w
int tr8_no_zex(Expression *&e)
{
    if (e->op == OP_ZEX && operand(e)->bitwidth == e->bitwidth)
    {
        mtag(operand(e), "simplified r8");
        return -replace_outputs(e, operand(e));
    }
    return 0;
}

// -9a. (& e (_ 0 w)) for width of e=w ==> (_ 0 w)
int tr9a_set_zero(Expression *&e)
{
    if (e->op == OP_BAND && rhs(e)->op == OP_CONST && value(rhs(e)) == "0")
    {
        mtag(rhs(e), "simplified r9a");
        return -replace_outputs(e, rhs(e));
    }
    return 0;
}

// -9b. (| e (_ 2^w-1 w)) ==> (_ 2^w-1 w)
int tr9b_set_one(Expression *&e)
{
    if (e->op == OP_BOR && rhs(e)->op == OP_CONST)
    {
        // constant should be positive
        if (value(rhs(e))[0] == '-')
            return 0;
        unsigned long long c = std::stoull(value(rhs(e)));
        unsigned long long nn = 1 << rhs(e)->bitwidth;
        if (c == nn-1)
        {
            mtag(rhs(e), "simplified r9b");
            return -replace_outputs(e, rhs(e));
        }
    }
    return 0;
}

// -9c. (& e (_ 2^w-1 w)) ==> e
int tr9c_discard_one(Expression *&e)
{
    if (e->op == OP_BAND && rhs(e)->op == OP_CONST)
    {
        // constant should be positive
        if (value(rhs(e))[0] == '-')
            return 0;
        unsigned long long c = std::stoull(value(rhs(e)));
        unsigned long long nn = 1 << rhs(e)->bitwidth;
        if (c == nn-1)
        {
            mtag(lhs(e), "simplified r9c");
            return -replace_outputs(e, lhs(e));
        }
    }
    return 0;
}

// -9d. (| e (_ 0 w)) ==> e
int tr9d_discard_zero(Expression *&e)
{
    if (e->op == OP_BOR && rhs(e)->op == OP_CONST && value(rhs(e)) == "0")
    {
        mtag(lhs(e), "simplified r9d");
        return -replace_outputs(e, lhs(e));
    }
    return 0;
}

// -9e. (| e e) ==> e, the same for '&', '=='
int tr9e_reduce_same(Expression *&e)
{
    if ((e->op == OP_BOR || e->op == OP_BAND || e->op == OP_EQ) &&
        (lhs(e) == rhs(e) ||
         (lhs(e)->op == OP_VAR && rhs(e)->op == OP_VAR &&
          signal(lhs(e)) == signal(rhs(e))) ||
         (lhs(e)->op == OP_CONST && rhs(e)->op == OP_CONST &&
          value(lhs(e)) == value(rhs(e)))))
    {
        mtag(lhs(e), "simplified r9e");
        return -replace_outputs(e, lhs(e));
    }
    return 0;
}

// -9f. (& e1 e2), for bitwidth = 1 ==> (&& e1 e2)
// -9g. (| e1 e2), for bitwidth = 1 ==> (|| e1 e2)
int tr9fg_transform_logic(Expression *&e)
{
    if ((e->op == OP_BOR || e->op == OP_BAND) &&
        e->bitwidth == 1)
    {
        ExpressionPool *pool;
        if (e->op == OP_BOR)
            pool = new ExprOr(lhs(e), rhs(e));
        else
            pool = new ExprAnd(lhs(e), rhs(e));
        return -replace_outputs(e, pool);
    }
    return 0;
}

// -10. (cond ((e) e1) (e2)) ==> (? e e1 e2)
int tr10_cond2ite(Expression *&e)
{
    if (e->op == OP_COND)
    {
        if (cond(e)->cases.size() == 1)
        {
            ExprITE *ite = new ExprITE(cond(e)->cases[0].first,
                    cond(e)->cases[0].second, cond(e)->def);
            mtag(ite, "simplified r10");
            return -replace_outputs(e, ite);
        }
    }
    return 0;
}

// -11. (? e (_ 1 1) (_ 0 1)) ==> e
int tr11_no_ite(Expression *&e)
{
    if (e->op == OP_ITE && e->bitwidth == 1)
    {
        if (ite(e)->e_true->op == OP_CONST && ite(e)->e_false->op == OP_CONST)
            if (value(ite(e)->e_true) == "1" && value(ite(e)->e_false) == "0")
            {
                mtag(ite(e)->cond, "simplified r11");
                return -replace_outputs(e, ite(e)->cond);
            }
    }
    return 0;
}

// -12. (? e (_ 0 1) (_ 1 0)) ==> (! e)
int tr12_no_ite(Expression *&e)
{
    if (e->op == OP_ITE && e->bitwidth == 1)
    {
        if (ite(e)->e_true->op == OP_CONST && ite(e)->e_false->op == OP_CONST)
            if (value(ite(e)->e_true) == "0" && value(ite(e)->e_false) == "1")
            {
                ExprNot *neq = new ExprNot(ite(e)->cond);
                mtag(neq, "simplified r12");
                return -replace_outputs(e, neq);
            }
    }
    return 0;
}

// 13. (? (! e1) e2 e3) ==> (? e1 e3 e2)
int tr13_inv_ite(Expression *&e)
{
    if (e->op == OP_ITE &&
        (ite(e)->cond->op == OP_NOT || ite(e)->cond->op == OP_INV))
    {
        Expression *t = ite(e)->e_false;
        Expression *f = ite(e)->e_true;
        ite(e)->e_false = ite(e)->e_true = nullptr;
        // set nullptr to prevent the case, when replacing operand in case of,
        // ite(e)->cond == ite(e)->e_true or ite(e)->e_false
        ExprRef holder(operand(ite(e)->cond));
        e->replace_operand(ite(e)->cond, operand(ite(e)->cond));
        ite(e)->e_true = t;
        ite(e)->e_false = f;
        mtag(e, "simplified r13");
        return 1;
    }
    return 0;
}

// -13a. (? cond c c) ==> c
int tr13a_no_ite(Expression *&e)
{
    if (e->op == OP_ITE &&
        (ite(e)->e_true == ite(e)->e_false ||
         (ite(e)->e_true->op == OP_VAR && ite(e)->e_false->op == OP_VAR &&
          signal(ite(e)->e_true) == signal(ite(e)->e_false)) ||
         (ite(e)->e_true->op == OP_CONST && ite(e)->e_false->op == OP_CONST &&
          value(ite(e)->e_true) == value(ite(e)->e_false))))
    {
        mtag(lhs(e), "simplified r13e");
        return -replace_outputs(e, ite(e)->e_true);
    }
    return 0;

}

// -14. (cond ((== v (_ a w)) e1)
//          ((== v (_ b w)) e2) ...
//          ((== v (_ n w)) en) (ed)) ==>
//    (switch v ((_ a w) e1)
//              ((_ b w) e2) ...
//              ((_ n w) en) (ed)) - mark v as weak-selector
int tr14_cond2switch(Expression *&e)
{
    if (e->op == OP_COND)
    {
        Signal *v = nullptr;
        for (auto c : cond(e)->cases)
        {
            if (c.first->op == OP_EQ &&
                lhs(c.first)->op == OP_VAR &&
                rhs(c.first)->op == OP_CONST)
            {
                if (v == nullptr)
                    v = signal(lhs(c.first));
                else if (signal(lhs(c.first)) != v)
                {
                    v = nullptr;
                    break;
                }
            }
            else
            {
                v = nullptr;
                break;
            }
        }
        if (v != nullptr)
        {
            ExprCase *ec = new ExprCase();
            ec->add_ctrl_expr(new ExprVariable(v));
            for (auto c : cond(e)->cases)
                ec->add_case(rhs(c.first), c.second);
            if (cond(e)->def)
                ec->add_default(cond(e)->def);
            mtag(ec, "simplified r14");
            return -replace_outputs(e, ec);
        }
    }
    return 0;
}

// 17. (&& e1 e2 ... (_ 1 1) ... en) ==> (&& e1 e2 ... en)
// 18. (&& e1 e2 ... (_ 0 1) ... en) ==> (_ 0 1)
int tr1718_and_zero_one_exclusion(Expression *&e)
{
    if (e->op == OP_AND)
    {
        ExpressionList to_delete;
        for (auto o : *pool(e))
            if (o->op == OP_CONST)
            {
                if (value(o) == "0")
                {
                    // rule #18
                    mtag(o, "simplified r18");
                    return -replace_outputs(e, o);
                }
                else
                if (value(o) == "1")
                    // rule #17
                    to_delete.push_back(o);
            }
        for (auto o : to_delete)
            pool(e)->erase_operand(o);
        mtag(e, "simplified r17");
        return to_delete.size();
    }
    return 0;
}

// 19. (|| e1 e2 ... (_ 1 1) ... en) ==> (_ 1 1)
// 20. (|| e1 e2 ... (_ 0 1) ... en) ==> (|| e1 e2 ... en)
int tr1920_or_zero_one_exclusion(Expression *&e)
{
    if (e->op == OP_OR)
    {
        ExpressionList to_delete;
        for (auto o : *pool(e))
            if (o->op == OP_CONST)
            {
                if (value(o) == "1")
                {
                    // rule #19
                    mtag(o, "simplified r19");
                    return -replace_outputs(e, o);
                }
                else
                if (value(o) == "0")
                    // rule #20
                    to_delete.push_back(o);
            }
        for (auto o : to_delete)
            pool(e)->erase_operand(o);
        mtag(e, "simplified r20");
        return to_delete.size();
    }
    return 0;
}

// 21. (&& e1 e2 ... (&& ex ... ey) ... en) ==> (&& e1 e2 ...ex...ey...en)
// 22. (|| e1 e2 ... (|| ex ... ey) ... en) ==> (|| e1 e2 ...ex...ey...en)
// 22a.(|. e1 e2 ... (|. ex ... ey) ... en) ==> (|. e1 e2 ...ex...ey...en)
int tr2122_nested_and_or(Expression *&e)
{
    if (e->op == OP_AND || e->op == OP_OR || e->op == OP_ORJOIN)
    {
        int changes = 0;
        bool next_round;
        do {
            next_round = false;
            for (auto i = pool(e)->operands.begin();
                    i != pool(e)->operands.end(); ++i)
            {
                Expression *o = *i;
                if (o->op == e->op)
                {
                    next_round = true;
                    // insert ex ... ey at position of (&& or (|| expression
                    pool(e)->operands.insert(i,
                            pool(o)->operands.begin(),
                            pool(o)->operands.end());
                    // bind ex ... ey with superior expression
                    for (auto nested_op : *pool(o))
                        nested_op->bind(e);
                    // remove nested (&& or (||
                    pool(e)->erase_operand(o);
                    changes++;
                    break;
                }
            }
        }
        while (next_round);
        if (changes)
            mtag(e, "simplified r21,r22,r22a");
        return changes;
    }
    return 0;
}

// -23. (! (== e1 e2)) ==> (!= e1 e2)
// -24. (! (!= e1 e2)) ==> (== e1 e2)
int tr2324_not_eq(Expression *&e)
{
    if (e->op == OP_NOT || e->op == OP_INV)
    {
        if (operand(e)->bitwidth != 1) return 0;
        if (operand(e)->op == OP_EQ)
        {
            ExprNotEqual *neq =
                new ExprNotEqual(lhs(operand(e)), rhs(operand(e)));
            mtag(neq, "simplified r23");
            return -replace_outputs(e, neq);
        }
        if (operand(e)->op == OP_NEQ)
        {
            ExprEqual *eq =
                new ExprEqual(lhs(operand(e)), rhs(operand(e)));
            mtag(eq, "simplified r24");
            return -replace_outputs(e, eq);
        }
    }
    return 0;
}

// -27. (!= (& e (_ c w)) (_ 0 w)) aka in C: (e & c != 0),
//      where c == 2^n for some integer n ==> ([.] e n)
// -28. (== (& e (_ c w)) (_ c w)) aka in C: (e & c == c),
//      where c == 2^n for some integer n ==> ([.] e n)
int tr2728_extract_bit_a(Expression *&e)
{
    if (//rule 27
        (e->op == OP_NEQ &&
        lhs(e)->op == OP_BAND && rhs(lhs(e))->op == OP_CONST &&
        rhs(e)->op == OP_CONST && value(rhs(e)) == "0") ||
        // rule 28
        (e->op == OP_EQ &&
        lhs(e)->op == OP_BAND && rhs(lhs(e))->op == OP_CONST &&
        rhs(e)->op == OP_CONST && value(rhs(e)) == value(rhs(lhs(e)))))

    {
        // c should be positive
        if (value(rhs(lhs(e)))[0] == '-')
            return 0;
        unsigned long long c = std::stoull(value(rhs(lhs(e))));
        // search for n, where c == 2^n
        unsigned long long nn = 1;
        unsigned n;
        for (n = 0; n < sizeof(unsigned long long)*8; n++)
        {
            if (c == nn)
                break;
            nn <<= 1;
        }
        if (c == nn)
        {
            // hit!
            ExprBitExtract *bex = new ExprBitExtract(lhs(lhs(e)), n);
            mtag(bex, "simplified r27,r28");
            return -replace_outputs(e, bex);
        }
    }
    return 0;
}

// -29. (>> (& e (_ c w)) (_ n x)), where c == 2^n ==>
//          (. (_ 0 w-n-1) ([.] e n))
int tr29_extract_bit_shift(Expression *&e)
{
    if (e->op == OP_SHR &&
        lhs(e)->op == OP_BAND && rhs(lhs(e))->op == OP_CONST &&
        rhs(e)->op == OP_CONST)
    {
        // c and n should be positive
        if (value(rhs(lhs(e)))[0] == '-' ||
            value(rhs(e))[0] == '-')
            return 0;
        unsigned long long c = std::stoull(value(rhs(lhs(e))));
        unsigned long long n = std::stoull(value(rhs(e)));
        if (c == 1UL<<n && 1UL<<n != 0)
        {
            // hit!
            Expression *rplc;
            if (rhs(lhs(e))->bitwidth - n - 1 == 0)
                rplc = new ExprBitExtract(lhs(lhs(e)), n);
            else
                rplc = new ExprConcatenate(
                    new ExprConst("0", rhs(lhs(e))->bitwidth - n - 1),
                    new ExprBitExtract(lhs(lhs(e)), n));
            mtag(rplc, "simplified r29");
            return -replace_outputs(e, rplc);
        }
    }
    return 0;
}

// -30. (& (>> e (_ c x)) (_ 1 w)) ==>
//          (. (_ 0 w-1) ([.] e c) for w-1>0, or
//          ([.] e c)
int tr30_extract_bit_shift(Expression *&e)
{
    if (e->op == OP_BAND &&
        lhs(e)->op == OP_SHR && rhs(lhs(e))->op == OP_CONST &&
        rhs(e)->op == OP_CONST && value(rhs(e)) == "1")
    {
        Expression *rplc;
        if (rhs(e)->bitwidth - 1 > 0)
            rplc = new ExprConcatenate(
                new ExprConst("0", rhs(e)->bitwidth - 1),
                new ExprBitExtract(lhs(lhs(e)), stoi(value(rhs(lhs(e))))));
        else
            rplc = new ExprBitExtract(lhs(lhs(e)), stoi(value(rhs(lhs(e)))));
        mtag(rplc, "simplified r30");
        return -replace_outputs(e, rplc);
    }
    return 0;
}

// -31. (== (>> e (_ c x)) (_ 1 w)) ==> ([.] e c)
int tr31_extract_bit_shift(Expression *&e)
{
    if (e->op == OP_EQ &&
        lhs(e)->op == OP_SHR && rhs(lhs(e))->op == OP_CONST &&
        rhs(e)->op == OP_CONST && value(rhs(e)) == "1")
    {
        ExprBitExtract *bex =
            new ExprBitExtract(lhs(lhs(e)), stoi(value(rhs(lhs(e)))));
        mtag(bex, "simplified r31");
        return -replace_outputs(e, bex);
    }
    return 0;
}

// -32. ([] e (_ c w)) ==> ([.] e c)
int tr32_extract_bit(Expression *&e)
{
    if (e->op == OP_IDX && rhs(e)->op == OP_CONST)
    {
        ExprBitExtract *bex =
            new ExprBitExtract(lhs(e), stoi(value(rhs(e))));
        mtag(bex, "simplified r32");
        return -replace_outputs(e, bex);
    }
    return 0;
}

// 33. ([.] ([..] e l u) c) ==> ([.] e l+c)
int tr33_no_slice(Expression *&e)
{
    if (e->op == OP_BITEX && operand(e)->op == OP_SLICE)
    {
        static_cast<ExprBitExtract*>(e)->index +=
            static_cast<ExprSlice*>(operand(e))->from;
        ExprRef holder(operand(e));
        mtag(operand(operand(e)), "simplified r33");
        e->replace_operand(operand(e), operand(operand(e)));
        return 1;
    }
    return 0;
}

// -34. ([.] ([.] e i) 0) ==> ([.] e i)
int tr34_no_bex(Expression *&e)
{
    if (e->op == OP_BITEX && operand(e)->op == OP_BITEX &&
        static_cast<ExprBitExtract*>(e)->index == 0)
    {
        mtag(operand(e), "simplified r34");
        return -replace_outputs(e, operand(e));
    }
    return 0;
}

// 35. (^^ (^^ e w1) w2) ==> (^^ e w2), the same for '^^.'
int tr35_ex_ex(Expression *&e)
{
    if ((e->op == OP_ZEX || e->op == OP_SEX) && operand(e)->op == e->op)
    {
        ExprRef holder(operand(e));
        mtag(operand(operand(e)), "simplified r35");
        e->replace_operand(operand(e), operand(operand(e)));
        return 1;
    }
    return 0;
}

// -36. (== (^^ e w) (_ 1 w)) ==> ([.] e 0), the same for '^^.'
//      (!= (^^ e w) (_ 0 w)) ==> ([.] e 0), the same for '^^.'
// -37. (== (^^ e w) (_ 0 w)) ==> (! ([.] e 0)), the same for '^^.'
//      (!= (^^ e w) (_ 1 w)) ==> (! ([.] e 0)), the same for '^^.'
int tr3637_extract_bit(Expression *&e)
{
    if ((e->op == OP_EQ || e->op == OP_NEQ) &&
        (lhs(e)->op == OP_SEX || lhs(e)->op == OP_ZEX) &&
        rhs(e)->op == OP_CONST && value(rhs(e)) == "1")
    {
        if ((e->op == OP_EQ && value(rhs(e)) == "1") ||
            (e->op == OP_NEQ && value(rhs(e)) == "0"))
        {
            Expression *bex = new ExprBitExtract(operand(lhs(e)), 0);
            mtag(bex, "simplified r36");
            return -replace_outputs(e, bex);
        }
        if ((e->op == OP_EQ && value(rhs(e)) == "0") ||
            (e->op == OP_NEQ && value(rhs(e)) == "1"))
        {
            Expression *bex = new ExprNot(
                new ExprBitExtract(operand(lhs(e)), 0));
            mtag(bex, "simplified r37");
            return -replace_outputs(e, bex);
        }
    }
    return 0;
}

// 38. ([..] (>> e (_ c w)) 0 u) ==> ([..] e c c+u)
int tr38_no_shift(Expression *&e)
{
    if (e->op == OP_SLICE &&
        operand(e)->op == OP_SHR && rhs(operand(e))->op == OP_CONST &&
        static_cast<ExprSlice*>(e)->from == 0)
    {
        ExprSlice *s = static_cast<ExprSlice*>(e);
        int c = stoi(value(rhs(operand(e))));
        s->to += c;
        s->from = c;
        ExprRef holder(operand(e));
        mtag(operand(operand(e)), "simplified r38");
        e->replace_operand(operand(e), operand(operand(e)));
        return 1;
    }
    return 0;
}

// -39. (| (| e1 e2) e3) ==> (|. e1 e2 e3)
int tr39_or_join(Expression *&e)
{
    if (e->op == OP_BOR && lhs(e)->op == OP_BOR)
    {
        ExprOrJoin *orj = new ExprOrJoin();
        orj->add_operand(lhs(lhs(e)));
        orj->add_operand(rhs(lhs(e)));
        orj->add_operand(rhs(e));
        return -replace_outputs(e, orj);
    }
    return 0;
}

// 40. (|. e1 e2 ... c ... ex ... c ... ey) ==> (|. e1 e2 ...c...ex...ey)
//     the same for '&&', '||'
int tr40_single_const(Expression *&e)
{
    vector<string> values;
    if (e->op == OP_AND || e->op == OP_OR || e->op == OP_ORJOIN)
    {

        int changes = 0;
        bool next_round;
        do {
            next_round = false;
            for (auto i = pool(e)->operands.begin();
                      i != pool(e)->operands.end(); ++i)
            {
                Expression *o = *i;
                if (o->op == OP_CONST)
                {
                    if (find(values.begin(), values.end(),
                        value(o)) != values.end())
                    {
                        next_round = true;
                        // found repeated c
                        pool(e)->erase_operand(o);
                        changes++;
                        break;
                    }
                    else
                        values.push_back(value(o));
                }
            }
        }
        while (next_round);
        if (changes)
            mtag(e, "simplified r40");
        return changes;
    }
    return 0;
}

// 41. (|. e) ==> e, the same for '||' and '&&'
int tr41_remove_single_pool(Expression *&e)
{
    if ((e->op == OP_AND || e->op == OP_OR || e->op == OP_ORJOIN) &&
        pool(e)->operands.size() == 1)
    {
        mtag(pool(e)->operands[0], "simplified r41");
        return -replace_outputs(e, pool(e)->operands[0]);
    }
    return 0;
}

} // anonymous namespace

const string usage_rel_with_constant(Expression *e)
{
    if ((e->op == OP_EQ || e->op == OP_NEQ ||
        e->op == OP_LT || e->op == OP_GT ||
        e->op == OP_LE || e->op == OP_GE ||
        e->op == OP_SLT || e->op == OP_SGT ||
        e->op == OP_SLE || e->op == OP_SGE) &&
        (rhs(e)->op == OP_CONST || lhs(e)->op == OP_CONST))
        return "rel_with_const";
    else
        return "";
}

const string usage_switch_ctrl(Expression *ref, Expression *e)
{
    if (e->op == OP_CASE &&
        static_cast<ExprCase*>(e)->ctrl == ref)
        return "switch_ctrl";
    else
        return "";
}

int SystemManipulation::count_usages(Signal *s)
{
    int usages = 0;
    for (Expression *ref : s->references)
    {
        if (ref->op == OP_VAR)
        {
            for (Expression *output : ref->outputs)
            {
                s->meta.node("used_by")["all"] += "|";
                s->meta.node("used_by")[
                    tbl_oper_to_str[output->op]] += "|";
                s->meta.node("class_used_by")[
                    tbl_oper_to_group[output->op]] += "|";
                string cls;
                cls = usage_rel_with_constant(output);
                if (!cls.empty())
                    s->meta.node("class_used_by")[cls] += "|";
                cls = usage_switch_ctrl(ref, output);
                if (!cls.empty())
                    s->meta.node("class_used_by")[cls] += "|";
                usages++;
            }
        }
        else
            // warning, should be via ExprVar only
            s->meta.tag("unexpected_direct_usage");
    }
    return usages;
}

int expression_uniform_fixpoint(Expression *&e, ExpressionList &, void *)
{
    int i = 0;
    int temp;
    i += tr1ab_eq_const(e);
    i += tr2_const_order(e);
    if ((temp = tr3_1bit_eq(e)) < 0) return -(i+(-temp));
    if ((temp = tr4_1bit_eq(e)) < 0) return -(i+(-temp));
    if ((temp = tr5ab_not_const(e)) < 0) return -(i+(-temp));
    if ((temp = tr6_not_not(e)) < 0) return -(i+(-temp));
    if ((temp = tr7_zex_const(e)) < 0) return -(i+(-temp));
    if ((temp = tr8_no_zex(e)) < 0) return -(i+(-temp));
    if ((temp = tr9a_set_zero(e)) < 0) return -(i+(-temp));
    if ((temp = tr9b_set_one(e)) < 0) return -(i+(-temp));
    if ((temp = tr9c_discard_one(e)) < 0) return -(i+(-temp));
    if ((temp = tr9d_discard_zero(e)) < 0) return -(i+(-temp));
    if ((temp = tr9e_reduce_same(e)) < 0) return -(i+(-temp));
    if ((temp = tr9fg_transform_logic(e)) < 0) return -(i+(-temp));
    if ((temp = tr10_cond2ite(e)) < 0) return -(i+(-temp));
    if ((temp = tr11_no_ite(e)) < 0) return -(i+(-temp));
    if ((temp = tr12_no_ite(e)) < 0) return -(i+(-temp));
    i += tr13_inv_ite(e);
    if ((temp = tr13a_no_ite(e)) < 0) return -(i+(-temp));
    if ((temp = tr14_cond2switch(e)) < 0) return -(i+(-temp));
    if ((temp = tr1718_and_zero_one_exclusion(e)) < 0) return -(i+(-temp));
    else i += temp;
    if ((temp = tr1920_or_zero_one_exclusion(e)) < 0) return -(i+(-temp));
    else i += temp;
    i += tr2122_nested_and_or(e);
    if ((temp = tr2324_not_eq(e)) < 0) return -(i+(-temp));
    if ((temp = tr2728_extract_bit_a(e)) < 0) return -(i+(-temp));
    if ((temp = tr29_extract_bit_shift(e)) < 0) return -(i+(-temp));
    if ((temp = tr30_extract_bit_shift(e)) < 0) return -(i+(-temp));
    if ((temp = tr31_extract_bit_shift(e)) < 0) return -(i+(-temp));
    if ((temp = tr32_extract_bit(e)) < 0) return -(i+(-temp));
    i += tr33_no_slice(e);
    if ((temp = tr34_no_bex(e)) < 0) return -(i+(-temp));
    i += tr35_ex_ex(e);
    if ((temp = tr3637_extract_bit(e)) < 0) return -(i+(-temp));
    i += tr38_no_shift(e);
    if ((temp = tr39_or_join(e)) < 0) return -(i+(-temp));
    i += tr40_single_const(e);
    if ((temp = tr41_remove_single_pool(e)) < 0) return -(i+(-temp));
    return i;
}

/// @param data    pointer to current FunNode
int update_fnode_inputs(Expression *&e, ExpressionList &,
        void *data)
{
    if (e->op == OP_VAR)
    {
        FunNode *fnode = reinterpret_cast<FunNode*>(data);
        fnode->inputs.insert(signal(e));
    }
    return 0;
}

/**
 * Routines to preserve model consistency
 */

void SystemManipulation::update_fnode_io()
{
    invalidates = false;
    for (auto fnode : system.fnodes)
    {
        fnode->inputs.clear();
        fnode->outputs.clear();
        for (ExprAssign *&assign : fnode->assignments)
        {
            ExprVariable *writing_ref = dynamic_cast<ExprVariable*>(assign->lhs);
            assert(writing_ref != nullptr);
            signal(writing_ref)->writing_expr = writing_ref;
            fnode->outputs.insert(signal(writing_ref));

            for_each_operand_dfs(assign->rhs, update_fnode_inputs, (void*)fnode);
        }
    }
}

ExprAssign *find_setting_assignment(Signal *s)
{
    for (Expression *ref : s->references)
        for (Expression *e : ref->outputs)
            if (e->op == OP_ASSIGN && lhs(e) == ref)
                return static_cast<ExprAssign*>(e);
    return nullptr;
}

/**
 * Generation of cone of influence
 */

void SystemManipulation::get_forward_expansion(Signal *s, SignalSet &open,
        SignalSet &closed)
{
    for (Expression *ref : s->references)
    {
        TraverseOutputs outputs_of_ref(ref,
               [](Expression *&e){return e->op == OP_ASSIGN;});

        for (Expression *assign : outputs_of_ref)
            if (lhs(assign) != ref)
                if (lhs(assign)->op == OP_VAR)
                {
                    Signal *signal_to_add = signal(lhs(assign));
                    if (!closed.contains(signal_to_add))
                    {
                        closed.insert(signal_to_add);
                        open.insert(signal_to_add);
                    }
                }
    }
}

void SystemManipulation::get_backward_expansion(Signal *s, SignalSet &open,
        SignalSet &closed)
{
    Expression *assign = find_setting_assignment(s);
    if (assign == nullptr)
        return;

    TraverseInputs inputs_of_assign(rhs(assign),
            [](Expression *&e){ return e->op == OP_VAR; });

    for (Expression *evar : inputs_of_assign)
    {
        Signal *signal_to_add = signal(evar);
        if (!closed.contains(signal_to_add))
        {
            closed.insert(signal_to_add);
            open.insert(signal_to_add);
        }
    }
}

void SystemManipulation::select_coi(const IdSet &id_open,
        int time_in_cycles, DataflowDir dir, ConeOfInfluence &coi)
{
    assert(dir == DF_FORWARD || dir == DF_BACKWARD);
    SignalSet open;
    for (auto id : id_open)
    {
        Signal *s = system.signals[id];
        ASSERT(s != nullptr, "unknown signal " << id);
        open.insert(s);
    }

    coi.clear();
    coi.dir = dir;

    for (int cycle = 0; cycle <= time_in_cycles; cycle++)
    {
        coi.cone_signals.push_back(SignalSet());
        SignalSet &closed = coi.cone_signals.back();

        coi.cone_storages.push_back(StorageSet());
        StorageSet &closed_storages = coi.cone_storages.back();

        coi.cone_exprs.push_back(GenerateSet<Expression*>());
        // TODO: make typedef
        GenerateSet<Expression*> &exprs = coi.cone_exprs.back();

        // initialize 'closed' aka cone_signals[cycle]
        if (cycle == 0)
            closed.merge(open);
        else
        {
            // include all signals which may be set in current cycle
            // e.g. for each storage which has been set before and its timings
            // fits to current time frame
            for (int i = 0; i < cycle; i++)
            {
                // pipeline distance
                tfint dist{cycle-i, cycle-i};
                if (dir == DF_BACKWARD)
                    dist = -dist;

                for (auto st : coi.cone_storages[i])
                    if (dir == DF_FORWARD) // forward coi
                        st->get_influenced_outputs(nullptr, dist, closed);
                    else // backward coi
                        st->get_influenced_inputs(nullptr, dist, closed);
            }
        }

        // expand set of signals
        open.clear();
        open.merge(closed);
        while (open.size())
        {
            Signal *top = *open.begin();
            open.erase(top);

            // connected signals to signal top, i.e., append new_sig to
            // open+closed, when:
            //
            //    ----top--->(FNODE, :=)---new_sig-->
            //
            // (the same for reverse direction)
            if (dir == DF_FORWARD)
            {
                get_forward_expansion(top, open, closed);
                for (Expression *ref : top->references)
                {
                    TraverseOutputs outputs_of_ref(ref,
                        [](Expression *&e){return e->op == OP_ASSIGN;});

                    for (Expression *assign : outputs_of_ref)
                        if (lhs(assign) != ref)
                            exprs.insert(assign);
                }
            }
            else
            {
                get_backward_expansion(top, open, closed);
                Expression *assign = find_setting_assignment(top);
                if (assign != nullptr)
                    exprs.insert(assign);
            }

            // connected storages to signal top, i.e.,
            //  case1: if the signal is connected to some STORAGE_X, add it.
            //
            //    ---top--->|STORAGE_X|
            //
            //  case2: and if the storage propagates with 0-delay, expand
            //    open+closed with propagated signals
            //
            //              |STORAGE_X|
            //    ---top--->|- - - - >|---0-delayed_output--->
            //
            // (the same for reverse direction)
            for (auto connection: top->connections)
                if (dir == DF_FORWARD)
                {
                    if (connection.storage->contains_input(top, TF_ALL))
                    {
                        // case1
                        closed_storages.insert(connection.storage);
                        // case2
                        SignalSet sigs_0delayed;
                        connection.storage->get_influenced_outputs(
                                top, tfint{0}, sigs_0delayed);
                        open.merge(sigs_0delayed);
                        closed.merge(sigs_0delayed);
                    }
                }
                else
                {
                    if (connection.storage->contains_output(top, TF_ALL))
                    {
                        // case1
                        closed_storages.insert(connection.storage);
                        // case2
                        SignalSet sigs_0delayed;
                        connection.storage->get_influenced_inputs(
                                top, tfint{0}, sigs_0delayed);
                        open.merge(sigs_0delayed);
                        closed.merge(sigs_0delayed);
                    }
                }
        }
    }

    return;
}

/**
 * Manipulations simplifying the model
 */

ExprAssign *SystemManipulation::find_single_assignment()
{
    // find the first useless assignment
    for (auto fnode : system.fnodes)
        for (ExprAssign *a : fnode->assignments)
            if (a->rhs->op == OP_VAR)
                return a;
    return nullptr;
}

ExprAssign *SystemManipulation::find_onebit_constant_assignment()
{
    // find the first useless assignment: variable := one-bit-constant
    // where variable is not connected to any port
    for (auto fnode : system.fnodes)
        for (ExprAssign *a : fnode->assignments)
            if (a->rhs->op == OP_CONST &&
                a->rhs->bitwidth == 1 &&
                static_cast<Signal*>(
                    static_cast<ExprVariable*>(a->lhs)->v
                    )->connections.empty())
                return a;
    return nullptr;
}

ExprAssign *SystemManipulation::find_singlepurpose_assignment()
{
    // find the first singlepurpose assignment (i.e. lhs variable is read at
    // most once)
    for (auto fnode : system.fnodes)
        for (ExprAssign *a : fnode->assignments)
        {
            ExprVariable *vref = static_cast<ExprVariable*>(a->lhs);
            Signal *s = static_cast<Signal*>(vref->v);
            if (s->connections.size())
                continue;
            int nref = 0;
            for (Expression *ref : vref->v->references)
            {
                nref += ref->outputs.size();
                // FIXME: use of ExprRef may lead to invalid nref count
                if (nref > 2) // one for 'rhs := ...', and one that uses it
                    break;
            }
            if (nref <= 2)
                return a;
        }
    return nullptr;
}

void SystemManipulation::replace_single_assignment(ExprAssign *as)
{
    /*      +---------+
     * Let  | a := b; |
     *      +---------+
     *
     * step 0: before replacement:
     *
     * ExprAssign as: lhs=v1, rhs=v4, outputs=(fnodeX_pool,)
     * Variable a: references=(v1,v2,v3)
     * Variable b: references=(v4,...)
     * ExprVariable v1: v=a, outputs=(as,e3)
     * ExprVariable v2: v=a, outputs=(e1,)
     * ExprVariable v3: v=a, outputs=(e2,)
     * ExprVariable v4: v=b, outputs=(as,eA,...,eB)
     * Expression e1: operands=(eX,...,eY,v2)
     * Expression e2: operands=(v3,eP,...,eQ)
     * Expression e3: operands=(v1,)
     *
     * step 1: after replacement:
     *
     * ExprAssign as: lhs=v1, rhs=v4, outputs=(fnodeX_pool,)
     * Variable a: references=(v1,)
     * Variable b: references=(v4,...)
     * ExprVariable v1: v=a, outputs=(as,)
     * ExprVariable v2: v=a, outputs=(,) -- removed
     * ExprVariable v3: v=a, outputs=(,) -- removed
     * ExprVariable v4: v=b, outputs=(as,eA,...,eB,e1,e2)
     * Expression e1: operands=(eX,...,eY,v4)
     * Expression e2: operands=(v4,eP,...,eQ)
     * Expression e3: operands=(v4,)
     *
     * step 2: removal of signal connections (if both a and b are signals)
     *
     * step 3: cleanup:
     *
     * ExprAssign as: lhs=v1, rhs=v4 -- removed
     * Expression fnodeX_pool aka fnode owner -- removed
     * Variable a: references=(v1,) -- removed
     * Variable b: references=(v4,...)
     * ExprVariable v1: v=a, outputs=(as,) -- removed
     * ExprVariable v4: v=b, outputs=(eA,...,eB,e1,e2)
     * Expression e1: operands=(eX,...,eY,v4)
     * Expression e2: operands=(v4,eP,...,eQ)
     * Expression e3: operands=(v4,)
     */
    Expression *replacement = as->rhs;

    // LHS is ExprVariable*: invariant required by ExprRTLAssignments
    Variable *orig_var = static_cast<ExprVariable*>(as->lhs)->v;

    // step 1: replacement (splitted into 1a and 1b because of
    // replace_operand()/remove()/delete invalidates iterator of
    // orig_var->references iterator)
    // 1a: setup set of all expressions referencing to the variable, i.e.
    //     set of pairs (X,Y) where X is the expression in which the
    //     replacement must be performed, and Y is the variable reference
    //     (ExprVariable), used by X, which must be replaced by 'replacement'
    vector<pair<Expression*,Expression*> > refs;
    for (Expression *vX : orig_var->references) // aka a.references={v1,v2,v3}
        // here, vX is ExprVariable*
        for (Expression *e : vX->outputs) // for all variable references...
            // aka {as,e1,e2,e3}
            if (e != as)  // ...excluded the one we come from, aka {e1,e2,e3}
                refs.push_back(make_pair(e,vX)); // vX \in {v2,v3} for v4
    // 1b: for each such an expression, replace the variable with the
    //     given replacement
    for (auto p : refs)
    {
        p.first->replace_operand(p.second, replacement);
    }

    // step 2: signal connection replacement, i.e.
    // any storage sa is connected to, it should be connected with sb
    Signal *sa = dynamic_cast<Signal*>(orig_var);
    Signal *sb;
    ExprVariable *vref = dynamic_cast<ExprVariable*>(replacement);
    if (vref)
        sb = dynamic_cast<Signal*>(vref->v);
    else
        sb = nullptr;
    if (sa && sb) // are both variables sa and sb signals?
    {
        for (SignalConnection &connection : sa->connections)
        {
            // reconnect sa with sb in the storage
            connection.storage->reconnect(connection.type, sa, sb);
            // add connection info into sb
            sb->connections.push_back(connection);
        }
    }
    if (sa)
        system.signals.erase(sa);

    // step 3: cleanup
    // if the owner of the 'as' is the pool of some fnode:
    if (as->outputs.size() == 1 && as->outputs[0]->op == OP_NOP)
    {
        ExprRTLAssignments *pool =
            static_cast<ExprRTLAssignments*>(as->outputs[0]);
        pool->remove_operand(as);
        // remove the fnode from the system if useless
        if (pool->operands.empty())
        {
            system.fnodes.erase(pool->owner);
            delete pool->owner;
        }
        else if (sa && sb)
        {
            replace_sig_in_set(sa, sb, pool->owner->inputs);
            replace_sig_in_set(sa, sb, pool->owner->outputs);
        }
    }
    if (sa && sa->references.empty())
        delete sa;

}

int SystemManipulation::for_each_operand_dfs(Expression *&root,
        ExprCallback cb, void *data)
{
    ExpressionList bt;

    // number of performed manipulations
    int m = 0; // temporal
    int manips = 0; // overall

    // open set of operand iterators
    vector<Expression::iterator> open;

    // manipulate over dfs-search-starting expression
    m = cb(root, bt, data);
    manips += abs(m);
    if (m < 0) return -manips;
    if (m > 0 && invalidates) return manips;

    // insert the first operand
    open.push_back(root->begin());
    bt.push_back(root);

    while (!open.empty())
    {
        Expression::iterator i = open.back();
        if (i == i.refexpr->end())
        {
            open.pop_back();
            bt.pop_back();
            if (!open.empty())
                ++open.back();
        }
        else
        {
            // call the manipulator over the expression behind iterator
            int m = cb(*i, bt, data);
            manips += abs(m);

            // stop if changed and invalidated
            if (m < 0) return -manips;
            if (m > 0 && invalidates) return manips;

            // continue deep search
            open.push_back(i->begin());
            bt.push_back(*i);
        }
    }
    return manips;
}

int SystemManipulation::remove_single_assignments()
{
    ExprAssign *as;
    int i = 0;
    while ((as = find_single_assignment()) != nullptr)
    {
        i++;
        replace_single_assignment(as);
    }
    while ((as = find_onebit_constant_assignment()) != nullptr)
    {
        i++;
        replace_single_assignment(as);
    }
    return i;
}

int SystemManipulation::remove_singlepurpose_assignments()
{
    ExprAssign *as;
    int i = 0;
    while ((as = find_singlepurpose_assignment()) != nullptr)
    {
        i++;
        replace_single_assignment(as);
    }
    return i;
}

void SystemManipulation::add_fork_fnode(SignalSet &sset, SignalConnection sc)
{
    assert(sset.size());
    Signal *ref = *sset.begin();
    Signal *join = new Signal;

    // make a new signal taking some properties from the first signal of the
    // given set
    join->id = system.get_unique_identifier(sc.storage->id + "_");
    join->bitwidth = ref->bitwidth;
    join->vt = ref->vt;
    join->dir = ref->dir;
    join->inputs = ref->inputs;
    join->inputs_name = ref->inputs_name;
    join->stage = ref->stage;
    join->add_connection(sc.type, sc.storage);

    // create a fork node
    FunNode *fork = new FunNode;
    fork->id = system.get_unique_identifier("fork");

    // insert new assignments to the fork node
    for (auto sig : sset)
    {
        ExprAssign *a = new ExprAssign{
            new ExprVariable{sig},
            new ExprVariable{join}};
        sig->owner = fork;
        fork->assignments.add_operand(a);
        sig->remove_connection(sc.type, sc.storage);
    }

    sset.clear();
    sset.insert(join);

    system.signals.insert(join);
    system.fnodes.insert(fork);
}

void SystemManipulation::add_fork_fnodes()
{
    for (auto reg : system.registers)
        if (reg->q.size() > 1)
            add_fork_fnode(reg->q, SignalConnection{CT_Q, reg});

    for (auto mem : system.memories)
        for (auto port : mem->ports)
            if (port->is_read() && port->data.size() > 1)
                add_fork_fnode(port->data, SignalConnection{CT_DATA, port});
}

void SystemManipulation::add_forward_fnode(Signal *sig, SignalConnection &sc)
{
    Signal *dummy = new Signal;
    dummy->id = system.get_unique_identifier(sc.storage->id + "_");
    dummy->bitwidth = sig->bitwidth;
    dummy->vt = sig->vt;
    dummy->dir = sig->dir;
    dummy->inputs = sig->inputs;
    dummy->inputs_name = sig->inputs_name;
    dummy->stage = sig->stage;
    dummy->meta.copy(sig->meta);

    dummy->add_connection(sc.type, sc.storage);
    sig->remove_connection(sc.type, sc.storage);
    sc.storage->reconnect(sc.type, sig, dummy);

    FunNode *forward = new FunNode;
    forward->id = system.get_unique_identifier("forward");

    ExprAssign *a = new ExprAssign{
        new ExprVariable{sig},
        new ExprVariable{dummy}};

    sig->owner = forward;
    forward->assignments.add_operand(a);
    system.signals.insert(dummy);
    system.fnodes.insert(forward);
}

void SystemManipulation::add_forward_fnodes()
{
    for (auto sig : system.signals)
    {
        if (sig->connections.size() > 1)
        {
            SignalConnection read_connection;
            // search for influential connection and check that it is the only
            // influential connection
            for (auto &connection : sig->connections)
                if (connection.is_read())
                {
                    assert(read_connection.type == CT_UNDEFINED);
                    read_connection = connection;
                }
            if (read_connection.type != CT_UNDEFINED)
                add_forward_fnode(sig, read_connection);
        }
    }
}

int SystemManipulation::make_uniform_fixpoint()
{
    tr_debug = debug;
    return for_each_operand(expression_uniform_fixpoint, nullptr, false);
}

int SystemManipulation::for_each_operand(ExprCallback cb, void *data,
        bool invalidates)
{
    int manipulations = 0;
    this->invalidates = invalidates;

    /**
     * NOTE: DANGER OF INFINITE LOOP:
     *  if 'cb' callback causes the restart of the loop, then it MUST ensure
     *  the system is changed so that the same 'cb' over the same expression is
     *  never called again.
     *
     *  If it happens (e.g. when analysing during debugging, where the he.. is
     *  the error), check the callback. It most probably causes either:
     *    - (directly unchanged) the system does not change, but returns !=0
     *      value, or
     *    - (undirectly unchanged) the system changes to the state it was somehow
     *      before, so another callback changes it again to the same state.
     */
    nextround:
    for (auto fnode : system.fnodes)
        for (ExprAssign *&assign : fnode->assignments)
        {
            int i = for_each_operand_dfs(assign->lhs, cb, data);
            manipulations += abs(i); // minus-sign means temporary invalidate
            if ((i == 0) || (i > 0 && !invalidates))
            {
                // continue search either if no manips. have been done, or
                // some have been done, but all (iterators) are valid
                i = for_each_operand_dfs(assign->rhs, cb, data);
                manipulations += abs(i);
            }
            if (i != 0 && (invalidates || i < 0))
                goto nextround;
        }
    return manipulations;
}

/**
 * Some heuristics over the model
 */

int SystemManipulation::count_signal_usages()
{
    int usages = 0;
    for (auto signal : system.signals)
    {
        usages += count_usages(signal);
    }
    return usages;
}

Expression *SystemManipulation::encapsulate_expr(Expression *e,
        string prefix, bool encapsulate_const, Signal **s)
{
    if (e->op == OP_VAR)
        return e;
    if (e->op == OP_CONST && !encapsulate_const)
        return e;

    // make a new signal
    Signal *signal = new Signal;
    signal->id = system.get_unique_identifier(prefix);
    signal->bitwidth = e->bitwidth;
    signal->meta.copy(e->meta);

    // make a new fnode for the signal
    FunNode *fnode = new FunNode;
    fnode->id = system.get_unique_identifier("f_" + signal->id, true);
    fnode->meta.addkv("comment", "encapsulation");

    // assign the expression to the new signal
    ExprAssign *a = new ExprAssign{
        new ExprVariable{signal}, e};

    // replace each expression output with reference to the new signal
    ExprVariable *ref = new ExprVariable{signal};
    replace_outputs(e, ref, a);
    if (ref->outputs.empty()) // expression 'e' is not connected
        delete ref;

    // connect signal, assignment, and fnode and signal+fnode to system
    signal->owner = fnode;
    fnode->assignments.add_operand(a);
    system.signals.insert(signal);
    system.fnodes.insert(fnode);

    // if interested in encapsulated signal
    if (s != nullptr)
        *s = signal;

    return nullptr;
}

Signal *SystemManipulation::encapsulate_expr_to_signal(Expression *e,
        string prefix)
{
    if (e->op == OP_VAR)
        return signal(e);

    Signal *signal;
    Expression *ret = encapsulate_expr(e, prefix, true, &signal);
    ASSERT(ret == nullptr, "unexpected control flow");

    return signal;
}

namespace {

int tr_ite2switch(Expression *&e, ExpressionList&, void*)
{
    if (e->op == OP_ITE)
    {
        ExprITE *ite = static_cast<ExprITE*>(e);
        ExprCase *sw = new ExprCase{};
        mtag(sw, "ite2switch");
        sw->add_ctrl_expr(ite->cond);
        sw->add_case(new ExprConst{"1", 1}, ite->e_true);
        sw->add_case(new ExprConst{"0", 1}, ite->e_false);
        return -replace_outputs(e, sw);
    }
    return 0;
}

int tr_encaps_switch_ctrl(Expression *&e, ExpressionList&, void *data)
{
    SystemManipulation *m = reinterpret_cast<SystemManipulation*>(data);
    if (e->op == OP_CASE)
    {
        ExprCase *sw = static_cast<ExprCase*>(e);
        if (sw->ctrl->op != OP_VAR)
        {
            m->encapsulate_expr_to_signal(sw->ctrl, "selector");
            return -1;
        }
    }
    return 0;
}

int tr_encaps_switch_cases(Expression *&e, ExpressionList&, void *data)
{
    SystemManipulation *m = reinterpret_cast<SystemManipulation*>(data);
    if (e->op == OP_CASE)
    {
        ExprCase *sw = static_cast<ExprCase*>(e);
        int manips = 0;
        for (auto c : sw->cases)
            if (m->encapsulate_expr(c.second, "sel_flow", false, nullptr)
                    == nullptr)
                manips++;
        return -manips;
    }
    return 0;
}

int tr_encaps_switch(Expression *&e, ExpressionList&, void *data)
{
    SystemManipulation *m = reinterpret_cast<SystemManipulation*>(data);
    if (e->op == OP_CASE)
    {
        // is the switch assigned to a variable?
        if (e->outputs.size() == 1 && e->outputs[0]->op == OP_ASSIGN)
            return 0;
        m->encapsulate_expr_to_signal(e, "vswitch");
        return -1;
    }
    return 0;
}

Expression *tr_cond2switch_genact(vector<Signal*> &actsignals, Expression *ccase)
{
    ExprAnd *rhs = new ExprAnd;
    for (auto s : actsignals)
        rhs->add_operand(new ExprNot{new ExprVariable{s}});
    rhs->add_operand(ccase);
    return rhs;
}

Expression *tr_cond2switch_gensel(vector<Signal*> &actsignals)
{
    ASSERT(actsignals.size() > 1, "unexpected few cases");
    Expression *rhs = nullptr;
    for (unsigned i = actsignals.size(); i > 0; i--)
    {
        Signal *s = actsignals[i-1];
        if (rhs == nullptr)
            rhs = new ExprVariable{s};
        else
        {
            ExprConcatenate *c = new ExprConcatenate{
                new ExprVariable{s}, rhs};
            rhs = c;
        }
    }
    return rhs;
}

int tr_cond2switch(Expression *&e, ExpressionList&, void *data)
{
    SystemManipulation *m = reinterpret_cast<SystemManipulation*>(data);
    if (e->op == OP_COND)
    {
        ExprCond *cond = static_cast<ExprCond*>(e);

        Signal *selector;
        if (cond->cases.size() == 1)
            selector = m->encapsulate_expr_to_signal(
                    cond->cases[0].first, "selector");
        else
        {
            vector<Signal*> actsignals;
            for (auto c : cond->cases)
            {
                // first case
                if (actsignals.empty())
                {
                    Signal *act =
                        m->encapsulate_expr_to_signal(c.first, "sel_case");
                    actsignals.push_back(act);
                }
                // rest
                else
                {
                    Expression *actrhs =
                        tr_cond2switch_genact(actsignals, c.first);
                    Signal *act =
                        m->encapsulate_expr_to_signal(actrhs, "sel_case");
                    actsignals.push_back(act);
                }
            }

            Expression *selectorrhs = tr_cond2switch_gensel(actsignals);
            selector = m->encapsulate_expr_to_signal(selectorrhs, "selector");
        }

        ExprCase *sw = new ExprCase;
        sw->add_ctrl_expr(new ExprVariable{selector});
        unsigned case_value = 1;
        for (auto c : cond->cases)
        {
            auto first = new ExprConst{to_string(case_value),
                selector->bitwidth};
            sw->add_case(first, c.second);
            case_value <<= 1;
        }
        if (cond->def)
            sw->add_default(cond->def);

        return -replace_outputs(e, sw);
    }
    return 0;
}

int tr_unifyselectors(Expression *&e, ExpressionList &bt, void *data)
{
    int temp;
    if ((temp = tr_ite2switch(e, bt, data)))
        return temp;
    if ((temp = tr_encaps_switch_ctrl(e, bt, data)))
        return temp;
    if ((temp = tr_encaps_switch_cases(e, bt, data)))
        return temp;
    if ((temp = tr_cond2switch(e, bt, data)))
        return temp;
    if ((temp = tr_encaps_switch(e, bt, data)))
        return temp;
    return 0;
}

}

void SystemManipulation::transform_ite_to_switch()
{
    tr_debug = debug;
    for_each_operand(tr_ite2switch, nullptr);
}

void SystemManipulation::transform_cond_to_switch()
{
    tr_debug = debug;
    for_each_operand(tr_cond2switch, this);
}

void SystemManipulation::unify_selectors()
{
    tr_debug = debug;
    for_each_operand(tr_unifyselectors, this);
    update_fnode_io();
}

void SystemManipulation::simplify()
{
    remove_single_assignments();
    remove_singlepurpose_assignments();
    make_uniform_fixpoint();
    add_fork_fnodes();
    add_forward_fnodes();
    // and once more, because fork and forward nodes may invalidates previous
    // simplification
    remove_single_assignments();
    remove_singlepurpose_assignments();
    make_uniform_fixpoint();
    add_fork_fnodes();
    add_forward_fnodes();
    update_fnode_io();

    erase_metas("who-reads", system.fnodes);
    for (auto fnode : system.fnodes)
        for (auto signal : fnode->outputs)
        {
            string comment = signal->meta.keyval("comment");
            if (comment.size())
                fnode->meta.addkv("comment", signal->id + "=" + comment);
        }
}

void SystemManipulation::get_signals_referenced_by(Expression *expr, SignalSet &sset)
{
    TraverseInputs inputs_of_expr(expr,
            [](Expression *&e){ return e->op == OP_VAR; });

    for (Expression *evar : inputs_of_expr)
        sset << signal(evar);
}

} // namespace vam
