/**
 * @desc RTL manipulation
 */
#ifndef _RTLMANIP_H
#define _RTLMANIP_H

#include <unordered_map>

#include "vamrtl.h"
#include "rtltraverse.h"

using namespace std;

namespace vam
{

enum DataflowDir
{
    DF_BACKWARD,
    DF_FORWARD,
};

/**
 * @brief Data structure for cone of influence. Contains vector of signals
 * influenced at a given time frame. The index to the vector represents the
 * given time frame.
 */
class ConeOfInfluence
{
public:
    DataflowDir dir;

    vector<SignalSet> cone_signals;
    vector<StorageSet> cone_storages;
    vector<GenerateSet<Expression*>> cone_exprs;

    void clear()
    {
        cone_signals.clear();
        cone_storages.clear();
        cone_exprs.clear();
    }
};

/**************************************************
 * SystemManipulation
 */
class SystemManipulation
{
public:
    System &system;

public:
    int step;
    int debug;

public:
    SystemManipulation(System &s) : system(s) { debug=0; }

    /**
     * @desc Update lists of inputs and outputs of all functional nodes.
     */
    void update_fnode_io();

    /**
     * @desc Return set of variables within the cone of influence.
     * @param open      set of initial signals from which to start searching
     * @param time_in_cycles    the length of coi in number of cycles
     * @param dir       direction of the cone
     * @param coi       target cone of influence.
     */
    void select_coi(const IdSet &open, int time_in_cycles, DataflowDir dir,
            ConeOfInfluence &coi);

    void expr_outputs_dfs_walkthrough(Expression *e,
        ExpressionList &bt, void *data);

    /**
     * @desc All assignments in the system which assigns a value of a signal to
     * another signal will be removed. Given signals will be also removed from
     * the system. Inputs and outputs of involved functional nodes will be
     * invalidated.
     * @return 1 if some replacement have been done
     */
    int remove_single_assignments();

    /**
     * @desc All assignments in the system which assigns some expression to a
     * signal which is used just by one expression will be removed.  Given
     * signal will be also removed from the system. Inputs and outputs of
     * involved functional nodes will be invalidated.
     * @return 1 if some replacement have been done
     */
    int remove_singlepurpose_assignments();

    /**
     * legend: '-' means to invalidate current walkthrough
     * 1a. (== (_ a w) (_ b w)) for a==b ==> (_ 1 1)
     * 1b. (== (_ a w) (_ b w)) for a!=b ==> (_ 0 1)
     * 2. (== (_ a w) e) ==> (== e (_ a w)), the same for '!=','&','|'
     * -3. (== e (_ 1 1)) and (!= e (_ 0 1)) ==> e
     * -4. (!= e (_ 1 1)) and (== e (_ 0 1)) ==> (! e)
     * -5a. (! (_ 0 1)) ==> (_ 1 1), the same for '~'
     * -5b. (! (_ 1 1)) ==> (_ 0 1), the same for '~'
     * -6. (! (! e)) ==> e, the same for '~'
     * -7. (^^ (_ x y) z) ==> (_ x z)
     * -8. (^^ e w) for width of e=w ==> e
     * -9a. (& e (_ 0 w)) for width of e=w ==> (_ 0 w)
     * -9b. (| e (_ 2^w-1 w)) ==> (_ 2^w-1 w)
     * -9c. (& e (_ 2^w-1 w)) ==> e
     * -9d. (| e (_ 0 w)) ==> e
     * -9e. (| e e) ==> e, the same for '&', '=='
     * -9f. (& e1 e2), for bitwidth = 1 ==> (&& e1 e2)
     * -9g. (| e1 e2), for bitwidth = 1 ==> (|| e1 e2)
     * -10. (cond ((e) e1) (e2)) ==> (? e e1 e2)
     * -11. (? e (_ 1 1) (_ 0 1)) ==> e
     * -12. (? e (_ 0 1) (_ 1 0)) ==> (! e)
     * 13. (? (! e1) e2 e3) ==> (? e1 e3 e2)
     * -13a. (? cond c c) ==> c
     * -14. (cond ((== v (_ a w)) e1)
     *          ((== v (_ b w)) e2) ...
     *          ((== v (_ n w)) en) (ed)) ==>
     *    (switch v ((_ a w) e1)
     *              ((_ b w) e2) ...
     *              ((_ n w) en) (ed)) - mark v as weak-selector
     * 15. (switch ee (eA e1) (eB e2) ... (eX en) (ed)), where
     *      width of 'switch' > 1 (i.e. it is not some combinational logic
     *      implemented using multiplexer) ==> mark ee as hard-selector
     * 16. (? v e1 e2) ==> mark v as weak-selector
     * 17. (&& e1 e2 ... (_ 1 1) ... en) ==> (&& e1 e2 ... en)
     * -18. (&& e1 e2 ... (_ 0 1) ... en) ==> (_ 0 1)
     * -19. (|| e1 e2 ... (_ 1 1) ... en) ==> (_ 1 1)
     * 20. (|| e1 e2 ... (_ 0 1) ... en) ==> (|| e1 e2 ... en)
     * 21. (&& e1 e2 ... (&& ex ... ey) ... en) ==> (&& e1 e2 ...ex...ey...en)
     * 22. (|| e1 e2 ... (|| ex ... ey) ... en) ==> (|| e1 e2 ...ex...ey...en)
     * 22a.(|. e1 e2 ... (|. ex ... ey) ... en) ==> (|. e1 e2 ...ex...ey...en)
     * -23. (! (== e1 e2)) ==> (!= e1 e2)
     * -24. (! (!= e1 e2)) ==> (== e1 e2)
     * cancelled: -25. (<< (_ a w1) w2) ==> (_ a w1+w2), the same for '<<.'
     * cancelled: -26. (<< e 0) ==> e, the same for '<<.' and '>>'
     * -27. (!= (& e (_ c w)) (_ 0 w)) aka in C: (e & c != 0),
     *      where c == 2^n for some integer n ==> ([.] e n)
     * -28. (== (& e (_ c w)) (_ c w)) aka in C: (e & c == c),
     *      where c == 2^n for some integer n ==> ([.] e n)
     * -29. (>> (& e (_ c w)) (_ n x)), where c == 2^n ==>
     *          (. (_ 0 w-n-1) ([.] e n)) for w-n-1>0, or
     *          ([.] e n)
     * -30. (& (>> e (_ c x)) (_ 1 w)) ==>
     *          (. (_ 0 w-1) ([.] e c) for w-1>0, or
     *          ([.] e c)
     * -31. (== (>> e (_ c x)) (_ 1 w)) ==> ([.] e c)
     * -32. ([] e (_ c w)) ==> ([.] e c)
     * 33. ([.] ([..] e l u) c) ==> ([.] e c)
     * -34. ([.] ([.] e i) 0) ==> ([.] e i)
     * 35. (^^ (^^ e w1) w2) ==> (^^ e w2), the same for '^^.'
     * -36. (== (^^ e w) (_ 1 w)) ==> ([.] e 0), the same for '^^.'
     *      (!= (^^ e w) (_ 0 w)) ==> ([.] e 0), the same for '^^.'
     * -37. (== (^^ e w) (_ 0 w)) ==> (! ([.] e 0)), the same for '^^.'
     *      (!= (^^ e w) (_ 1 w)) ==> (! ([.] e 0)), the same for '^^.'
     * -38. ([..] (>> e (_ c w)) 0 u) ==> ([..] e c c+u)
     * -39. (| (| e1 e2) e3) ==> (|. e1 e2 e3)
     * 40. (|. e1 e2 ... c ... ex ... c ... ey) ==> (|. e1 e2 ...c...ex...ey)
     *     the same for '&&', '||'
     * 41. (|. e) ==> e, the same for '||' and '&&'
     */
    int make_uniform_fixpoint();

    /**
     * @desc DFS walkthrough all expressions and their operands within the
     * system. Call 'f' manipulator for each operand. If the manipulator
     * returns positive number (meaning that some manipulation has been
     * done over the expression), and 'invalidates' is true, DFS completely
     * restarts. Negative number returned by a manipulator means that some
     * (abs(return_value)) manipulations have been done and DFS should
     * completely restart regardless 'invalidates' value. Return value 0 of a
     * manipulator means that no manipulation has been done, thus DFS should
     * continue with walkthrough.
     * @param   f       Expression manipulator
     * @param   data    Pointer to callback data
     * @param   invalidates true means no further search within manipulated
     * expression (default: false)
     * @return number of performed manipulations.
     */
    int for_each_operand(ExprCallback f, void *data, bool invalidates = false);

    int for_each_operand_dfs(Expression *&e, ExprCallback cb, void *data);

    /**
     * @desc Create histogram of signal usage. Mark each use of each signal
     * in the meta of a given signal. The meta node name will be "used_by" and
     * each child node contains key-value pair, where key is the name of
     * expression operation reading the signal value and the length of value
     * (the length of a string) is the number of usages in such kind of
     * expression. E.g. (used_by (OP_OR "IIIII")) stands for 5 usages by OR
     * operations.
     * @return number of overall usages
     */
    int count_signal_usages();

    /**
     * @brief   Encapsulates an expression to a signal. Creates a new fun. node
     *          assigning the expression to a new signal. Replace all
     *          references to the expression with the reference to a signal.
     *          If expr is reference to a signal, do not bother with
     *          encapsulating, just return the signal.
     * @param   e       expression to encapsulate
     * @param   prefix  prefix of the name for the new signal
     * @return  a new signal
     */
    Signal *encapsulate_expr_to_signal(Expression *e, string prefix);

    /**
     * @brief   Encapsulates an expression to a signal. Does not encapsulate
     *          if expression is already a variable reference, or, if
     *          encapsualet_const is not set, expression is a contant.
     * @param   e       expression to encapsulate
     * @param   prefix  prefix of the name for the new signal
     * @param   encapsulate_const   if true, encapsulate even constant
     * @param   signal  pointer to a signal pointer. If non null and
     *                  encapsulation happens, store the pointer to a new
     *                  signal here.
     * @return  Value of e if no ecapsulation happens. nullptr if encapsulation
     * is performed.
     */
    Expression *encapsulate_expr(Expression *e, string prefix,
            bool encapsulate_const, Signal **signal);

    /**
     * @desc Add fork-fnode after any storage with a set of output signals,
     * i.e. for a connection:
     * @verbatim
       |storage1|-+-s1--->
                  +-s2--->
                  +-s3--->
       @endverbatim
     * add a fnode and connecting signal such that:
     * @verbatim
       |storage1|---storage1_q--->(fork_storage1_q)---s1--->
                                             \    \---s2--->
                                              \-------s3--->
       @endverbatim
     */
    void add_fork_fnodes();

    /**
     * @brief   Helper function for add_fork_fnodes.
     * @param   sset    set of signals connected to the same output.
     * @param   sc      common output of signals.
     */
    void add_fork_fnode(SignalSet &sset, SignalConnection sc);

    /**
     * @desc Add forward-fnodes between any two storages with direct connection,
     * i.e. for a connection
     * @verbatim
       |storage1|---signal1-->|storage2|, add a fnode and connecting signal,
       @endverbatim
     * such that,
     * @verbatim
       |storage1|---storage1_q-->(fnode_signal1)---signal1--->|storage2|
       @endverbatim
     */
    void add_forward_fnodes();

    /**
     * @brief   Helper function for add_forward_fnodes.
     * @param   sig     signal directly connecting two storages
     * @param   sc      connection from storage to sig
     */
    void add_forward_fnode(Signal *sig, SignalConnection &sc);

    /**
     * @brief   Transforms each ITE to a corresponding switch statement.
     */
    void transform_ite_to_switch();

    /**
     * @brief   Encapsulates each conditional statement (cond) to a
     *          corresponding switch statement.
     */
    void transform_cond_to_switch();

    /**
     * @brief   Uniforms selectors, i.e.:
     *  1. transforms ITE to SWITCH,
     *  2. transforms COND to SWITCH,
     *  3. encapsulates switch control expressions to a variable,
     *  4. encapsulates each switch to a new variable.
     */
    void unify_selectors();

    /**
     * @brief   Perform more general simplification consisting of:
     *  1. remove_single_assignments();
     *  2. remove_singlepurpose_assignments();
     *  3. make_uniform_fixpoint();
     *  4. add_forward_fnodes();
     *  5. add_fork_fnodes();
     *  6. update_fnode_io();
     *  7. updating fnode comment, which variable it sets
     */
    void simplify();

    /**
     * @brief   Returns set of signals referenced by the expression. Note the
     *          direction of dataflow is not considered, i.e. signals on LHS or
     *          RHS of an assignment expression is considered in the same way.
     * @param   expr    expression
     * @param   sset    set of signals to be filled by referenced signals
     */
    void get_signals_referenced_by(Expression *expr, SignalSet &sset);

protected:
    ExprAssign *find_single_assignment();
    ExprAssign *find_onebit_constant_assignment();
    ExprAssign *find_singlepurpose_assignment();
    void replace_single_assignment(ExprAssign *as);

    bool invalidates;

    int count_usages(Signal *s);

    void get_forward_expansion(Signal *s, SignalSet &open, SignalSet &closed);
    void get_backward_expansion(Signal *s, SignalSet &open, SignalSet &closed);
};

ExprAssign *find_setting_assignment(Signal *s);

inline Expression *&lhs(Expression *e)
{
    return static_cast<BinaryExpression*>(e)->lhs;
}

inline Expression *&rhs(Expression *e)
{
    return static_cast<BinaryExpression*>(e)->rhs;
}

inline Expression *&operand(Expression *e)
{
    return static_cast<UnaryExpression*>(e)->expr;
}

inline ExpressionPool *pool(Expression *e)
{
    return static_cast<ExpressionPool*>(e);
}

inline Variable *&var(Expression *e)
{
    return static_cast<ExprVariable*>(e)->v;
}

inline Signal *signal(Expression *e)
{
    return static_cast<Signal*>(static_cast<ExprVariable*>(e)->v);
}

inline ExprConst *constant(Expression *e)
{
    return static_cast<ExprConst*>(e);
}

inline ExprITE *ite(Expression *e)
{
    return static_cast<ExprITE*>(e);
}

inline ExprCond *cond(Expression *e)
{
    return static_cast<ExprCond*>(e);
}

inline string &value(Expression *e)
{
    return constant(e)->value;
}

} // namespace vam

#endif
