/**
 * @desc Traverse algorithms and classes for manipulating RTL models.
 */
#include "rtltraverse.h"

namespace vam
{

// static singletons of end() iterators
template<>
TraverseOutputs::iterator TraverseOutputs::end_singleton{true};
template<>
TraverseInputs::iterator TraverseInputs::end_singleton{true};

TraverseOutputIterator &TraverseOutputIterator::operator++()
{
    while (1)
    {
        // no more iterator is in the backtrace
        if (itbt.empty())
        {
            // set me as the end() iterator
            end = true;
            return *this;
        }

        // iterator to current target expression
        Expression::outputs_iterator curr_i = itbt.back();
        // expression, owner of the iterator
        Expression *curr_e = bt.back();
        
        // if the current iterator points to the end of the expression
        if (curr_i == curr_e->outputs.end())
        {
            /* curr_i is represented by ===>
             * curr_e is represented by ||n||
             *
             * |n1| ---> ||n2|| ===> end
             *    \---->  |n3| (maybe end)
             *
             *    to
             *
             * ||n1|| ---> |n2| ---> ...
             *      \====> |n3| (maybe end)
             */
            bt.pop_back();
            itbt.pop_back();
            if (itbt.empty())
                continue;
            ++itbt.back();
            if (itbt.back() == bt.back()->outputs.end())
                continue;
        }
        else
        // if curr_i points to non-leaf node
        if (!(*curr_i)->outputs.empty())
        {
            /* curr_i is represented by ===> pointing to it
             * curr_e is represented by ||n||
             *
             * ||n1|| ===> |n2| ---> |n3|
             *
             *    to
             *
             *  |n1| ---> ||n2|| ===> |n3|
             */
            bt.push_back(*curr_i);
            itbt.push_back((*curr_i)->outputs.begin());
        }
        else
        // curr_i points to 
        if (curr_i != curr_e->outputs.end())
        {
            /* curr_i is represented by ===> pointing to it
             * curr_e is represented by ||n||
             *
             * ||n1|| ===> |n2|
             *      \----> |n3| (maybe end)
             *
             *    to
             *
             * ||n1|| ---> |n2|
             *      \====> |n3| (maybe end)
             */
            ++itbt.back();
            if (itbt.back() == bt.back()->outputs.end())
                continue;
        }

        if (filter == nullptr || filter(*itbt.back()))
            return *this;
    }
}

void TraverseInputIterator::move_to_begin()
{
    // reset the state to the begin()
    if (bt.size() != 1 || itbt.size())
    {
        bt.erase(bt.begin()+1, bt.end());
        itbt.clear();
        end = false;
    }
    // move to the first valid item
    if (filter && !filter(bt.back()))
        ++(*this);
}

TraverseInputIterator &TraverseInputIterator::operator++()
{
    if (end)
        return *this;
    
    // if it is the first item (i.e. points to the root node)
    if (itbt.empty())
    {
        // move on to the first item of the root node
        itbt.push_back(bt.back()->begin());
        
        // if the root has no item, finish as end()
        if (itbt.back() == bt.back()->end())
        {
            end = true;
            return *this;
        }
        
        // if the first item conforms to the filter, return successfully
        if (filter == nullptr || filter(*itbt.back()))
            return *this;

        // else continue with dfs
    }

    // dfs searching for conforming item
    while (1)
    {
        if (itbt.empty())
        {
            // set me as the end() iterator
            end = true;
            return *this;
        }

        // iterator to current target expression
        Expression::iterator curr_i = itbt.back();
        // expression, owner of the iterator
        Expression *curr_e = bt.back();
        
        if (curr_i == curr_e->end())
        {
            /* curr_i is represented by ===>
             * curr_e is represented by ||n||
             *
             * |n1| ---> ||n2|| ===> end
             *    \---->  |n3| (maybe end)
             *
             *    to
             *
             * ||n1|| ---> |n2| ---> ...
             *      \====> |n3| (maybe end)
             */
            itbt.pop_back();
            bt.pop_back();
            if (itbt.empty())
                continue;
            ++itbt.back();
            if (itbt.back() == bt.back()->end())
                continue;
        }
        else
        if ((*curr_i)->begin() != (*curr_i)->end())
        {
            /* curr_i is represented by ===> pointing to it
             * curr_e is represented by ||n||
             *
             * ||n1|| ===> |n2| ---> |n3|
             *
             *    to
             *
             *  |n1| ---> ||n2|| ===> |n3|
             */
            bt.push_back(*curr_i);
            itbt.push_back((*curr_i)->begin());
        }
        else
        if (curr_i != curr_e->end())
        {
            /* curr_i is represented by ===> pointing to it
             * curr_e is represented by ||n||
             *
             * ||n1|| ===> |n2|
             *      \----> |n3| (maybe end)
             *
             *    to
             *
             * ||n1|| ---> |n2|
             *      \====> |n3| (maybe end)
             */
            ++itbt.back();
            if (itbt.back() == bt.back()->end())
                continue;
        }

        if (filter == nullptr || filter(*itbt.back()))
            return *this;
    }
}

} // namespace vam
