/**
 * @desc Traverse algorithms and classes for manipulating RTL models.
 */
#ifndef _RTLTRAVERSE_H
#define _RTLTRAVERSE_H
#include "vamrtl.h"

using namespace std;

namespace vam {

/**
 * @brief   Expression callback for DFS-walkthrough (either expression outputs
 * or its operands.
 * @param   e       visited expression
 * @param   bt      backtrace (from DFS-init expression)
 * @param   data    unspecified local data passed to DFS algorithm.
 * @return  Semantics depends on walkthrough algorithm.
 */
using ExprCallback = std::function<
    int(Expression*& e, ExpressionList& bt, void* data)>;

/**
 * @brief   Filter callback for TraverseOutputs/TraverseInputs.
 * @param   e   expression being filtered
 * @return  true if expression holds a property being traversed.
 */
using TraverseFilter = std::function<bool(Expression*& e)>;

/****************************************************************************
 * Traverse Outputs
 */

/**
 * @desc    TraverseOutputIterator is iterator for TraverseOutput virtual
 * container. The iterator traverses through all outputs of a given expression
 * and subsequents outputs. TraverseOutputIterator implements DFS search when
 * moving to the next iterated value. If the RTL structure is modified,
 * iterator is invalidated. Each traversed expression is asked by a callback
 * (so called filter) if it fits requirements of traversed outputs.
 */
class TraverseOutputIterator
{
public:
    /// backtrace from root expression to current expression
    ExpressionList bt;
    /// iterator backtrace
    vector<Expression::outputs_iterator> itbt;
    /// filter callback
    TraverseFilter filter;
    /// flag of finished iteration
    bool end;

    TraverseOutputIterator(Expression *root, TraverseFilter filter) :
        filter(filter), end(false)
    {
        bt.push_back(root);
    }

    TraverseOutputIterator(bool end_flag) : end(true)
    {
        (void)end_flag;
    }

    bool operator!=(const TraverseOutputIterator &i)
    {
        // if both are end() iterators, they don't differ
        if (end && i.end)
            return false;
        // if both are not end(), they differ only if the
        // structures differ
        else if (!end && !i.end)
            return !((bt == i.bt) && (itbt == i.itbt));
        // if just one is the end() iterator, they differ
        else
            return true;
    }

    void move_to_begin()
    {
        itbt.push_back(bt.back()->outputs.begin());
    }

    TraverseOutputIterator &operator++();

    Expression *&operator*()
    {
        return *itbt.back();
    }

    Expression *&operator->()
    {
        return *itbt.back();
    }
};

/**
 * @desc    TraverseInputIterator is very similar to TraverseOutputIterator
 * but traverses through all expression operands instead of all its outputs.
 */
class TraverseInputIterator
{
public:
    /// backtrace from root expression to current expression
    ExpressionList bt;
    /// iterator backtrace
    vector<Expression::iterator> itbt;
    /// filtering callback
    TraverseFilter filter;
    /// flag of finished iteration
    bool end;

    TraverseInputIterator(Expression *root, TraverseFilter filter) :
        filter(filter), end(false)
    {
        bt.push_back(root);
    }

    // ctor just for end() iterator
    TraverseInputIterator(bool end_flag) : end(true)
    {
        (void)end_flag;
    }

    /**
     * @brief   Moves the iterator to the begin() of the container.
     * @post    *this == container.begin()
     */
    void move_to_begin();

    bool operator!=(const TraverseInputIterator &i)
    {
        // if both are end() iterators, they don't differ
        if (end && i.end)
            return false;
        // if both are not end(), they differ only if the
        // structures differ
        else if (!end && !i.end)
            return !((bt == i.bt) && (itbt == i.itbt));
        // if just one is the end() iterator, they differ
        else
            return true;
    }

    TraverseInputIterator &operator++();

    Expression *&operator*()
    {
        if (itbt.empty())
            return bt.back();
        else
            return *itbt.back();
    }

    Expression *&operator->()
    {
        if (itbt.empty())
            return bt.back();
        else
            return *itbt.back();
    }
};

/**
 * @brief   Template of virtual container of inputs or outputs of the
 * expression, cf. TraverseOutputs and TraverseInputs specializations.
 */
template<typename IteratorType>
class TraverseIO
{
public:
    using iterator = IteratorType;

private:
    /// starting expression
    Expression *root;

    /// filtering callback
    TraverseFilter filter;

    /// end() iterator
    static iterator end_singleton;

public:
    /**
     * @brief   Create virtual container.
     * @param   root    given expression for which start traversing.
     * @param   filter  filtering callback
     *          (default: nullptr, means no filter).
     * @pre     root != nullptr
     */
    TraverseIO(Expression *root, TraverseFilter filter = nullptr)
        : root(root), filter(filter)
    {
        assert(root != nullptr);
    }

    iterator begin()
    {
        iterator begin_iterator{root, filter};
        begin_iterator.move_to_begin();
        return begin_iterator;
    }

    const iterator &end()
    {
        return end_singleton;
    }
};

/**
 * @brief Virtual container of outputs of a given expression. Outputs may be
 * constrained by a filtering callback. Traversing is performed by provided
 * iterators. Each instance of TraverseOutput container provides just a single
 * iterator (returned by begin()). Another call to begin() return invalid
 * iterators.
 *
 * @example
 *      TraverseOutputs all_logic(this_expression,
 *              [](Expression *&e) {
 *                  return e->op_is_logic();
 *              });
 *      for (auto logic_expr : all_logic)
 *      {
 *          do_something_with(logic_expr);
 *      }
 */
using TraverseOutputs = TraverseIO<TraverseOutputIterator>;

/*
 * @brief Virtual container very similar to TraverseOutputs but traverses
 * through all expression operands.
 *
 */
using TraverseInputs = TraverseIO<TraverseInputIterator>;

}

#endif
