/**
 * @desc Time frame manipulations
 */
#ifndef _TFRAMES_H
#define _TFRAMES_H

#include <limits.h>

/**
 * @brief   tframe represents integer specifing a time frame. The only
 * supported operations are addition and substraction. Moreover a special value
 * representing infinity and undefined value is provided. Whenever too big
 * number is assigned (or additive operation is applied too many times), time
 * frame value becomes infinite (tframe::_TF_INF). Whenever minus and plus
 * infinite number is added, the resulting number is undefined
 * (tframe::_TF_NAN) and no arithmetic operation does not change it.
 */
class tframe
{
public:
    int value;
    /**
     * @brief
     *      tframe i = 0;
     *      while (i > -TF_INF)
     *          i -= 100;
     */
    static const int _TF_INF = (SHRT_MAX << 8) | 0xff;
    static const int BOUND_INF = SHRT_MAX;
    static const int _TF_NAN = INT_MIN;

    tframe(int i = 0, bool nan = false) : value(i)
    {
        if (nan)
            value = _TF_NAN;
        else
            normalize();
    }

    tframe(const tframe &t) : value(t.value) {}

    void normalize()
    {
        if (value >= BOUND_INF)
            value = _TF_INF;
        else if (value != _TF_NAN && value <= -BOUND_INF)
            value = -_TF_INF;
    }

    void setnan()
    {
        value = _TF_NAN;
    }

    bool isnan() const
    {
        return value == _TF_NAN;
    }

    tframe &operator++()
    {
        ++value;
        normalize();
        return *this;
    }

    tframe operator++(int)
    {
        tframe result(*this);
        ++value;
        normalize();
        return result;
    }

    tframe &operator--()
    {
        --value;
        normalize();
        return *this;
    }

    tframe operator--(int)
    {
        tframe result(*this);
        --value;
        normalize();
        return result;
    }

    tframe &operator+=(short int i)
    {
        value += i;
        normalize();
        return *this;
    }

    tframe &operator+=(const tframe &t)
    {
        if (value == _TF_NAN || t.value == _TF_NAN)
            value = _TF_NAN;
        else if (value == _TF_INF && t.value == -_TF_INF)
            value = _TF_NAN;
        else if (value == -_TF_INF && t.value == _TF_INF)
            value = _TF_NAN;
        else if (value == _TF_INF && t.value == _TF_INF)
            return *this;
        else if (value == -_TF_INF && t.value == -_TF_INF)
            return *this;
        else
            value += t.value;
        normalize();
        return *this;
    }

    tframe &operator-=(short int i)
    {
        value -= i;
        normalize();
        return *this;
    }

    tframe &operator-=(const tframe &t)
    {
        if (value == _TF_NAN || t.value == _TF_NAN)
            value = _TF_NAN;
        else if (value == _TF_INF && t.value == _TF_INF)
            value = _TF_NAN;
        else if (value == -_TF_INF && t.value == -_TF_INF)
            value = _TF_NAN;
        else if (value == _TF_INF && t.value == -_TF_INF)
            return *this;
        else if (value == -_TF_INF && t.value == _TF_INF)
            return *this;
        else
            value -= t.value;
        normalize();
        return *this;
    }

    tframe operator+(short int i) const
    {
        tframe result(*this);
        if (!isnan())
            result += i;
        return result;
    }

    tframe operator+(const tframe &t) const
    {
        if (isnan())
            return tframe(*this);
        else if (t.isnan())
            return tframe(t);
        else
        {
            tframe result(*this);
            result += t;
            return result;
        }
    }

    tframe operator-() const
    {
        if (!isnan())
            return tframe(-value);
        else
            return tframe(value);
    }

    tframe operator-(short int i) const
    {
        tframe result(*this);
        if (!isnan())
            result -= i;
        return result;
    }

    tframe operator-(const tframe &t) const
    {
        if (isnan())
            return tframe(*this);
        else if (t.isnan())
            return tframe(t);
        else
        {
            tframe result(*this);
            result -= t;
            return result;
        }
    }

    tframe &operator=(short int i)
    {
        value = i;
        return *this;
    }

    bool operator==(short int i) const
    { return (value == i); }

    bool operator==(const tframe &t) const
    { return value == t.value; }

    bool operator!=(short int i) const
    { return !(*this == i); }

    bool operator!=(const tframe &t) const
    { return !(*this == t); }

    bool operator>(short int i) const
    { return value > i; }

    bool operator>=(short int i) const
    { return value >= i; }

    bool operator<(short int i) const
    { return value < i; }

    bool operator<=(short int i) const
    { return value <= i; }

    bool operator>(const tframe &t) const
    { return value > t.value; }
    
    bool operator>=(const tframe &t) const
    { return value >= t.value; }

    bool operator<(const tframe &t) const
    { return value < t.value; }

    bool operator<=(const tframe &t) const
    { return value <= t.value; }

    operator short int() const
    {
        if (value == _TF_INF)
            return SHRT_MAX;
        else if (value == -_TF_INF)
            return SHRT_MIN;
        else
            return value;
    }

};

extern const tframe TF_INF;
extern const tframe TF_NAN;

/**
 * @brief Interval of time frames. Defines time frame with the interval.
 * Typical use is to define write delays on storages, e.g. registers -- write
 * delay is represented by interval [1;1] i.e. one cycle. Interval of receiving
 * data from bus can be represented by [2;TF_INF] meaning that the response
 * will arive between 2 and infinitely many cycles.
 */
class tfint
{
public:
    tframe min;
    tframe max;

    tfint() : min(0), max(0) {}

    tfint(tframe value) : min(value), max(value) {}

    tfint(tframe min, tframe max) : min(min), max(max)
    {
        normalize();
    }

    tfint(int min, int max) : min(min), max(max)
    {
        normalize();
    }

    tfint &normalize()
    {
        if (max < min)
        {
            tframe tmp = min;
            min = max;
            max = tmp;
        }
        return *this;
    }

    tfint &set(int _min, int _max)
    {
        min = _min;
        max = _max;
        return normalize();
    }

    tfint &set_empty()
    {
        min.setnan();
        max.setnan();
        return *this;
    }

    bool empty() const
    {
        return min.isnan() || max.isnan();
    }

    bool contains(int i) const
    {
        return min <= (short)i && max >= (short)i;
    }

    bool contains(const tfint &t) const
    {
        return min <= t.min && t.min <= max &&
               min <= t.max && t.max <= max;
    }

    tfint &operator=(int i)
    {
        min = i;
        max = i;
        return *this;
    }

    bool operator==(int i) const
    {
        return contains(i);
    }

    bool operator==(const tfint &t) const
    {
        return (min == t.min) && (max == t.max);
    }

    tfint operator+(const tfint &t) const
    {
        tfint result(*this);
        result.min += t.min;
        result.max += t.max;
        return result;
    }

    tfint operator-() const
    {
        if (empty())
        {
            tfint result;
            result.min.setnan();
            return result;
        }
        return tfint(-max, -min);
    }

};

/// both lhs and rhs should overlap or empty interval is returned
inline tfint tf_union(const tfint &lhs, const tfint &rhs)
{
    tfint result(lhs);
    if (rhs.min < result.min)
        result.min = rhs.min;
    if (rhs.max > result.max)
        result.max = rhs.max;
    if ((lhs.max <= rhs.min && lhs.min <= rhs.min) ||
        (rhs.max <= lhs.min && rhs.min <= lhs.min))
        result.set_empty();
    return result;
}

inline tfint tf_intersect(const tfint &lhs, const tfint &rhs)
{
    if (lhs.min <= rhs.min && rhs.min <= lhs.max)
        return tfint(rhs.min, lhs.max);
    else if (rhs.min <= lhs.min && lhs.min <= rhs.max)
        return tfint(lhs.min, rhs.max);
    else
        return tfint{}.set_empty();
}

extern const tfint TF_INSTANT;
extern const tfint TF_ALL;

#endif
