/**
 * @brief VAM language demo/sanity test.
 */
#include "vam.h"
#include "vamrtl.h"
#include "rtlmanip.h"
#include "language.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <typeinfo>
#include <string>
#include <boost/algorithm/string.hpp> // replace_all

using namespace std;
using namespace vam;

ostringstream debug;
int debugflag = 0;

/**
 * Test case procedure. Test function `tc' should print its debug info to 
 * `debug' string stream.
 */
void t(void (*tc)(), string name, bool verbose = false)
{
    verbose = verbose || debugflag != 0;
    try {
        debug.str("");
        debug.clear();
        tc();
        cerr << "TEST " << left << setw(20) << name << " passed" << endl;
    }
    catch(VAMException e) {
        cerr << "TEST " << left << setw(20) << name << " FAILED:" << endl << 
            e.msg << endl;
    }
    if (verbose && debug.str().size())
    {
        string d = debug.str();
        if (d[d.size()-1] == '\n')
            d.erase(d.size()-1,1);
        boost::replace_all(d, "\n", "\n    D: ");
        cout << "    D: " << d << endl;
        debug.str("");
        debug.clear();
    }
}

void meta()
{
    Meta m(MT_NODE, "");
    *m.add(MT_NODE, "loc") << "xyz.codal" << 123;
    *m.add(MT_NODE, "loc") << "xyz.codal" << 528;
    auto l = m.find("loc");
    debug << "first loc = " << (*l)[1] << endl;
    l = m.find("loc", l);
    debug << "second loc = " << (*l)[1] << endl;
    RTLVisitor ov;
    m.Accept(&ov);
    debug << ov.out.str() << endl;
}

void assign()
{
    Variable v1("v1", 4);
    Variable v2("v2", 4);
    Variable v3("v3", 4);
    ExprAssign e(new ExprVariable(&v1), new ExprVariable(&v2));
    *e.meta.add(MT_NODE, "loc") << "some_src_file.cc" << 123;
    RTLVisitor ov;
    e.Accept(&ov);
    debug << ov.out.str() << endl;
}

void expr1()
{
    Variable freg_d("fetchD", 32);
    Variable freg_q("fetchQ", 32);
    Variable freg_we("fetchWE", 1);
    auto a1 = new ExprAssign(
            new ExprVariable(&freg_we),
            new ExprNot(
            new ExprEqual(
                new ExprSlice(
                    new ExprVariable(&freg_q),
                    0,3), 
                new ExprConst("15", 4))));
    *a1->meta.add(MT_NODE, "reason") 
        << "load a new instruction iff no \"halt\"";
    RTLVisitor ov;
    a1->Accept(&ov);
    debug << ov.out.str() << endl;
    delete a1;
}

void expr2()
{
    Variable freg_d("fetchD", 32);
    Variable freg_q("fetchQ", 32);
    Variable freg_we("fetchWE", 1);
    auto c = new ExprCase();
    c->add_ctrl_expr(new ExprSlice(
            new ExprVariable(&freg_q),
            4,5));
    c->add_case(new ExprConst("0",2), new ExprConst("0",32));
    c->add_case(new ExprConst("1",2), new ExprConst("-1",32));
    c->add_case(new ExprConst("2",2), new ExprConst("1",32));
    c->add_case(new ExprConst("3",2), new ExprVariable(&freg_q));
    c->add_default(new ExprConst("0",32));
    auto a = new ExprAssign(new ExprVariable(&freg_d), c); 
    RTLVisitor ov;
    a->Accept(&ov);
    debug << ov.out.str() << endl;
    delete a;
}

void rtl1()
{
    Signal *d = new Signal("fetchD", 32);
    Signal *q = new Signal("fetchQ", 32);
    Signal *we = new Signal("fetchWE", 1);

    Register r("fetch", 32);
    r.add_d(d);
    r.add_q(q);
    r.add_we(we);

    auto a1 = new ExprAssign(
            new ExprVariable(we),
            new ExprNot(
            new ExprEqual(
                new ExprSlice(
                    new ExprVariable(q),
                    0,3), 
                new ExprConst("15", 4))));
    *a1->meta.add(MT_NODE, "reason") 
        << "load a new instruction iff no \"halt\"";
    
    RTLVisitor ov;

    d->Accept(&ov); ov.out << endl;

    r.remove_d();
    d->Accept(&ov); ov.out << endl;
    
    q->Accept(&ov); ov.out << endl;
    we->Accept(&ov); ov.out << endl;
    a1->Accept(&ov); 
    debug << ov.out.str() << endl;
    delete a1;
    delete d;
    delete q;
    delete we;
}

void rtl2()
{
    Signal *d = new Signal("r0D", 32);
    Signal *q = new Signal("r0Q", 32);
    Signal *we = new Signal("r0WE", 1);
    Register r("r0", 32);
    r.add_d(d);
    r.add_q(q);
    r.add_we(we);

    RTLVisitor ov;
    // show register
    r.Accept(&ov); ov.out << endl;
    // show D signal
    d->Accept(&ov); ov.out << endl;
    // remove D from register from D and add it twice to Q
    r.remove_d();
    r.add_q(d);
    r.add_q(d);
    // show D again (should have just one connection)
    d->Accept(&ov); ov.out << endl;
    // show Q signal
    q->Accept(&ov); ov.out << endl;
    // show WE signal
    we->Accept(&ov); ov.out << endl;
    // show register again
    r.Accept(&ov);

    debug << ov.out.str() << endl;
    delete d;
    delete q;
    delete we;
}

void system1()
{
    System sys;
    Signal *d = sys.insert(new Signal("r0_D", 32));
    Signal *q = sys.insert(new Signal("r0_Q", 32));
    Signal *we = sys.insert(new Signal("r0_WE", 1));
    Register *r = sys.insert(new Register("r0", 32));
    r->connect(d, q, we);

    RTLVisitor v;
    sys.Accept(&v);
    debug << v.out.str() << endl;
}

void lexer()
{
    Lexer l("(sig r0 1 (vt control (meta \"as\ndf\")))");
    Token t = l.next_token();
    while (t.type != T_EOF)
    {
        debug << t.row << ":" << t.col << ": token " << t.type << ", value=\"" 
            << t.value << "\"\n";
        t = l.next_token();
    }
}

void parse1()
{
    RTLParser p;
    System s;
    p.parse_str("(model (meta 42 (loc 5)) modelname"
            "(pipeline-stages fe id ex wb (meta x)))", &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse2()
{
    RTLParser p;
    System s;
    p.parse_str("(model modelname"
            "(pipeline-stages fe id ex wb)"
            "(sig v1 1 (vt control))"
            "(sig v2 32 (vt data))"
            "(sig v3 32 (vt data) (stage id)))", &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse3()
{
    RTLParser p;
    System s;
    p.parse_str("(model modelname"
            "(pipeline-stages fe id ex wb)"
            "(sig v1 1)"
            "(sig v2 32)"
            "(sig v3 32 (stage id))"
            "(reg r0 32 (we v1) (d v2) (q v3)))", &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse_mem()
{
    RTLParser p;
    System s;
    p.parse_str("(model modelname\n"
            "(pipeline-stages fe id ex wb)\n"
            "(sig v1 1)\n"
            "(sig v2 3)\n"
            "(sig v3 32 (stage id))\n"
            "(mem regs 32 3 8 (ports \n"
            "(port p0 (dir ro) (timing 0) (en v1) (addr v2) (data v3)\n"
            "  (control v3) (stage id))\n"
            ")))\n", &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse_fnode_unary()
{
    RTLParser p;
    System s;
    p.parse_str(
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (_ -1 3))\n"
  /*  8 */  "(:= v3 (! v1))\n"
  /*  9 */  "(:= v1 (-- (~ (_ 0 1))))\n"
  /* 10 */  ")))\n"
            , &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse_fnode_binary()
{
    RTLParser p;
    System s;
    p.parse_str(
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (| (_ 1 1) (_ 0 1)))\n"
  /*  8 */  "(:= v3 (& v2 v3))\n"
  /*  9 */  "(:= v1 ([] (>>. (- (_ 0 7) (^^ (<< v2 (_ 6 3)) 7)) (_ 6 3)) 0))\n"
  /* 10 */  ")))\n"
            , &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse_fnode_special()
{
    RTLParser p;
    System s;
    p.parse_str(
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(sig v4 1)\n"
  /*  7 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  8 */  "(:= v2 (? v3 (_ 1 1) (_ 0 1)))\n"
  /*  9 */  "(:= v3 (switch v2 ((_ 0 1) v1) ((_ 1 1) v2) (v3)))\n"
  /* 10 */  "(:= v4 (cond ((_ 0 1) v1) ((_ 1 1) v2) (v3)))\n"
  /* 11 */  "(:= v1 ([] ([..] (^^. v2 8) 0 3) 1))\n"
  /* 12 */  ")))\n"
            , &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void parse_fnode_abstract()
{
    RTLParser p;
    System s;
    p.parse_str(
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (!! 1 (_ 1 1) (_ 0 1)))\n"
  /*  8 */  "(:= v3 (~~ 1 v3))\n"
  /*  9 */  "(:= v1 (<> 1 (][ 1 (-+ 1 (~~ 1 v2)))))\n"
  /* 10 */  ")))\n"
            , &s);
    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

/*
void parse_system_dup()
{
    RTLParser p;
    System s;
    p.parse_str(
            "(model modelname\n"
            "(pipeline-stages fe id ex wb)\n"
            "(sig v1 1)\n"
            "(sig v2 1)\n"
            "(sig v3 1 (stage id))\n"
            "(sig v4 1)\n"
            "(sig v5 3)\n"
            "(sig v6 32 (stage id))\n"
            "(reg r0 1 (we v1) (d v2) (q v3))\n"
            "(mem regs 32 3 8 \n"
            "(ports \n"
            "(port p0 (dir ro) (timing 0) (en v4) (addr v5) (data v6)\n"
            "  (control v6) (stage id))\n"
            "))\n"
            "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
            "(:= v2 (!! (_ 1 1) (_ 0 1)))\n"
            "(:= v3 (~~ v3))\n"
            "(:= v1 (<> (][ (-+ (~~ v2)))))\n"
            "))\n"
            ")\n"
            , &s);
    System *dup = s.duplicate();
    RTLVisitor v;
    dup->Accept(&v);
    debug << v.out.str() << endl;

    delete dup;
}
*/

inline string to_str(int val, int length)
{
    ostringstream s;
    s << setfill('0') << setw(length) << val;
    return s.str();
}

void parse_real_example()
{
    RTLParser p;
    System s;
    p.parse_from_file("test-real-example.vam", &s);
    RTLVisitor v;
    s.Accept(&v);
    ofstream o("test-parsed-example.vam");
    o << v.out.str() << endl;
    o.close();
}

void manip1()
{
    RTLParser p;
    System s;
    p.parse_from_file("test-real-example.vam", &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_single_assignments();
    m.update_fnode_io();
    {
        RTLVisitor v;
        s.Accept(&v);
        ofstream o("test-no-single-assignments.vam");
        o << v.out.str() << endl;
        o.close();
    }
    m.remove_singlepurpose_assignments();
    m.update_fnode_io();
    {
        RTLVisitor v;
        s.Accept(&v);
        ofstream o("test-single.vam");
        o << v.out.str() << endl;
        o.close();
    }
}

void manip2()
{
    RTLParser p;
    System s;
    p.parse_str(
            "(model modelname\n"
            "(pipeline-stages xx)\n"
            "(sig v1 1)\n"
            "(sig v2 1)\n"
            "(sig v3 1)\n"
            "(sig v4 1)\n"
            "(sig v5 1)\n"
            "(sig v6 1)\n"
            "(reg r0 1 (we v1) (d v2) (q v3))\n"
            "(fnode f0 (input v3) (output v1) (stage xx) (assign \n"
            "(:= v2 v4)\n"
            "(:= v1 (|| v2 v3))"
            "))\n"
            ")\n"
            , &s);
    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_single_assignments();

    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void manip3()
{
    RTLParser p;
    System s;
    p.parse_str(
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1 (vt control) (connections (r0 we) ) )\n"
            "  (sig v3 1 (vt data ) (connections (r0 q) ) )\n"
            "  (sig v4 1 (vt unknown) )\n"
            "  (sig v5 1 (vt unknown) (connections (r0 d) ) )\n"
            "  (reg r0 1 (we v1) (d v5) (q v3 ) )\n"
            "  (fnode f0 (input v3 v4 v5 ) (output v1 v4 ) (stage xx) (assign \n"
            "(:= v1 (|| v5 v4 ) ) \n"
            "(:= v4 (&& v3 v5 ) ) \n"
            ")) \n"
            ")\n"
            , &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_singlepurpose_assignments();
    m.update_fnode_io();

    RTLVisitor v;
    s.Accept(&v);
    debug << v.out.str() << endl;
}

void manip4()
{
    RTLParser p;
    System s;
    p.parse_from_file("test-single.vam", &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    m.step = 0;
    m.remove_single_assignments();
    m.remove_singlepurpose_assignments();
    m.make_uniform_fixpoint();
    m.update_fnode_io();
    RTLVisitor v;
    s.Accept(&v);

    ofstream o("test-uniformed.vam");
    o << v.out.str() << endl;
    o.close();
}

void histogram()
{
    RTLParser p;
    System s;
    p.parse_from_file("test-uniformed.vam", &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    int usages = m.count_signal_usages();

    RTLVisitor v;
    s.Accept(&v);
    ofstream o("test-usages.vam");
    o << v.out.str() << endl;
    o.close();

    debug << "variable usages: " << usages << endl;
    debug << "no. of signals:  " << s.signals.size() << endl;
    debug << "usage ratio:     " << 1.0*usages/s.signals.size() << endl;
}

struct TestCases {
    const char *testname;
    void (*tstfun)();
} testcases[] = {
    {"meta", meta},
    {"assign", assign},
    {"expr1", expr1},
    {"rtl1", rtl1},
    {"rtl2", rtl2},
    {"system1", system1},
    {"lexer", lexer},
    {"parse_meta", parse1},
    {"parse_signals", parse2},
    {"parse_registers", parse3},
    {"parse_memory", parse_mem},
    {"parse_fnode_unary", parse_fnode_unary},
    {"parse_fnode_binary", parse_fnode_binary},
    {"parse_fnode_special", parse_fnode_special},
    {"parse_fnode_abstract", parse_fnode_abstract},
//    {"parse_system_dup", parse_system_dup},
    {"parse_real_example", parse_real_example},
    {"manip1", manip1},
    {"manip2", manip2},
    {"manip3", manip3},
    {"manip4", manip4},
    {"histogram", histogram},
    {NULL, NULL}, // stop
};

int main(int argc, char *argv[])
{
    if (argc > 1 && string(argv[1]) == "--help")
    {
        cerr << "syntax: " << argv[0] 
                << "[--debug] test_name [test_name ...]" << endl
             << "    or: " << argv[0] << " --list" << endl;
        return 0;
    }

    if (argc > 1 && string(argv[1]) == "--list")
    {
        // list all test cases
        cout << "known test cases:" << endl;
        for (int i = 0; testcases[i].testname != NULL; i++)
            cout << "  " << testcases[i].testname << endl;
        return 0;
    }

    int current_arg = 0;
    if (argc > 1 + current_arg && string(argv[1+current_arg]) == "--debug")
    {
        debugflag++;
        current_arg++;
    }

    if (argc > 1 + current_arg && string(argv[1+current_arg]) == "--all")
    {
        for (int i = 0; testcases[i].testname != NULL; i++)
            t(testcases[i].tstfun, testcases[i].testname);
    }
    else
    while (argc > 1 + current_arg)
    {
        string required_test_name;
        required_test_name = string(argv[1+current_arg]);
        // find the test and run it
        int i;
        for (i = 0; testcases[i].testname != NULL; i++)
            if (required_test_name == testcases[i].testname)
            {
                // perform the test
                t(testcases[i].tstfun, testcases[i].testname);
                break;
            }
        if (testcases[i].testname == NULL)
            cout << "test " << required_test_name << " not found" << endl;
        current_arg++;
    }

    return 0;
}
