/**
 * @desc RTL frame over VAM
 */
#include "vamrtl.h"
#include "language.h"
#include <iostream>
#include <iomanip>

namespace vam
{

/****************************************************************************
 * RTLNode
 ***************************************************************************/
void RTLNode::get_influenced_inputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    (void)s;
    (void)t;
    (void)sset;
}

void RTLNode::get_influenced_outputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    (void)s;
    (void)t;
    (void)sset;
}

/// default implementation is ineffective
bool RTLNode::contains_input(Signal *s, const tfint &t)
{
    SignalSet sset;
    get_influenced_inputs(nullptr, t, sset);
    return sset.contains(s);
}

/// default implementation is ineffective
bool RTLNode::contains_output(Signal *s, const tfint &t)
{
    SignalSet sset;
    get_influenced_outputs(nullptr, t, sset);
    return sset.contains(s);
}

/****************************************************************************
 * SignalConnection
 ***************************************************************************/
bool SignalConnection::is_read()
{
    if (type == CT_UNDEFINED || type == CT_UNKNOWN)
        return false;
    if (type < CT_DATA)
        return true;
    if (type == CT_DATA)
    {
        MemPort *port = dynamic_cast<MemPort*>(storage);
        if (port && port->is_read())
            return true;
    }
    return false;
}

bool SignalConnection::is_write()
{
    if (type == CT_UNDEFINED || type == CT_UNKNOWN)
        return false;
    if (type > CT_DATA)
        return true;
    if (type == CT_DATA)
    {
        MemPort *port = dynamic_cast<MemPort*>(storage);
        if (port && port->is_write())
            return true;
    }
    return false;
}

/****************************************************************************
 * Signal
 ***************************************************************************/
void Signal::Accept(VAMVisitor *v)
{
    RTLVisitor *rv = dynamic_cast<RTLVisitor*>(v);
    if (rv)
        rv->VisitSignal(this);
    else
        v->VisitVariable(this);
}

void Signal::add_connection(int type, Storage *owner)
{
    connections.push_back(SignalConnection(type, owner));
}

set<Storage*> Signal::find_connections(int type)
{
    set<Storage*> s;
    for (auto c : connections)
        if (c.type == type)
            s.insert(c.storage);
    return s;
}

set<int> Signal::find_connections(Storage *storage)
{
    set<int> s;
    for (auto c : connections)
        if (c.storage == storage)
            s.insert(c.type);
    return s;
}

void Signal::remove_connection(int type, Storage *storage)
{
    for (auto i = connections.begin(); i != connections.end(); i++)
        if (i->type == type && i->storage == storage)
        {
            connections.erase(i);
            break;
        }
}

Storage::Storage() {}
Storage::Storage(string _id, int bw) : id(_id), bitwidth(bw) {}
Storage::~Storage() {}

/****************************************************************************
 * Register
 ***************************************************************************/
Register::~Register() {}

void Register::Accept(RTLVisitor *v)
{
    v->VisitRegister(this);
}

void Register::get_influenced_inputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    // t.contains(-1) === alias: !intersection(-write_timing, t).empty()
    if (t.contains(-1) && ( s == nullptr || q.contains(s)))
    {
        if (clr)
            sset.insert(clr);
        if (stall)
            sset.insert(stall);
        if (we)
            sset.insert(we);
        if (d)
            sset.insert(d);
    }
    if (t.contains(0) && (s == nullptr || q.contains(s)))
        if (rst)
            sset.insert(rst);
}

void Register::get_influenced_outputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    if (t.contains(1) && (
            s == nullptr ||
            s == clr ||
            s == stall ||
            s == we ||
            s == d))
        sset.merge(q);
    else if (t.contains(0) && (s == nullptr || s == rst))
        sset.merge(q);
}

bool Register::contains_input(Signal *s, const tfint &t)
{
    return (t.contains(0) &&
                s == rst) ||
           (t.contains(-1) && (
                (s == d) ||
                (s == we) ||
                (s == stall) ||
                (s == clr)));
}

bool Register::contains_output(Signal *s, const tfint &t)
{
    return t.contains(1) && q.contains(s);
}

void Register::add_d(Signal *s)
{
    ASSERT(d == nullptr, "D for " << id << " already set.");
    d = s;
    s->set_data();
    s->add_connection(CT_D, this);
}

void Register::add_we(Signal *s)
{
    ASSERT(we == nullptr, "WE for " << id << " already set.");
    we = s;
    s->set_control();
    s->add_connection(CT_WE, this);
}

void Register::add_q(Signal *s)
{
    if (!q.contains(s))
    {
        q.insert(s);
        s->set_data();
        s->add_connection(CT_Q, this);
    }
}

void Register::add_stall(Signal *s)
{
    ASSERT(stall == nullptr, "STALL for " << id << " already set.");
    stall = s;
    s->set_control();
    s->add_connection(CT_STALL, this);
}

void Register::add_clr(Signal *s)
{
    ASSERT(clr == nullptr, "CLR for " << id << " already set.");
    clr = s;
    s->set_control();
    s->add_connection(CT_CLR, this);
}

void Register::add_rst(Signal *s)
{
    ASSERT(rst == nullptr, "RESET for " << id << " already set.");
    rst = s;
    s->set_control();
    s->add_connection(CT_RST, this);
}

Signal *Register::remove_signal(Signal *s)
{
    if (s == d)
        return remove_d();
    else if (s == we)
        return remove_we();
    else if (s == stall)
        return remove_stall();
    else if (s == clr)
        return remove_clr();
    else if (s == rst)
        return remove_rst();
    else
        return remove_q(s);
}

Signal *Register::remove_d()
{
    Signal *s = d;
    d = nullptr;
    if (s)
        s->remove_connection(CT_D, this);
    return s;
}

Signal *Register::remove_we()
{
    Signal *s = we;
    d = nullptr;
    if (s)
        s->remove_connection(CT_WE, this);
    return s;
}

Signal *Register::remove_q(Signal *s)
{
    if (s == nullptr)
        if (q.size() == 1)
        {
            Signal *s = *q.begin();
            q.clear();
            s->remove_connection(CT_Q, this);
            return s;
        }
        else
            return nullptr;
    else
    {
        if (q.contains(s))
        {
            q.erase(s);
            s->remove_connection(CT_Q, this);
            return s;
        }
        else
            return nullptr;
    }
}

Signal *Register::remove_stall()
{
    Signal *s = stall;
    d = nullptr;
    if (s)
        s->remove_connection(CT_STALL, this);
    return s;
}

Signal *Register::remove_clr()
{
    Signal *s = clr;
    d = nullptr;
    if (s)
        s->remove_connection(CT_CLR, this);
    return s;
}

Signal *Register::remove_rst()
{
    Signal *s = rst;
    d = nullptr;
    if (s)
        s->remove_connection(CT_RST, this);
    return s;
}

void Register::reconnect(int type, Signal *sold, Signal *snew)
{
    switch (type)
    {
        case CT_D:     d = snew; break;
        case CT_WE:    we = snew; break;
        case CT_Q:  replace_sig_in_set(sold, snew, q); break;
        case CT_STALL: stall = snew; break;
        case CT_CLR:   clr = snew; break;
        case CT_RST:   rst = snew; break;
    }
}

/// Check connections of signals and their bitwidth
void Register::check_bitwidth()
{
    ASSERT(d, "register " << id << " incomplete (missing D)");
    ASSERT(d->bitwidth == bitwidth, "register " << id <<
            ": invalid D bitwidth " << d->bitwidth);
    for (auto s : q)
        ASSERT(s->bitwidth == bitwidth, "register " << id << ": " <<
            "invalid Q bitwidth " << s->bitwidth << " of signal " << s->id);
    if (we)
        ASSERT(we->bitwidth == 1, "register " << id <<
            ": invalid WE bitwidth " << we->bitwidth);
    if (stall)
        ASSERT(stall->bitwidth == 1, "register " << id <<
            ": invalid STALL bitwidth " << stall->bitwidth);
    if (clr)
        ASSERT(clr->bitwidth == 1, "register " << id <<
            ": invalid CLEAR bitwidth " << clr->bitwidth);
    if (rst)
        ASSERT(rst->bitwidth == 1, "register " << id <<
            ": invalid RESET bitwidth " << rst->bitwidth);
}

/****************************************************************************
 * Register file port (MemPort)
 ***************************************************************************/
MemPort::~MemPort() {}

void MemPort::Accept(RTLVisitor *v)
{
    v->VisitMemPort(this);
}

namespace { // private namespace

// auxiliary function for filling set of signals of signal of the given memory
// port.
void fillup_signals_from_memport(MemPort *p, SignalSet &s,
        bool with_data_signals)
{
    if (p->enable)
        s.insert(p->enable);
    if (p->address)
        s.insert(p->address);
    s.merge(p->control);
    if (with_data_signals)
        s.merge(p->data);
}

}

void MemPort::get_influenced_inputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    if (is_read())
    {
        if (tf_intersect(-read_timing, t).empty())
            return;

        if (s != nullptr && !data.contains(s))
            return;

        fillup_signals_from_memport(this, sset, false);

        if (!owner)
            return;

        for (MemPort *port : owner->ports)
            if (port->is_write())
            {
                tfint propag_delay = -(port->write_timing + read_timing);
                if (tf_intersect(propag_delay, t).empty())
                    continue;
                fillup_signals_from_memport(port, sset, true);
            }
    }
    else if (is_write())
    {
        if (s == nullptr)
            if (!tf_intersect(-write_timing, t).empty())
                fillup_signals_from_memport(this, sset, true);
    }
}

void MemPort::get_influenced_outputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    if (is_write())
    {
        if (tf_intersect(write_timing, t).empty())
            return;

        if (s != nullptr && s != enable && s != address &&
                !data.contains(s) &&
                !control.contains(s))
            return;

        if (!owner)
            return;

        for (MemPort *port : owner->ports)
            if (port->is_read())
            {
                tfint propag_delay = write_timing + port->read_timing;
                if (tf_intersect(propag_delay, t).empty())
                    continue;
                sset.merge(port->data);
            }
    }
    else if (is_read())
    {
        if (s == nullptr || s == enable || s == address ||
               control.contains(s))
            if (!tf_intersect(read_timing, t).empty())
                sset.merge(data);
    }
}

SignalConnectionType MemPort::connects(Signal *s)
{
    if (s == nullptr)
        return CT_UNDEFINED;
    if (s == enable)
        return CT_EN;
    if (s == address)
        return CT_A;
    if (data.contains(s))
        return CT_DATA;
    if (control.contains(s))
        return CT_UNKNOWN;
    return CT_UNDEFINED;
}

void MemPort::add_data(Signal *s)
{
    if (!data.contains(s))
    {
        data.insert(s);
        s->set_data();
        s->add_connection(CT_DATA, this);
    }
}

void MemPort::add_enable(Signal *s)
{
    ASSERT(enable == nullptr, "EN for " << id << " already set.");
    enable = s;
    s->set_control();
    s->add_connection(CT_EN, this);
}

void MemPort::add_address(Signal *s)
{
    ASSERT(address == nullptr, "ADDRESS for " << id << " already set.");
    address = s;
    s->set_control();
    s->add_connection(CT_A, this);
}

void MemPort::add_control(Signal *s)
{
    if (!control.contains(s))
    {
        control.insert(s);
        s->set_control();
        s->add_connection(CT_UNKNOWN, this);
    }
}

Signal *MemPort::remove_signal(Signal *s)
{
    if (s == enable)
        return remove_enable();
    else if (s == address)
        return remove_address();
    else
        if (remove_data(s) != nullptr)
            return s;
        else
            return remove_control(s);
}

Signal *MemPort::remove_enable()
{
    Signal *s = enable;
    enable = nullptr;
    if (s)
        s->remove_connection(CT_EN, owner);
    return s;
}

Signal *MemPort::remove_address()
{
    Signal *s = address;
    enable = nullptr;
    if (s)
        s->remove_connection(CT_A, owner);
    return s;
}

Signal *MemPort::remove_data(Signal *s)
{
    if (s == nullptr)
    {
        if (data.contains(s))
        {
            data.erase(s);
            s->remove_connection(CT_DATA, owner);
            return s;
        }
    }
    return nullptr;
}

Signal *MemPort::remove_control(Signal *s)
{
    if (s == nullptr)
    {
        if (control.contains(s))
        {
            control.erase(s);
            s->remove_connection(CT_UNKNOWN, owner);
            return s;
        }
    }
    return nullptr;
}

void MemPort::reconnect(int type, Signal *sold, Signal *snew)
{
    switch (type)
    {
        case CT_DATA:   replace_sig_in_set(sold, snew, data); break;
        case CT_EN:     enable = snew; break;
        case CT_A:      address = snew; break;
        case CT_UNKNOWN:    replace_sig_in_set(sold, snew, control); break;
    }
}

void MemPort::check_bitwidth()
{
    ASSERT(owner, "incomplete register file port (owner missing)");
    ASSERT(address, "incomplete register file port (address missing)");
    ASSERT(data.size(), "incomplete register file port (data missing)");
    ASSERT(enable, "incomplete register file port (enable missing)");
    for (auto d : data)
        ASSERT(d->bitwidth == owner->bitwidth,
            "incompatible data bitwidth of memory port " << id << endl <<
            "memory data width: " << owner->bitwidth <<
            ", data signal '" << d->id << "' width: " << d->bitwidth);
    ASSERT(enable->bitwidth == 1, "incompatible enable bitwidth " <<
            enable->bitwidth);
    ASSERT(address->bitwidth == owner->addr_bitwidth,
            "incompatible address bitwidth " << address->bitwidth <<
            ", required " << owner->addr_bitwidth);
}

/****************************************************************************
 * Memory
 ***************************************************************************/
Memory::~Memory()
{
    for (auto p : ports)
        delete p;
}

void Memory::Accept(RTLVisitor *v)
{
    v->VisitMemory(this);
}

void Memory::get_influenced_inputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    for (MemPort *port : ports)
        port->get_influenced_inputs(s, t, sset);
}

void Memory::get_influenced_outputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    for (MemPort *port : ports)
        port->get_influenced_outputs(s, t, sset);
}

void Memory::add_port(MemPort *port)
{
    port->owner = this;
    port->check_bitwidth();
    ports.push_back(port);
}

void Memory::add_rst(Signal *s)
{
    ASSERT(rst == nullptr, "RESET for " << id << " already set.");
    rst = s;
    s->set_control();
    s->add_connection(CT_RST, this);
}

void Memory::reconnect(int type, Signal *sold, Signal *snew)
{
    if (type == CT_RST)
        rst = snew;
    else
        for (auto port : ports)
            port->reconnect(type, sold, snew);
}

void Memory::check_bitwidth()
{
    ASSERT(size > 0, "incomplete memory " << id << " (size=" << size << ")");
    ASSERT(bitwidth > 0, "incomplete memory " << id <<
            " (bitwidth=" << bitwidth << ")");
    int w = 1;
    while (1 << w < size) w++;
    ASSERT(addr_bitwidth >= w, "memory " << id <<
            ": address bitwidth (" << addr_bitwidth <<
            ") incompatible with memory size (" << size << ")");
    if (rst)
        ASSERT(rst->bitwidth == 1, "memory " << id <<
            ": invalid RESET bitwidth " << rst->bitwidth);
    for (auto p : ports)
        p->check_bitwidth();
}

/****************************************************************************
 * Functional node
 ***************************************************************************/
/********************************************
 * ExprRTLAssignments
 */
ExprRTLAssignments::~ExprRTLAssignments()
{
    for (auto o : operands)
        remove(o);
}

void ExprRTLAssignments::add_operand(ExprAssign *o)
{
    ASSERT(o->lhs->op == OP_VAR, "unsupported LHS (operator " << o->lhs->op <<
            ") of an assignment");
    operands.push_back(static_cast<ExprAssign*>(o->bind(this)));
    reset_bitwidth();

    // set signal writing expr accordingly
    Signal *s = static_cast<Signal*>(static_cast<ExprVariable*>(o->lhs)->v);
    s->writing_expr = static_cast<ExprVariable*>(o->lhs);
}

void ExprRTLAssignments::remove_operand(ExprAssign *o)
{
    auto i = find(operands.begin(), operands.end(), o);
    if (i != operands.end())
    {
        remove(*i);
        operands.erase(i);
    }
}

void ExprRTLAssignments::replace_operand(Expression *o, Expression *r)
{
    ASSERT(r->op == OP_ASSIGN, "unsupported replacement, expected assignment");
    for (auto &expr : operands)
        if (expr == o)
        {
            remove(expr);
            expr = static_cast<ExprAssign*>(r->bind(this));
        }
}

/********************************************
 * FunNode
 */
void FunNode::Accept(RTLVisitor *v)
{
    v->VisitFunNode(this);
}

void FunNode::get_influenced_inputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    if (t.contains(0))
        if (s == nullptr || outputs.contains(s))
            sset.merge(inputs);
}

void FunNode::get_influenced_outputs(Signal *s, const tfint &t,
        SignalSet &sset)
{
    if (t.contains(0))
        if (s == nullptr || inputs.contains(s))
            sset.merge(outputs);
}

bool FunNode::contains_input(Signal *s, const tfint &t)
{
    return t.contains(0) && inputs.contains(s);
}

bool FunNode::contains_output(Signal *s, const tfint &t)
{
    return t.contains(0) && outputs.contains(s);
}

/****************************************************************************
 * System
 ***************************************************************************/
void System::Accept(RTLVisitor *v)
{
    v->VisitSystem(this);
}

bool System::id_exists(const string &id)
{
    for (auto sig : signals)
        if (id == sig->id)
            return true;
    for (auto reg : registers)
        if (reg->id == id)
            return true;
    for (auto mem : memories)
    {
        if (mem->id == id)
            return true;
        for (auto port : mem->ports)
            if (port->id == id)
                return true;
    }
    for (auto fnode : fnodes)
        if (fnode->id == id)
            return true;

    return false;
}

string System::get_unique_identifier(string prefix,
            bool try_bare)
{
    // approximate number of rtl objects (no memports)
    int num_of_objects = signals.size() + registers.size() +
        memories.size() + fnodes.size();
    if (unique_counter < num_of_objects)
        unique_counter = num_of_objects;
    string id;

    // try prefix as the id itself
    if (try_bare)
    {
        id = prefix;
        if (!id_exists(id))
            return id;
    }

    while (1) // do not like do-while
    {
        id = prefix + std::to_string(unique_counter++);
        if (id_exists(id))
            continue;
        return id;
    }
}

/****************************************************************************
 * System
 ***************************************************************************/
void RTLObject::Accept(RTLVisitor *v)
{
    switch (type)
    {
        case EXPR:
            return expr->Accept(v);
        case SIGNAL:
            return signal->Accept(v);
        case REGISTER:
            return reg->Accept(v);
        case MEMPORT:
            return port->Accept(v);
        case MEMORY:
            return mem->Accept(v);
        case SYSTEM:
            return system->Accept(v);
        case STORAGE:
            if (dynamic_cast<Register*>(storage))
                return reg->Accept(v);
            else if (dynamic_cast<MemPort*>(storage))
                return port->Accept(v);
            else if (dynamic_cast<Memory*>(storage))
                return mem->Accept(v);
        case FNODE:
            return fnode->Accept(v);
        case NONE:
            return;
    }
}

string RTLObject::get_typestr()
{
    switch (type)
    {
        case EXPR: return "expression";
        case SIGNAL: return "signal";
        case REGISTER: return "register";
        case MEMPORT: return "memport";
        case MEMORY: return "memory";
        case SYSTEM: return "system";
        case STORAGE: return "storage";
        case FNODE: return "fnode";
        default:
                    return "none";
    }
}

void set_meta(RTLObject ro, const string& key, const string& val)
{
    Meta* meta = nullptr;
    switch(ro.type)
    {
        case RTLObject::SIGNAL:   meta = &ro.signal->meta; break;
        case RTLObject::REGISTER: meta = &ro.reg->meta;    break;
        case RTLObject::MEMPORT:  meta = &ro.port->meta;   break;
        case RTLObject::FNODE:    meta = &ro.fnode->meta;  break;
        case RTLObject::STORAGE:  meta = &ro.storage->meta;break;
        case RTLObject::EXPR:     meta = &ro.expr->meta;   break;
        case RTLObject::MEMORY:   meta = &ro.mem->meta;    break;
        case RTLObject::SYSTEM:   meta = &ro.system->meta; break;
        default:
            cerr << "warning: set_meta(): Unexpected type of RTLObject occurred!"
                << endl;
            return;
    }
    meta->find_erase(key);
    meta->addkv(key, val);
}

string get_meta(RTLObject ro, const string& key)
{
    Meta* meta = nullptr;
    switch(ro.type)
    {
        case RTLObject::SIGNAL:   meta = &ro.signal->meta; break;
        case RTLObject::REGISTER: meta = &ro.reg->meta;    break;
        case RTLObject::MEMPORT:  meta = &ro.port->meta;   break;
        case RTLObject::FNODE:    meta = &ro.fnode->meta;  break;
        case RTLObject::STORAGE:  meta = &ro.storage->meta;break;
        case RTLObject::EXPR:     meta = &ro.expr->meta;   break;
        case RTLObject::MEMORY:   meta = &ro.mem->meta;    break;
        case RTLObject::SYSTEM:   meta = &ro.system->meta; break;
        default:
            cerr << "warning: get_meta(): Unexpected type of RTLObject occurred!"
                << endl;
            return string("");
    }
    Meta* node = meta->find(key);
    if (node == nullptr)
        return string("");

    return meta->keyval_nocreat(key);
}

bool is_micro(RTLObject ro)
{
    return ro.type == RTLObject::REGISTER &&
        ro.reg->stall == nullptr &&
        ro.reg->meta.find("architectural") == nullptr;
}

bool is_pipe(RTLObject ro)
{
    return ro.type == RTLObject::REGISTER &&
        ro.reg->stall != nullptr;
}

bool is_arch(RTLObject ro)
{
    return ro.type == RTLObject::MEMPORT ||
        (ro.type == RTLObject::REGISTER &&
        ro.reg->stall == nullptr &&
        ro.reg->meta.find("architectural") != nullptr);
}

} // namespace vam
