/**
 * @desc RTL frame over VAM
 */
#ifndef _VAMRTL_H
#define _VAMRTL_H

#include "vam.h"
#include "tframes.h"
#include "generateset.h"
#include <set>
#include <map>
#include <array>
#include <algorithm> // find
#include <boost/algorithm/string.hpp> // replace_all

using namespace std;

namespace vam
{

class RTLException : public VAMException
{
public:
    RTLException(string m) noexcept : VAMException(m) {}
    const char *what() const noexcept { return "RTLException"; }
};

#define RTLTHROW(msg) \
    { ostringstream _throw_ss; \
      _throw_ss << msg; \
      throw RTLException(_throw_ss.str()); }

#define RTLTHROWLOC(loc, msg) \
    { ostringstream _throw_ss; \
      _throw_ss << msg; \
      string d = _throw_ss.str(); \
      if (d[d.size()-1] == '\n') \
          d.erase(d.size()-1,1); \
      boost::replace_all(d, "\n", "\n" + loc); \
      throw RTLException(loc + d); }

class RTLVisitor;
// acceptable default definition
#define RTLAcceptor(type) void type::Accept(RTLVisitor *v) { v->Visit(this); }

class Signal;
class Storage;
class Register;
class Memory;
class MemPort;
class FunNode;
class System;

/// Set of signals.
using SignalSet = GenerateSet<Signal*>;
/// Set of registers.
using RegisterSet = GenerateSet<Register*>;
/// Set of memory ports, memories, and other storages
using StorageSet = GenerateSet<Storage*>;
/// Set of identifiers.
using IdSet = GenerateSet<string>;

/// OR-ed data access direction
enum DataAccessDir
{
    DA_UNDEFINED    = 0,
    DA_READ         = 0x1,
    DA_WRITE        = 0x2,
    DA_READWRITE    = 0x3, // DA_READ | DA_WRITE
};

/**
 * \{
 * @brief Auxiliary inline functions for methods of objects using `dir'
 * (DataAccessDir).
 */
inline bool da_is_read(int dir) { return dir & DA_READ; }
inline bool da_is_write(int dir) { return dir & DA_WRITE; }
inline void da_set_read(int &dir) { dir |= DA_READ; }
inline void da_set_write(int &dir) { dir |= DA_WRITE; }
/// \}

/// OR-ed value types
enum ValueType
{
    VT_UNDEFINED    = 0,
    VT_DATA         = 0x1,
    VT_ADDRESS      = 0x2,
    VT_CONTROL      = 0x4,
};

/**
 * \{
 * @brief Auxiliary inline functions for methods of objects using `vt'
 * value type.
 */
inline bool vt_is_data(int vt) { return vt & VT_DATA; }
inline bool vt_is_address(int vt) { return vt & VT_ADDRESS; }
inline bool vt_is_control(int vt) { return vt & VT_CONTROL; }
inline void vt_set_data(int &vt) { vt |= VT_DATA; }
inline void vt_set_address(int &vt) { vt |= VT_ADDRESS; }
inline void vt_set_control(int &vt) { vt |= VT_CONTROL; }
/// \}

class Signal;

/**
 * @desc   A node of RTL design abstracting functional node or storage with
 * some inputs and outputs. To understand the principle, consider signal
 * assignment which assigns a signal value from some expression. This can be
 * represented as functional node with inputs as all signals referenced in the
 * expression, and output containing just the assigned signal. Another kind of
 * RTL node is storage, e.g. registers or memory ports. For instance, register
 * inputs such as D, WE, and RST signals are featured by time frame interval
 * within which they have influence on the value of the register. In
 * particular, a simple flipflop outputs has the dependency function:
 * $Q_t = function_of(Q_{t-1}, D_{t-1}, WE_{t-1}, RST_t)$, where $Q_t$ means
 * value of Q in time frame t.
 *
 * This can be represented by the set of node (register) inputs and outputs.
 * Given a time interval [-1;0] which is relative to 't' in previous equation,
 * inputs of the register are {Q,D,WE,RST}. Given the time interaval
 * [-INF;-1], the inputs are just {Q,D,WE,RST}.
 *
 * \example Example of time frame dependency on input/output sets of signals
 * of pipeline register (WE, CLR, and STALL are synchronous).
 * @verbatim
   time frame: -2 | -1 |  0 | +1 | +2
              ----|--+-|--+-|----|----
                     |    |
   inputs:           |    +--- RST
                     +----|--- CLR, STALL, WE, D, Q
                          |
   outputs:               +--- Q
   @endverbatim
 */
class RTLNode
{
public:
    virtual ~RTLNode() noexcept = default;

    /**
     * @brief Fills the set of signals representing inputs of the node
     * (functional or storage node) which influence (within time interval t)
     * the given signal.
     * @param   s       output signal which depends on the inputs (nullptr
     *                  means any output signal)
     * @param   t       time interval
     * @param   sset    signal set to be filled with node inputs
     */
    virtual void get_influenced_inputs(Signal *s, const tfint &t,
            SignalSet &sset);

    /**
     * @brief Fills the set of signals representing outputs of the node
     * (functional or storage node) which depends (within time interval t)
     * on the given signal.
     * @param   s       input signal which influence the outputs (nullptr
     *                  means any intput signal)
     * @param   t       time interval
     * @param   sset    signal set to be filled with node outputs
     */
    virtual void get_influenced_outputs(Signal *s, const tfint &t,
            SignalSet &sset);

    /**
     * @brief Returns true if the signal is a part of inputs of the given
     * node.
     */
    virtual bool contains_input(Signal *s, const tfint &t = TF_ALL);

    /**
     * @brief Returns true if the signal is a part of outputs of the given
     * node.
     */
    virtual bool contains_output(Signal *s, const tfint &t = TF_ALL);

    /// \}
};

/****************************************************************************
 * Signal
 ***************************************************************************/

/**
 * \{
 * @desc    Each signal can be connected to several storages and their ports
 * such as D, Q, WE, etc. Such a connection (SignalConnection) is represented
 * by a type (SignalConnectionType) and storage identification (a pointer to
 * the Storage object). Since there are various kinds of storages,
 * SignalConnectionType enumerates every possible type of connection. For
 * instance, for connection carrying data, there are CT_D representing data to
 * be written to a storage, CT_Q representing data to be read from a storage,
 * and more general CT_DATA representing data which may be read from or
 * written to a storage.
 */

enum SignalConnectionType
{
    CT_UNDEFINED,
    CT_Q                          = 1, // read data
    CT_DATA,    // read+write data
                // CT_DATA also used as separator of read and write connection
                // types
    CT_D      , // write data
    CT_WE     ,  // write enable
    CT_RE     ,  // read enable
    CT_EN     ,  // enable (read or write)
    CT_WA     ,  // write address
    CT_RA     ,  // read address
    CT_A      ,  // address (read or write)
    CT_STALL  ,  // stall (for registers or pipeline stage)
    CT_CLR    ,  // clear (for registers or pipeline stage)
    CT_RST    ,  // reset
    CT_UNKNOWN,  // unknown
};

class SignalConnection
{
public:
    int type;
    Storage *storage;
public:
    SignalConnection(int t = CT_UNDEFINED, Storage *s = nullptr) :
        type(t), storage(s) {}

    /**
     * @brief  Returns true, if the connection represents reading from
     * the storage.
     */
    bool is_read();

    /**
     * @brief  Returns true, if the connection represents writing to
     * the storage.
     */
    bool is_write();
};
/// \}

/// Signal access within its owner functional node.
enum FunNodeAccess
{
    FNA_PRIVATE,    // signal is used by the owner only
    FNA_PUBLIC,     // signal is an output of the functional node
};

enum VariableTypeRTL
{
    VT_SIGNAL = VT_GENERAL + 1,
};

/**
 * @brief   Signal is an envelope over vam::Variable.
 */
class Signal : public Variable
{
public:
    /// value type
    int vt;
    /// direction of data flow
    int dir;
    /// influential inputs
    set<Storage*> inputs;
    set<string> inputs_name; // when generating model from input file, where
                             // storages are unknown
    /// owner stage
    int stage;

    /**
     * @brief Connections to storages. Thic class member can be manipulated
     * (i) directly, (ii) through methods add_connection(),
     * find_connections(), or (iii) by methods of Storage objects
     * `remove_signal()' and similar (cf. Register, MemPort, and Memory
     * classes).
     */
    vector<SignalConnection> connections;

    /**
     * @brief The owner is the functional node responsible for setting the
     * signal value. nullptr means that the information is not present.
     */
    FunNode *owner = nullptr;

    /// Access type to the signal wrt. owner.
    FunNodeAccess faccess = FNA_PUBLIC;

    /**
     * @brief   Expression reference writing to this signal If not set but
     * should be, SystemManipulation::update_fnode_io() must be called.
     */
    ExprVariable *writing_expr = nullptr;

public:
    Signal(string id = "", int bw = -1) :
        Variable(id, bw), vt(VT_UNDEFINED), dir(DA_UNDEFINED), stage(-1),
        owner(nullptr), faccess(FNA_PRIVATE)
    {
        var_type = VT_SIGNAL;
    }

    void Accept(VAMVisitor *v) override;

    /// \{
    /// value type manipulation
    bool is_data() { return vt_is_data(vt); }
    bool is_address() { return vt_is_address(vt); }
    bool is_control() { return vt_is_control(vt); }
    void set_data() { vt_set_data(vt); }
    void set_address() { vt_set_address(vt); }
    void set_control() { vt_set_control(vt); }
    /// \}

    /// \{
    /// direction manipulation
    bool is_read() { return da_is_read(dir); }
    bool is_write() { return da_is_write(dir); }
    void set_read() { da_set_read(dir); }
    void set_write() { da_set_write(dir); }
    /// \}

    /// \{
    /// connection manipulation
    void add_connection(int type, Storage *storage);
    set<Storage*> find_connections(int type);
    set<int>      find_connections(Storage *storage);
    void remove_connection(int type, Storage *storage);
    /// \}
};

/****************************************************************************
 * Storage super class
 ***************************************************************************/

/**
 * @brief Auxiliary function. Intended to be used by Storage::reconnect()
 * methods.
 * @param   o   Signal to be replaced
 * @param   n   Replacement signal
 * @param   s   set of signals in which the replacement should be done
 */
inline void replace_sig_in_set(Signal *o, Signal *n, SignalSet &s)
{
    if (s.contains(o))
    {
        s.erase(o);
        s.insert(n);
    }
}

class Storage : public RTLNode
{
protected:
    Storage();

public:
    string id;
    int bitwidth;
    tfint read_timing;
    tfint write_timing;
    Meta meta;
public:
    Storage(string _id, int bw);
    virtual ~Storage();

    /**
     * @brief Assigns a new signal (snew) to a port specified by type
     * (SignalConnectionType). Replace the old one (sold) if existing. No
     * consistency control.
     */
    virtual void reconnect(int type, Signal *sold, Signal *snew) = 0;

};

/****************************************************************************
 * Register
 ***************************************************************************/

/**
 * @brief Register represents one register with optional stall (STALL),
 * clear (CLR), and reset (RST) signals.  STALL, CLR, and RST are
 * asynchronous. Q_t (value of signal Q in time frame t):
 * Q_t = function(Q_{t-1}, D_{t-1}, WE_{t-1}, STALL_t, CLT_t, RST_t)
 */
class Register : public Storage
{
public:
    Signal *d;
    Signal *we;
    SignalSet q;
    Signal *stall;
    Signal *clr;
    Signal *rst;
    int stage;

public:
    Register() :
        d(nullptr), we(nullptr), stall(nullptr), clr(nullptr), rst(nullptr),
        stage(-1)
    {
        read_timing.set(0,0);
        write_timing.set(1,1);
    }

    Register(string id, int bitwidth)
        : Storage(id, bitwidth),
          d(nullptr), we(nullptr), stall(nullptr), clr(nullptr), rst(nullptr),
          stage(-1)
    {
        read_timing.set(0,0);
        write_timing.set(1,1);
    }

    ~Register();

    void Accept(RTLVisitor *v);

    void get_influenced_inputs(Signal *s, const tfint &t,
            SignalSet &sset);
    void get_influenced_outputs(Signal *s, const tfint &t,
            SignalSet &sset);
    bool contains_input(Signal *s, const tfint &t);
    bool contains_output(Signal *s, const tfint &t);

    /// \{
    /// Add signal and set its value type accordingly.
    void add_d(Signal *s);
    void add_we(Signal *s);
    void add_q(Signal *s);
    void add_stall(Signal *s);
    void add_clr(Signal *s);
    void add_rst(Signal *s);

    /// Connect signals to D, Q, and WE ports
    void connect(Signal *_d, Signal *_q, Signal *_we)
        { add_d(_d); add_q(_q); add_we(_we); }
    /// Connect signals to D, Q, WE, STALL, and CLR ports
    void connect_pipeline(Signal *_d, Signal *_q, Signal *_we,
        Signal *_s, Signal *_c)
        { add_d(_d); add_q(_q); add_we(_we); add_stall(_s); add_clr(_c); }

    /// Remove connection between register and signal.
    Signal *remove_signal(Signal *s);
    Signal *remove_d();
    Signal *remove_we();
    /**
     * @brief Removes Q signal. If s is nullptr, and size of `q' is 1, remove
     * the only Q signal.
     */
    Signal *remove_q(Signal *s = nullptr);
    Signal *remove_stall();
    Signal *remove_clr();
    Signal *remove_rst();
    /// \}

    /// Check connections of signals and their bitwidth
    void check_bitwidth();

    void reconnect(int type, Signal *sold, Signal *snew);

};

/****************************************************************************
 * Memory
 ***************************************************************************/

/**
 * @brief Memory port. \n
 * Dependency function for write port is: $owner_content = function_of(
 * enable_{t-W}, address_{t-W}, data_{t-W}, control_{t-W})$
 * where W is write_timing interval.
 * Dependency function for read port is: $data_t = function_of(
 * enable_{t-R}, address_{t-R}, control_{t-R})$
 */
class MemPort : public Storage
{
public:
    Memory *owner;
    /// valid values are DA_READ and DA_WRITE only
    int dir;
    Signal *enable;
    Signal *address;
    SignalSet data;
    SignalSet control;
    int stage;

public:

    MemPort(Memory *o = nullptr) :
        Storage("", 0), owner(o), dir(DA_UNDEFINED),
        enable(nullptr), address(nullptr), stage(-1) {}
    ~MemPort();

    void Accept(RTLVisitor *v);

    /**
     * \{
     * RTLNode virtual methods */
    void get_influenced_inputs(Signal *s, const tfint &t,
            SignalSet &sset);
    void get_influenced_outputs(Signal *s, const tfint &t,
            SignalSet &sset);
    /// \}

    bool is_read() { return da_is_read(dir); }
    bool is_write() { return da_is_write(dir); }
    void set_read() { da_set_read(dir); }
    void set_write() { da_set_write(dir); }

    SignalConnectionType connects(Signal *s);

    /// \{
    /// Add signal and set its value type accordingly.
    void add_data(Signal *s);
    void add_enable(Signal *s);
    void add_address(Signal *s);
    void add_control(Signal *s);

    /// Remove connection between memory port and signal.
    Signal *remove_signal(Signal *s);
    Signal *remove_enable();
    Signal *remove_address();
    Signal *remove_data(Signal *s = nullptr);
    Signal *remove_control(Signal *s = nullptr);
    /// \}
    //

    void reconnect(int type, Signal *sold, Signal *snew);

    /**
     * @brief Checks bitwidth of all signals wrt. owner memory.
     * Throws VAMException if an error occurs.
     */
    void check_bitwidth();
};

class Memory : public Storage
{
public:
    using MemPorts = vector<MemPort*>;

    /// address bitwidth
    int addr_bitwidth;
    /// number of registers in the file
    int size;
    MemPorts ports;
    Signal *rst;

public:
    Memory() : addr_bitwidth(-1), size(-1), rst(nullptr) {}
    ~Memory();

    void Accept(RTLVisitor *v);

    /**
     * \{
     * RTLNode virtual methods */
    void get_influenced_inputs(Signal *s, const tfint &t,
            SignalSet &sset);
    void get_influenced_outputs(Signal *s, const tfint &t,
            SignalSet &sset);
    /// \}

    /**
     * @brief Add port to register file. Check its bitwidths. Throws
     * VAMException if an error occurs.
     */
    void add_port(MemPort *port);

    /**
     * @brief Connect a signal to RESET port.
     */
    void add_rst(Signal *s);

    void reconnect(int type, Signal *sold, Signal *snew);

    /**
     * @brief Check bit widths and size settings on all ports.
     */
    void check_bitwidth();

};

/****************************************************************************
 * Functional node
 ***************************************************************************/

class FunNode;

class ExprRTLAssignments : public Expression
{
public:
    vector<ExprAssign*> operands;
    FunNode *owner;
public:
    ExprRTLAssignments(FunNode *_owner = nullptr)
    {
        op = OP_NOP;
        owner = _owner;
    }
    ~ExprRTLAssignments();
    void add_operand(ExprAssign *o);
    void remove_operand(ExprAssign *o);
    void replace_operand(Expression *o, Expression *r);

    using iterator = vector<ExprAssign*>::iterator;
    iterator begin() { return operands.begin(); }
    iterator end() { return operands.end(); }
};

class FunNode : public RTLNode
{
public:
    string id;
    SignalSet inputs;
    SignalSet outputs;
    int stage;
    ExprRTLAssignments assignments;
    Meta meta;

public:
    FunNode() : stage(-1), assignments(this) {}

    ~FunNode() {}

    void Accept(RTLVisitor *v);

    void get_influenced_inputs(Signal *s, const tfint &t,
            SignalSet &sset);
    void get_influenced_outputs(Signal *s, const tfint &t,
            SignalSet &sset);
    bool contains_input(Signal *s, const tfint &t);
    bool contains_output(Signal *s, const tfint &t);

    int count_assignments()
    {
        return assignments.operands.size();
    }

    ExprAssign *get_assignment(int index)
    {
        return assignments.operands[index];
    }
};

/****************************************************************************
 * @desc Container of RTL elements of a given type. Includes iterators and
 * methods for insertion, deletion, and searching (via RTL element
 * identifier).
 ***************************************************************************/

template<class T>
class NamedElementContainer
{
protected:
    set<T*> _s;
    map<string, T*> symbols;

public:
    using iterator = typename set<T*>::iterator;
    using const_iterator = typename set<T*>::const_iterator;

    iterator begin() { return _s.begin(); }
    const_iterator cbegin() const { return _s.cbegin(); }
    iterator end() { return _s.end(); }
    const_iterator cend() const { return _s.cend(); }

    bool empty() const { return _s.empty(); }
    unsigned size() const { return _s.size(); }

public:
    T *insert(T *s)
    {
        if (_s.insert(s).second)
        {
            auto ret = symbols.insert(pair<string,T*>(s->id, s));
            if (ret.second == false)
                THROW("ambiguous symbol '" << s->id << "'");
            return s;
        }
        else
            THROW("element already existed");
        return nullptr;
    }

    T *find(string id)
    {
        auto i = symbols.find(id);
        if (i != symbols.end())
            return i->second;
        else
            return nullptr;
    }

    T *operator[](string id)
    {
        return find(id);
    }

    void erase(T *s)
    {
        _s.erase(s);
        auto i = symbols.find(s->id);
        if (i != symbols.end())
            symbols.erase(i);
    }

    T *update(T *s)
    {
        erase(s);
        return insert(s);
    }

    void delete_all()
    {
        for (auto s : _s)
            delete s;
        _s.clear();
        symbols.clear();
    }

};

using Signals = NamedElementContainer<Signal>;
using Registers = NamedElementContainer<Register>;
using Memories = NamedElementContainer<Memory>;
using FunNodes = NamedElementContainer<FunNode>;

/**
 * @brief   Erases all meta nodes corresponding to a given nodename
 * in each item of the given container.
 * @param   nodename    node name to be erased
 * @param   container   container of items (e.g. set of signals)
 */
template<class T>
void erase_metas(const string &nodename, T &container)
{
    for (auto obj : container)
        obj->meta.find_erase(nodename);
}

/****************************************************************************
 * Overall pipipelined system
 ***************************************************************************/

class System
{
public:
    string name;

    vector<string> stages;

    Signals signals;
    Registers registers;
    Memories memories;
    FunNodes fnodes;

    Meta meta;

private:
    System(const System &s);
    System(System&&s);
public:
    System() = default;
    ~System()
    {
        fnodes.delete_all();
        registers.delete_all();
        memories.delete_all();
        signals.delete_all();
    }

    void Accept(RTLVisitor *v);

    /**
     * @brief   Finds a stage
     * @param   id  stage identificator
     * @return  stage order (the first stage has value 0)
     */
    int find_stage(string id)
    {
        auto pos = find(stages.begin(), stages.end(), id);
        if (pos != stages.end())
            return pos - stages.begin();
        else
            return -1;
    }

    Signal *insert(Signal *s) { return signals.insert(s); }
    Register *insert(Register *s) { return registers.insert(s); }
    Memory *insert(Memory *s) { return memories.insert(s); }
    FunNode *insert(FunNode *f) { return fnodes.insert(f); }

    /**
     * @brief   Clear the system.
     */
    void clear()
    {
        name.clear();
        stages.clear();
        Meta tmp; swap(meta, tmp);
        fnodes.delete_all();
        registers.delete_all();
        memories.delete_all();
        signals.delete_all();
    }

    /**
     * @brief   Finds per system unique identifier.
     */
    int unique_counter = 0;

    /**
     * @brief   Finds and returns per system unique identifier
     * @param   prefix      identifier prefix (default: "obj_")
     * @param   try_bare    if true, then try the prefix as the identifier
     *                      itself first. If the identifier already exists,
     *                      append the unique number.
     * @return  unique identifier (no signal, register, memory, memory
     *          port or functional node exists with the same identifier).
     * @complexity linear with number of objects in system
     */
    string get_unique_identifier(string prefix = "obj_",
            bool try_bare = false);
protected:
    /**
     * @brief   Helper for get_unique_identifier().
     */
    bool id_exists(const string &id);
};

/**
 * @brief   Transforms set of signals to set of their identifiers.
 * @param   ss  set of signals.
 * @param   ids reference of set of identifiers to be filled.
 */
inline void signalset_to_idset(const SignalSet &ss, IdSet &ids)
{
    for (auto s : ss)
        ids << s->id;
}

class RTLObject
{
public:
    enum type_t {
        NONE = 0,
        SIGNAL = 1,
        STORAGE = 2,
        REGISTER = 3,
        MEMORY = 4,
        MEMPORT = 5,
        FNODE = 6,
        EXPR = 7,
        SYSTEM = 8,
    } type;
    union {
        void *pointer;
        Signal *signal;
        Storage *storage;
        Register *reg;
        Memory *mem;
        MemPort *port;
        FunNode *fnode;
        Expression *expr;
        System *system;
    };
    RTLObject() { type = NONE; signal = nullptr; }
    RTLObject(Signal *s)
    {
        type = SIGNAL;
        signal = s;
    }
    RTLObject(Storage *s)
    {
        storage = s;
        if (dynamic_cast<Register*>(s))
            type = REGISTER;
        else if (dynamic_cast<Memory*>(s))
            type = MEMORY;
        else if (dynamic_cast<MemPort*>(s))
            type = MEMPORT;
        else
            type = STORAGE;

    }
    RTLObject(Register *r)
    {
        type = REGISTER;
        reg = r;
    }
    RTLObject(Memory *m)
    {
        type = MEMORY;
        mem = m;
    }
    RTLObject(MemPort *p)
    {
        type = MEMPORT;
        port = p;
    }
    RTLObject(FunNode *f)
    {
        type = FNODE;
        fnode = f;
    }
    RTLObject(Expression *e)
    {
        type = EXPR;
        expr = e;
    }
    RTLObject(System *s)
    {
        type = SYSTEM;
        system = s;
    }
    Signal *operator=(Signal *s)
    {
        type = SIGNAL;
        return signal = s;
    }
    Storage *operator=(Storage *s)
    {
        type = STORAGE;
        return storage = s;
    }
    Register *operator=(Register *r)
    {
        type = REGISTER;
        return reg = r;
    }
    Memory *operator=(Memory *m)
    {
        type = MEMORY;
        return mem = m;
    }
    MemPort *operator=(MemPort *p)
    {
        type = MEMPORT;
        return port = p;
    }
    FunNode *operator=(FunNode *f)
    {
        type = FNODE;
        return fnode = f;
    }
    Expression *operator=(Expression *e)
    {
        type = EXPR;
        return expr = e;
    }
    System *operator=(System *s)
    {
        type = SYSTEM;
        return system = s;
    }
    bool operator==(RTLObject o)
    {
        return (type == o.type) && (pointer == o.pointer);
    }
    void Accept(RTLVisitor *v);
    string get_typestr();
};

void set_meta(RTLObject ro, const string& key, const string& val);
string get_meta(RTLObject ro, const string& key);
bool is_micro(RTLObject ro);
bool is_pipe(RTLObject ro);
bool is_arch(RTLObject ro);

} // namespace vam
#endif
