/**
 * @brief   High-level test suite for RTL elements.
 */

#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>
#include "vaml_test.h"

BOOST_AUTO_TEST_SUITE(rtl_elements)

BOOST_AUTO_TEST_CASE(rtl1)
{
    Signal *d = new Signal{"fetchD", 32};
    Signal *q = new Signal{"fetchQ", 32};
    Signal *we = new Signal{"fetchWE", 1};

    Register r{"fetch", 32};
    r.add_d(d);
    r.add_q(q);
    r.add_we(we);

    auto a1 = new ExprAssign{
            new ExprVariable{we},
            new ExprNot{
            new ExprEqual{
                new ExprSlice{
                    new ExprVariable{q},
                    0,3}, 
                new ExprConst{"15", 4}}}};
    *a1->meta.add(MT_NODE, "reason") 
        << "load a new instruction iff no halt";
    
    RTLVisitor ov;

    d->Accept(&ov); ov.out << endl;

    r.remove_d();
    d->Accept(&ov); ov.out << endl;
    
    q->Accept(&ov); ov.out << endl;
    we->Accept(&ov); ov.out << endl;
    a1->Accept(&ov); 

    BOOST_REQUIRE_EQUAL(strip(ov.out.str()), strip(
            "(sig fetchD 32 (vt data ) (connections (fetch d) ) )"
            "(sig fetchD 32 (vt data ) )"
            "(sig fetchQ 32 (vt data ) (connections (fetch q) ) )"
            "(sig fetchWE 1 (vt control) (connections (fetch we) ) )"
            "(:= fetchWE (! (== ([..] fetchQ 0 3 ) (_ 15 4) ) ) "
            "(meta (reason \"load a new instruction iff no halt\" ) ) ) "
                ));

    // teardown
    delete a1;
    delete d;
    delete q;
    delete we;
}

BOOST_AUTO_TEST_CASE(rtl2)
{
    Signal *d = new Signal{"r0D", 32};
    Signal *q = new Signal{"r0Q", 32};
    Signal *we = new Signal{"r0WE", 1};
    Register r{"r0", 32};
    r.add_d(d);
    r.add_q(q);
    r.add_we(we);

    RTLVisitor ov;
    // show register
    r.Accept(&ov); ov.out << endl;
    // show D signal
    d->Accept(&ov); ov.out << endl;
    // remove D from register from D and add it twice to Q
    r.remove_d();
    r.add_q(d);
    r.add_q(d);
    // show D again (should have just one connection)
    d->Accept(&ov); ov.out << endl;
    // show Q signal
    q->Accept(&ov); ov.out << endl;
    // show WE signal
    we->Accept(&ov); ov.out << endl;
    // show register again
    r.Accept(&ov);

    BOOST_REQUIRE_EQUAL(strip(ov.out.str()), strip(
           "(reg r0 32 (we r0WE) (d r0D) (q r0Q ) )"
           "(sig r0D 32 (vt data ) (connections (r0 d) ) )"
           "(sig r0D 32 (vt data ) (connections (r0 q) ) )"
           "(sig r0Q 32 (vt data ) (connections (r0 q) ) )"
           "(sig r0WE 1 (vt control) (connections (r0 we) ) )"
           "(reg r0 32 (we r0WE) (d N/A) (q r0D r0Q ) )"
                ));
    // teardown
    delete d;
    delete q;
    delete we;
}

BOOST_AUTO_TEST_CASE(system1)
{
    System sys;
    Signal *d = sys.insert(new Signal{"r0_D", 32});
    Signal *q = sys.insert(new Signal{"r0_Q", 32});
    Signal *we = sys.insert(new Signal{"r0_WE", 1});
    Register *r = sys.insert(new Register{"r0", 32});
    r->connect(d, q, we);

    RTLVisitor v;
    sys.Accept(&v);
    
    BOOST_REQUIRE_EQUAL(strip(v.out.str()), strip(
           "(model "
           "  (pipeline-stages)"
           "  (sig r0_D 32 (vt data ) (connections (r0 d) ) )"
           "  (sig r0_Q 32 (vt data ) (connections (r0 q) ) )"
           "  (sig r0_WE 1 (vt control) (connections (r0 we) ) )"
           "  (reg r0 32 (we r0_WE) (d r0_D) (q r0_Q ) )"
           ")"
            ));
}

BOOST_AUTO_TEST_SUITE_END()
