/**
 * @brief   High-level test suite for RTL input/output.
 */

#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>
#include "vaml_test.h"
#include <array>

BOOST_AUTO_TEST_SUITE(rtl_language)

BOOST_AUTO_TEST_CASE(lexer)
{
    static struct {
        unsigned row; 
        unsigned col; 
        int type; 
        string value;
    } tokens[] = {
        {1, 1, T_LPAR, "("},
        {1, 2, T_IDENT, "sig"},
        {1, 6, T_IDENT, "r0"},
        {1, 9, T_CONST, "1"},
        {1, 11, T_LPAR, "("},
        {1, 12, T_IDENT, "vt"},
        {1, 15, T_IDENT, "control"},
        {1, 23, T_LPAR, "("},
        {1, 24, T_IDENT, "meta"},
        {1, 29, T_STRING, "as\ndf"},
        {2, 4, T_RPAR, ")"},
        {2, 5, T_RPAR, ")"},
        {2, 6, T_RPAR, ")"},
        {2, 7, T_EOF, ""},
    };
    Lexer l{"(sig r0 1 (vt control (meta \"as\ndf\")))"};
    Token t = l.next_token();
    int i = 0;
    while (i < 14)
    {
        BOOST_REQUIRE_MESSAGE(t.row == tokens[i].row,
                "token #" << i << " row=" << t.row);
        BOOST_REQUIRE_MESSAGE(t.col == tokens[i].col,
                "token #" << i << " col=" << t.col);
        BOOST_REQUIRE_MESSAGE(t.type == tokens[i].type,
                "token #" << i << " type=" << t.type);
        BOOST_REQUIRE_MESSAGE(t.value == tokens[i].value,
                "token #" << i << " value=" << t.value);
        t = l.next_token();
        i++;
    }
    BOOST_REQUIRE_EQUAL(i, 14);
}

BOOST_AUTO_TEST_CASE(parse1)
{
    RTLParser p;
    System s;
    p.parse_str("(model (meta 42 (loc 5)) modelname"
            "(pipeline-stages fe id ex wb (meta x)))", &s);
    RTLVisitor v;
    s.Accept(&v);

    BOOST_REQUIRE_EQUAL(strip(v.out.str()), strip(
        "(model modelname"
        "(pipeline-stages fe id ex wb)"
        "(meta 42 (loc 5 ) \"x\" ))"
                ));
}

BOOST_AUTO_TEST_CASE(parse2)
{
    string input = "(model modelname"
        "(pipeline-stages fe id ex wb)"
        "(sig v1 1 (vt control))"
        "(sig v2 32 (vt data))"
        "(sig v3 32 (vt data) (stage id)))";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v;
    s.Accept(&v);
    BOOST_REQUIRE_EQUAL(strip(v.out.str()), strip(input));
}

BOOST_AUTO_TEST_CASE(parse3)
{
    string input = "(model modelname"
            "(pipeline-stages fe id ex wb)"
            "(sig v1 1)"
            "(sig v2 32)"
            "(sig v3 32 (stage id))"
            "(reg r0 32 (we v1) (d v2) (q v3)))";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v;
    s.Accept(&v);
    
    string expected = "(model modelname"
          "(pipeline-stages fe id ex wb)"
          "(sig v1 1 (vt control) (connections (r0 we) ) )"
          "(sig v2 32 (vt data ) (connections (r0 d) ) )"
          "(sig v3 32 (vt data ) (stage id) (connections (r0 q) ) )"
          "(reg r0 32 (we v1) (d v2) (q v3 ) ) )";
    BOOST_REQUIRE_EQUAL(strip(v.out.str()), strip(expected));
}

BOOST_AUTO_TEST_CASE(parse_mem)
{
    string input = "(model modelname\n"
          "(pipeline-stages fe id ex wb)\n"
          "(sig v1 1)\n"
          "(sig v2 3)\n"
          "(sig v3 32 (stage id))\n"
          "(mem regs 32 3 8 (ports \n"
          "(port p0 (dir ro) (timing 0) (en v1) (addr v2) (data v3)\n"
          "  (control v3) (stage id))\n"
          ")))\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v;
    s.Accept(&v);

    string expected = "(model modelname"
              "(pipeline-stages fe id ex wb)"
              "(sig v1 1 (vt control) (connections (p0 en) ) )"
              "(sig v2 3 (vt control) (connections (p0 address) ) )"
              "(sig v3 32 (vt data control) (stage id) "
                "(connections (p0 data) (p0 unknown) ) )"
              "(mem regs 32 3 8 (ports (port p0 (dir ro) (timing 0)"
              "(en v1) (addr v2) (data v3 ) (control v3 ) (stage id) ) ) ) )";
    BOOST_REQUIRE_EQUAL(strip(v.out.str()), strip(expected));
}

BOOST_AUTO_TEST_CASE(parse_fnode_unary)
{
    string input =
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (_ -1 3))\n"
  /*  8 */  "(:= v3 (! v1))\n"
  /*  9 */  "(:= v1 (-- (~ (_ 0 1))))\n"
  /* 10 */  ")))\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v{&s};
    string expected =
        "(sig v1 1 (vt unknown) )"
        "(sig v2 1 (vt unknown) )"
        "(sig v3 1 (vt unknown) (stage id))"
        "  (fnode f0 (input v1 ) (output v3 ) (stage ex) "
        "    (assign (:= v2 (_ -1 3) ) (:= v3 (! v1 ) ) "
        "            (:= v1 (-- (~ (_ 0 1) ) ) )     )) ";
    
    // just 3 signals
    BOOST_REQUIRE(s.signals.size() == 3);
    Signal *signal;
    for (auto id : {"v1", "v2", "v3"})
    {
        signal = s.signals[id];
        BOOST_REQUIRE(signal != nullptr);
        signal->Accept(&v);
    }
    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    fnode->Accept(&v);
    BOOST_REQUIRE_MESSAGE(strip(v.out.str()) == strip(expected),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(parse_fnode_binary)
{
    string input =
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (| (_ 1 1) (_ 0 1)))\n"
  /*  8 */  "(:= v3 (& v2 v3))\n"
  /*  9 */  "(:= v1 ([] (>>. (- (_ 0 7) (^^ (<< v2 (_ 6 3)) 7)) (_ 6 3)) 0))\n"
  /* 10 */  ")))\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v{&s};
    string expected =
        "  (sig v1 1 (vt unknown) )"
        "  (sig v2 1 (vt unknown) )"
        "  (sig v3 1 (vt unknown) (stage id) )"
        "  (fnode f0 (input v1 ) (output v3 ) (stage ex) "
        "    (assign (:= v2 (| (_ 1 1) (_ 0 1) ) ) "
        "            (:= v3 (& v2 v3 ) ) "
        "            (:= v1 ([] (>>. (- (_ 0 7) (^^ (<< v2 (_ 6 3) ) 7 ) )"
        "                (_ 6 3) ) 0 ) ) ))";
    // just 3 signals
    BOOST_REQUIRE(s.signals.size() == 3);
    Signal *signal;
    for (auto id : {"v1", "v2", "v3"})
    {
        signal = s.signals[id];
        BOOST_REQUIRE(signal != nullptr);
        signal->Accept(&v);
    }
    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    fnode->Accept(&v);
    BOOST_REQUIRE_MESSAGE(strip(v.out.str()) == strip(expected),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(parse_fnode_special)
{
    string input =
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(sig v4 1)\n"
  /*  7 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  8 */  "(:= v2 (? v3 (_ 1 1) (_ 0 1)))\n"
  /*  9 */  "(:= v3 (switch v2 ((_ 0 1) v1) ((_ 1 1) v2) (v3)))\n"
  /* 10 */  "(:= v4 (cond ((_ 0 1) v1) ((_ 1 1) v2) (v3)))\n"
  /* 11 */  "(:= v1 ([] ([..] (^^. v2 8) 0 3) 1))\n"
  /* 12 */  ")))\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v{&s};
    string expected =
        "  (sig v1 1 (vt unknown) )"
        "  (sig v2 1 (vt unknown) )"
        "  (sig v3 1 (vt unknown) (stage id) )"
        "  (sig v4 1 (vt unknown) )"
        "  (fnode f0 (input v1 ) (output v3 ) (stage ex) "
        "    (assign (:= v2 (? v3 (_ 1 1) (_ 0 1) ) )"
        "            (:= v3 (switch v2 ((_ 0 1) v1 ) "
        "               ((_ 1 1) v2 ) (v3 ) ) ) "
        "           (:= v4 (cond ((_ 0 1) v1 ) ((_ 1 1) v2 ) (v3 ) ) ) "
        "           (:= v1 ([] ([..] (^^. v2 8 ) 0 3 ) 1 ) )     )) ";
    // just 3 signals
    BOOST_REQUIRE(s.signals.size() == 4);
    Signal *signal;
    for (auto id : {"v1", "v2", "v3", "v4"})
    {
        signal = s.signals[id];
        BOOST_REQUIRE(signal != nullptr);
        signal->Accept(&v);
    }
    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    fnode->Accept(&v);
    BOOST_REQUIRE_MESSAGE(strip(v.out.str()) == strip(expected),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(parse_fnode_abstract)
{
    string input =
  /*  1 */  "(model modelname\n"
  /*  2 */  "(pipeline-stages fe id ex wb)\n"
  /*  3 */  "(sig v1 1)\n"
  /*  4 */  "(sig v2 1)\n"
  /*  5 */  "(sig v3 1 (stage id))\n"
  /*  6 */  "(fnode f0 (input v1) (output v3) (stage ex) (assign \n"
  /*  7 */  "(:= v2 (!! 1 (_ 1 1) (_ 0 1)))\n"
  /*  8 */  "(:= v3 (~~ 1 v3))\n"
  /*  9 */  "(:= v1 (<> 1 (][ 1 (-+ 1 (~~ 1 v2)))))\n"
  /* 10 */  ")))\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    RTLVisitor v{&s};
    string expected =
        "  (sig v1 1 (vt unknown) )"
        "  (sig v2 1 (vt unknown) )"
        "  (sig v3 1 (vt unknown) (stage id) )"
        "  (fnode f0 (input v1 ) (output v3 ) (stage ex) "
        "    (assign (:= v2 (!! 1 (_ 1 1) (_ 0 1) ))"
        "           (:= v3 (~~ 1 v3 ) ) "
        "           (:= v1 (<> 1 (][ 1 (-+ 1 (~~ 1 v2 ) ) ) ) ) )) ";
    // just 3 signals
    BOOST_REQUIRE(s.signals.size() == 3);
    Signal *signal;
    for (auto id : {"v1", "v2", "v3"})
    {
        signal = s.signals[id];
        BOOST_REQUIRE(signal != nullptr);
        signal->Accept(&v);
    }
    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    fnode->Accept(&v);
    BOOST_REQUIRE_MESSAGE(strip(v.out.str()) == strip(expected),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(parse_real_example)
{
    RTLParser p;
    System s;
    try {
        p.parse_from_file("test-real-example.vam", &s);
    }
    catch (VAMException &e)
    {
        BOOST_ERROR("parsing real example");
    }
    BOOST_REQUIRE_EQUAL(s.signals.size(), 1702);
    BOOST_REQUIRE_EQUAL(s.registers.size(), 23);
    BOOST_REQUIRE_EQUAL(s.memories.size(), 3);
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 1675);
}

BOOST_AUTO_TEST_SUITE_END()
