/**
 * @brief   High-level test suite for RTL manipulations.
 * @pre     Valid RTL model: test-real-example.vam
 */

#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>
#include "vaml_test.h"
#include <boost/algorithm/string/predicate.hpp>

BOOST_AUTO_TEST_SUITE(rtl_manipulation)

string log_model(string test_name, System &s)
{
    RTLVisitor v{&s};
    s.Accept(&v);

    string model_msg = "output model:\n" + v.out.str();
    BOOST_TEST_MESSAGE(test_name + " " + model_msg);

    return model_msg;
}

BOOST_AUTO_TEST_CASE(manipulation1)
{
    RTLParser p;
    System s;
    p.parse_from_file("test-real-example.vam", &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_single_assignments();
    m.update_fnode_io();

    for (auto fnode : s.fnodes)
        for (auto assign : fnode->assignments)
            BOOST_REQUIRE(assign->rhs->op != OP_VAR);

    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 293);
    
    m.remove_singlepurpose_assignments();
    m.update_fnode_io();

    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 98);
}

BOOST_AUTO_TEST_CASE(manipulation2)
{
    string input =
            "(model modelname\n"
            "(pipeline-stages xx)\n"
            "(sig v1 1)\n"
            "(sig v2 1)\n"
            "(sig v3 1)\n"
            "(sig v4 1)\n"
            "(sig v5 1)\n"
            "(sig v6 1)\n"
            "(reg r0 1 (we v1) (d v2) (q v3))\n"
            "(fnode f0 (input v3) (output v1) (stage xx) (assign \n"
            "(:= v2 v4)\n"
            "(:= v1 (|| v2 v3))"
            "))\n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_single_assignments();

    // signal v2 should be removed
    BOOST_REQUIRE(s.signals["v2"] == nullptr);

    log_model("manipulation2", s);

    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    RTLVisitor v{&s};
    fnode->Accept(&v);
    string expected =
        "  (fnode f0 (input v3 ) (output v1 ) (stage xx) "
        "    (assign (:= v1 (|| v4 v3 ) )     )) ";
    BOOST_REQUIRE_MESSAGE(strip(v.out.str()) == strip(expected),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(manipulation3)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1 (vt control) (connections (r0 we) ) )\n"
            "  (sig v3 1 (vt data ) (connections (r0 q) ) )\n"
            "  (sig v4 1 (vt unknown) )\n"
            "  (sig v5 1 (vt unknown) (connections (r0 d) ) )\n"
            "  (reg r0 1 (we v1) (d v5) (q v3 ) )\n"
            "  (fnode f0 (input v3 v4 v5 ) (output v1 v4 ) (stage xx) (assign \n"
            "(:= v1 (|| v5 v4 ) ) \n"
            "(:= v4 (&& v3 v5 ) ) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);

    SystemManipulation m(s);
    m.debug = debugflag;
    m.remove_singlepurpose_assignments();
    m.update_fnode_io();

    log_model("manipulation3", s);

    // signal v4 should be removed
    BOOST_REQUIRE(s.signals["v4"] == nullptr);

    // just one f0 fnode
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    RTLVisitor v{&s};
    fnode->Accept(&v);
    string expected1 =
        "  (fnode f0 (input v5 v3 ) (output v1 ) (stage xx) "
        "    (assign (:= v1 (|| v5 (&& v3 v5 ) ) ) )) ";
    string expected2 =
        "  (fnode f0 (input v3 v5 ) (output v1 ) (stage xx) "
        "    (assign (:= v1 (|| v5 (&& v3 v5 ) ) ) )) ";
    BOOST_REQUIRE_MESSAGE(
            strip(v.out.str()) == strip(expected1) ||
            strip(v.out.str()) == strip(expected2),
        "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(encapsulate_expr1)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (? v2 v3 (? v4 v5 v3))) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    auto fnode = s.fnodes["f0"];
    auto a = fnode->get_assignment(0);
    auto a_rhs = a->rhs;
    Signal *signal = nullptr;
    
    // execute
    auto ret = m.encapsulate_expr(a_rhs, "s", false, &signal);

    // verify
    log_model("encapsulate_expr1", s);

    BOOST_REQUIRE(s.fnodes.size() == 2);
    BOOST_REQUIRE(s.signals.size() == 6);
    BOOST_REQUIRE(ret == nullptr); 
    BOOST_REQUIRE(signal != nullptr);
    BOOST_REQUIRE(signal->references.size() == 2);

    Expression *signal_rhs = nullptr;
    for (auto e : signal->references)
        if (e->outputs[0]->op == OP_ASSIGN)
        {
            signal_rhs = static_cast<ExprAssign*>(e->outputs[0])->rhs;
            break;
        }
    BOOST_REQUIRE(signal_rhs != nullptr);
    BOOST_REQUIRE(signal_rhs == a_rhs);
    BOOST_REQUIRE(a->rhs->op == OP_VAR);
    BOOST_REQUIRE(static_cast<ExprVariable*>(a->rhs)->v == signal);
}

BOOST_AUTO_TEST_CASE(encapsulate_expr2)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (? v2 v3 (? v4 v5 v3))) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    auto fnode = s.fnodes["f0"];
    auto a = fnode->get_assignment(0);
    
    // execute
    auto ret = m.encapsulate_expr(a->rhs, "s", false, nullptr);

    // verify
    BOOST_REQUIRE(s.fnodes.size() == 2);
    BOOST_REQUIRE(s.signals.size() == 6);
    BOOST_REQUIRE(ret == nullptr); 
}

BOOST_AUTO_TEST_CASE(encapsulate_expr3)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (_ 0 1)) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    auto fnode = s.fnodes["f0"];
    auto a = fnode->get_assignment(0);
    
    // execute
    auto ret = m.encapsulate_expr(a->rhs, "s", false, nullptr);

    // verify
    log_model("encapsulate_expr3", s);
    BOOST_REQUIRE(ret == a->rhs); 
    BOOST_REQUIRE(s.fnodes.size() == 1);
    BOOST_REQUIRE(s.signals.size() == 5);
}

BOOST_AUTO_TEST_CASE(encapsulate_expr4)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (_ 0 1)) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    auto fnode = s.fnodes["f0"];
    auto a = fnode->get_assignment(0);
    
    // execute
    auto ret = m.encapsulate_expr(a->rhs, "s", true, nullptr);

    // verify
    log_model("encapsulate_expr4", s);
    BOOST_REQUIRE(ret == nullptr); 
    BOOST_REQUIRE(s.fnodes.size() == 2);
    BOOST_REQUIRE(s.signals.size() == 6);
}

BOOST_AUTO_TEST_CASE(encapsulate_expr_to_signal)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (_ 0 1)) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    auto fnode = s.fnodes["f0"];
    auto a = fnode->get_assignment(0);
    
    // execute
    auto signal = m.encapsulate_expr_to_signal(a->rhs, "s");

    // verify
    log_model("encapsulate_expr_to_signal", s);
    BOOST_REQUIRE_MESSAGE(boost::starts_with(signal->id, "s"), signal->id); 
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 2);
    BOOST_REQUIRE_EQUAL(s.signals.size(), 6);
}

BOOST_AUTO_TEST_CASE(ite2switch1)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (? v2 v3 (? v4 v5 (&& v3 v5))) ) \n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    // execute
    m.transform_ite_to_switch();

    // verify
    log_model("ite2switch1", s);
    BOOST_REQUIRE(s.fnodes.size() == 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    BOOST_REQUIRE_EQUAL(fnode->count_assignments(), 1);
    RTLVisitor v{&s};
    fnode->get_assignment(0)->Accept(&v);
    string expected = "(:= v1 (switch v2 "
        "   ((_ 1 1) v3)"
        "   ((_ 0 1) (switch v4 "
        "       ((_ 1 1) v5)"
        "       ((_ 0 1) (&& v3 v5))"
        "   ))"
        "))";
    BOOST_REQUIRE_MESSAGE(
            strip(v.out.str()) == strip(expected),
            "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(ite2switch2)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (&& v2 (? (~ v3) (+ v4 v5) (! v2))))\n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    // execute
    m.transform_ite_to_switch();

    // verify
    log_model("ite2switch2", s);
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    BOOST_REQUIRE_EQUAL(fnode->count_assignments(), 1);
    RTLVisitor v{&s};
    fnode->get_assignment(0)->Accept(&v);
    string expected = "(:= v1 (&& v2 (switch (~ v3)"
        "   ((_ 1 1) (+ v4 v5))"
        "   ((_ 0 1) (! v2))"
        "   )))";
    BOOST_REQUIRE_MESSAGE(
            strip(v.out.str()) == strip(expected),
            "output: \n" << v.out.str());
}

BOOST_AUTO_TEST_CASE(normalization)
{
    System s;
    {
        // setup
        RTLParser p;
        p.parse_from_file("test-real-example.vam", &s);
        SystemManipulation m(s);
        m.remove_single_assignments();
        m.remove_singlepurpose_assignments();
        m.update_fnode_io();
    }

    // excercise
    SystemManipulation m(s);
    m.debug = debugflag;
    m.step = 0;
    m.remove_single_assignments();
    m.remove_singlepurpose_assignments();
    m.make_uniform_fixpoint();
    m.update_fnode_io();

    BOOST_REQUIRE_EQUAL(s.signals.size(), 125);
    BOOST_REQUIRE_EQUAL(s.registers.size(), 23);
    BOOST_REQUIRE_EQUAL(s.memories.size(), 3);
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 98);
}

BOOST_AUTO_TEST_CASE(unify_selectors1)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (? v2 v3 v4))\n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    m.unify_selectors();

    // verify
    log_model("unify_selectors1", s);
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 1);
    auto fnode = s.fnodes["f0"];
    BOOST_REQUIRE(fnode != nullptr);
    BOOST_REQUIRE_EQUAL(fnode->count_assignments(), 1);

    RTLVisitor v{&s};
    auto a = fnode->get_assignment(0);
    a->Accept(&v);
    string output = "output:\n" + v.out.str();

    BOOST_CHECK_MESSAGE(a->rhs->op == OP_CASE, output);
    
    string expected = "(:= v1 (switch v2 ((_ 1 1) v3 ) ((_ 0 1) v4 )))";
    BOOST_REQUIRE_MESSAGE(
            strip(v.out.str()) == strip(expected), output);
}

template<typename T>
T *rtl_object_starts_with(NamedElementContainer<T> &container, string prefix)
{
    for (T *o : container)
        if (boost::starts_with(o->id, prefix))
            return o;
    return nullptr;
}

BOOST_AUTO_TEST_CASE(unify_selectors2)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 (? v2 v3 (? v4 v5 v2)))\n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    BOOST_CHECK(true);
    m.unify_selectors();

    // verify
    log_model("unify_selectors2", s);
    BOOST_REQUIRE_EQUAL(s.signals.size(), 6);
    BOOST_REQUIRE(s.fnodes["f0"] != nullptr);
    BOOST_REQUIRE_EQUAL(s.fnodes["f0"]->count_assignments(), 1);
    Signal *signal_helper = rtl_object_starts_with(s.signals, "sel_flow");
    BOOST_REQUIRE(signal_helper != nullptr);
    FunNode *fnode = s.fnodes["f_" + signal_helper->id];
    BOOST_REQUIRE(fnode != nullptr);
    BOOST_REQUIRE_EQUAL(fnode->count_assignments(), 1);
    
    RTLVisitor v{&s};
    s.Accept(&v);
    string model_msg = "output model:\n" + v.out.str();

    for (auto f : s.fnodes)
        BOOST_REQUIRE_MESSAGE(f->get_assignment(0)->rhs->op == OP_CASE,
                model_msg);
}

BOOST_AUTO_TEST_CASE(unify_selectors3)
{
    string input =
            "(model modelname\n"
            "  (pipeline-stages xx)\n"
            "  (sig v1 1)\n"
            "  (sig v2 1)\n"
            "  (sig v3 1)\n"
            "  (sig v4 1)\n"
            "  (sig v5 1)\n"
            "  (fnode f0 (input) (output) (assign \n"
            "(:= v1 "
            "   (cond "
            "       ((&& v1 v2) (_ 0 1))"
            "       ((|| v2 v3) (! v5))"
            "       ((! (&& v1 v3)) v5)"
            "   )"
            ")\n"
            ")) \n"
            ")\n";
    RTLParser p;
    System s;
    p.parse_str(input, &s);
    SystemManipulation m(s);

    BOOST_CHECK(true);
    m.unify_selectors();

    // verify
    log_model("unify_selectors3", s);
    BOOST_REQUIRE_EQUAL(s.signals.size(), 10);
    BOOST_REQUIRE_EQUAL(s.fnodes.size(), 6);

    string model_msg = log_model("unify_selectors3", s);

    BOOST_REQUIRE(s.fnodes["f0"] != nullptr);
    BOOST_REQUIRE_EQUAL(s.fnodes["f0"]->count_assignments(), 1);
    Signal *signal_helper = rtl_object_starts_with(s.signals, "sel_flow");
    BOOST_REQUIRE(signal_helper != nullptr);
    FunNode *fnode = s.fnodes["f_" + signal_helper->id];
    BOOST_REQUIRE(fnode != nullptr);
    BOOST_REQUIRE_EQUAL(fnode->count_assignments(), 1);
    

    for (auto f : s.fnodes)
    {
        BOOST_REQUIRE_MESSAGE(f->count_assignments() == 1, model_msg);
        if (f->id == "f0")
        {
            BOOST_REQUIRE_MESSAGE(f->get_assignment(0)->rhs->op == OP_CASE,
                    model_msg);
        }
        else if (boost::starts_with(f->id, "f_selector"))
        {
            BOOST_REQUIRE_MESSAGE(f->get_assignment(0)->rhs->op == OP_CONCAT,
                    model_msg);
        }
        else if (boost::starts_with(f->id, "f_sel_"))
        {
            BOOST_REQUIRE_MESSAGE(f->get_assignment(0)->rhs->op_is_logic(),
                    model_msg);
        }
        else
            BOOST_FAIL("\nunexpected fnode identifier: " + f->id);
    }
}

BOOST_AUTO_TEST_SUITE_END()
