/**
 * @brief   High-level test suite for VAM elements.
 */

#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>
#include "vaml_test.h"

BOOST_AUTO_TEST_SUITE(vam_elements)

BOOST_AUTO_TEST_CASE(meta)
{
    Meta m{MT_NODE, ""};
    *m.add(MT_NODE, "loc") << "xyz.codal" << 123;
    *m.add(MT_NODE, "loc") << "xyz.codal" << 528;

    auto l = m.find("loc");
    BOOST_REQUIRE_EQUAL((*l)[1], "123");

    l = m.find("loc", l);
    BOOST_REQUIRE_EQUAL((*l)[1], "528");

    RTLVisitor ov;
    m.Accept(&ov);
    BOOST_REQUIRE_EQUAL(strip(ov.out.str()),
            strip("(meta (loc \"xyz.codal\" 123)(loc \"xyz.codal\" 528))"));
}

BOOST_AUTO_TEST_CASE(assign)
{
    Variable v1{"v1", 4};
    Variable v2{"v2", 4};
    Variable v3{"v3", 4};
    ExprAssign e{new ExprVariable{&v1}, new ExprVariable{&v2}};
    *e.meta.add(MT_NODE, "loc") << "some_src_file.cc" << 123;
    RTLVisitor ov;
    e.Accept(&ov);
    BOOST_REQUIRE_EQUAL(strip(ov.out.str()),
            strip("(:= v1 v2 (meta (loc \"some_src_file.cc\" 123 ) ) )"));
}

BOOST_AUTO_TEST_CASE(expr1)
{
    Variable freg_d{"fetchD", 32};
    Variable freg_q{"fetchQ", 32};
    Variable freg_we{"fetchWE", 1};
    auto a1 = new ExprAssign{
            new ExprVariable{&freg_we},
            new ExprNot{
            new ExprEqual{
                new ExprSlice{
                    new ExprVariable{&freg_q},
                    0,3}, 
                new ExprConst{"15", 4}}}};
    *a1->meta.add(MT_NODE, "reason") 
        << "load a new instruction iff no halt";
    RTLVisitor ov;
    a1->Accept(&ov);
    BOOST_REQUIRE_EQUAL(strip(ov.out.str()),
            strip("(:= fetchWE (! (== ([..] fetchQ 0 3 ) (_ 15 4) ) )"
                "(meta (reason \"load a new instruction iff no halt\" ) ) )"));

    // teardown
    delete a1;
}

BOOST_AUTO_TEST_CASE(expr2)
{
    Variable freg_d{"fetchD", 32};
    Variable freg_q{"fetchQ", 32};
    Variable freg_we{"fetchWE", 1};
    auto c = new ExprCase();
    c->add_ctrl_expr(new ExprSlice{
            new ExprVariable{&freg_q},
            4,5});
    c->add_case(new ExprConst{"0",2}, new ExprConst{"0",32});
    c->add_case(new ExprConst{"1",2}, new ExprConst{"-1",32});
    c->add_case(new ExprConst{"2",2}, new ExprConst{"1",32});
    c->add_case(new ExprConst{"3",2}, new ExprVariable{&freg_q});
    c->add_default(new ExprConst{"0",32});
    auto a = new ExprAssign{new ExprVariable{&freg_d}, c}; 
    RTLVisitor ov;
    a->Accept(&ov);

    BOOST_REQUIRE_EQUAL(strip(ov.out.str()),
            strip("(:= fetchD (switch "
                "([..] fetchQ 4 5 ) ((_ 0 2) (_ 0 32) ) ((_ 1 2) (_ -1 32) ) "
                "((_ 2 2) (_ 1 32) ) ((_ 3 2) fetchQ ) ((_ 0 32) ) ) )"));
    // teardown
    delete a;
}

BOOST_AUTO_TEST_CASE(expr_replacement1)
{
    auto a = new ExprAnd{new ExprConst{"1", 1}, new ExprConst{"1", 1}};
    auto o = new ExprOr{a, new ExprConst{"1", 1}};
    auto ite = new ExprITE{a, o, o};
    ExprRef holder{ite};

    auto rplc = new ExprConst{"0", 1};

    replace_outputs(a, rplc, ite);

    BOOST_REQUIRE(o->operands[0] == rplc);
    BOOST_REQUIRE(ite->cond == a);
    BOOST_REQUIRE(a->outputs.size() == 1);
}

BOOST_AUTO_TEST_CASE(expr_replacement2)
{
    auto a = new ExprAnd{new ExprConst{"1", 1}, new ExprConst{"1", 1}};
    auto o = new ExprOr{a, new ExprConst{"1", 1}};
    auto ite = new ExprITE{a, o, o};
    ExprRef holder{ite};

    auto rplc = new ExprConst{"0", 1};

    replace_outputs(a, rplc);

    BOOST_REQUIRE(o->operands[0] == rplc);
    BOOST_REQUIRE(ite->cond == rplc);
}

BOOST_AUTO_TEST_SUITE_END()
